/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <drm/drm_modes.h>
#include "common/gb_common.h"
#include "kms/gb_pcie_map.h"
#include "kms/device/gb_ip.h"
#include "common/gb_procfs.h"
#include "ip/gb_dp.h"

/* version format(3 sections): xx.xx.xx (major.minor.build). e.g., v0.1.20 */
#define DIGIT_TOTAL_NUM    6  /* version data in register. e.g., 0x000120 */
#define DIGIT_NUM_PER_SECTION   2  /* number of hex digit in each section */
#define SECTION_NUM    3
#define HEX_DIGIT_SHIFT    4
#define VERSION_STR_LEN_MAX 10

struct proc_dir_entry *proc_parent;

static int gb_hex_to_version(uint32_t reg_val, char* version)
{
	uint8_t sect_arr[SECTION_NUM] = {0, 0, 0};
	uint8_t sect = 0;
	int i,j = 0;
	if(version==NULL) {
		return -1;
	}
	for(i=0; i < DIGIT_TOTAL_NUM; i++) {
		/*get 2-bit dec digits from 2-bit hex digits of each section*/
		if(i%DIGIT_NUM_PER_SECTION) {
			sect = sect + ((reg_val>>(i*HEX_DIGIT_SHIFT))&0xf) * 10;
			sect_arr[j++] = sect;
		} else {
			sect = 0;
			sect = ((reg_val>>(i*HEX_DIGIT_SHIFT))&0xf);
		}
	}
	sprintf(version, "%d.%d.%d", sect_arr[2], sect_arr[1], sect_arr[0]);
	gb_printf(KERN_INFO, "verr:%s\n", version);
	return 0;
}

static int gb_version_show(struct seq_file *m, void *v)
{
	char mcu_version[VERSION_STR_LEN_MAX] = {0,};
	char bios_version[VERSION_STR_LEN_MAX] = {0,};
	struct version_str gb_version;
	struct gbdc_pcie_info *gbpcie_info = m->private;
	void __iomem *pcie_base = gbpcie_info->pci_mmio_bar[0];
	struct gb_pcie_info *gb_pcie = get_pcie_gl_info();
	iowrite32(0x60000000, gbpcie_info->pci_mmio_bar[4] + 0xd0c);
	gbpcie_info->gb_cops->get_version(pcie_base, &gb_version);
	gb_hex_to_version(gb_version.mcu_vid, mcu_version);
	gb_hex_to_version(gb_version.bios_vid, bios_version);
	get_conversion_chip_version(gb_pcie);
	seq_printf(m, "driver\t: %s\n", STR2(GB_VERSION));
	seq_printf(m, "vbios\t: V%s-%x\n", mcu_version, gb_version.mcu_vdata);
	seq_printf(m, "option rom\t: V%s-%x\n", bios_version, gb_version.bios_vdata);
      if (gb_pcie->f_info.version_info & HDMI_VERSION_INFO)
                seq_printf(m, "HDMI firmware\t: V%x.%x\n",
                           gb_pcie->f_info.hdmi_firmware_maj,
                           gb_pcie->f_info.hdmi_firmware_min);
        if (gb_pcie->f_info.version_info & VGA_VERSION_INFO)
                seq_printf(m, "VGA firmware\t: V%x.%x\n",
                           gb_pcie->f_info.vga_firmware_maj,
                           gb_pcie->f_info.vga_firmware_min);
	return 0;
}

static int gb_version_open(struct inode *inode, struct file *file)
{
	struct gbdc_pcie_info *gbpcie_info;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	gbpcie_info = PDE_DATA(inode);
#else
	gbpcie_info = pde_data(inode);
#endif
	gb_printf(KERN_INFO, "chen gb open proc--\n");

	return single_open(file, gb_version_show, gbpcie_info);

}
static int gb_version_release(struct inode *inode, struct file *file)
{
	int res = single_release(inode, file);
	return res;

}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
static const struct file_operations exports_proc_ops = {
	.open = gb_version_open,
	.read = seq_read,
	.release=gb_version_release,
	.owner= THIS_MODULE,
};
#else
static const struct proc_ops exports_proc_ops = {
	.proc_open = gb_version_open,
	.proc_read = seq_read,
	.proc_release = gb_version_release,
};
#endif

int gb_create_proc_entry(struct gbdc_pcie_info *gbpcie_info)
{
	gb_printf(KERN_INFO, "gb enter proc entry-\n");

	remove_proc_subtree("gb", NULL);
	proc_parent = proc_mkdir("gb", NULL);
	if(!proc_parent){
		gb_printf(KERN_ERR, "error creating proc entry\n");
		return -ENOMEM;
	}

	proc_parent = proc_create_data("version", 0, proc_parent, &exports_proc_ops, gbpcie_info);
	if(!proc_parent){
		remove_proc_subtree("gb", NULL);
		return -ENOMEM;
	}

	gb_printf(KERN_INFO, "gb create proc entry-\n");
	return 0;
}

void gb_remove_proc(void)
{
	remove_proc_subtree("gb", NULL);
}
