// SPDX-License-Identifier: GPL-2.0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include "gb_firmware_upgrade.h"
#include "gb_device.h"

#define UPGRADE_MCU_TEST_CMD_BASE 'U'
#define UPGRADE_MCU_CMD_START _IOW(UPGRADE_MCU_TEST_CMD_BASE, 1u, int)
#define UPGRADE_MCU_CMD_GET_PROGRESS _IOW(UPGRADE_MCU_TEST_CMD_BASE, 2u, int)
#define UPGRADE_MCU_CMD_SET_IMAGE_INFO _IOW(UPGRADE_MCU_TEST_CMD_BASE, 3u, int)

static struct gb_fw_upgrade *mcu_infos;

#define PCIE_UPGRADE_MCU_STR                ("upgrade_mcu")

#define UPGRADE_BUFFER_SIZE (1024 * 4)    /* 4KB */
//#define MCU_VRAM_START_ADDR (0xB0000000 )
#define GB02_UPGRADE_MCU_SUB_IMAGE_SIZE	(0x100000) /* 1MB */
#define MCU_SUB_IMAGE_FILENAME_OFFSET	(0x20000)
#define MCU_SUB_IMAGE_ADDR_OFFSET		(0x20100)
#define MCU_SUB_IMAGE_SIZE_OFFSET		(0x20200)
#define MCU_SUB_IMAGE_CRC16_OFFSET		(0x20300)
#define MCU_SUB_IMAGE_CHECKSUM_OFFSET	(0x20400)
#define MCU_SUB_IMAGE_TYPE_OFFSET		(0x20500)

ssize_t mcu_upgrade_img_read(struct gb_fw_upgrade *mcu_ctrl, unsigned int offset, void *buffer, size_t len)
{
    struct gb_device *gbdev = mcu_ctrl->gbdev;
    int ddr_bar_id = 2;
    void *base_addr = gbdev->gb_pcie->pci_bars[ddr_bar_id].mmio + MCU_VRAM_START_ADDR + offset;

    memcpy(buffer, base_addr, len);

    return len;
}

ssize_t mcu_upgrade_img_write(struct gb_fw_upgrade *mcu_ctrl, unsigned int offset, void *buffer, size_t len)
{
    struct gb_device *gbdev = mcu_ctrl->gbdev;
    int ddr_bar_id = 2;
    void *base_addr = gbdev->gb_pcie->pci_bars[ddr_bar_id].mmio + MCU_VRAM_START_ADDR + offset;

    memcpy(base_addr, buffer, len);

    return len;
}

static int gb02_fw_open(struct inode *inode, struct file *filp)
{
    filp->private_data = mcu_infos;
	return 0;
}

static int gb02_fw_release(struct inode *inode, struct file *filp)
{
    filp->private_data = NULL;
	return 0;
}
#define VMEM_FLAGS (VM_READ|VM_WRITE|VM_EXEC)
static int gb02_fw_mmap(struct file *filp, struct vm_area_struct *vma)
{
    unsigned long off;
    unsigned long phys;
    unsigned long vsize;
    unsigned long psize;
    int rv;
    struct gb_device *gbdev = mcu_infos->gbdev;
    off = vma->vm_pgoff << PAGE_SHIFT;

    /* BAR physical address */
    phys = pci_resource_start(gbdev->gb_pcie->pdev, 0) + off;
    vsize = vma->vm_end - vma->vm_start;
    /* complete resource */
    psize = pci_resource_end(gbdev->gb_pcie->pdev, 0) -
        pci_resource_start(gbdev->gb_pcie->pdev, 0) + 1 - off;

    // printk("mmap(): mmap->bar:0\n");
    // printk("off = 0x%lx, vsize 0x%lu, psize 0x%lu.\n", off, vsize, psize);
    // printk("start = 0x%llx\n", gbdev->gb_pcie->pci_bars[0].base);
    // printk("phys = 0x%lx\n", phys);

    if (vsize > psize)
        return -EINVAL;
    /*
    * pages must not be cached as this would result in cache line sized
    * accesses to the end point
    */
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
    /*
    * prevent touching the pages (byte access) for swap-in,
    * and prevent the pages from being swapped out
    */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	vma->vm_flags |= VMEM_FLAGS;
#else
	vm_flags_set(vma, VMEM_FLAGS);
#endif
    /* make MMIO accessible to user space */
    rv = io_remap_pfn_range(vma, vma->vm_start, phys >> PAGE_SHIFT,
            vsize, vma->vm_page_prot);
    // printk("vma=0x%p, vma->vm_start=0x%lx, phys=0x%lx, size=%lu = %d\n",
    //     vma, vma->vm_start, phys >> PAGE_SHIFT, vsize, rv);

    if (rv)
        return -EAGAIN;
    return 0;
}

static ssize_t
gb02_fw_read(struct file *filp, char __user *buf, size_t count, loff_t *ppos)
{
    char *kbuffer;
    ssize_t bytes_read = 0;
    struct gb_fw_upgrade *mcu_ctrl = (struct gb_fw_upgrade *)(filp->private_data);

    kbuffer = kmalloc(min_t(size_t, count, UPGRADE_BUFFER_SIZE), GFP_KERNEL);
    if (!kbuffer)
        return -ENOMEM;

    while (bytes_read == 0) {
        ssize_t need = min_t(unsigned long, count, UPGRADE_BUFFER_SIZE);

        bytes_read = mcu_upgrade_img_read(mcu_ctrl, bytes_read, kbuffer, need);
        if (bytes_read != 0)
            break;

        if (filp->f_flags & O_NONBLOCK) {
            bytes_read = -EAGAIN;
            break;
        }

	    if (bytes_read > 0 && copy_to_user(buf, kbuffer, bytes_read))
    	    bytes_read = -EFAULT;

		bytes_read += bytes_read;
	}

    kfree(kbuffer);
    return bytes_read;
}

static ssize_t
gb02_fw_write(struct file *filp, const char __user *buf,
              size_t count, loff_t *ppos)
{
    char *kbuffer;
    ssize_t bytes_written = 0;
    ssize_t wrote;
    struct gb_fw_upgrade *mcu_ctrl = (struct gb_fw_upgrade *)(filp->private_data);

    kbuffer = kmalloc(min_t(size_t, count, UPGRADE_BUFFER_SIZE), GFP_KERNEL);
    if (!kbuffer)
        return -ENOMEM;

    while (bytes_written < count) {
        ssize_t n = min_t(unsigned long, count - bytes_written, UPGRADE_BUFFER_SIZE);

        if (copy_from_user(kbuffer, buf + bytes_written, n)) {
            bytes_written = -EFAULT;
            break;
        }

        wrote = mcu_upgrade_img_write(mcu_ctrl, bytes_written, kbuffer, n);
        if (wrote <= 0) {
            if (!bytes_written)
                bytes_written = wrote;
            break;
        }

        bytes_written += wrote;

        if (filp->f_flags & O_NONBLOCK) {
            if (!bytes_written)
                bytes_written = -EAGAIN;
            break;
        }
    }

    kfree(kbuffer);
    return bytes_written;
}

static int gb02_upgrade_mcu_ioctl_start(struct gb_fw_upgrade *mcu_ctrl)
{
    u32 val = 0;
    struct gb_device *gbdev = mcu_ctrl->gbdev;

    base_reg_write(MCU_SYSCTRL_BASE_ADDR, gbdev->gb_pcie->pci_bars[UPGRADE_CFG_REG_BAR].mmio, CFG_MCU_XIP_ADDR_OFFSET);

    /* tell MCU to upgrade sub image */

    val = base_reg_read(gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE5_OFFSET);
    val &= ~0xffff;
    val |= 0x5a;
    base_reg_write(val, gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE5_OFFSET);

    mcu_ctrl->upgrade_result = T_UPGRADE_MCU_RESULT_UPGRADING;
	return 0;
}

static int gb02_upgrade_mcu_ioctl_get_progress(struct gb_fw_upgrade *mcu_ctrl)
{
    u32 val = 0;
    struct gb_device *gbdev = mcu_ctrl->gbdev;

    /* set SPARE4 register */
    base_reg_write(MCU_SYSCTRL_BASE_ADDR, gbdev->gb_pcie->pci_bars[UPGRADE_CFG_REG_BAR].mmio, CFG_MCU_XIP_ADDR_OFFSET);

    /* get the progress of upgrade MCU's sub image */
    val = base_reg_read(gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE5_OFFSET);
    mcu_ctrl->upgrade_result = ((val >> 8) & 0xff);
    return mcu_ctrl->upgrade_result;
}

static int gb02_upgrade_mcu_ioctl_set_image_info(struct gb_fw_upgrade *mcu_ctrl, struct upgrade_image_info_t *image_info)
{
	u32 blk_nbr = 0;
    struct gb_device *gbdev = mcu_ctrl->gbdev;
    int ddr_bar_id = 2;
	void __iomem *base_addr = gbdev->gb_pcie->pci_bars[ddr_bar_id].mmio + MCU_VRAM_START_ADDR;

	if (image_info == NULL)
		return -1;

	mcu_ctrl->upgrade_result = T_UPGRADE_MCU_RESULT_NO_UPGRADE;
	/* FILENAME */
	gb_printf(KERN_DEBUG, "filename : %s, addr : 0x%llx, size : 0x%x, type : %d\n",
			image_info->filename, image_info->addr, image_info->size, image_info->type);

	memcpy(base_addr + MCU_SUB_IMAGE_FILENAME_OFFSET, image_info->filename, strlen(image_info->filename));
	*((char *)(base_addr + MCU_SUB_IMAGE_FILENAME_OFFSET + strlen(image_info->filename))) = '\0';

	/* ADDR */
	base_reg_write(image_info->addr & 0x1FFFFFFF, base_addr, MCU_SUB_IMAGE_ADDR_OFFSET);

	blk_nbr = (image_info->addr)/(512*MB);

	gb_printf(KERN_DEBUG, "512M_blk_number:%d--0x%x",blk_nbr, blk_nbr);

	/* write 512M block index to sysctrl_SPARE3 6000020c register spare3*/
	base_reg_write(blk_nbr, gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE3_OFFSET);

	/* SIZE */
	base_reg_write(image_info->size, base_addr, MCU_SUB_IMAGE_SIZE_OFFSET);

	/* CRC16 */
	base_reg_write(image_info->crc16, base_addr, MCU_SUB_IMAGE_CRC16_OFFSET);

	/* CHECKSUM */
	base_reg_write(image_info->checksum, base_addr, MCU_SUB_IMAGE_CHECKSUM_OFFSET);

	/* IMAGE TYPE */
	base_reg_write(image_info->type, base_addr, MCU_SUB_IMAGE_TYPE_OFFSET);

	return 0;
}

static long gb02_fw_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    int ret = 0;
	struct gb_fw_upgrade *mcu_ctrl = filp->private_data;
	struct upgrade_image_info_t image_info;

	gb_printf(KERN_INFO, "enter ioctrl!\n");

    switch (cmd) {
    case UPGRADE_MCU_CMD_START:
		gb_printf(KERN_INFO, "=====Upgrade MCU's SUB start!======\n");
        ret = gb02_upgrade_mcu_ioctl_start(mcu_ctrl);
        break;
    case UPGRADE_MCU_CMD_GET_PROGRESS:
		gb_printf(KERN_INFO, "=====Get the progress of upgrade MCU's sub start!======\n");
        ret = gb02_upgrade_mcu_ioctl_get_progress(mcu_ctrl);
        break;
	case UPGRADE_MCU_CMD_SET_IMAGE_INFO:
		gb_printf(KERN_INFO, "=====Set the upgrade image info of MCU start!======\n");
		memset(&image_info, 0, sizeof(image_info));
		if (copy_from_user(&image_info, (struct upgrade_image_info_t *)arg,
			sizeof(struct upgrade_image_info_t)) ) {
			gb_printf(KERN_ERR, "%s:%d:copy_from_user() invalid!\n", __func__, __LINE__);
			return -EFAULT;
		}

        ret = gb02_upgrade_mcu_ioctl_set_image_info(mcu_ctrl, &image_info);
		break;
    default:
        break;
    }

	return 0;
}

static __poll_t gb02_fw_poll(struct file *filp, poll_table *wait)
{
    struct gb_fw_upgrade *mcu_ctrl = filp->private_data;
	__poll_t mask = 0;
    int val = 0;

    /* for FPGA, comment out such codes */
    val = gb02_upgrade_mcu_ioctl_get_progress(mcu_ctrl);
    gb_printf(KERN_INFO, "%s upgrade state : %d\n",__func__,val);
    if(val == T_UPGRADE_MCU_RESULT_OK) {
        mcu_ctrl->upgrade_flag = 0;
        mcu_ctrl->upgrade_result = T_UPGRADE_MCU_RESULT_OK;
        mask |= EPOLLIN | EPOLLRDNORM;
    }

    return mask;
}

static struct file_operations gb02_fw_fops = {
    .owner = THIS_MODULE,
    .open = gb02_fw_open,
    .read = gb02_fw_read,
    .write = gb02_fw_write,
    .release = gb02_fw_release,
    .unlocked_ioctl = gb02_fw_ioctl,
    .mmap = gb02_fw_mmap,
    .poll = gb02_fw_poll,
};
static struct miscdevice gb02_fw_misc =
{
	.fops       = &gb02_fw_fops,
	.name       = "gb02_fw",
	.minor      = MISC_DYNAMIC_MINOR,
	.mode       = S_IRWXUGO,
};

irqreturn_t gb02_mcu2pcie_handler(int irq, void *arg)
{
    struct gb_fw_upgrade *mcu_ctrl = mcu_infos;
    struct gb_device *gbdev = mcu_ctrl->gbdev;
    u32 status = 0, val = 0;

	gb_printf(KERN_DEBUG, "%s:%d:enter.\n", __func__, __LINE__);

    base_reg_write(INT_CONTROLLER_BASE_ADDR, gbdev->gb_pcie->pci_bars[UPGRADE_CFG_REG_BAR].mmio, CFG_SYS_PERI_ADDR_OFFSET);
    status = base_reg_read(gbdev->gb_pcie->pci_bars[INT_CONTROLLER_BAR].mmio, INT_CONTROLLER_OFFSET + INTC_IRT_STATUS_H_OFFSET);
    if (status & (1 << (44 - 32))) {
        /* MCU upgrade finished */
        val = base_reg_read(gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_BOOT_MODE_OFFSET);
        val &= ~BOOT_MODE_MCU2PCIE_SW_INT;
        base_reg_write(val, gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_BOOT_MODE_OFFSET);

        val = base_reg_read(gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE5_OFFSET);
        val &= ~0xffff;
        val |= (2 << 8);
        base_reg_write(val, gbdev->gb_pcie->pci_bars[MCU_ACCESS_BAR].mmio, SYSCTL_SPARE5_OFFSET);
    }

    mcu_ctrl->upgrade_flag = 1;
    wake_up_interruptible(&mcu_ctrl->wait);

    return IRQ_HANDLED;
}


static int proc_upgrade_mcu_show(struct seq_file *m, void *v)
{
	struct gb_fw_upgrade *mcu_ctrl = (struct gb_fw_upgrade *)m->private;
	upgrade_mcu_result_e result = mcu_ctrl->upgrade_result;

	seq_printf(m, "%d\n", result);

	return 0;
}

static int proc_upgrade_mcu_open(struct inode *inode, struct file *file)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 17, 0)
	struct gb_fw_upgrade *mcu_ctrl = (struct gb_fw_upgrade *)PDE_DATA(file_inode(file));
#else
	struct gb_fw_upgrade *mcu_ctrl = (struct gb_fw_upgrade *)pde_data(file_inode(file));
#endif

    return single_open(file, proc_upgrade_mcu_show, mcu_ctrl);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
static const struct file_operations upgrade_mcu_proc_fops = {
    .owner      = THIS_MODULE,
    .open       = proc_upgrade_mcu_open,
    .read       = seq_read,
    .llseek     = seq_lseek,
    .release    = single_release,
};
#else
static const struct proc_ops upgrade_mcu_proc_fops = {
    .proc_open       = proc_upgrade_mcu_open,
    .proc_read       = seq_read,
    .proc_lseek      = seq_lseek,
    .proc_release    = single_release,
};
#endif

static int upgrade_mcu_procfs_node_create(struct gb_fw_upgrade *mcu_ctrl)
{
	char proc_name[15] = {0};

	snprintf(proc_name, sizeof(proc_name), "%s", "upgrade_mcu");

	mcu_ctrl->procdir = proc_mkdir(proc_name, NULL);
	if(!mcu_ctrl->procdir) {
		gb_printf(KERN_ERR, "create the procfs node(%s) failed.\n", proc_name);
		return -1;
	}

	proc_create_data("upgrade_result",  S_IRUGO, mcu_ctrl->procdir, &upgrade_mcu_proc_fops, (void *)mcu_ctrl);

	return 0;
}

int gb_firmware_upgrade_init(struct gb_device *gbdev, struct gb_asic_dev_info *gb_dev_info)
{
	int ret = 0;
	struct gb_fw_upgrade *mcu_ctrl;
	struct device *dev = gbdev->dev;

	gb_printf(KERN_INFO, "Start to init GB02 firmware's upgrade\n");

	mcu_ctrl = devm_kzalloc(dev, sizeof(*mcu_ctrl), GFP_KERNEL);
	if (!mcu_ctrl) {
		gb_printf(KERN_ERR, "%s:kmalloc failed\n", __func__);
		return -ENOMEM;
	}

	mcu_ctrl->gbdev = gbdev;
    mcu_ctrl->dev_info = gb_dev_info;
    init_waitqueue_head(&mcu_ctrl->wait);
    mcu_ctrl->upgrade_flag = 0;
    mcu_ctrl->upgrade_result = T_UPGRADE_MCU_RESULT_NO_UPGRADE;

	ret = misc_register(&gb02_fw_misc);
	if (ret < 0) {
		gb_printf(KERN_ERR, "%s:Failed to register as misc device:%d\n",
			__func__, ret);
		return ret;
	}

	/* create procfs node */
	ret = upgrade_mcu_procfs_node_create(mcu_ctrl);
	if(ret) {
		gb_printf(KERN_ERR, "upgrade_mcu_procfs_node_create() failed:%d\n", ret);
		goto err_misc_deregister;
	}

	/* enable mcu2pcie sw interrupt */
    /* for FPGA, disable such codes */
#if 0
    base_reg_write(INT_CONTROLLER_BASE_ADDR, gbdev->gb_pcie->pci_bars[UPGRADE_CFG_REG_BAR].mmio, CFG_SYS_PERI_ADDR_OFFSET);
	val = base_reg_read(gbdev->gb_pcie->pci_bars[INT_CONTROLLER_BAR].mmio, INT_CONTROLLER_OFFSET + INTC_IRQ_INTEN_H_OFFSET);
	val |= 1 << (44 - 32);
	base_reg_write(val, gbdev->gb_pcie->pci_bars[INT_CONTROLLER_BAR].mmio, INT_CONTROLLER_OFFSET + INTC_IRQ_INTEN_H_OFFSET);
#endif

    mcu_infos = mcu_ctrl;

    return 0;

err_misc_deregister:
	misc_deregister(&gb02_fw_misc);

	return ret;
}

void gb_pci_upgrade_remove(struct gb_device *gbdev)
{
	struct gb_fw_upgrade *mcu_ctrl = mcu_infos;

	if (mcu_ctrl->procdir) {
		proc_remove(mcu_ctrl->procdir);
	}

	misc_deregister(&gb02_fw_misc);
    mcu_infos = NULL;
}

void gb_pci_upgrade_shutdown(struct gb_device *gbdev)
{
	//struct gb_fw_upgrade *mcu_ctrl = mcu_infos;
#if 0
    u32 val = 0, i = 0;
#endif
    gb_printf(KERN_DEBUG, "%s:%d:enter\n", __func__, __LINE__);

    /* tell MCU to reboot */
    //base_reg_write(MCU_SYSCTRL_BASE_ADDR, mcu_ctrl->pci_mmio_bar[DMA_CFG_REG_BAR4], CFG_MCU_XIP_ADDR_OFFSET);
    /* hot reset */
    //base_reg_write(SYSCTL_SPARE2_HOT_RESET, mcu_ctrl->pci_mmio_bar[MCU_ACCESS_BAR], SYSCTL_SPARE2_OFFSET);
    /* MCU reset */
    //base_reg_write(SYSCTL_SPARE3_REBOOT_VALUE, mcu_ctrl->pci_mmio_bar[MCU_ACCESS_BAR], SYSCTL_SPARE3_OFFSET);

#if 0
    /* set SPARE4 register */
    base_reg_write(0, mcu_ctrl->pci_mmio_bar[MCU_ACCESS_BAR], SYSCTL_SPARE4_OFFSET);
    mdelay(100);

    for(i = 0;i < 1000;i++) {
        val = base_reg_read(mcu_ctrl->pci_mmio_bar[MCU_ACCESS_BAR], SYSCTL_SPARE4_OFFSET);
        if (val == SYSCTL_SPARE4_MCU_BOOT_FINISHED_VALUE)
            break;
        mdelay(10);
        gb_printf(KERN_INFO, "%s:%d:val = 0x%x\n", __func__, __LINE__, val);
    }
#endif
}
