/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/bitfield.h>
#include <linux/bitmap.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/iopoll.h>
#include <linux/pci.h>

#include "common/gb_uk.h"
#include "gb_device.h"
#include "gb_gpu.h"
#include "gb_regs.h"
#include "gb_gpu_irq.h"


irqreturn_t gb_gpu_irq_func(int irq, void *data)
{
	irqreturn_t ret = IRQ_NONE;
	u32 gpu_status;
	struct gb_device *gbdev = data;

	gpu_status = reg_read(gbdev, GPU_IRQ_STATUS);
	if (!gpu_status) {
		gb_printf(KERN_INFO,
			"Error: triggerd gpu_status irq once gpu: 0x%x\n",
			gpu_status);
		return IRQ_NONE;
	}

	if (gpu_status)
		ret = gb_gpu_irq_handler(gbdev, gpu_status);

	return ret;
}

irqreturn_t gb_gpu_irq_handler(struct gb_device *gbdev, u32 state)
{
	u32 fault_status = gpu_read(gbdev, GPU_FAULTSTATUS);

	if (state & (GPU_FAULT | MULTIPLE_GPU_FAULTS)) {
		u64 address = (u64) gpu_read(gbdev, GPU_FAULTADDRESS_HI) << 32;
		address |= gpu_read(gbdev, GPU_FAULTADDRESS_LO);

		dev_warn(gbdev->dev, "GPU Fault 0x%08x (%s) at 0x%016llx\n",
			 fault_status & 0xFF, gb_exception_name(gbdev, fault_status),
			 address);

		if (state & MULTIPLE_GPU_FAULTS)
			dev_warn(gbdev->dev, "There were multiple GPU faults - some have not been reported\n");

		gpu_write(gbdev, GPU_IRQ_MASK, 0);
	}

	gpu_write(gbdev, GPU_IRQ_CLEAR, state);

	return IRQ_HANDLED;
}

void gb_gpu_disable_irq(struct gb_device *gbdev)
{
	gpu_write(gbdev, GPU_IRQ_CLEAR, GPU_IRQ_REG_ALL);
	gpu_write(gbdev, GPU_IRQ_MASK, 0);
}

int gb_gpu_soft_reset(struct gb_device *gbdev)
{
	int ret;
	u32 val;
	u32 stage_status, mmu_status, gpu_status;
	u32 stage_rawstats;
	gpu_write(gbdev, GPU_IRQ_MASK, 0);
	gpu_write(gbdev, GPU_IRQ_CLEAR, RESET_COMPLETED);
	gpu_write(gbdev, GPU_COMMAND, GPU_COMMAND_SOFT_RESET);

	gb_printf(KERN_INFO, "===%s start===\n", __func__);

	ret = readl_relaxed_poll_timeout(gbdev->gpu_reg_base + GPU_IRQ_RAWSTAT,
		val, val & RESET_COMPLETED, 100, 10000);

	stage_status = reg_read(gbdev, STAGE_IRQ_STATUS);
	mmu_status = reg_read(gbdev, MMU_IRQ_STATUS);
	gpu_status = reg_read(gbdev, GPU_IRQ_STATUS);
	stage_rawstats = reg_read(gbdev, STAGE_IRQ_RAWSTAT);

	gb_printf(KERN_INFO, "IRQ info: gpu: 0x%x mmu: %x stage: %x stage_rawstats:%x\n",
		gpu_status, mmu_status, stage_status, stage_rawstats);

	if (ret) {
		dev_err(gbdev->dev, "gpu soft reset timed out\n");
		return ret;
	}

	gpu_write(gbdev, GPU_IRQ_CLEAR, GPU_IRQ_REG_ALL);
	gpu_write(gbdev, GPU_IRQ_MASK, GPU_IRQ_REG_ALL);
	gb_printf(KERN_INFO, "===%s end===\n", __func__);

	return 0;
}

static void gb_gpu_init_quirks(struct gb_device *gbdev)
{
	u32 quirks = 0;

	quirks = gpu_read(gbdev, TILER_CONFIG);

	gpu_write(gbdev, TILER_CONFIG, quirks);

	quirks = gpu_read(gbdev, L2_MMU_CONFIG);

	quirks &= ~(L2_MMU_CONFIG_LIMIT_EXTERNAL_READS |
		    L2_MMU_CONFIG_LIMIT_EXTERNAL_WRITES);

	gpu_write(gbdev, L2_MMU_CONFIG, quirks);
}

static void gb_gpu_init_features(struct gb_device *gbdev)
{
	u32 gpu_id, num_ss, major, minor, status, rev;
	int i;

	gbdev->features.l2_features = gpu_read(gbdev, L2_FEATURES);
	gbdev->features.tiler_features = gpu_read(gbdev, TILER_FEATURES);
	gbdev->features.mem_features = gpu_read(gbdev, MEM_FEATURES);
	gbdev->features.mmu_features = gpu_read(gbdev, MMU_FEATURES);
	gbdev->features.thread_features = gpu_read(gbdev, THREAD_FEATURES);
	gbdev->features.max_threads = gpu_read(gbdev, THREAD_MAX_THREADS);
	gbdev->features.thread_max_workgroup_sz = gpu_read(gbdev, THREAD_MAX_WORKGROUP_SIZE);
	gbdev->features.thread_max_barrier_sz = gpu_read(gbdev, THREAD_MAX_BARRIER_SIZE);
	gbdev->features.coherency_features = gpu_read(gbdev, COHERENCY_FEATURES);
	gbdev->features.texture_features[0] = gpu_read(gbdev, TEXTURE_FEATURES_0);
	gbdev->features.texture_features[1] = gpu_read(gbdev, TEXTURE_FEATURES_1);
	gbdev->features.texture_features[2] = gpu_read(gbdev, TEXTURE_FEATURES_2);

	gbdev->features.as_present = gpu_read(gbdev, AS_PRESENT);
	/* we use as[NUM_THREAD_AS] for perf & gpu utilization */
	gbdev->features.as_present &= (~(0x1 << NUM_THREAD_AS));

	gbdev->features.ss_present = gpu_read(gbdev, SS_PRESENT);
	num_ss = hweight32(gbdev->features.ss_present);
	for (i = 0; i < num_ss; i++)
		gbdev->features.ss_features[i] = gpu_read(gbdev, SS_FEATURES_REG(i));

	gbdev->features.shader_present = gpu_read(gbdev, SHADER_PRESENT_LO);
	gbdev->features.shader_present |= (u64)gpu_read(gbdev, SHADER_PRESENT_HI) << 32;

	gbdev->features.tiler_present = gpu_read(gbdev, TILER_PRESENT_LO);
	gbdev->features.tiler_present |= (u64)gpu_read(gbdev, TILER_PRESENT_HI) << 32;

	gbdev->features.l2_present = gpu_read(gbdev, L2_PRESENT_LO);
	gbdev->features.l2_present |= (u64)gpu_read(gbdev, L2_PRESENT_HI) << 32;
	gbdev->features.nr_core_groups = hweight64(gbdev->features.l2_present);

	gbdev->features.stack_present = 0;

	gbdev->features.revision = GENBU02_PCI_DEVICE_ID;//GENBU01_PCI_DEVICE_ID;
	gbdev->features.id = GENBU02_PCI_VENDOR_ID;//GENBU01_PCI_VENDOR_ID;

	major = (gbdev->features.revision >> 12) & 0xf;
	minor = (gbdev->features.revision >> 4) & 0xff;
	status = gbdev->features.revision & 0xf;
	rev = gbdev->features.revision;

	gpu_id = gbdev->features.id;

	dev_info(gbdev->dev, "Features: L2:0x%08x Tiler:0x%08x Mem:0x%0x MMU:0x%08x AS:0x%x JS:0x%x",
		 gbdev->features.l2_features,
		 gbdev->features.tiler_features,
		 gbdev->features.mem_features,
		 gbdev->features.mmu_features,
		 gbdev->features.as_present,
		 gbdev->features.ss_present);

	dev_info(gbdev->dev, "shader_present=0x%0llx l2_present=0x%0llx",
		 gbdev->features.shader_present, gbdev->features.l2_present);
}

void gb_gpu_L2_cache_power_on(void)
{
	struct gb_pcie_info *pcie_dev = get_pcie_gl_info();
	struct gb_device *gbdev = pcie_dev->gbdev;

	gpu_write(gbdev, L2_PWRON_LO,
		gbdev->features.l2_present & 0xffffffff);
	gpu_write(gbdev, L2_PWRON_HI,
		gbdev->features.l2_present & 0xffffffff);
}

void gb_gpu_L2_cache_power_off(void)
{
	struct gb_pcie_info *pcie_dev = get_pcie_gl_info();
	struct gb_device *gbdev = pcie_dev->gbdev;

	gpu_write(gbdev, L2_PWROFF_LO, 0xffffffff);
	gpu_write(gbdev, L2_PWROFF_HI, 0xffffffff);
}

void gb_gpu_power_on(struct gb_device *gbdev)
{
	int ret;
	u32 val;

	/* Just turn on everything for now */
	gpu_write(gbdev, L2_PWRON_LO,
		gbdev->features.l2_present & 0xffffffff);
	/* L2_READY_LO only support 1, no matter how many l2 are ready */
	ret = readl_relaxed_poll_timeout(gbdev->gpu_reg_base + L2_READY_LO,
		val, val == 1, 100, 1000);
		//val, val == gbdev->features.l2_present, 100, 1000);

	gpu_write(gbdev, STACK_PWRON_LO, gbdev->features.stack_present);
	ret |= readl_relaxed_poll_timeout(gbdev->gpu_reg_base + STACK_READY_LO,
		val, val == gbdev->features.stack_present, 100, 1000);

	gpu_write(gbdev, SHADER_PWRON_LO, gbdev->features.shader_present);
	ret |= readl_relaxed_poll_timeout(gbdev->gpu_reg_base + SHADER_READY_LO,
		val, val == gbdev->features.shader_present, 100, 1000);

	gpu_write(gbdev, TILER_PWRON_LO, gbdev->features.tiler_present);
	ret |= readl_relaxed_poll_timeout(gbdev->gpu_reg_base + TILER_READY_LO,
		val, val == gbdev->features.tiler_present, 100, 1000);

	if (ret)
		dev_err(gbdev->dev, "error powering up gpu");
}

void gb_gpu_power_off(struct gb_device *gbdev)
{
	gpu_write(gbdev, TILER_PWROFF_LO, 0xffffffff);
	gpu_write(gbdev, SHADER_PWROFF_LO, 0xffffffff);
	gpu_write(gbdev, STACK_PWROFF_LO, 0xffffffff);
	gpu_write(gbdev, L2_PWROFF_LO, 0xffffffff);
}

int gb_gpu_init(struct gb_device *gbdev)
{
	int err;

	err = gb_gpu_soft_reset(gbdev);
	if (err)
		return err;

	gb_gpu_init_features(gbdev);

	dma_set_mask_and_coherent(gbdev->dev,
		DMA_BIT_MASK(FIELD_GET(0xff00, gbdev->features.mmu_features)));

	gb_gpu_init_quirks(gbdev);
	gb_gpu_power_on(gbdev);

	return 0;
}

void gb_gpu_release(struct gb_device *gbdev)
{
	gb_gpu_power_off(gbdev);
}

u32 gb_gpu_get_latest_flush_id(struct gb_device *gbdev)
{
	return gpu_read(gbdev, LATEST_FLUSH);
}
