/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/atomic.h>
#include <linux/version.h>
#include <linux/bitfield.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/iopoll.h>
#if KERNEL_VERSION(5, 3, 0) < LINUX_VERSION_CODE
#include <linux/io-pgtable.h>
#endif
#include <linux/iommu.h>
#include <linux/pci.h>
#include <linux/pm_runtime.h>
#include <linux/shmem_fs.h>
#include <linux/sizes.h>

#include "gb_device.h"
#include "gb_mmu.h"
#include "gb_gem.h"
#include "gb_regs.h"
#include "gb_stage.h"
#include "gb_pages_alloc.h"
#include "gb_gpu_irq.h"
#include "common/gb_common.h"

static int wait_ready(struct gb_device *gbdev, u32 as_nr)
{
	int ret;
	u32 val;

	ret = readl_relaxed_poll_timeout_atomic(gbdev->gpu_reg_base +
	 AS_STATUS(as_nr), val, !(val & AS_STATUS_AS_ACTIVE), 10, 10000);

	if (ret)
		dev_err(gbdev->dev, "%d AS_ACTIVE bit stuck\n", as_nr);

	return ret;
}

static int write_cmd(struct gb_device *gbdev, u32 as_nr, u32 cmd)
{
	int status;

	status = wait_ready(gbdev, as_nr);
	if (!status)
		mmu_write(gbdev, AS_COMMAND(as_nr), cmd);
	else {
		gb_printf(KERN_ERR, "ERROR.......\n");
	}

	return status;
}

static void lock_region(struct gb_device *gbdev, u32 as_nr,
			u64 iova, size_t size)
{
	u8 region_width;
	u64 region = iova & GPU_PAGE_MASK;

	/* The size is encoded as ceil(log2) minus(1), which may be calculated
	 * with fls. The size must be clamped to hardware bounds.
	 */
	size = max_t(u64, size, GPU_PAGE_SIZE);
	region_width = fls64(size - 1) - 1;
	region |= region_width;

	mmu_write(gbdev, AS_LOCKADDR_LO(as_nr), region & 0xFFFFFFFFUL);
	mmu_write(gbdev, AS_LOCKADDR_HI(as_nr), (region >> 32) & 0xFFFFFFFFUL);
	write_cmd(gbdev, as_nr, AS_COMMAND_LOCK);
}


static int mmu_hw_do_operation_locked(struct gb_device *gbdev, int as_nr,
				      u64 iova, size_t size, u32 op)
{
	if (as_nr < 0) {
		// dev_WARN(gbdev->dev, "as_nr is a invalid value\n");
		return 0;
	}

	if (op != AS_COMMAND_UNLOCK)
		lock_region(gbdev, as_nr, iova, size);

	/* Run the MMU operation */
	write_cmd(gbdev, as_nr, op);

	/* Wait for the flush to complete */
	return wait_ready(gbdev, as_nr);
}

/* reference to gb_mmu_hw_do_operation */
int mmu_hw_do_operation(struct gb_device *gbdev, int as_nr,
			       u64 iova, size_t size, u32 op)
{
	int ret;
	unsigned long flags;

	spin_lock_irqsave(&gbdev->mmu_hw_lock, flags);
	ret = mmu_hw_do_operation_locked(gbdev, as_nr, iova, size, op);
	spin_unlock_irqrestore(&gbdev->mmu_hw_lock, flags);

	return ret;
}

int mmu_hw_do_operation_all(struct gb_device *gbdev,
			       u64 iova, size_t size, u32 op)
{
	int i, ret;

	for (i = 0; i < NUM_THREAD_AS; i++)
		ret = mmu_hw_do_operation(gbdev, i, iova, size, op);

	return ret;
}

#ifdef MMU_ENABLED
static u64 gb_set_mmu_memattr(void)
{
	u64 memattr = 0;

	memattr = (AS_MEMATTR_IMPL_DEF_CACHE_POLICY <<
		  (AS_MEMATTR_INDEX_IMPL_DEF_CACHE_POLICY * 8)) |
		  (AS_MEMATTR_FORCE_TO_CACHE_ALL    <<
		  (AS_MEMATTR_INDEX_FORCE_TO_CACHE_ALL * 8)) |
		  (AS_MEMATTR_WRITE_ALLOC           <<
		  (AS_MEMATTR_INDEX_WRITE_ALLOC * 8)) |
		  (AS_MEMATTR_AARCH64_OUTER_IMPL_DEF   <<
		  (AS_MEMATTR_INDEX_OUTER_IMPL_DEF * 8)) |
		  (AS_MEMATTR_AARCH64_OUTER_WA  << (AS_MEMATTR_INDEX_OUTER_WA * 8));

	return memattr;
}
#endif

static u64 gb_set_mmu_transcfg(void)
{
	u64 transcfg;

#ifdef MMU_ENABLED
	transcfg = AS_TRANSCFG_ADRMODE_AARCH64_4K;
#else
	transcfg = AS_TRANSCFG_ADRMODE_IDENTITY;
#endif

	return transcfg;
}

static void gb_mmu_irq_enable(struct gb_device *gbdev, int as_num)
{
	u32 irq_mask;
	unsigned long flags;

	spin_lock_irqsave(&gbdev->hw_irq_lock, flags);
	irq_mask = mmu_read(gbdev, MMU_IRQ_MASK);
	mmu_write(gbdev, MMU_IRQ_MASK, irq_mask | BIT(as_num) | BIT(as_num + 16));
	spin_unlock_irqrestore(&gbdev->hw_irq_lock, flags);
	gb_printf(KERN_INFO, "%s: enable %d MMU irq\n", __func__, as_num);
}

static void gb_mmu_irq_disable(struct gb_device *gbdev, int as_num)
{
	u32 irq_mask;
	unsigned long flags;

	spin_lock_irqsave(&gbdev->hw_irq_lock, flags);
	irq_mask = mmu_read(gbdev, MMU_IRQ_MASK);
	mmu_write(gbdev, MMU_IRQ_MASK, irq_mask & ~(BIT(as_num) | BIT(as_num + 16)));
	spin_unlock_irqrestore(&gbdev->hw_irq_lock, flags);
	gb_printf(KERN_INFO, "%s: disable %d MMU irq\n", __func__, as_num);
}

void gb_mmu_enable(struct gb_device *gbdev, int as_nr)
{
	u64 transcfg = gb_set_mmu_transcfg();
#ifdef MMU_ENABLED
	u64 transtab, memattr;
#endif

#ifdef MMU_ENABLED
	transtab = (u64)gbdev->mmu_mode->pgd & AS_TRANSTAB_BASE_MASK;
	memattr = gb_set_mmu_memattr();
	mmu_write(gbdev, AS_TRANSTAB_LO(as_nr), transtab & 0xffffffffUL);
	mmu_write(gbdev, AS_TRANSTAB_HI(as_nr), transtab >> 32);

	mmu_write(gbdev, AS_MEMATTR_LO(as_nr), memattr & 0xffffffffUL);
	mmu_write(gbdev, AS_MEMATTR_HI(as_nr), memattr >> 32);
#endif

	mmu_write(gbdev, AS_TRANSCFG_LO(as_nr), transcfg & 0xffffffffUL);
	mmu_write(gbdev, AS_TRANSCFG_HI(as_nr), transcfg >> 32);

	gb_mmu_irq_enable(gbdev, as_nr);

	write_cmd(gbdev, as_nr, AS_COMMAND_UPDATE);
}

static void gb_mmu_disable(struct gb_device *gbdev, u32 as_nr)
{
	gb_mmu_irq_disable(gbdev, as_nr);
	mmu_hw_do_operation(gbdev, as_nr, 0, ~0UL, AS_COMMAND_FLUSH_MEM);

	mmu_write(gbdev, AS_TRANSTAB_LO(as_nr), 0);
	mmu_write(gbdev, AS_TRANSTAB_HI(as_nr), 0);

	mmu_write(gbdev, AS_MEMATTR_LO(as_nr), 0);
	mmu_write(gbdev, AS_MEMATTR_HI(as_nr), 0);

	write_cmd(gbdev, as_nr, AS_COMMAND_UPDATE);
}

void gb_mmu_controll_all(struct gb_device *gbdev, int enabled)
{
	int as;

	for (as = 0; as < NUM_THREAD_AS; as++) {
		if (enabled)
			gb_mmu_enable(gbdev, as);
		else
			gb_mmu_disable(gbdev, as);
	}
}

int gb_mmu_as_get(void)
{
	static int as_nr = 0;
	as_nr++;
	if (as_nr >= NUM_THREAD_AS)
		as_nr = 0;
	return as_nr;
}

void gb_mmu_reset(struct gb_device *gbdev)
{
	mmu_write(gbdev, MMU_IRQ_CLEAR, ~0);
}

void gb_mmu_flush_range(struct gb_device *gbdev, int as,
				     u64 iova, size_t size, bool sync)
{
	u32 op = 0;

	pm_runtime_get_noresume(gbdev->dev);

	/* Flush the PTs only if we're already awake */
	if (pm_runtime_active(gbdev->dev)) {
	}
	if (sync)
		op = AS_COMMAND_FLUSH_MEM;
	else
		op = AS_COMMAND_FLUSH_PT;
	mmu_hw_do_operation(gbdev, as, iova, size, op);

	pm_runtime_put_sync_autosuspend(gbdev->dev);
}

int gb_mmu_pgd_init(struct gb_device *gbdev)
{
	struct gb_page *p_pgd;
	struct gb_mmu_mode *mmu_mode = gbdev->mmu_mode;
	u64 *p_vaddr;
	int i;

	mmu_mode->mmu_teardown_pages = kmalloc(PAGE_SIZE * 4, GFP_KERNEL);
	if (NULL == mmu_mode->mmu_teardown_pages)
		return -ENOMEM;

	mutex_lock(&mmu_mode->mmu_lock);
	p_pgd = gb_alloc_mmu_page(gbdev);
	if (!p_pgd) {
		mutex_unlock(&mmu_mode->mmu_lock);
		goto alloc_free;
	}
	mutex_unlock(&mmu_mode->mmu_lock);

	p_vaddr = gb_kmap(p_pgd);
	if (NULL == p_vaddr) {
		goto alloc_free;
	}

	for (i = 0; i < GB_MMU_PAGE_ENTRIES; i++) {
		mmu_mode->entry_invalidate(&p_vaddr[i]);
	}

	mmu_mode->pgd = gb_page_to_phys(p_pgd);
	//gb_printf(KERN_ERR, "%s alloc pgd 0x%llx\n", __func__, mmu_mode->pgd);

	return 0;

alloc_free:
	gb_err(gbdev->dev, "%s alloc failed\n", __func__);
	if (p_pgd)
		gb_free_mmu_page(p_pgd, gbdev);

	kfree(mmu_mode->mmu_teardown_pages);

	return -ENOMEM;
}

void gb_mmu_pgd_release(struct gb_device *gbdev)
{
	struct gb_mmu_mode *mmu_mode;
	mmu_mode = gbdev->mmu_mode;

	pm_runtime_get_noresume(gbdev->dev);

	if (pm_runtime_active(gbdev->dev))
		mmu_mode->mmu_controll(gbdev, MMU_DISABLE);

	pm_runtime_put_autosuspend(gbdev->dev);
	gb_mmu_free_pg_table(gbdev);
	gb_printf(KERN_INFO, "%s x used_mmu_pages: %d\n", __func__, atomic_read(&gbdev->gb_page_ctrl.used_mmu_pages));
}

struct gb_gem_object *get_bo_by_addr(struct gb_device *gbdev, u64 addr)
{
	int i;
	u64 offset;
	struct gb_gem_object *bo = NULL;
	struct gb_mmu_mode *mmu_mode;
	struct drm_mm_node *node;
	mmu_mode = gbdev->mmu_mode;
	offset = addr >> PAGE_SHIFT;

	mutex_lock(&mmu_mode->mm_lock);
	for (i = 0; i < GB_MAX_VA_MM; i++) {
		drm_mm_for_each_node(node, &mmu_mode->mm[i]) {
			if (offset >= node->start &&
		    	offset < (node->start + node->size)) {
				bo = drm_mm_node_to_gb_bo(node);
				mutex_unlock(&mmu_mode->mm_lock);
				goto out;
			}
		}
	}
	mutex_unlock(&mmu_mode->mm_lock);
out:
	return bo;
}

static int gb_mmu_fault_addr(struct gb_device *gbdev, int as, u64 addr)
{
	int i, ret = 0;
	unsigned long flags;
	struct gb_gem_object *bo;
	struct gb_file_priv *gb_priv;

	bo = get_bo_by_addr(gbdev, addr);
	if (!bo) {
		gb_printf(KERN_ERR, "%s: APP in %d AS can't found bo by addr 0x%llx\n",
			__func__, as, addr);
		ret = -ENOENT;
		goto fail;
	}

	if (!bo->is_heap_growable) {
		dev_WARN(gbdev->dev, "matching BO 0x%llx is not heap type (GPU VA = %llx)",
				(long long )bo, bo->base.node.start << PAGE_SHIFT);
		gb_printf(KERN_ERR, "fault addr: 0x%llx bo start: 0x%llx nr_pages: 0x%x bo end: 0x%llx\n",
			  addr, bo->base.node.start<<PAGE_SHIFT, bo->base.nr_pages,
			  (bo->base.node.start << PAGE_SHIFT) + bo->base.nr_pages * PAGE_SIZE);
		ret = -EINVAL;
		goto fail;
	}
#if 0
	list_for_each_entry(gb_heap_bo, &bo->gb_base.ttm_bo_root, ttm_bo_list) {
		if (addr >= gb_heap_bo->gb_va_start && addr <
			gb_heap_bo->gb_va_start + gb_heap_bo->grow_nr_pages * GPU_PAGE_SIZE) {
			printk("[%s] found heapbo already mappped addr 0x%llx as:%d\n", __func__, addr, as);
			return ret;
		}
	}
#endif

	gb_priv = bo->private_data;
	ret = gb_ttm_create_heap_bo(bo, gbdev, gb_priv, addr, 1, as);
	return ret;

fail:
	for (i = 0; i < NUM_JOB_SLOTS; i++) {
		spin_lock_irqsave(&gbdev->ss->stages_lock, flags);
		if (gbdev->stages[i] && (gbdev->stages[i]->as == as)) {
			/* TODO for fault job*/
			gbdev->stages[i]->reset_time = STAGE_RESET_TIME_MAX + 1;
			dma_fence_signal_locked(gbdev->stages[i]->irq_done_fence);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 4, 131)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0))
#if (defined(GB_KYLIN_SERV_TERCEL)) && (LINUX_VERSION_CODE == KERNEL_VERSION(4, 19, 90))
			mod_delayed_work(system_wq,
					&gbdev->ss->queue[i].sched.work_tdr, 0);
#else
			mod_delayed_work(system_wq,
					&gbdev->stages[i]->base.work_tdr, 0);
#endif
#else
			mod_delayed_work(system_wq,
					&gbdev->ss->queue[i].sched.work_tdr, 0);
#endif
			gbdev->stages[i] = NULL;
		}
		spin_unlock_irqrestore(&gbdev->ss->stages_lock, flags);
	}

	return ret;
}

static const char *access_type_name(struct gb_device *gbdev,
		u32 fault_status)
{
	switch (fault_status & AS_FAULTSTATUS_ACCESS_TYPE_MASK) {
	case AS_FAULTSTATUS_ACCESS_TYPE_ATOMIC:
		return "ATOMIC";
	case AS_FAULTSTATUS_ACCESS_TYPE_READ:
		return "READ";
	case AS_FAULTSTATUS_ACCESS_TYPE_WRITE:
		return "WRITE";
	case AS_FAULTSTATUS_ACCESS_TYPE_EX:
		return "EXECUTE";
	default:
		WARN_ON(1);
		return NULL;
	}
}

irqreturn_t gb_mmu_irq_handler(int irq, void *data)
{
	struct gb_device *gbdev = data;

	if (!reg_read(gbdev, MMU_IRQ_STATUS))
		return IRQ_NONE;

	mmu_write(gbdev, MMU_IRQ_MASK, 0);
	return IRQ_WAKE_THREAD;
}

irqreturn_t gb_mmu_irq_handler_thread(int irq, void *data)
{
	struct gb_device *gbdev = data;
	u32 status = reg_read(gbdev, MMU_IRQ_RAWSTAT);
	int i, ret;

	/*TODO save IRQ_MASK value for restore instead of ALL slot*/
	while (status) {
		for (i = 0; status && i < NUM_THREAD_AS; i++) {
			u32 mask = BIT(i) | BIT(i + 16);
				u64 addr;
			u32 fault_status;
			u32 exception_type;
			u32 access_type;
			u32 source_id;

			if (!(status & mask))
				continue;

			fault_status = mmu_read(gbdev, AS_FAULTSTATUS(i));
			addr = mmu_read(gbdev, AS_FAULTADDRESS_LO(i));
			addr |= (u64)mmu_read(gbdev, AS_FAULTADDRESS_HI(i)) << 32;
			gb_printf(KERN_INFO, "IRQ mmu fault_status: 0x%x, addr: 0x%llx status 0x%x\n", fault_status, addr, status);

			exception_type = fault_status & 0xFF;
			access_type = (fault_status >> 8) & 0x3;
			source_id = (fault_status >> 16);

		/* Page fault only */
			if ((status & mask) == BIT(i)) {
				WARN_ON(exception_type < 0xC1 || exception_type > 0xC4);
				ret = gb_mmu_fault_addr(gbdev, i, addr);
				if (!ret) {
					status &= ~mask;
					continue;
				}
			}

			dev_err(gbdev->dev,
				"Unhandled Page fault in AS%d at VA 0x%016llX\n"
				"raw fault status: 0x%X\n"
				"decoded fault status: %s\n"
				"exception type 0x%X: %s\n"
				"access type 0x%X: %s\n"
				"source id 0x%X\n",
				i, addr,
				fault_status,
				(fault_status & (1 << 10) ? "DECODER FAULT" : "SLAVE FAULT"),
				exception_type, gb_exception_name(gbdev, exception_type),
				access_type, access_type_name(gbdev, fault_status),
				source_id);

			mmu_write(gbdev, MMU_IRQ_CLEAR, mask);
			status &= ~mask;
		}
		status = reg_read(gbdev, MMU_IRQ_RAWSTAT);
	}
	mmu_write(gbdev, MMU_IRQ_CLEAR, (0xFF << 16 || 0xFF));
	mmu_write(gbdev, MMU_IRQ_MASK, ~0);
	return IRQ_HANDLED;
};

void gb_mmu_free_pgd(struct gb_device *gbdev)
{
	struct gb_page *p;
	struct gb_mmu_mode *mmu_mode = gbdev->mmu_mode;

	if (mmu_mode->mmu_teardown_pages) {
		kfree(mmu_mode->mmu_teardown_pages);
		mmu_mode->mmu_teardown_pages = NULL;
	}

	p = gb_phys_to_page(gbdev, mmu_mode->pgd);

	if (!p) {
		gb_err(gbdev->dev, "Error: %s gb_page is NULL\n", __func__);
		return;
	}
	gb_free_mmu_page(p, gbdev);

}

void gb_mmu_release(struct gb_device *gbdev)
{
	unsigned long flags;
	gb_mmu_pgd_release(gbdev);
	gb_mmu_free_pgd(gbdev);

	spin_lock_irqsave(&gbdev->hw_irq_lock, flags);
	mmu_write(gbdev, MMU_IRQ_MASK, 0);
	spin_unlock_irqrestore(&gbdev->hw_irq_lock, flags);

	drm_mm_takedown(&gbdev->mmu_mode->mm[GB_LOW_VA_MM]);
	drm_mm_takedown(&gbdev->mmu_mode->mm[GB_HIGH_VA_MM]);
}

int gb_mmu_init(struct gb_device *gbdev)
{
	int ret;
	gbdev->mmu_mode = gb_mmu_mode_get_64bit();
	gbdev->mmu_mode->gbdev = gbdev;
	mutex_init(&gbdev->mmu_mode->mm_lock);
	mutex_init(&gbdev->mmu_mode->mmu_lock);

	drm_mm_init(&gbdev->mmu_mode->mm[GB_LOW_VA_MM], GPU_VA_START >> PAGE_SHIFT,
		    GPU_VA_LOW_MAX_SIZE >> PAGE_SHIFT);
	drm_mm_init(&gbdev->mmu_mode->mm[GB_HIGH_VA_MM], GPU_VA_HIGH_START >> PAGE_SHIFT,
	            GPU_VA_MAX_SIZE >> PAGE_SHIFT);
	//pr_err("[%s] start init\n", __func__);

	ret = gb_mmu_pgd_init(gbdev);
	if (ret) {
		goto pgd_init_fail;
	}

	gbdev->mmu_mode->mmu_controll(gbdev, MMU_AS_ENABLED);
	return ret;

pgd_init_fail:
	drm_mm_takedown(&gbdev->mmu_mode->mm[GB_LOW_VA_MM]);
	drm_mm_takedown(&gbdev->mmu_mode->mm[GB_HIGH_VA_MM]);
	return ret;
}
