/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/delay.h>
#include <generated/uapi/linux/version.h>
#include "gb_device.h"
#include "gb_mmu.h"

enum gb_share_attr_bits {
        /* (1ULL << 8) bit is reserved */
        SHARE_BOTH_BITS = (2ULL << 8),  /* inner and outer shareable coherency */
        SHARE_INNER_BITS = (3ULL << 8)  /* inner shareable coherency */
};


static inline void page_table_entry_set(u64 *pte, u64 phy)
{
	*pte = phy;
}

static void mmu_disable_as(struct gb_device *gbdev, int as_nr)
{
}

static void mmu_enable_as(struct gb_device *gbdev, int as_nr)
{
}

static phys_addr_t pte_to_phy_addr(u64 entry)
{
	if (!(entry & 1))
		return 0;

	return entry & ~(GPU_PAGE_SIZE - 1);
}

static int ate_is_valid(u64 ate)
{
	return ((ate & ENTRY_TYPE_MASK) == ENTRY_IS_ATE);
}

static int pte_is_valid(u64 pte)
{
	return ((pte & ENTRY_TYPE_MASK) == ENTRY_IS_PTE);
}

/*
 * Map GB_REG flags to MMU flags
 */
static u64 get_mmu_flags(unsigned long flags)
{
	u64 mmu_flags = 0;

	// fixme: currently the flags parameter is always reset
	flags = GB_REG_GPU_WR | GB_REG_SHARE_IN;

	/* store mem_attr index as 4:2 (macro called ensures 3 bits already) */
	mmu_flags = GB_REG_MEMATTR_VALUE(flags) << 2;

	/* Set access flags - note that AArch64 stage 1 does not support
	 * write-only access, so we use read/write instead
	 */
	if (flags & GB_REG_GPU_WR)
		mmu_flags |= ENTRY_ACCESS_RW;
	else if (flags & GB_REG_GPU_RD)
		mmu_flags |= ENTRY_ACCESS_RO;

	/* nx if requested */
	mmu_flags |= (flags & GB_REG_GPU_NX) ? ENTRY_NX_BIT : 0;

	if (flags & GB_REG_SHARE_BOTH) {
		/* inner and outer shareable */
		mmu_flags |= SHARE_BOTH_BITS;
	} else if (flags & GB_REG_SHARE_IN) {
		/* inner shareable coherency */
		mmu_flags |= SHARE_INNER_BITS;
	}

	return mmu_flags;
}

static void entry_set_ate(u64 *entry, phys_addr_t phy, unsigned long flags)
{
	u64 mmu_flags = get_mmu_flags(flags);
	page_table_entry_set(entry, (phy & ~0xFFF) | mmu_flags | ENTRY_ACCESS_BIT | ENTRY_IS_ATE);
}

static void entry_set_pte(u64 *entry, phys_addr_t phy)
{
	page_table_entry_set(entry, (phy & ~0xFFF) |
			ENTRY_ACCESS_BIT | ENTRY_IS_PTE);
}

static void entry_invalidate(u64 *entry)
{
	page_table_entry_set(entry, ENTRY_IS_INVAL);
}

static struct gb_mmu_mode mmu64_mode = {
	.mmu_controll = gb_mmu_controll_all,
	.disable_as = mmu_disable_as,
	.enable_as = mmu_enable_as,
	.pte_to_phy_addr = pte_to_phy_addr,
	.ate_is_valid = ate_is_valid,
	.pte_is_valid = pte_is_valid,
	.entry_set_ate = entry_set_ate,
	.entry_set_pte = entry_set_pte,
	.entry_invalidate = entry_invalidate
};

struct gb_mmu_mode *gb_mmu_mode_get_64bit(void)
{
	return &mmu64_mode;
}
