/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/dma-mapping.h>
#include <linux/highmem.h>
#include <linux/spinlock.h>
#include <linux/atomic.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5, 12, 0))
#include <linux/timekeeping32.h>
#else
#include <linux/timekeeping.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
#include <linux/vmalloc.h>
#endif
#include "gb_mmu.h"
#include "gb_gem.h"
#include "gb_device.h"
#include "gb_pages_alloc.h"
#include "common/gb_common.h"
#ifdef GB_DMA_ON
#include "ip/gb_edma.h"
#endif

#define NOT_DIRTY false
#define NOT_RECLAIMED false

static size_t find_u64_zeros(u64 word,
	size_t num_remain, u32 *bit_record, u64 total, size_t nr_pages)
{
	int i;

	if (word == ~0UL)
		return num_remain;

	for (i = 0; i < BITS_PER_LONG; i++) {
		if (!word & 1ULL) {
			bit_record[nr_pages - num_remain] = total + i;
			num_remain--;
			if (!num_remain)
				return 0;
		}
		word >>= 1UL;
	}

	return num_remain;
}

/*
 * find size of size_bit_record zero bits from 0~max_cnt
 * bits in usage and record bits nums in bit_record
 */
static int gb_find_zero_bits(volatile u64 *usage,
	u32 max_cnt, size_t size_bit_record, u32 *bit_record)
{
	u64 total;
	size_t num_remain = size_bit_record;
	size_t num_tmp;
	u64 usage_tmp;

	for (total = 0; total < max_cnt; total += 64, ++usage) {
		num_tmp = num_remain;
		usage_tmp = *usage;
		num_remain = find_u64_zeros(usage_tmp, num_tmp, bit_record, total, size_bit_record);
		if (!num_remain)
			return 0;
	}

	return -ENOMEM;
}

static u32 gb_find_highest_set(volatile u64 *usage, u32 max_cnt)
{
	int i;
	for (i = (max_cnt + 63) / 64 - 1; i >= 0; i--) {
		if (usage[i] != 0) {
			return (i * 64 + fls64(usage[i]));
		}
	}

	return 0;
}

static void gb_set_page_usage_bit(u32 bits, volatile u64 *usage)
{
	u32 idx = bits / BITS_PER_LONG;
	u64 bit = bits % BITS_PER_LONG;
	usage[idx] |= (1UL << bit);
//	gb_verb_printf(KERN_INFO "idx: %d bit: %d usage[0]: 0x%lx, BITS_PER_LONG: %d\n", idx, bit, usage[0], BITS_PER_LONG);
}

static void gb_clear_page_usage_bit(int bits, volatile u64 *usage)
{
	int idx = bits / BITS_PER_LONG;
	u64 bit = bits % BITS_PER_LONG;
//	gb_verb_printf("bits: %d idx: %d bit: %d, usage[idx]: %lx\n", bits, idx, bit, usage[idx]);
	u64 tmp = usage[idx];
	usage[idx] &= (~(1UL << bit));
	if (tmp == usage[idx]) {
	//	gb_verb_printf(KERN_INFO "BBBTTTSSS invalid setting!\n");
	}
//	gb_verb_printf("idx: %d bit: %d, usage[idx]: %lx\n", idx, bit, usage[idx]);
}

// GPU device physical address
phys_addr_t gb_page_to_phys(struct gb_page *p)
{
	if (p) {
		return p->gpu_phy;
	} else {
		gb_err(NULL, "Error: %s gb_page is NULL\n", __func__);
	}

	return 0;
}

struct gb_page *gb_phys_to_page(struct gb_device *gbdev, phys_addr_t phys)
{
	int pfn;

	if ((phys < VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info)) ||
		(phys >=
		 VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info)) ||
		(phys < gbdev->gb_page_ctrl.mmu_lowest_addr &&
		 phys >= gbdev->gb_page_ctrl.page_highest_addr)) {
		gb_printf(KERN_INFO, "AAAAA+++++++ After gb_phys_to_page\n");
		gb_err(gbdev->dev, "Error: %s %d bad mmu phys: 0x%llx mmu_low:0x%llx pg_high:0x%llx\n, normal %llx MMU_TOP %llx ",
			   __func__, __LINE__, phys,
			   gbdev->gb_page_ctrl.mmu_lowest_addr,
			   gbdev->gb_page_ctrl.page_highest_addr,
			   VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info),
			   VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info));
		return NULL;
	}

	/* for mmu pages */
	if (phys >= gbdev->gb_page_ctrl.mmu_lowest_addr) {
		pfn =
		 (VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info)
		  - 1 - phys) >> GPU_PAGE_SHIFT;
		return gbdev->mmu_top_page + pfn;
	}

	/* for normal pages */
	pfn =
	 (phys - VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info)) >>
	  PAGE_SHIFT;
	return gbdev->global_gb_page + pfn;
}

u64 *gb_kmap(struct gb_page *p)
{
	struct gb_asic_dev_info *gb_dev_info = p->gbdev->gb_pcie->gb_dev_info;
	int ddr_bar_id = get_genbu_ddr_bar_id(gb_dev_info);
	u64 *ret_p = NULL;

	if (p == NULL) {
		gb_err(NULL, "ERROR: %s p is NULL\n", __func__);
		return NULL;
	}
	if (p->pfn == -1 || p->gpu_phy == -1) {
		gb_err(NULL, "Error: %s invalid pfn.", __func__);
		return NULL;
	}

	if (p->gpu_phy >= p->gbdev->gb_page_ctrl.mmu_lowest_addr) {
		// stand for mmu page
		ret_p =
		 (u64 *)((void *)p->gbdev->gb_pcie->pci_bars[ddr_bar_id].mmio +
		  VRAM_FOR_MMU_PAGE_TOP_BASE(gb_dev_info) -
		   (((p->pfn)+1)<<GPU_PAGE_SHIFT));

		return ret_p;
	} else {
		ret_p =
		 (u64 *)((void *)p->gbdev->gb_pcie->pci_bars[ddr_bar_id].mmio +
		  VRAM_FOR_NORMAL_PG_BASE(gb_dev_info) +
		   ((p->pfn)<<PAGE_SHIFT));
		return ret_p;
	}
}

/*
 * Set info for struct gb_page
 * nr_pages: number of pages in gpu pg size
 * bit_record: pfn bit record in gb_find_first_zero_bit
 */
static void gb_mem_set_gb_page(struct gb_device *gbdev,
	size_t nr_pages, phys_addr_t *pages, u32 *bit_record)
{
	struct gb_page *gb_page = NULL;
	u32 gpu_pfn;
	size_t i;
	int k;

	spin_lock(&gbdev->gb_page_ctrl.ddr_lock);

	/*
	 * nr_pages is number of page in gpu pg size, but gb_page
	 * is for page in cpu page size, so use PG_SIZE_MULTI to
	 * control loop
	 */
	for (i = 0; i < nr_pages; i += PG_SIZE_MULTI) {
		gpu_pfn = bit_record[i/PG_SIZE_MULTI];

		gb_set_page_usage_bit(gpu_pfn, gbdev->gb_page_ctrl.usage);
		atomic_add_return(1, &(gbdev->gb_page_ctrl.used_pages));

		smp_mb();

		gb_page = gbdev->global_gb_page + gpu_pfn;
		gb_page->pfn = gpu_pfn;
		gb_page->gbdev = gbdev;
		/* one page 16K on MIPS */
		gb_page->gpu_phy =
		 VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info) +
		  ((gb_page->pfn) << PAGE_SHIFT);
		for (k = 0; k < PG_SIZE_MULTI; k++){
			pages[i+k] = gb_page->gpu_phy + k * GPU_PAGE_SIZE;
		}
	}

	/*
	 * refresh page_highest_addr, the last pfn in bit_record
	 * must be the highest address in pfns of bit_record, so
	 * use it directly to compare with old page_highest_addr
	 */
	if (gb_page->gpu_phy + PAGE_SIZE > gbdev->gb_page_ctrl.page_highest_addr)
		gbdev->gb_page_ctrl.page_highest_addr = gb_page->gpu_phy + PAGE_SIZE;

	spin_unlock(&gbdev->gb_page_ctrl.ddr_lock);
}

int gb_alloc_pages(struct gb_device *gbdev, size_t nr_pages, phys_addr_t *pages)
{
	u64 search_max_pfn;
	u64 vram_end_pfn;
	u64 vram_end_addr;
	u32 *bit_record = NULL;
	int vmalloc_flag = false;
	int ret = 0;
	size_t nr_pg_in_cpu_pgsz = (nr_pages + PG_SIZE_MULTI - 1) / PG_SIZE_MULTI;

	if (nr_pages == 0) {
		gb_printf(KERN_DEBUG, "%s alloc 0 page, return", __func__);
		return 0;
	}

	/* record allocated pfns in gb_find_first_zero_bit */
	bit_record = (u32 *)kmalloc(nr_pg_in_cpu_pgsz * sizeof(u32), GFP_KERNEL);
	if (unlikely(!bit_record)) {
		bit_record = (u32 *)vmalloc(nr_pg_in_cpu_pgsz * sizeof(u32));
		if (!bit_record) {
			gb_err(gbdev->dev, "ERROR: %s %d no enough cpu ddr space:alloc size: 0x%zx\n",
				__func__, __LINE__, nr_pg_in_cpu_pgsz * sizeof(u32));
			return -ENOMEM;
		}

		vmalloc_flag = true;
	}

	mutex_lock(&gbdev->vram_mutex);
	/*
	 * pfn for vram pg must be less than lowest pfn of mmu pg,
	 * they can't be overlaped
	 */
	search_max_pfn = (gbdev->gb_page_ctrl.mmu_lowest_addr -
	 VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info)) / PAGE_SIZE;
	ret = gb_find_zero_bits(gbdev->gb_page_ctrl.usage,
		search_max_pfn, nr_pg_in_cpu_pgsz, bit_record);
	if (ret) {
		gb_printf(KERN_ERR, "ERROR in %s mmu_lowest_addr:0x%llx\n",
		 __func__, gbdev->gb_page_ctrl.mmu_lowest_addr);
		gb_printf(KERN_ERR, "ERROR pg_highest_addr:0x%llx, 0x%llx, 0x%llx\n",
		 gbdev->gb_page_ctrl.page_highest_addr,
		  VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info) -
		   ((gb_find_highest_set(gbdev->gb_page_ctrl.mmu_usage,
		    MMU_PG_USG_BITS_LEN(gbdev->gb_pcie->gb_dev_info) * 64))
			 << GPU_PAGE_SHIFT),
			VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info) +
			 ((gb_find_highest_set(gbdev->gb_page_ctrl.usage,
			  VRAM_PG_USG_BITS_LEN(gbdev->gb_pcie->gb_dev_info)
			   * 64)) << PAGE_SHIFT));
		goto out;
	}

	vram_end_pfn = bit_record[nr_pages - 1];
	vram_end_addr = VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info) +
	 (vram_end_pfn << PAGE_SHIFT);
	if (vram_end_addr >= gbdev->gb_page_ctrl.mmu_lowest_addr) {
		gb_printf(KERN_ERR, "%s mmu_lowest_addr:0x%llx vram_end_addr:0x%llx \n", __func__,
			gbdev->gb_page_ctrl.mmu_lowest_addr, vram_end_addr);
		ret = -ENOMEM;
		goto out;
	}

	gb_mem_set_gb_page(gbdev, nr_pages, pages, bit_record);

out:
	mutex_unlock(&gbdev->vram_mutex);

	if (vmalloc_flag)
		vfree(bit_record);
	else
		kfree(bit_record);

	return ret;
}

void gb_free_pages(struct gb_page *gb_page, struct gb_device *gbdev)
{
	/*
	 * this pfn is the highest pfn for vram page plus 1
	 * for example if vram page only used pfn 0,
	 * then this pfn is 1
	 */
	u32 pfn;
	void *ptr;

	ptr = gb_kmap(gb_page);
	if (ptr) {
		// gb_printf(KERN_INFO, "%s page free gb_page->pfn: 0x%x\n",__func__, gb_page->pfn);
	} else {
		gb_err(gbdev->dev, "Error:%s gb_page is invalid!\n", __func__);

		return;
	}
	memset(ptr, 0, GPU_PAGE_SIZE);

	spin_lock(&gbdev->gb_page_ctrl.ddr_lock);
	gb_clear_page_usage_bit(gb_page->pfn, gbdev->gb_page_ctrl.usage);
	atomic_sub_return(1, &(gbdev->gb_page_ctrl.used_pages));

	/* if highest addr is freed, then refresh it */
	if (gb_page->gpu_phy + PAGE_SIZE == gbdev->gb_page_ctrl.page_highest_addr) {
		pfn = gb_find_highest_set(gbdev->gb_page_ctrl.usage, gb_page->pfn);

		gbdev->gb_page_ctrl.page_highest_addr =
			VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info) +
			 (pfn << PAGE_SHIFT);
	}
	spin_unlock(&gbdev->gb_page_ctrl.ddr_lock);

	memset(gb_page, 0, sizeof(struct gb_page));
	gb_page->pfn = -1;
	gb_page->gpu_phy = -1;
}

struct gb_page *gb_alloc_mmu_page(struct gb_device *gbdev)
{
	struct gb_page *gb_page;
	u32 gpu_pfn;
	u64 search_max_pfn;
	u64 mmu_end_addr;
	int ret;

	mutex_lock(&gbdev->vram_mutex);

	search_max_pfn =
	 (VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info) -
	  gbdev->gb_page_ctrl.page_highest_addr) / GPU_PAGE_SIZE;
	ret = gb_find_zero_bits(gbdev->gb_page_ctrl.mmu_usage,
		search_max_pfn, 1, &gpu_pfn);
	if (ret) {
		gb_printf(KERN_ERR, "ERROR in %s!!! mmu_lowest_addr:0x%llx, pg_highest_addr:0x%llx, 0x%llx, 0x%llx\n",
		 __func__, gbdev->gb_page_ctrl.mmu_lowest_addr,
		  gbdev->gb_page_ctrl.page_highest_addr,
		   VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info) -
			 ((gb_find_highest_set(gbdev->gb_page_ctrl.mmu_usage,
			  MMU_PG_USG_BITS_LEN(gbdev->gb_pcie->gb_dev_info) * 64))
			   << GPU_PAGE_SHIFT),
			VRAM_FOR_NORMAL_PG_BASE(gbdev->gb_pcie->gb_dev_info) +
			 ((gb_find_highest_set(gbdev->gb_page_ctrl.usage,
			  VRAM_PG_USG_BITS_LEN(gbdev->gb_pcie->gb_dev_info)
			   * 64)) << PAGE_SHIFT));
		mutex_unlock(&gbdev->vram_mutex);
		return NULL;
	}

	mmu_end_addr =
	 VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info) -
	  ((gpu_pfn + 1) << GPU_PAGE_SHIFT);
	if (mmu_end_addr < gbdev->gb_page_ctrl.page_highest_addr) {
		gb_printf(KERN_ERR, "%s mmu alloc err, page_highest_addr:0x%llx mmu_end_addr:0x%llx\n", __func__,
			gbdev->gb_page_ctrl.page_highest_addr, mmu_end_addr);
		mutex_unlock(&gbdev->vram_mutex);
		return NULL;
	}

	spin_lock(&gbdev->gb_page_ctrl.ddr_lock);

	gb_set_page_usage_bit(gpu_pfn, gbdev->gb_page_ctrl.mmu_usage);
	gb_page = gbdev->mmu_top_page + gpu_pfn;
	memset(gb_page, 0, sizeof(struct gb_page));
	atomic_add_return(1, &(gbdev->gb_page_ctrl.used_mmu_pages));

	smp_mb();

	gb_page->gpu_phy = mmu_end_addr;
	gb_page->pfn = gpu_pfn;
	gb_page->gbdev = gbdev;
	//pr_err("[%s] phy %llx low %llx\n", __func__, gb_page->gpu_phy, gbdev->gb_page_ctrl.mmu_lowest_addr);
	if (gb_page->gpu_phy < gbdev->gb_page_ctrl.mmu_lowest_addr)
		gbdev->gb_page_ctrl.mmu_lowest_addr = gb_page->gpu_phy;

	spin_unlock(&gbdev->gb_page_ctrl.ddr_lock);

	// gb_verb_printf(KERN_INFO "%s gpu_phy: 0x%llx pfn: %u\n", __func__, gb_page->gpu_phy, gb_page->pfn);
	mutex_unlock(&gbdev->vram_mutex);

	return gb_page;
}

void gb_free_mmu_page(struct gb_page *gb_page, struct gb_device *gbdev)
{
	void *ptr;
	u32 pfn;

	ptr = gb_kmap(gb_page);
	if (!ptr) {
		gb_err(gbdev->dev, "Error:%s gb_page is invalid!\n", __func__);

		return;
	}
	memset_io(ptr, 0, GPU_PAGE_SIZE);
	//dump_stack();
	spin_lock(&gbdev->gb_page_ctrl.ddr_lock);
	gb_clear_page_usage_bit(gb_page->pfn, gbdev->gb_page_ctrl.mmu_usage);
	atomic_sub_return(1, &(gbdev->gb_page_ctrl.used_mmu_pages));

	/* if lowest addr is freed, then refresh it */
	if (gb_page->gpu_phy == gbdev->gb_page_ctrl.mmu_lowest_addr) {
		pfn = gb_find_highest_set(gbdev->gb_page_ctrl.mmu_usage, gb_page->pfn);
		//pr_err("[%s] low %llx phy %llx\n", __func__, gbdev->gb_page_ctrl.mmu_lowest_addr, gb_page->gpu_phy);
		gbdev->gb_page_ctrl.mmu_lowest_addr =
		 VRAM_FOR_MMU_PAGE_TOP_BASE(gbdev->gb_pcie->gb_dev_info) -
			 (pfn << GPU_PAGE_SHIFT);
	}
	spin_unlock(&gbdev->gb_page_ctrl.ddr_lock);
	memset(gb_page, 0, sizeof(struct gb_page));
	gb_page->pfn = -1;
	gb_page->gpu_phy = -1;
}

int gb_vram_alloc_pages(struct gb_device *gbdev, size_t nr_pages, phys_addr_t *pages)
{
	return gb_alloc_pages(gbdev, nr_pages, pages);
}

void gb_vram_free_pages(struct gb_device *gbdev, size_t nr_pages, phys_addr_t *pages)
{
	struct gb_page *p;
	size_t i = 0;
	size_t k;

	/* Free any remaining pages to kernel */
	for (; i < nr_pages; i+=PG_SIZE_MULTI) {
		if (unlikely(!pages[i])) {
			gb_err(gbdev->dev, "BUG.... in %s\n", __func__);

			continue;
		}

		p = gb_phys_to_page(gbdev, pages[i]);
		if (!p) {
			gb_err(gbdev->dev, "Error: gb_page is NULL %s\n", __func__);
			return;
		}
		gb_free_pages(p, gbdev);

		pages[i] = 0;
		for (k = 0; k < PG_SIZE_MULTI; k++) {
			pages[i+k] = 0;
		}
	}
}

#ifdef GB_DMA_ON
int gb_get_mem_status(struct gb_context *gbctx, struct gb_uk_get_mem_status *mem)
{
	struct gb_page_ctrl *page_ctrl;

	page_ctrl = &gbctx->gbdev->gb_page_ctrl;

	mem->used_pages = atomic_read(&page_ctrl->used_pages);
	mem->used_mmu_pages = atomic_read(&page_ctrl->used_mmu_pages);
	gb_verb_printf(KERN_INFO "used_pages: %d used_mmu_pages: 0x%x pg_high: 0x%llx mmu_low: 0x%llx\n",
		mem->used_pages, mem->used_mmu_pages,
		page_ctrl->page_highest_addr, page_ctrl->mmu_lowest_addr);

	return 0;
}
GB_EXPORT_TEST_API(gb_get_mem_status);
#endif
