/*
 * Copyright (c) 2016 Synopsys, Inc.
 *
 * Synopsys DP TX Linux Software Driver and documentation (hereinafter,
 * "Software") is an Unsupported proprietary work of Synopsys, Inc. unless
 * otherwise expressly agreed to in writing between Synopsys and you.
 *
 * The Software IS NOT an item of Licensed Software or Licensed Product under
 * any End User Software License Agreement or Agreement for Licensed Product
 * with Synopsys or any supplement thereto. You are permitted to use and
 * redistribute this Software in source and binary forms, with or without
 * modification, provided that redistributions of source code must retain this
 * notice. You may not view, use, disclose, copy or distribute this file or
 * any information contained herein except pursuant to this license grant from
 * Synopsys. If you do not agree with this notice, including the disclaimer
 * below, then you are not authorized to use the Software.
 *
 * THIS SOFTWARE IS BEING DISTRIBUTED BY SYNOPSYS SOLELY ON AN "AS IS" BASIS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE HEREBY DISCLAIMED. IN NO EVENT SHALL SYNOPSYS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "gb_dp_dptx.h"

/*
 * Core Access Layer
 *
 * Provides low-level register access to the DPTX core.
 */

/**
 * dptx_intr_en() - Enables interrupts
 * @dptx: The dptx struct
 * @bits: The interrupts to enable
 *
 * This function enables (unmasks) all interrupts in the INTERRUPT
 * register specified by @bits.
 */
static void dptx_intr_en(struct dptx *dptx, u32 bits)
{
	u32 ien;

	dptx_dbg(dptx, "%s:\n", __func__);
	ien = dptx_readl(dptx, DPTX_IEN);
	ien |= bits;
	dptx_writel(dptx, DPTX_IEN, ien);
	//dptx_writel(dptx, 0x60000, 1);
}

/**
 * dptx_intr_dis() - Disables interrupts
 * @dptx: The dptx struct
 * @bits: The interrupts to disable
 *
 * This function disables (masks) all interrupts in the INTERRUPT
 * register specified by @bits.
 */
static void dptx_intr_dis(struct dptx *dptx, u32 bits)
{
	u32 ien;

	dptx_dbg(dptx, "%s:\n", __func__);
	ien = dptx_readl(dptx, DPTX_IEN);
	ien &= ~bits;
	dptx_writel(dptx, DPTX_IEN, ien);
}

/**
 * dptx_global_intr_en() - Enables top-level interrupts
 * @dptx: The dptx struct
 *
 * Enables (unmasks) all top-level interrupts.
 */
void dptx_global_intr_en(struct dptx *dptx)
{
	dptx_intr_en(dptx, DPTX_IEN_ALL_INTR &
		~(DPTX_ISTS_AUX_REPLY | DPTX_ISTS_AUX_CMD_INVALID |
			DPTX_IEN_TYPE_C));
}

/**
 * dptx_global_intr_dis() - Disables top-level interrupts
 * @dptx: The dptx struct
 *
 * Disables (masks) all top-level interrupts.
 */
void dptx_global_intr_dis(struct dptx *dptx)
{
	dptx_intr_dis(dptx, DPTX_IEN_ALL_INTR);
}

/**
 * dptx_soft_reset() - Performs a core soft reset
 * @dptx: The dptx struct
 * @bits: The components to reset
 *
 * Resets specified parts of the core by writing @bits into the core
 * soft reset control register and clearing them 10-20 microseconds
 * later.
 */
void dptx_soft_reset(struct dptx *dptx, u32 bits)
{
	u32 rst;

	bits &= (DPTX_SRST_CTRL_ALL);

	/* Set reset bits */
	rst = dptx_readl(dptx, DPTX_SRST_CTRL);
	rst |= bits;
	dptx_writel(dptx, DPTX_SRST_CTRL, rst);

	usleep_range(10000, 20000);

	/* Clear reset bits */
	rst = dptx_readl(dptx, DPTX_SRST_CTRL);
	rst &= ~bits;
	dptx_writel(dptx, DPTX_SRST_CTRL, rst);
}

/**
 * dptx_soft_reset_all() - Reset all core modules
 * @dptx: The dptx struct
 */
void dptx_soft_reset_all(struct dptx *dptx)
{
	dptx_soft_reset(dptx, DPTX_SRST_CTRL_ALL);
}

void dptx_phy_soft_reset(struct dptx *dptx)
{
	dptx_soft_reset(dptx, DPTX_SRST_CTRL_PHY);
}

/**
 * dptx_core_init_phy() - Initializes the DP TX PHY module
 * @dptx: The dptx struct
 *
 * Initializes the PHY layer of the core. This needs to be called
 * whenever the PHY layer is reset.
 */
void dptx_core_init_phy(struct dptx *dptx)
{
	u32 phyifctrl;

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_WIDTH;
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}


/**
* dptx_sink_enabled_ssc() - Returns true, if sink is enabled ssc
* @dptx: The dptx struct
*
*/
bool dptx_sink_enabled_ssc(struct dptx *dptx)
{
	u8 byte;

	dptx_read_dpcd(dptx, DP_MAX_DOWNSPREAD, &byte);

	return byte & 1;
}

/**
 * dptx_core_program_ssc() - Move phy to P3 state and programs SSC
 * @dptx: The dptx struct
 *
 * Enables SSC should be called during hot plug.
 *
 */
int dptx_core_program_ssc(struct dptx *dptx, bool sink_ssc)
{
	u32 phyifctrl;
        u8  retval;

	/* Enable 4 lanes, before programming SSC */
	dptx_phy_set_lanes(dptx, 4);

	// Move PHY to P3 to program SSC
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl |= (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
		dptx_err(dptx, "Timed out waiting for PHY BUSY\n");
	        return retval;
	}

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
        if(dptx->ssc_en && sink_ssc)
	        phyifctrl &= ~DPTX_PHYIF_CTRL_SSC_DIS;
	else
		phyifctrl |= DPTX_PHYIF_CTRL_SSC_DIS;

        dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
	        dptx_err(dptx, "Timed out waiting for PHY BUSY\n");
	        return retval;
	 }

	return 0;
}

/**
 * dptx_check_dptx_id() - Check value of DPTX_ID register
 * @dptx: The dptx struct
 *
 * Returns True if DPTX core correctly identifyed.
 */
bool dptx_check_dptx_id(struct dptx *dptx)
{
	u32 dptx_id;

	dptx_id = dptx_readl(dptx, DPTX_ID);
	if (dptx_id != ((DPTX_ID_DEVICE_ID << DPTX_ID_DEVICE_ID_SHIFT) |
			DPTX_ID_VENDOR_ID))
		return false;

	return true;
}

/**
* dptx_enable_ssc() - Enables SSC based on automation request,
*		      if DPTX controller enables ssc
* @dptx: The dptx struct
*
*/
void dptx_enable_ssc(struct dptx *dptx)
{
	bool sink_ssc = dptx_sink_enabled_ssc(dptx);
	if(sink_ssc)
		dev_dbg(dptx->dev, "%s: SSC enable on the sink side\n", __func__);
	else
		dev_dbg(dptx->dev, "%s: SSC disabled on the sink side\n", __func__);
	dptx_core_program_ssc(dptx, sink_ssc);
}

/* zfl:文档第3步，获取dptx配置参数 */
void dptx_init_hwparams(struct dptx *dptx)
{
	u32 reg;

	reg = dptx_readl(dptx, DPTX_CONFIG1);

	/* Num MST streams */
	dptx->streams = (reg & DPTX_CONFIG1_NUM_STREAMS_MASK) >>
		DPTX_CONFIG1_NUM_STREAMS_SHIFT;

	/* Combo PHY     zfl?这参数干啥的，代码里也没用*/
	dptx->hwparams.gen2phy = !!(reg & DPTX_CONFIG1_GEN2_PHY);

	/* DSC   zfl:显示视频压缩技术*/
	dptx->hwparams.dsc = !!(reg & DPTX_CONFIG1_DSC_EN);

	/* Multi pixel mode 	zfl?这个是？*/
	switch ((reg & DPTX_CONFIG1_MP_MODE_MASK) >> DPTX_CONFIG1_MP_MODE_SHIFT) {
	default:
	case DPTX_CONFIG1_MP_MODE_SINGLE:
		dptx->hwparams.multipixel = DPTX_MP_SINGLE_PIXEL;
		break;
	case DPTX_CONFIG1_MP_MODE_DUAL:
		dptx->hwparams.multipixel = DPTX_MP_DUAL_PIXEL;
		break;
	case DPTX_CONFIG1_MP_MODE_QUAD:
		dptx->hwparams.multipixel = DPTX_MP_QUAD_PIXEL;
		break;
	}
}

/**
 * dptx_core_init() - Initializes the DP TX core
 * @dptx: The dptx struct
 *
 * Initialize the DP TX core and put it in a known state.
 */
int dptx_core_init(struct dptx *dptx, bool is_edp)
{
	char str[15];
	u32 version;
	u32 hpd_ien;
	u32 reg;

	/* Reset the core */
	dptx_soft_reset_all(dptx);

	reg = dptx_readl(dptx, DPTX_CCTL);
	if (is_edp) {
		/* enable edp */
		reg |= (1 << 27);
	}
	/* disable fec */
	reg &= (~(1 << 26));
	/* reset SCALE DOWN */
	reg &= (~(0x01FF << 16));
	/* disable fast train */
	reg &= (~(0x0001 << 2));
	/* ENABLE_SCRAMBLE */
	reg &= (~(0x0001 << 0));
	//CCTL
	dptx_writel(dptx, DPTX_CCTL, reg);
#if 0
	/* CFG PHY POWERDOWN */
	reg = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	reg |= (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
	dptx_writel(dptx, DPTX_PHYIF_CTRL, reg);
	ret = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (ret) {
		dptx_err(dptx, "%s %d Timed out waiting for PHY BUSY\n",
			__func__, __LINE__);
		return ret;
	}

	/* CFG PHY POWERON */
	reg = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	reg &= (~DPTX_PHYIF_CTRL_LANE_PWRDOWN_MASK);// P0
	dptx_writel(dptx, DPTX_PHYIF_CTRL, reg);
#endif
	/* Reset the core */
//	dptx_soft_reset_all(dptx);
#if 0
	/* Reset the phy */
	dptx_phy_soft_reset(dptx);
	ret = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (ret) {
		dptx_err(dptx, "%s %d Timed out waiting for PHY BUSY\n",
			__func__, __LINE__);
		return ret;
	}
#endif
	/* Check the core version */
	memset(str, 0, sizeof(str));
	version = dptx_readl(dptx, DPTX_VER_NUMBER);
	str[0] = (version >> 24) & 0xff;
	str[1] = '.';
	str[2] = (version >> 16) & 0xff;
	str[3] = (version >> 8) & 0xff;
	str[4] = version & 0xff;

	version = dptx_readl(dptx, DPTX_VER_TYPE);
	str[5] = '-';
	str[6] = (version >> 24) & 0xff;
	str[7] = (version >> 16) & 0xff;
	str[8] = (version >> 8) & 0xff;
	str[9] = version & 0xff;

	dptx_dbg(dptx, "Core version: %s\n", str);
	dptx->version = version;
	dptx_core_init_phy(dptx);

	/* Enable all HPD interrupts */
	//hpd_ien = dptx_readl(dptx, DPTX_HPD_IEN);
	hpd_ien = (DPTX_HPD_IEN_IRQ_EN | DPTX_HPD_IEN_HOT_PLUG_EN |
		    DPTX_HPD_IEN_HOT_UNPLUG_EN);

	dptx_writel(dptx, DPTX_HPD_IEN, hpd_ien);

#ifdef DP_HDCP_ENABLE
	/* Mask interrupt related to HDCP22 GPIO output status */
	reg = dptx_readl(dptx, DPTX_HDCP_API_INT_MSK);
	reg |= DPTX_HDCP22_GPIOINT;
	dptx_writel(dptx, DPTX_HDCP_API_INT_MSK, reg);
#endif

	/* Enable all top-level interrupts */
	dptx_global_intr_en(dptx);
#if 0
	/* AUX_250US_CNT_LIMIT */
	dptx_writel(dptx, DPTX_AUX_250US_CNT, 0x000000fa);
	/* AUX_2000US_CNT_LIMIT */
	dptx_writel(dptx, DPTX_AUX_2000US_CNT, 0x000007d0);
	/* AUX_100000US_CNT_LIMIT */
	dptx_writel(dptx, DPTX_AUX_100000US_CNT, 0x000186a0);
#endif
	return 0;
}

/**
 * dptx_core_deinit() - Deinitialize the core
 * @dptx: The dptx struct
 *
 * Disable the core in preparation for module shutdown.
 */
int dptx_core_deinit(struct dptx *dptx)
{
	dptx_global_intr_dis(dptx);
	dptx_soft_reset_all(dptx);
	return 0;
}

/*
 * PHYIF core access functions
 */

unsigned int dptx_phy_get_lanes(struct dptx *dptx)
{
	u32 phyifctrl;
	u32 val;

	dptx_dbg(dptx, "%s:\n", __func__);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	val = (phyifctrl & DPTX_PHYIF_CTRL_LANES_MASK) >>
		DPTX_PHYIF_CTRL_LANES_SHIFT;

	return (1 << val);
}

void dptx_phy_set_lanes(struct dptx *dptx, unsigned int lanes)
{
	u32 phyifctrl;
	u32 val;

	dptx_dbg(dptx, "%s: lanes=%d\n", __func__, lanes);

	switch (lanes) {
	case 1:
		val = 0;
		break;
	case 2:
		val = 1;
		break;
	case 4:
		val = 2;
		break;
	default:
		dptx_err(dptx, "Invalid number of lanes %d\n", lanes);
		return;
	}

	phyifctrl = 0;
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_LANES_MASK;
	phyifctrl |= (val << DPTX_PHYIF_CTRL_LANES_SHIFT);
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

void dptx_phy_set_rate(struct dptx *dptx, unsigned int rate)
{
	u32 phyifctrl;

	dptx_dbg(dptx, "%s: rate=%d\n", __func__, rate);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
#ifdef CONFIG_GB02_CTS
	switch (rate) {
	case DPTX_PHYIF_CTRL_RATE_RBR:

		dptx_writel(dptx, 0xc14, 0x00000001);
		dptx_writel(dptx, 0xc18, 0x00500051);
		dptx_writel(dptx, 0xc1c, 0x001e600a);
		dptx_writel(dptx, 0xc20, 0x00000000);
	//	dptx_writel(dptx, 0xc3c, 0x00000006);
                dptx_writel(dptx, 0xc3c, 0x00000000);
	//	dptx_writel(dptx, 0xc40, 0x01010014);
                dptx_writel(dptx, 0xc40, 0x0101002f);
		dptx_writel(dptx, 0xc44, 0x00000777);
                break;
	case DPTX_PHYIF_CTRL_RATE_HBR:

                dptx_writel(dptx, 0xc14, 0x00000001);
		dptx_writel(dptx, 0xc18, 0x00400051);
		dptx_writel(dptx, 0xc1c, 0x0018a01e);
		dptx_writel(dptx, 0xc20, 0x00000000);
		dptx_writel(dptx, 0xc3c, 0x00000006);
                dptx_writel(dptx, 0xc3c, 0x00000000);
	//	dptx_writel(dptx, 0xc40, 0x01010014);
                dptx_writel(dptx, 0xc40, 0x0101002f);
		dptx_writel(dptx, 0xc44, 0x00000777);

		break;
	case DPTX_PHYIF_CTRL_RATE_HBR2:

                dptx_writel(dptx, 0xc14, 0x00000001);
		dptx_writel(dptx, 0xc18, 0x00400051);
		dptx_writel(dptx, 0xc1c, 0x0018a01e);
		dptx_writel(dptx, 0xc20, 0x00000000);
		dptx_writel(dptx, 0xc3c, 0x00000006);
                dptx_writel(dptx, 0xc3c, 0x00000000);
	//	dptx_writel(dptx, 0xc40, 0x01010014);
                dptx_writel(dptx, 0xc40, 0x0101002f);
		dptx_writel(dptx, 0xc44, 0x00000777);

		break;
	case DPTX_PHYIF_CTRL_RATE_HBR3:

		dptx_writel(dptx, 0xc14, 0x00000001);
		dptx_writel(dptx, 0xc18, 0x00100051);
		dptx_writel(dptx, 0xc1c, 0x0015c045);
		dptx_writel(dptx, 0xc20, 0x00000000);
		dptx_writel(dptx, 0xc3c, 0x00000006);
                dptx_writel(dptx, 0xc3c, 0x00000000);
	//	dptx_writel(dptx, 0xc40, 0x01010014);
                dptx_writel(dptx, 0xc40, 0x0101002f);
		dptx_writel(dptx, 0xc44, 0x00000777);
		break;
	default:
		dptx_err(dptx, "Invalid PHY rate %d\n", rate);
                dptx_writel(dptx, 0xc14, 0x00000001);
		dptx_writel(dptx, 0xc18, 0x00500051);
		dptx_writel(dptx, 0xc1c, 0x001e600a);
		dptx_writel(dptx, 0xc20, 0x00000000);
		dptx_writel(dptx, 0xc3c, 0x00000006);
                dptx_writel(dptx, 0xc3c, 0x00000000);
	//	dptx_writel(dptx, 0xc40, 0x01010014);
		dptx_writel(dptx, 0xc40, 0x0101002f);
		dptx_writel(dptx, 0xc44, 0x00000777);
		break;
	}
#endif
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_RATE_MASK;
	phyifctrl |= rate << DPTX_PHYIF_CTRL_RATE_SHIFT;
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

unsigned int dwc_phy_get_rate(struct dptx *dptx)
{
	u32 phyifctrl;
	u32 rate;

	dptx_dbg(dptx, "%s:\n", __func__);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	rate = (phyifctrl & DPTX_PHYIF_CTRL_RATE_MASK) >>
		DPTX_PHYIF_CTRL_RATE_SHIFT;

	return rate;
}

int dptx_phy_wait_busy(struct dptx *dptx, unsigned int lanes)
{
	unsigned int count;
	u32 phyifctrl;
	u32 mask = 0;

	dptx_dbg(dptx, "%s: lanes=%d\n", __func__, lanes);

	switch (lanes) {
	case 4:
		mask |= DPTX_PHYIF_CTRL_BUSY(3);
		mask |= DPTX_PHYIF_CTRL_BUSY(2);
		mask |= DPTX_PHYIF_CTRL_BUSY(1);
		mask |= DPTX_PHYIF_CTRL_BUSY(0);
		break;
	case 2:
		mask |= DPTX_PHYIF_CTRL_BUSY(1);
		mask |= DPTX_PHYIF_CTRL_BUSY(0);
		break;
	case 1:
		mask |= DPTX_PHYIF_CTRL_BUSY(0);
		break;
	default:
		dptx_err(dptx, "Invalid number of lanes %d\n", lanes);
		break;
	}

	count = 0;

	while (1) {
		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

		if (!(phyifctrl & mask))
			break;

		count++;
		if (count > 1000) {
			dptx_warn(dptx, "%s: PHY BUSY timed out\n", __func__);
			return -EBUSY;
		}

		udelay(5);
	}

	return 0;
}

void dptx_phy_set_pre_emphasis(struct dptx *dptx,
			       unsigned int lane,
			       unsigned int level)
{
	u32 phytxeq;

	dptx_dbg(dptx, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

	if (lane > 3) {
		dptx_err(dptx, "Invalid lane %d", lane);
		return;
	}

	if (level > 3) {
		dptx_err(dptx, "Invalid pre-emphasis level %d, using 3", level);
		level = 3;
	}

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	phytxeq &= ~DPTX_PHY_TX_EQ_PREEMP_MASK(lane);
	phytxeq |= (level << DPTX_PHY_TX_EQ_PREEMP_SHIFT(lane)) &
		DPTX_PHY_TX_EQ_PREEMP_MASK(lane);

	dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}
void dptx_phy_set_pre_emphasis_vsw(struct dptx *dptx,
			       unsigned int lane,
			       unsigned int pre,
			       unsigned int vsw)
{
	u32 eq_main;
	u32 eq_post;
	u32 vboost_lvl;

	eq_main = 0x209000; //0x9002
	eq_post = 0x309000; //0x9003
	vboost_lvl = 0x30000c;
	printk(KERN_INFO"%s-%d: lane=%d, pre=0x%x vsw=0x%x\n",
			__func__, __LINE__, lane, pre, vsw);

	if (lane > 3) {
		dptx_err(dptx, "Invalid lane %d", lane);
		return;
	}

	if (pre > 3) {
		dptx_err(dptx, "Invalid pre-emphasis level %d, using 3", pre);
		pre = 3;
	}

        switch (lane) {
	     case 0:
		     eq_main = 0x201000;
		     eq_post = 0x301000;
		     break;
             case 1:
		     eq_main = 0x201100;
		     eq_post = 0x301100;
		     break;
             case 2:
		     eq_main = 0x201200;
		     eq_post = 0x301200;
		     break;
	     case 3:
		     eq_main = 0x201300;
		     eq_post = 0x301300;
		     break;
             default:
	             break;
          }
	switch (pre){
	case 0:
		switch (vsw) {
		case 0:
			dptx_writel(dptx, 0x20000+eq_main, 0x99F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2040);
	                dptx_writel(dptx, 0x20000+vboost_lvl, 0x0240);
                      //  dptx_writel(dptx, 0x20000+vboost_lvl, 0x0280);
			break;
		case 1:
			dptx_writel(dptx, 0x20000+eq_main, 0xA5F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2040);
			dptx_writel(dptx, 0x20000+vboost_lvl, 0x0240);
			break;
		case 2:
			dptx_writel(dptx, 0x20000+eq_main, 0xB1F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2040);
			dptx_writel(dptx, 0x20000+vboost_lvl, 0x0240);
			break;
		case 3:
			dptx_writel(dptx, 0x20000+eq_main, 0xB1F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2040);
			dptx_writel(dptx, 0x20000+vboost_lvl, 0x03C0);
			break;
		}
		break;
	case 1:
		switch (vsw) {
		case 0:
			dptx_writel(dptx, 0x20000+eq_main, 0x9FF8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2640);
			dptx_writel(dptx, 0x20000+vboost_lvl,0x0240);
			break;
		case 1:
			dptx_writel(dptx, 0x20000+eq_main, 0xA9F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2840);
			dptx_writel(dptx, 0x20000+vboost_lvl,0x0280);

			break;
		case 2:
			dptx_writel(dptx, 0x20000+eq_main, 0xA9F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2840);
			dptx_writel(dptx, 0x20000+vboost_lvl,0x03C0);
			break;
		}
		break;
	case 2:
		switch (vsw) {
		case 0:
			dptx_writel(dptx, 0x20000+eq_main, 0xA5F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2C40);
			dptx_writel(dptx, 0x20000+vboost_lvl, 0x0280);
			break;
		case 1:
			dptx_writel(dptx, 0x20000+eq_main, 0xA5F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x2C40);
			dptx_writel(dptx, 0x20000+vboost_lvl, 0x03C0);
			break;
		}
		break;
	case 3:
		switch (vsw) {
		case 0:
			dptx_writel(dptx, 0x20000+eq_main, 0xA1F8);
			dptx_writel(dptx, 0x20000+eq_post, 0x3040);
			dptx_writel(dptx, 0x20000+vboost_lvl,0x03C0);
			break;
		}
		break;
	default:
		break;
	}
#if 0
	gb_printf(KERN_INFO"%s-%d: eq_main = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0x20000+eq_main));
	gb_printf(KERN_INFO"%s-%d: eq_post = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0x20000+eq_post));
	gb_printf(KERN_INFO"%s-%d: dptx_c14 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc14));
	gb_printf(KERN_INFO"%s-%d: dptx_c18 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc18));
        gb_printf(KERN_INFO"%s-%d: dptx_c1c = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc1c));
	gb_printf(KERN_INFO"%s-%d: dptx_c20 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc20));
        gb_printf(KERN_INFO"%s-%d: dptx_c3c = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc3c));
	gb_printf(KERN_INFO"%s-%d: dptx_c40 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc40));
        gb_printf(KERN_INFO"%s-%d: dptx_c44 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xc44));
	gb_printf(KERN_INFO"%s-%d: dptx_a00 = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0xa00));
        gb_printf(KERN_INFO"%s-%d: dpphy_12004c = 0x%x\n", __func__, __LINE__,
			dptx_readl(dptx, 0x12004c));
#endif
}

void dptx_phy_set_vswing(struct dptx *dptx,
			 unsigned int lane,
			 unsigned int level)
{
	u32 phytxeq;

	dptx_dbg(dptx, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

	if (lane > 3) {
		dptx_err(dptx, "Invalid lane %d", lane);
		return;
	}

	if (level > 3) {
		dptx_err(dptx, "Invalid vswing level %d, using 3", level);
		level = 3;
	}

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	phytxeq &= ~DPTX_PHY_TX_EQ_VSWING_MASK(lane);
	phytxeq |= (level << DPTX_PHY_TX_EQ_VSWING_SHIFT(lane)) &
		DPTX_PHY_TX_EQ_VSWING_MASK(lane);

	dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}

void dptx_phy_set_pattern(struct dptx *dptx,
			  unsigned int pattern)
{
	u32 phyifctrl = 0;

	dptx_dbg(dptx, "%s: Setting PHY pattern=0x%x\n", __func__, pattern);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_TPS_SEL_MASK;
	phyifctrl |= ((pattern << DPTX_PHYIF_CTRL_TPS_SEL_SHIFT) &
		      DPTX_PHYIF_CTRL_TPS_SEL_MASK);
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

void dptx_phy_enable_xmit(struct dptx *dptx, unsigned int lanes, bool enable)
{
	u32 phyifctrl;
	u32 mask = 0;

	dptx_dbg(dptx, "%s: lanes=%d, enable=%d\n", __func__, lanes, enable);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

	switch (lanes) {
	case 4:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(3);
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(2);
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(1);
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(0);
		break;
	case 2:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(1);
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(0);
		break;
	case 1:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(0);
		break;
	default:
		dptx_err(dptx, "Invalid number of lanes %d\n", lanes);
		break;
	}

	if (enable)
		phyifctrl |= mask;
	else
		phyifctrl &= ~mask;

	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

int dptx_phy_rate_to_bw(unsigned int rate)
{
	switch (rate) {
	case DPTX_PHYIF_CTRL_RATE_RBR:
		return DP_LINK_BW_1_62;
	case DPTX_PHYIF_CTRL_RATE_HBR:
		return DP_LINK_BW_2_7;
	case DPTX_PHYIF_CTRL_RATE_HBR2:
		return DP_LINK_BW_5_4;
	case DPTX_PHYIF_CTRL_RATE_HBR3:
		return DP_LINK_BW_8_1;
	case DPTX_PHYIF_CTRL_EDP_RATE_R216:
	case DPTX_PHYIF_CTRL_EDP_RATE_R243:
	case DPTX_PHYIF_CTRL_EDP_RATE_R324:
	case DPTX_PHYIF_CTRL_EDP_RATE_R432:
		return DP_LINK_RATE_TABLE;
	default:
		dptx_err(dptx, "Invalid rate 0x%x\n", rate);
		return -EINVAL;
	}
}

u8 dptx_bw_to_phy_rate(unsigned int bw)
{
	switch (bw) {
	case DP_LINK_BW_1_62:
		return DPTX_PHYIF_CTRL_RATE_RBR;
	case DP_LINK_BW_2_7:
		return DPTX_PHYIF_CTRL_RATE_HBR;
	case DP_LINK_BW_5_4:
		return DPTX_PHYIF_CTRL_RATE_HBR2;
	case DP_LINK_BW_8_1:
		return DPTX_PHYIF_CTRL_RATE_HBR3;
	default:
		dptx_err(dptx, "Invalid bw 0x%x\n", bw);
		return -EINVAL;
	}
}
