/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef __DPTX_DRIVER_H__
#define __DPTX_DRIVER_H__

#include <linux/irq.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
#include <drm/drm_dp_helper.h>
#include <drm/drm_dp_mst_helper.h>
#else
#include <drm/display/drm_dp.h>
#include <drm/display/drm_dp_mst_helper.h>
#endif
#include <drm/drm_fixed.h>
//#include "drm_dp_helper_additions.h"

//#define GB02_PAL_DP
#define DPTX_RECEIVER_CAP_SIZE		0x100
#define DPTX_SDP_NUM		0x10
#define DPTX_SDP_LEN	0x9
#define DPTX_SDP_SIZE	(9 * 4)
//#define DPTX_TYPE_C
#define PARSE_EST_TIMINGS_FROM_BYTE3

//DSC definitions coming from RC files of DSC C model
#define RC_MODEL_SIZE			8192
#define INITIAL_OFFSET			6144
#define INITIAL_DELAY			512
#define PIXELS_PER_GROUP		3
#define PIXEL_HOLD_DELAY		5
#define PIXEL_GROUP_DELAY		3
#define MUXER_INITIAL_BUFFERING_DELAY	9
#define DSC_MAX_AUX_ORIG_WIDTH		24
#define PIXEL_FLATNESSBUF_DELAY		DSC_MAX_AUX_ORIG_WIDTH
#define FLATNESS_MIN_QP			3
#define	FLATNESS_MAX_QP			12
#define RC_EDGE_FACTOR			6
#define RC_QUANT_INCR_LIMIT0		11
#define RC_QUANT_INCR_LIMIT1		11
#define RC_TGT_OFFSET_HIGH		3
#define RC_TGT_OFFSET_LOW		3
#define INITIAL_SCALE_VALUE_SHIFT	3
#define OFFSET_FRACTIONAL_BITS		11

struct rc_range_param {
	int minQP	:5;
	int maxQP	:5;
	int offset	:6;
};


// Rounding up to the nearest multiple of a number
#define ROUND_UP_TO_NEAREST(numToRound, mult) \
	((((numToRound+(mult)-1) / (mult)) * (mult)))


#include "gb_dp_avgen.h"
#include "gb_dp_reg.h"
#include "gb_dp_dbg.h"

struct dptx;

/* The max rate and lanes supported by the core */
#define DPTX_MAX_LINK_RATE DPTX_PHYIF_CTRL_RATE_HBR
#define DPTX_MAX_LINK_LANES 4

/* The default rate and lanes to use for link training */
#define DPTX_DEFAULT_LINK_RATE DPTX_MAX_LINK_RATE
#define DPTX_DEFAULT_LINK_LANES DPTX_MAX_LINK_LANES

/**
 * struct dptx_link - The link state.
 * @status: Holds the sink status register values.
 * @trained: True if the link is successfully trained.
 * @rate: The current rate that the link is trained at.
 * @lanes: The current number of lanes that the link is trained at.
 * @preemp_level: The pre-emphasis level used for each lane.
 * @vswing_level: The vswing level used for each lane.
 */
struct dptx_link {
	u8 status[DP_LINK_STATUS_SIZE];
	bool trained;
	u8 rate;
	u8 lanes;
	u8 preemp_level[4];
	u8 vswing_level[4];
};

enum established_timings {
	DMT_640x480_60hz,
	DMT_800x600_60hz,
	DMT_1024x768_60hz,
	NONE
};

/**
 * struct dptx_aux - The aux state
 * @sts: The AUXSTS register contents.
 * @data: The AUX data register contents.
 * @event: Indicates an AUX event ocurred.
 * @abort: Indicates that the AUX transfer should be aborted.
 */
struct dptx_aux {
	u32 sts;
	u32 data[4];
	atomic_t abort;
};

struct sdp_header {
	u8 HB0;
	u8 HB1;
	u8 HB2;
	u8 HB3;
} __packed;

struct sdp_full_data {
	u8 en;
	u32 payload[9];
	u8 blanking;
	u8 cont;
} __packed;

#define DPTX_HDCP_REG_DPK_CRC_ORIG	0x331c1169
#define DPTX_HDCP_MAX_AUTH_RETRY	10

struct hdcp_aksv {
	u32 lsb;
	u32 msb;
};

struct hdcp_dpk {
	u32 lsb;
	u32 msb;
};

struct hdcp_params {
	struct hdcp_aksv aksv;
	struct hdcp_dpk dpk[40];
	u32 enc_key;
	u32 crc32;
	u8 auth_fail_count;
	int hdcp13_is_en;
	int hdcp22_is_en;
};

/**
 * struct dptx - The representation of the DP TX core
 * @mutex: dptx mutex
 * @base: Base address of the registers
 * @irq: IRQ number
 * @version: Contents of the IP_VERSION register
 * @max_rate: The maximum rate that the controller supports
 * @max_lanes: The maximum lane count that the controller supports
 * @dev: The struct device
 * @root: The debugfs root
 * @regset: The debugfs regset
 * @vparams: The video params to use
 * @aparams: The audio params to use
 * @hparams: The HDCP params to use
 * @waitq: The waitq
 * @shutdown: Signals that the driver is shutting down and that all
 *            operations should be aborted.
 * @c_connect: Signals that a HOT_PLUG or HOT_UNPLUG has occurred.
 * @sink_request: Signals the a HPD_IRQ has occurred.
 * @rx_caps: The sink's receiver capabilities.
 * @edid: The sink's EDID.
 * @sdp_list: The array of SDP elements
 * @aux: AUX channel state for performing an AUX transfer.
 * @link: The current link state.
 * @multipixel: Controls multipixel configuration. 0-Single, 1-Dual, 2-Quad.
 */
typedef struct dptx {
	struct mutex mutex; /* generic mutex for dptx */

	struct {
		u8 multipixel;
		u8 streams;
		bool gen2phy;
		bool dsc;
	} hwparams;

	void __iomem *base;
	int irq;

	u32 version;
	int index;
	u8 max_rate;
	u8 max_lanes;
	bool ycbcr420;
	u8 streams;

	bool mst;
	int active_mst_links;

	struct drm_dp_mst_topology_mgr mst_mgr;  //sahakyan

	bool cr_fail;// harutk need to remove

	u8 multipixel;
	u8 bstatus;

	bool ssc_en;
	bool fec;
	bool dsc;

	int vcp_id;
	u16 pbn;
	bool need_rad;
	bool logic_port;
	u8 rad_port;
	u8 port[2];

	bool dummy_dtds_present;
	enum established_timings selected_est_timing;

	struct device *dev;
	struct dentry *root;
	struct debugfs_regset32 *regset;

	struct video_params vparams;
	struct audio_params aparams;
	struct audio_short_desc audio_desc;
	struct hdcp_params hparams;

	wait_queue_head_t waitq;

	atomic_t shutdown;
	atomic_t c_connect;
	atomic_t sink_request;

	u8 rx_caps[DPTX_RECEIVER_CAP_SIZE];

	u8 *edid;
	u8 *edid_second;
#define DPTX_DEFAULT_EDID_BUFLEN 128

	struct sdp_full_data sdp_list[DPTX_SDP_NUM];
	struct dptx_aux aux;
	struct dptx_link link;
	struct drm_device *drm_dev;
	struct delayed_work hotplug_work;
	bool is_edp;
	bool edid_getted;
}dptx_s;

/*
 * Core interface functions
 */
int dptx_core_init(struct dptx *dptx, bool is_edp);
void dptx_init_hwparams(struct dptx *dptx);
bool dptx_check_dptx_id(struct dptx *dptx);
void dptx_core_init_phy(struct dptx *dptx);
int dptx_core_program_ssc(struct dptx *dptx, bool sink_ssc);
bool dptx_sink_enabled_ssc(struct dptx *dptx);
void dptx_enable_ssc(struct dptx *dptx);

int dptx_core_deinit(struct dptx *dptx);
void dptx_soft_reset(struct dptx *dptx, u32 bits);
void dptx_soft_reset_all(struct dptx *dptx);
void dptx_phy_soft_reset(struct dptx *dptx);

irqreturn_t dptx_irq(int irq, void *dev);
irqreturn_t dptx_threaded_irq(int irq, void *dev);

void dptx_global_intr_en(struct dptx *dp);
void dptx_global_intr_dis(struct dptx *dp);
void dptx_dsc_enable(struct dptx *dptx);

/*
 * PHY IF Control
 */
void dptx_phy_set_lanes(struct dptx *dptx, unsigned int num);
unsigned int dptx_phy_get_lanes(struct dptx *dptx);
void dptx_phy_set_rate(struct dptx *dptx, unsigned int rate);
unsigned int dwc_phy_get_rate(struct dptx *dptx);
int dptx_phy_wait_busy(struct dptx *dptx, unsigned int lanes);
void dptx_phy_set_pre_emphasis_vsw(struct dptx *dptx,
			       unsigned int lane,
			       unsigned int pre,
			       unsigned int vsw);
void dptx_phy_set_pre_emphasis(struct dptx *dptx,
	unsigned int lane, unsigned int level);
void dptx_phy_set_vswing(struct dptx *dptx,
	unsigned int lane, unsigned int level);
void dptx_phy_set_pattern(struct dptx *dptx,
			  unsigned int pattern);
void dptx_phy_enable_xmit(struct dptx *dptx, unsigned int lane, bool enable);

int dptx_phy_rate_to_bw(unsigned int rate);
u8 dptx_bw_to_phy_rate(unsigned int bw);

int dptx_link_reduce_rate(struct dptx *dptx);
int dptx_link_training_start(struct dptx *dptx);
int dptx_link_reduce_lanes(struct dptx *dptx);
int dptx_link_training_init(struct dptx *dptx,
				   u8 rate,
				   u8 lanes);
/*
 * AUX Channel
 */

#define DPTX_AUX_TIMEOUT 2000

int dptx_read_bytes_from_i2c(struct dptx *dptx,
	unsigned int device_addr, u8 *bytes, u32 len);

int dptx_i2c_address_only(struct dptx *dptx,
	unsigned int device_addr);

int dptx_write_bytes_to_i2c(struct dptx *dptx,
	unsigned int device_addr, u8 *bytes, u32 len);

int __dptx_read_dpcd(struct dptx *dptx, u32 addr, u8 *byte);
int __dptx_write_dpcd(struct dptx *dptx, u32 addr, u8 byte);

int __dptx_read_bytes_from_dpcd(struct dptx *dptx,
				unsigned int reg_addr,
				u8 *bytes, u32 len);

int __dptx_write_bytes_to_dpcd(struct dptx *dptx,
			       unsigned int reg_addr,
			       u8 *bytes, u32 len);

#ifndef DPTX_DEBUG_DPCD_CMDS
static inline int dptx_read_dpcd(struct dptx *dptx, u32 addr, u8 *byte)
{
	return __dptx_read_dpcd(dptx, addr, byte);
}

static inline int dptx_write_dpcd(struct dptx *dptx, u32 addr, u8 byte)
{
	return __dptx_write_dpcd(dptx, addr, byte);
}

static inline int dptx_read_bytes_from_dpcd(struct dptx *dptx,
					    unsigned int reg_addr,
					    u8 *bytes, u32 len)
{
	return __dptx_read_bytes_from_dpcd(dptx, reg_addr, bytes, len);
}

static inline int dptx_write_bytes_to_dpcd(struct dptx *dptx,
					   unsigned int reg_addr,
					   u8 *bytes, u32 len)
{
	return __dptx_write_bytes_to_dpcd(dptx, reg_addr, bytes, len);
}

#else
#define dptx_read_dpcd(_dptx, _addr, _byteptr) ({		\
	int _ret = __dptx_read_dpcd(_dptx, _addr, _byteptr);	\
	dptx_dbg(dptx, "%s: DPCD Read %s(0x%03x) = 0x%02x\n",	\
		 __func__, #_addr, _addr, *(_byteptr));		\
	_ret;							\
})

#define dptx_write_dpcd(_dptx, _addr, _byte) ({			\
	int _ret;						\
	dptx_dbg(dptx, "%s: DPCD Write %s(0x%03x) = 0x%02x\n",	\
		 __func__, #_addr, _addr, _byte);		\
	_ret = __dptx_write_dpcd(_dptx, _addr, _byte);		\
	_ret;							\
})

char *__bytes_str(u8 *bytes, unsigned int n);

#define dptx_read_bytes_from_dpcd(_dptx, _addr, _b, _len) ({	\
	int _ret;						\
	char *_str;						\
	_ret = __dptx_read_bytes_from_dpcd(_dptx,		\
					   _addr, _b, _len);	\
	_str = __bytes_str(_b, _len);				\
	dptx_dbg(dptx, "%s: Read %llu bytes from %s(0x%02x) = [ %s ]\n", \
		 __func__, (u64)_len, #_addr, _addr, _str);		\
	_ret;							\
})

#define dptx_write_bytes_to_dpcd(_dptx, _addr, _b, _len) ({	\
	int _ret;						\
	char *_str = __bytes_str(_b, _len);			\
	dptx_dbg(dptx, "%s: Writing %llu bytes to %s(0x%02x) = [ %s ]\n", \
		 __func__, (u64)_len, #_addr, _addr, _str);		\
	_ret = __dptx_write_bytes_to_dpcd(_dptx, _addr, _b, _len); \
	_ret;							\
})

#endif

/*
 * Link training
 */
int dptx_link_training(struct dptx *dptx, u8 rate, u8 lanes);
int dptx_link_check_status(struct dptx *dptx);

/*
 * Register read and write functions
 */
static inline u32 __dptx_readl(struct dptx *dp, char const *func,
	int line, u32 offset)
{
	u32 data = readl(dp->base + offset);
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: READ: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif

	return data;
}
static inline u32 __dpphy_readl(struct dptx *dp, char const *func,
	int line, u32 offset)
{
	u32 data = readl(dp->base + 0x20000 + offset);
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: dpphyREAD: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif

	return data;
}


static inline void __dptx_writel(struct dptx *dp, char const *func, int line,
	u32 offset, u32 data)
{
#ifdef DPTX_DEBUG_REG
	u32 rdata = 0;
	dptx_dbg(dp, "%s:%d: WRITE: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif
	writel(data, dp->base + offset);
#ifdef DPTX_DEBUG_REG
	rdata = __dptx_readl(dp, func, line, offset);
	if (rdata != data)
		gb_printf(KERN_ERR, "%s-%d:ERR,base:0x%p,addr:0x%x,data:0x%x,rdata:0x%x\n",
		 func, line, dp->base, offset, data, rdata);
#endif
}
static inline void __dpphy_writel(struct dptx *dp, char const *func, int line,
	u32 offset, u32 data)
{
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: ~~~~~~~dpphy WRITE: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base + 0x20000, offset, data);
#endif
	writel(data, dp->base + 0x20000 + offset);
}

#define dptx_readl(_dptx, _offset) ({ \
	__dptx_readl(_dptx, __func__, __LINE__, _offset); \
})
#define dpphy_readl(_dptx, _offset) ({ \
	__dpphy_readl(_dptx, __func__, __LINE__, _offset); \
})

#define dptx_writel(_dptx, _offset, _data) do { \
	__dptx_writel(_dptx, __func__, __LINE__, _offset, _data); \
} while (0)
#define dpphy_writel(_dptx, _offset, _data) do { \
	__dpphy_writel(_dptx, __func__, __LINE__, _offset, _data); \
} while (0)

/*
 * Wait functions
 */
#define dptx_wait(_dptx, _cond, _timeout)				\
	({								\
		int __retval;						\
		__retval = wait_event_interruptible_timeout(		\
			_dptx->waitq,					\
			((_cond) || (atomic_read(&_dptx->shutdown))),	\
			msecs_to_jiffies(_timeout));			\
		if (atomic_read(&_dptx->shutdown)) {			\
			__retval = -ESHUTDOWN;				\
		}							\
		else if (!__retval) {					\
			__retval = -ETIMEDOUT;				\
		}							\
		__retval;						\
	})
#define dptx_link_wait(_dptx, _cond, _timeout)				\
	({								\
		int __retval;						\
		__retval = wait_event_interruptible_timeout(		\
			_dptx->waitq,					\
			(_cond),	\
			msecs_to_jiffies(_timeout));			\
		__retval;						\
	})

void dptx_notify(struct dptx *dptx);
void dptx_notify_shutdown(struct dptx *dptx);

/* Type C */
void dptx_flip_logic(struct dptx *dptx);

/*
 * Debugfs
 */
void dptx_debugfs_init(struct dptx *dptx);
void dptx_debugfs_exit(struct dptx *dptx);

void dptx_fill_sdp(struct dptx *dptx, struct sdp_full_data *data);
int dptx_configure_hdcp13(struct dptx *dptx);
void dptx_en_dis_hdcp13(struct dptx *dptx, u8 enable);
void dptx_init_hdcp_keys(struct dptx *dptx);

int dptx_en_hdcp22(struct dptx *dptx);
void dptx_en_dis_hdcp22(struct dptx *dptx, u8 enable);

int dptx_en_mst(struct dptx *dptx, u8 enable);
int dptx_en_fec(struct dptx *dptx, u8 enable);
int dptx_en_dsc(struct dptx *dptx, u8 enable);

struct dptx *dptx_get_handle(void);

int dptx_aux_rw_bytes(struct dptx *dptx,
		      bool rw,
		      bool i2c,
		      u32 addr,
		      u8 *bytes,
		      unsigned int len);
int dptx_read_bytes_from_i2c(struct dptx *dptx,
			     u32 device_addr,
			     u8 *bytes,
			     u32 len);
int dptx_write_bytes_to_i2c(struct dptx *dptx,
			    u32 device_addr,
			    u8 *bytes,
			    u32 len);
int dptx_aux_rw(struct dptx *dptx,
			bool rw,
			bool i2c,
			bool mot,
			bool addr_only,
			u32 addr,
			u8 *bytes,
			unsigned int len);
static inline int dptx_video_pixel_freq_change(struct dptx *dptx, u32 frequency, int stream)
{
	return 0;
};
int dptx_audio_clock_freq_change(struct dptx *dptx, u32 audio_clock_freq,
	u16 oversample_factor);
/*
 * HDCP2.2
 */
void dptx_write_hdcp22_test_keys(struct dptx *dptx);

//#define DPTX_COMBO_PHY
// To enable audio during hotplug
void dptx_audio_sdp_en(struct dptx *dptx);

void dptx_audio_timestamp_sdp_en(struct dptx *dptx);
void dptx_link_set_preemp_vswing(struct dptx *dptx);
int dptx_link_cr(struct dptx *dptx);
int dptx_link_ch_eq(struct dptx *dptx);
/* EDID Audio Data Block */
#define AUDIO_TAG		1
#define VIDEO_TAG		2
#define EDID_TAG_MASK		GENMASK(7, 5)
#define EDID_TAG_SHIFT		5
#define EDID_SIZE_MASK		GENMASK(4, 0)
#define EDID_SIZE_SHIFT		0

/* Established timing blocks */
#define ET1_800x600_60hz	BIT(0)
#define ET1_800x600_56hz	BIT(1)
#define ET1_640x480_75hz	BIT(2)
#define ET1_640x480_72hz	BIT(3)
#define ET1_640x480_67hz	BIT(4)
#define ET1_640x480_60hz	BIT(5)
#define ET1_720x400_88hz	BIT(6)
#define ET1_720x400_70hz	BIT(7)

#define ET2_1280x1024_75hz	BIT(0)
#define ET2_1024x768_75hz	BIT(1)
#define ET2_1024x768_70hz	BIT(2)
#define ET2_1024x768_60hz	BIT(3)
#define ET2_1024x768_87hz	BIT(4)
#define ET2_832x624_75hz	BIT(5)
#define ET2_800x600_75hz	BIT(6)
#define ET2_800x600_72hz	BIT(7)

#define ET3_1152x870_75hz	BIT(7)

#endif
