/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/dma-mapping.h>
#include <linux/highmem.h>
#include <linux/spinlock.h>
#include <linux/shrinker.h>
#include <linux/atomic.h>
#include <linux/version.h>
#include <linux/irqreturn.h>
#include <linux/pci.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5, 12, 0))
#include <linux/timekeeping32.h>
#else
#include <linux/timekeeping.h>
#endif
#include <linux/vmalloc.h>
#include "gb_edma.h"
#include "edma_reg.h"
#include "gpu/gb_device.h"
#include "common/xt.h"
#include "common/gb_common.h"

static void gb02_chan_irq_clear(struct pcie_edma_chan *chan, u32 status)
{
    chan_reg_write32(chan, HDMA_INT_CLEAR_OFF, status);
}

static void gb02_hdmac_irq_clear(struct pcie_edma_chan *chan)
{
    struct hdma_ctrl_info *chip = chan->chip;

    hdma_reg_write(HDMAC_INT_CLEAR_BIT, chip->hdmac_regs, HDMAC_INT_CLEAR_OFFSET);;
}

irqreturn_t gb_edma_irq_handler(int irq, void *arg)
{
    struct hdma_ctrl_info *dma_ctrl = (struct hdma_ctrl_info *)arg;
    u32 status = 0;
    u32 i = 0;
    struct pcie_edma_chan *chan;

    gb_printf(KERN_DEBUG, "%s:%d:enter.\n", __func__, __LINE__);

    /* Poll, clear and process every read chanel interrupt status */
    for (i = 0; i < GB_HDMA_MAX_CH; i++) {
        chan = &(dma_ctrl->rd_chan[i]);
        status = chan_reg_read32(chan, HDMA_INT_STATUS_OFF);
        /* gb_printf(KERN_INFO, "%s %u HDMA_IRQ status: 0x%04x\n",
            __func__, i, status); */

        gb02_chan_irq_clear(chan, status);
        if (status) {
            gb02_hdmac_irq_clear(chan);
        }

        if (status & HDMA_INT_STATUS_ERROR) {
            /* error */
            gb_printf(KERN_ERR, "HDMA read channel %u error : 0x%x\n", i,
                (status & HDMA_INT_STATUS_ERROR) >> 3);
        }
        if (status & HDMA_INT_STATUS_ABORT) {
            /* abort */
            gb_printf(KERN_ERR, "HDMA read channel %u abort\n", i);
        }
        if (status & HDMA_INT_STATUS_WATERMARK) {
            /* watermark */
            gb_printf(KERN_ERR, "HDMA read channel %u watermark\n", i);
        }
        if (status & HDMA_INT_STATUS_STOP) {
            /* done */
           gb_printf(KERN_ERR, "HDMA read channel %u done\n", i);
        }
    }

    for (i = 0; i < GB_HDMA_MAX_CH; i++) {
        chan = &(dma_ctrl->wr_chan[i]);
        status = chan_reg_read32(chan, HDMA_INT_STATUS_OFF);
        /* gb_printf(KERN_INFO, "%s %u HDMA_IRQ status: 0x%04x\n",
            __func__, i, status); */

        gb02_chan_irq_clear(chan, status);
        if (status) {
            gb02_hdmac_irq_clear(chan);
        }

        if (status & HDMA_INT_STATUS_ERROR) {
            /* error */
            gb_printf(KERN_ERR, "HDMA write channel %u error : 0x%x\n", i,
                (status & HDMA_INT_STATUS_ERROR) >> 3);
        }
        if (status & HDMA_INT_STATUS_ABORT) {
            /* abort */
            gb_printf(KERN_ERR, "HDMA write channel %u abort\n", i);
        }
        if (status & HDMA_INT_STATUS_WATERMARK) {
            /* watermark */
            gb_printf(KERN_ERR, "HDMA write channel %u watermark\n", i);
        }
        if (status & HDMA_INT_STATUS_STOP) {
            /* done */
           gb_printf(KERN_ERR, "HDMA write channel %u done\n", i);
        }
    }

	/* TODO: send event to UMD */
    return IRQ_HANDLED;
}

static void __iomem *gb_edma_alloc_ll_table(struct gb_pcie_info *pcie_info,
	unsigned long num_dma_entry, phys_addr_t *ll_table_addr)
{
	unsigned long bitmap_no;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3, t4;
#else
	struct timespec64 t1, t2, t3, t4;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2, time_use3;
#endif
	void __iomem *ll_va = NULL;
	int ddr_bar_id;
	int bitmap_cnt;
	unsigned long *bitmap;
	unsigned long bitmap_maxno;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	ddr_bar_id = pcie_info->gb_dev_info->ddr_bar_id;
	bitmap_cnt = 1 + num_dma_entry *
		DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32 / GPU_PAGE_SIZE;
	bitmap = pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap;
	bitmap_maxno = pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap_maxno;

	mutex_lock(&pcie_info->edma_para.ll_bitmap.lock);
	bitmap_no = bitmap_find_next_zero_area(bitmap,
					bitmap_maxno,
					0, bitmap_cnt, 0);

	//print(KERN_INFO "lldebug max:%lu, no:%lu, num_dma_entry:%lu, cnt:%d\n",
	//		bitmap_maxno, bitmap_no, num_dma_entry, bitmap_cnt);
	if (bitmap_no >= bitmap_maxno) {
			mutex_unlock(&pcie_info->edma_para.ll_bitmap.lock);
			return NULL;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	bitmap_set(bitmap, bitmap_no, bitmap_cnt);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif
	mutex_unlock(&pcie_info->edma_para.ll_bitmap.lock);

	*ll_table_addr = LL_TABLE_START_ADDR +
		(bitmap_no << GPU_PAGE_SHIFT);
	ll_va = (void *)pcie_info->pci_bars[ddr_bar_id].mmio +
		LL_TABLE_START_ADDR +
		(bitmap_no << GPU_PAGE_SHIFT);
#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
#endif
	memset_io(ll_va, 0, GPU_PAGE_SIZE * bitmap_cnt);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t4);
#else
	ktime_get_real_ts64(&t4);
#endif
#ifdef GB_TIME_DEBUG
	time_use3=(t4.tv_sec-t3.tv_sec)*1000000+(t4.tv_nsec-t3.tv_nsec)/1000;
	/* gb_printf(KERN_INFO, "%s cost:1:%ld 2:%ld 3:%ld\n",
		__func__, time_use1, time_use2, time_use3); */
#endif
	return ll_va;
}

static void gb_edma_free_ll_table(struct gb_pcie_info *pcie_info,
	unsigned long num_dma_entry, phys_addr_t ll_table_addr)
{
	int bitmap_cnt = 1 + num_dma_entry *
		DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32 / GPU_PAGE_SIZE;
	unsigned long *bitmap =
		pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap;
	unsigned long bitmap_no = (ll_table_addr -
		LL_TABLE_START_ADDR) >>
		GPU_PAGE_SHIFT;

	mutex_lock(&pcie_info->edma_para.ll_bitmap.lock);
	bitmap_clear(bitmap, bitmap_no, bitmap_cnt);
	mutex_unlock(&pcie_info->edma_para.ll_bitmap.lock);
}

static int gb_edma_get_free_channel(struct gb_pcie_info *pcie_info,
	bool dma_write)
{
	int i = 0;
	int timeout = 0;
	unsigned int *channel_usage;
	unsigned int channel_status = 0;
	unsigned int int_status = 0;
	struct hdma_ctrl_info *dma_ctrl = pcie_info->edma_para.dma_ctrl;
	struct pcie_edma_chan *chan = NULL;

	if (dma_write) {
		chan = dma_ctrl->wr_chan;
		channel_usage = &pcie_info->edma_para.wr_channel_usage;
	} else {
		chan = dma_ctrl->rd_chan;
		channel_usage = &pcie_info->edma_para.rd_channel_usage;
	}

	do {
		for (i = 0; i < GB_HDMA_MAX_CH; i++) {
			channel_status = chan_reg_read32(&chan[i], HDMA_STATUS_OFF);
			channel_status &= HDMA_STATUS_MASK_BITS;
			int_status = chan_reg_read32(&chan[i], HDMA_INT_STATUS_OFF);

			if ((channel_status == HDMA_CH_RESERVED) ||
				(channel_status == HDMA_CH_STOPPED) ||
				!(int_status & (HDMA_INT_STATUS_ABORT | HDMA_INT_STATUS_ERROR))) {
				/*
				 * If channel i was occupied by other process,
				 * once we get this channel, there is at least
				 * 1 process will lose channel i stop sign, this
				 * process would assume DMA timeout, something
				 * goes wrong.
				 */
				if (!(*channel_usage & (1 << i))) {
					*channel_usage |= (1 << i);
					return i;
				}
			}
		}

		usleep_range(50, 100);
		timeout++;
	} while (timeout < 5000);

	return GB_HDMA_MAX_CH;
}

void gb_edma_reg_dumps(struct pcie_edma_chan *chan, bool dma_write)
{
	gb_printf(KERN_ERR, "====== HDMA %s REGISTER DUMP =====\n", dma_write? "WRCH" : "RDCH");
	gb_printf(KERN_ERR, "HDMA_LLP_LOW_OFF:    0x%08x\n", chan_reg_read32(chan, HDMA_LLP_LOW_OFF));
	gb_printf(KERN_ERR, "HDMA_LLP_HIGH_OFF:   0x%08x\n", chan_reg_read32(chan, HDMA_LLP_HIGH_OFF));
	gb_printf(KERN_ERR, "HDMA_XFERSIZE_OFF:   0x%08x\n", chan_reg_read32(chan, HDMA_XFERSIZE_OFF));
	gb_printf(KERN_ERR, "HDMA_SAR_LOW_OFF:    0x%08x\n", chan_reg_read32(chan, HDMA_SAR_LOW_OFF));
	gb_printf(KERN_ERR, "HDMA_SAR_HIGH_OFF:   0x%08x\n", chan_reg_read32(chan, HDMA_SAR_HIGH_OFF));
	gb_printf(KERN_ERR, "HDMA_DAR_LOW_OFF:    0x%08x\n", chan_reg_read32(chan, HDMA_DAR_LOW_OFF));
	gb_printf(KERN_ERR, "HDMA_DAR_HIGH_OFF:   0x%08x\n", chan_reg_read32(chan, HDMA_DAR_HIGH_OFF));
	gb_printf(KERN_ERR, "HDMA_CONTROL1_OFF:   0x%08x\n", chan_reg_read32(chan, HDMA_CONTROL1_OFF));
	gb_printf(KERN_ERR, "HDMA_STATUS_OFF:     0x%08x\n", chan_reg_read32(chan, HDMA_STATUS_OFF));
	gb_printf(KERN_ERR, "HDMA_INT_STATUS_OFF: 0x%08x\n", chan_reg_read32(chan, HDMA_INT_STATUS_OFF));
	gb_printf(KERN_ERR, "=====================================\n");
}

static int gb_edma_get_dma_done(struct gb_pcie_info *pcie_info,
	int channel_id, bool dma_write)
{
	int timeout = 0;
	unsigned int channel_status = 0;
	struct hdma_ctrl_info *dma_ctrl = pcie_info->edma_para.dma_ctrl;
	struct pcie_edma_chan *chan = NULL;
	u32 status = 0;

	if (dma_write)
		chan = &(dma_ctrl->wr_chan[channel_id]);
	else
		chan = &(dma_ctrl->rd_chan[channel_id]);

	do {
		channel_status = chan_reg_read32(chan, HDMA_STATUS_OFF);
		channel_status &= HDMA_STATUS_MASK_BITS;
		if (channel_status == HDMA_CH_STOPPED)
			return 0;

		if (channel_status == HDMA_CH_ABORT) {
			gb_printf(KERN_ERR, "%s %d channel_id:%d dma trans aborted!\n",
				__func__, __LINE__, channel_id);
			goto err_handle;
		}

		usleep_range(50, 100);
		timeout++;
	} while (timeout < 5000);

	gb_printf(KERN_ERR, "%s %d channel_id:%d dma trans timeout!\n",
		__func__, __LINE__, channel_id);

err_handle:
	gb_edma_reg_dumps(chan, dma_write);
	status = chan_reg_read32(chan, HDMA_INT_STATUS_OFF);
	gb02_chan_irq_clear(chan, status);
	if (status)
		gb02_hdmac_irq_clear(chan);

	return -1;
}

static int gb_edma_ll_mode_start(struct gb_pcie_info *pcie_info,
	struct dma_trans_info *dma_info)
{
	unsigned int value = 0, i;
	unsigned int channel_id = 0;
	int ret = 0;
	u64 ll_addr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3, t4, t5;
#else
	struct timespec64 t1, t2, t3, t4, t5;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2, time_use3, time_use4;
#endif
	bool dma_write;
	struct pcie_edma_chan *chan = NULL;
	struct hdma_ctrl_info *dma_ctrl = pcie_info->edma_para.dma_ctrl;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	dma_write = dma_info->dma_dir == GB_DMA_FROM_DEVICE ? true : false;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	mutex_lock(dma_write ? &pcie_info->edma_para.channel_wr_lock :
				&pcie_info->edma_para.channel_rd_lock);
	channel_id = gb_edma_get_free_channel(pcie_info, dma_write);
	if (channel_id >= GB_HDMA_MAX_CH) {
		mutex_unlock(dma_write ? &pcie_info->edma_para.channel_wr_lock :
						&pcie_info->edma_para.channel_rd_lock);
		gb_err(&pcie_info->pdev->dev, "%s %d dma get free channel failed!\n",
			__func__, __LINE__);
		return -1;
	}
	ll_addr = 0x800000000 + dma_info->ll_table_addr;
	/* HDMAC interrupt pass through */
	value = hdma_reg_read(dma_ctrl->hdmac_regs, HDMAC_INT_MODE_OFFSET);
	value |= HDMAC_INT_MODE_HDMA_INT_PASS_THROUGH_BIT;
	hdma_reg_write(value, dma_ctrl->hdmac_regs, HDMAC_INT_MODE_OFFSET);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif
	if (dma_write) {
		/* DMA write channel */
		chan = &(dma_ctrl->wr_chan[channel_id]);
		/* DMA Write Interrupt Enable */
		value = chan_reg_read32(chan, HDMA_INT_SETUP_OFF);
		value &= ~(HDMA_INT_SETUP_STOP_MASK_BIT | HDMA_INT_SETUP_ABORT_MASK_BIT);
		value &= ~(HDMA_INT_SETUP_RSIE_BIT | HDMA_INT_SETUP_RAIE_BIT);
		value |= (HDMA_INT_SETUP_LSIE_BIT | HDMA_INT_SETUP_LAIE_BIT);
		chan_reg_write32(chan, HDMA_INT_SETUP_OFF, value);

		/* DMA Write Channel Enable */
		chan_reg_write32(chan, HDMA_EN_OFF, HDMA_EN_ENABLE_BIT);

		/* DMA Channel Control 1 */
		chan_reg_write32(chan, HDMA_CONTROL1_OFF, HDMA_CONTROL1_LLEN_BIT);
		chan_reg_write32(chan, HDMA_CYCLE_OFF, HDMA_CYCLE_STATE_BIT);

		/* DMA Linked List Pointer */
		chan_reg_write32(chan, HDMA_LLP_LOW_OFF, lower_32_bits(ll_addr));
		chan_reg_write32(chan, HDMA_LLP_HIGH_OFF, upper_32_bits(ll_addr));

		/* DMA Write Doorbell */
		chan_reg_write32(chan, HDMA_DOORBELL_OFF, HDMA_DOORBELL_START_BIT);
	} else {
		/* DMA read channel */
		chan = &(dma_ctrl->rd_chan[channel_id]);
		/* DMA read Interrupt Enable */
		value = chan_reg_read32(chan, HDMA_INT_SETUP_OFF);
		value &= ~(HDMA_INT_SETUP_STOP_MASK_BIT | HDMA_INT_SETUP_ABORT_MASK_BIT);
		value &= ~(HDMA_INT_SETUP_RSIE_BIT | HDMA_INT_SETUP_RAIE_BIT);
		value |= (HDMA_INT_SETUP_LSIE_BIT | HDMA_INT_SETUP_LAIE_BIT);
		chan_reg_write32(chan, HDMA_INT_SETUP_OFF, value);

		/* DMA Read Channel Enable */
		chan_reg_write32(chan, HDMA_EN_OFF, HDMA_EN_ENABLE_BIT);

		/* DMA Channel Control 1 */
		chan_reg_write32(chan, HDMA_CONTROL1_OFF, HDMA_CONTROL1_LLEN_BIT);
		chan_reg_write32(chan, HDMA_CYCLE_OFF, HDMA_CYCLE_STATE_BIT);

		/* DMA Linked List Pointer */
		chan_reg_write32(chan, HDMA_LLP_LOW_OFF, lower_32_bits(ll_addr));
		chan_reg_write32(chan, HDMA_LLP_HIGH_OFF, upper_32_bits(ll_addr));

		/* DMA Read Doorbell */
		chan_reg_write32(chan, HDMA_DOORBELL_OFF, HDMA_DOORBELL_START_BIT);
	}

	mutex_unlock(dma_write ? &pcie_info->edma_para.channel_wr_lock :
				 &pcie_info->edma_para.channel_rd_lock);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t4);
#else
	ktime_get_real_ts64(&t4);
#endif
	ret = gb_edma_get_dma_done(pcie_info, channel_id, dma_write);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t5);
#else
	ktime_get_real_ts64(&t5);
#endif
#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
	time_use3 = (t4.tv_sec - t3.tv_sec) * 1000000 +
		(t4.tv_nsec - t3.tv_nsec)/1000;
	time_use4 = (t5.tv_sec - t4.tv_sec) * 1000000 +
		(t5.tv_nsec - t4.tv_nsec)/1000;
#endif
	//clear channel usage since CPU got dma done status
	mutex_lock(dma_write ? &pcie_info->edma_para.channel_wr_lock :
				 &pcie_info->edma_para.channel_rd_lock);
	if (dma_write)
		pcie_info->edma_para.wr_channel_usage &= (0 << channel_id);
	else
		pcie_info->edma_para.rd_channel_usage &= (0 << channel_id);
	mutex_unlock(dma_write ? &pcie_info->edma_para.channel_wr_lock :
				 &pcie_info->edma_para.channel_rd_lock);

	/* DMA Write Channel Disable */
	chan_reg_write32(chan, HDMA_EN_OFF, ~HDMA_EN_ENABLE_BIT);
#ifdef GB_TIME_DEBUG
	gb_printf(KERN_INFO, "%s, get channel time:%ld per register time:%ld  trans time: %ld %s\n", __func__, time_use2, time_use3, time_use4, dma_write ? "aaa" : "bbb");
#endif
	gb_edma_free_ll_table(pcie_info, dma_info->num_dma_entry,
		dma_info->ll_table_addr);
	for (i = 0; i < dma_info->num_dma_pages; i++) {
		dma_unmap_page(&pcie_info->pdev->dev, dma_info->dma_buffer[i],
			PAGE_SIZE, !dma_info->dma_dir);
	}
	vfree(dma_info->dma_buffer);
	return ret;
}

static void gb_edma_gpu_set_ll_read_table(void __iomem *ll_table_addr,
	struct gb_pcie_info *pcie_info, struct dma_trans_info *dma_info)
{
	int offset_base, offset, i, j;
	dma_addr_t src_addr, src_addr_base;
	phys_addr_t dst_addr, dst_addr_base;
	int last_page_num = dma_info->num_dma_pages - 1;
	u64 table_size = dma_info->num_dma_pages * PAGE_SIZE / LL_READ_STEP *
		DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
	void *table_addr = kvmalloc(table_size, GFP_KERNEL);
	memset(table_addr, 0x0, table_size);

	if (!table_addr)
		table_addr = ll_table_addr;
	dma_info->dma_buffer = vmalloc(sizeof(dma_addr_t) * dma_info->num_dma_pages);
	for (i = 0; i < dma_info->num_dma_pages; i++) {
		src_addr_base = dma_map_page(&pcie_info->pdev->dev,
			dma_info->cpu_pages[i],
			0, PAGE_SIZE, DMA_TO_DEVICE);
		dma_info->dma_buffer[i] = src_addr_base;
		dst_addr_base = 0x800000000 + dma_info->gpu_phys[PG_SIZE_MULTI * i];
		offset_base = i * PAGE_SIZE / LL_READ_STEP *
			DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;

		for (j = 0; j < PAGE_SIZE / LL_READ_STEP; j++) {
			src_addr = src_addr_base + j * LL_READ_STEP;
			dst_addr = dst_addr_base + j * LL_READ_STEP;
			offset = offset_base + j *
				DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
			/* enable interrupt for the last element */
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_CTRL_OFF) =
				((i == (dma_info->num_dma_pages - 1)) && (j == (PAGE_SIZE / LL_READ_STEP - 1))) ?
				(DMA_LL_MODE_ELEMENT_CTRL_CB | DMA_LL_MODE_ELEMENT_CTRL_LIE) :
					HDMA_CYCLE_BIT_BIT;
			if ((i == last_page_num) && (dma_info->last_page_size != 0))
				*(u32 *)(table_addr + offset +
					DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
					dma_info->last_page_size;
			else
				*(u32 *)(table_addr + offset +
					DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
					LL_READ_STEP;
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_SAR_LOW_OFF) =
				lower_32_bits(src_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_SAR_HIGH_OFF) =
				upper_32_bits(src_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_DAR_LOW_OFF) =
				lower_32_bits(dst_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_DAR_HIGH_OFF) =
				upper_32_bits(dst_addr);
		}
	}

	if (table_addr != ll_table_addr) {
		memcpy_toio(ll_table_addr, table_addr, table_size);
		kvfree(table_addr);
	}
}

/* VRAM ---> DDR */
static void gb_edma_gpu_set_ll_write_table(void __iomem *ll_table_addr,
	struct gb_pcie_info *pcie_info, struct dma_trans_info *dma_info)
{
	int offset_base, offset, i, j;
	dma_addr_t src_addr, src_addr_base;
	phys_addr_t dst_addr, dst_addr_base;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3;
#else
	struct timespec64 t1, t2, t3;
#endif
	int last_page_num = dma_info->num_dma_pages - 1;
	u64 table_size = dma_info->num_dma_pages * PAGE_SIZE / LL_WRITE_STEP *
		DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
	void *table_addr = kvmalloc(table_size, GFP_KERNEL);
	memset(table_addr, 0x0, table_size);

#ifdef GB_TIME_DEBUG
	long time_use1, time_use2;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	if (!table_addr)
		table_addr = ll_table_addr;
	dma_info->dma_buffer = vmalloc(sizeof(dma_addr_t) * dma_info->num_dma_pages);
	for (i = 0; i < dma_info->num_dma_pages; i++) {
		src_addr_base = 0x800000000 + dma_info->gpu_phys[PG_SIZE_MULTI * i];
		dst_addr_base = dma_map_page(&pcie_info->pdev->dev,
			dma_info->cpu_pages[i],
			0, PAGE_SIZE, DMA_FROM_DEVICE);
		dma_info->dma_buffer[i] = dst_addr_base;
		offset_base = i * PAGE_SIZE / LL_WRITE_STEP *
			DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;

		for (j = 0; j < PAGE_SIZE / LL_WRITE_STEP; j++) {
			src_addr = src_addr_base + j * LL_WRITE_STEP;
			dst_addr = dst_addr_base + j * LL_WRITE_STEP;
			offset = offset_base + j *
				DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;

			/* enable interrupt for the last element */
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_CTRL_OFF) =
				((i == (dma_info->num_dma_pages - 1)) && (j == (PAGE_SIZE / LL_WRITE_STEP - 1))) ?
				(DMA_LL_MODE_ELEMENT_CTRL_CB | DMA_LL_MODE_ELEMENT_CTRL_LIE) :
					HDMA_CYCLE_BIT_BIT;
			if ((i == last_page_num) && (dma_info->last_page_size != 0))
				*(u32 *)(table_addr + offset +
					DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
					dma_info->last_page_size;
			else
				*(u32 *)(table_addr + offset +
					DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
					LL_WRITE_STEP;
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_SAR_LOW_OFF) =
				lower_32_bits(src_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_SAR_HIGH_OFF) =
				upper_32_bits(src_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_DAR_LOW_OFF) =
				lower_32_bits(dst_addr);
			*(u32 *)(table_addr + offset +
				DMA_LL_MODE_ELEMENT_DAR_HIGH_OFF) =
				upper_32_bits(dst_addr);
		}
	}
#ifdef GB_TIME_DEBUG
	for (i = 0; i < table_size / 24 ; i++) {
			gb_printf(KERN_INFO, "flag %lx", ((u32 *)table_addr)[i * 6]);
			gb_printf(KERN_INFO, "szie %lx", ((u32 *)table_addr)[i * 6 + 1]);
			gb_printf(KERN_INFO, "sar1 %lx", ((u32 *)table_addr)[i * 6 + 2]);
			gb_printf(KERN_INFO, "sar2 %lx", ((u32 *)table_addr)[i * 6 + 3]);
			gb_printf(KERN_INFO, "dar1 %lx", ((u32 *)table_addr)[i * 6 + 4]);
			gb_printf(KERN_INFO, "dar2 %lx", ((u32 *)table_addr)[i * 6 + 5]);
	}
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	if (table_addr != ll_table_addr) {
		memcpy_toio(ll_table_addr, table_addr, table_size);
		kvfree(table_addr);
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif
#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
	/* gb_printf(KERN_INFO, "%s cost:1:%ld 2:%ld\n",
		__func__, time_use1, time_use2); */
#endif
}

static int gb_edma_gpu_ll_mode_prepare(struct gb_pcie_info *pcie_info,
	struct dma_trans_info *dma_info)
{
	void __iomem *base_addr;
	phys_addr_t ll_table_addr = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3;
#else
	struct timespec64 t1, t2, t3;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2;
#endif
	unsigned long num_dma_entry;

	if (dma_info->dma_dir == GB_DMA_TO_DEVICE) {
		num_dma_entry = dma_info->num_dma_pages *
			PAGE_SIZE / LL_READ_STEP;
		num_dma_entry++;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t1);
#else
		ktime_get_real_ts64(&t1);
#endif
		base_addr = gb_edma_alloc_ll_table(pcie_info,
			num_dma_entry, &ll_table_addr);
		if (!base_addr) {
			gb_printf(KERN_ERR, "%s %d alloc ll table failed!\n", __func__, __LINE__);
			return -1;
		}
		dma_info->num_dma_entry = num_dma_entry;
		dma_info->ll_table_addr = ll_table_addr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t2);
#else
		ktime_get_real_ts64(&t2);
#endif
		gb_edma_gpu_set_ll_read_table(base_addr, pcie_info, dma_info);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t3);
#else
		ktime_get_real_ts64(&t3);
#endif
	} else {
		num_dma_entry = dma_info->num_dma_pages * PAGE_SIZE / LL_WRITE_STEP;
		num_dma_entry++;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t1);
#else
		ktime_get_real_ts64(&t1);
#endif
		base_addr = gb_edma_alloc_ll_table(pcie_info, num_dma_entry, &ll_table_addr);
		if (!base_addr) {
			gb_printf(KERN_ERR, "%s %d alloc ll table failed!\n", __func__, __LINE__);
			return -1;
		}
		dma_info->num_dma_entry = num_dma_entry;
		dma_info->ll_table_addr = ll_table_addr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t2);
#else
		ktime_get_real_ts64(&t2);
#endif
		gb_edma_gpu_set_ll_write_table(base_addr, pcie_info, dma_info);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
		getnstimeofday(&t3);
#else
		ktime_get_real_ts64(&t3);
#endif
	}

#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
	/* gb_printf(KERN_INFO, "%s cost:1:%ld 2:%ld\n",
		__func__, time_use1, time_use2); */
#endif
	return 0;
}

int gb_edma_gpu_start(struct device *dev,
	struct dma_trans_info *dma_info,
	struct gb_pcie_info *pcie_info)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3;
#else
	struct timespec64 t1, t2, t3;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	/* first, prepare the memory of SAR */
	if (gb_edma_gpu_ll_mode_prepare(pcie_info, dma_info)) {
		gb_err(dev, "%s %d dma prepare failed!\n",
			__func__, __LINE__);
		return -1;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	/* second, start DMA transferring */
	mutex_lock(&pcie_info->edma_para.dma_lock);
	if (gb_edma_ll_mode_start(pcie_info, dma_info)) {
		mutex_unlock(&pcie_info->edma_para.dma_lock);
		gb_err(dev, "%s %d dma transfer failed!\n",
			__func__, __LINE__);
		return -1;
	}
	mutex_unlock(&pcie_info->edma_para.dma_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif
#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
gb_printf(KERN_INFO, "%s cost gb_edma_gpu_ll_mode_prepare:%ld  gb_edma_ll_mode_start:%ld size %lld\n",
                __func__, time_use1, time_use2, (long long)(dma_info->num_dma_pages * 4096));
#endif
	return 0;
}

static void gb_edma_drm_set_ll_read_table(void __iomem *ll_table_va,
	phys_addr_t ll_table_pa,
	struct gb_pcie_info *pcie_info,
	struct genbu_dma_trans_to_fb *dma_info,
	struct page **user_pages, phys_addr_t gpu_phy)
{
	int i, j, ll_cnt;
	dma_addr_t src_addr;
	phys_addr_t dst_addr, dst_addr_base;
	u64 cpu_va_row_start, row_pg_offset, ll_offset, ll_offset_base = 0;
	u32 width_sz, height_sz, plane_width_sz, pg_num;
	u32 size, cur_size, ll_total_cnt = 0;
	u64 table_size = dma_info->num_dma_entry *
		DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
	void *table_addr = kzalloc(table_size, GFP_KERNEL);

	if (!table_addr)
		table_addr = ll_table_va;

	width_sz = 4 * dma_info->width;
	height_sz = 4 * dma_info->height;
	plane_width_sz = 4 * dma_info->plane_width;

	for (i = 0; i < dma_info->height; i++) {
		cpu_va_row_start = dma_info->cpu_va + i * plane_width_sz;
		row_pg_offset = (cpu_va_row_start & (PAGE_SIZE - 1));
		ll_cnt = DIV_ROUND_UP(row_pg_offset + width_sz, PAGE_SIZE);
		pg_num = (cpu_va_row_start - (dma_info->cpu_va & PAGE_MASK)) /
					PAGE_SIZE;

		src_addr = row_pg_offset + dma_map_page(&pcie_info->pdev->dev,
			user_pages[pg_num], 0, PAGE_SIZE, DMA_TO_DEVICE);
		dst_addr = 0x800000000 + gpu_phy + i * plane_width_sz;
		ll_offset = ll_offset_base;

		*(u32 *)(table_addr + ll_offset +
			DMA_LL_MODE_ELEMENT_CTRL_OFF) =
			DMA_LL_MODE_ELEMENT_CTRL_CB;
		*(u32 *)(table_addr + ll_offset +
			DMA_LL_MODE_ELEMENT_SAR_LOW_OFF) =
			lower_32_bits(src_addr);
		*(u32 *)(table_addr + ll_offset +
			DMA_LL_MODE_ELEMENT_SAR_HIGH_OFF) =
			upper_32_bits(src_addr);
		*(u32 *)(table_addr + ll_offset +
			DMA_LL_MODE_ELEMENT_DAR_LOW_OFF) =
			lower_32_bits(dst_addr);
		*(u32 *)(table_addr + ll_offset +
			DMA_LL_MODE_ELEMENT_DAR_HIGH_OFF) =
			upper_32_bits(dst_addr);
		if (ll_cnt == 1) {
			*(u32 *)(table_addr + ll_offset +
				DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) = width_sz;
			/* enable interrupt for the last element */
			*(u32 *)(table_addr + ll_offset +
				DMA_LL_MODE_ELEMENT_CTRL_OFF) =
				(i == (dma_info->height - 1)) ?
				(DMA_LL_MODE_ELEMENT_CTRL_CB | DMA_LL_MODE_ELEMENT_CTRL_LIE) :
				HDMA_CYCLE_BIT_BIT;
		} else {
			*(u32 *)(table_addr + ll_offset +
				DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
				PAGE_SIZE - row_pg_offset;
			size = width_sz - (PAGE_SIZE - row_pg_offset);
			dst_addr_base = dst_addr + (PAGE_SIZE - row_pg_offset);
			for (j = 1; j < ll_cnt; j++) {
				cur_size = size;
				if (size > PAGE_SIZE)
					cur_size = PAGE_SIZE;
				size -= cur_size;

				src_addr = dma_map_page(pcie_info->gbdev->dev,
					user_pages[pg_num + j], 0,
					PAGE_SIZE, DMA_TO_DEVICE);
				dst_addr = dst_addr_base + (j - 1) * PAGE_SIZE;
				ll_offset = ll_offset_base +
					j * DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
				/* enable interrupt for the last element */
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_CTRL_OFF) =
					((i == (dma_info->height - 1)) && (j == (ll_cnt - 1))) ?
					(DMA_LL_MODE_ELEMENT_CTRL_CB | DMA_LL_MODE_ELEMENT_CTRL_LIE) :
					HDMA_CYCLE_BIT_BIT;
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_XFER_SIZE_OFF) =
					cur_size;
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_SAR_LOW_OFF) =
					lower_32_bits(src_addr);
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_SAR_HIGH_OFF) =
					upper_32_bits(src_addr);
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_DAR_LOW_OFF) =
					lower_32_bits(dst_addr);
				*(u32 *)(table_addr + ll_offset +
					DMA_LL_MODE_ELEMENT_DAR_HIGH_OFF) =
					upper_32_bits(dst_addr);
			}
		}
		ll_total_cnt += ll_cnt;
		ll_offset_base += ll_cnt * DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32;
	}
/*
	for (i = 0; i < ll_total_cnt ; i++) {
			gb_printf(KERN_INFO, "flag %lx", ((u32 *)table_addr)[i * 6]);
			gb_printf(KERN_INFO, "szie %lx", ((u32 *)table_addr)[i * 6 + 1]);
			gb_printf(KERN_INFO, "sar1 %lx", ((u32 *)table_addr)[i * 6 + 2]);
			gb_printf(KERN_INFO, "sar2 %lx", ((u32 *)table_addr)[i * 6 + 3]);
			gb_printf(KERN_INFO, "dar1 %lx", ((u32 *)table_addr)[i * 6 + 4]);
			gb_printf(KERN_INFO, "dar2 %lx", ((u32 *)table_addr)[i * 6 + 5]);
	}
*/
	if (table_addr != ll_table_va) {
		memcpy_toio(ll_table_va, table_addr,
			ll_total_cnt * DWC_PCIE_DMA_LL_ELEMENT_LENGTH_U32);
		kfree(table_addr);
	}
}

static int gb_edma_drm_ll_mode_prepare(struct gb_pcie_info *pcie_info,
	struct genbu_dma_trans_to_fb *dma_info,
	struct page **user_pages, phys_addr_t gpu_phy)
{
	void __iomem *ll_table_va;
	phys_addr_t ll_table_pa = 0;
	u32 entry_num_one_row;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3;
#else
	struct timespec64 t1, t2, t3;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2;
#endif
	unsigned long num_dma_entry;

	entry_num_one_row = DIV_ROUND_UP(4 * dma_info->width, PAGE_SIZE) + 1;
	num_dma_entry = dma_info->height * entry_num_one_row + 1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	ll_table_va = gb_edma_alloc_ll_table(pcie_info, num_dma_entry, &ll_table_pa);
	if (!ll_table_va) {
		gb_printf(KERN_ERR, "%s %d alloc ll table failed!\n", __func__, __LINE__);
		return -1;
	}

	dma_info->num_dma_entry = num_dma_entry;
	dma_info->ll_table_addr = ll_table_pa;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	gb_edma_drm_set_ll_read_table(ll_table_va, ll_table_pa,
		pcie_info, dma_info, user_pages, gpu_phy);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif

#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
	/* gb_printf(KERN_INFO, "%s cost:1:%ld 2:%ld\n",
		__func__, time_use1, time_use2); */
#endif
	return 0;
}

int gb_edma_drm_start(struct device *dev,
	struct genbu_dma_trans_to_fb *dma_info,
	phys_addr_t gpu_phy,
	struct page **user_pages,
	struct gb_pcie_info *pcie_info)
{
	struct dma_trans_info dma_start_info;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	struct timespec t1, t2, t3;
#else
	struct timespec64 t1, t2, t3;
#endif
#ifdef GB_TIME_DEBUG
	long time_use1, time_use2;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t1);
#else
	ktime_get_real_ts64(&t1);
#endif
	/* first, prepare the memory of SAR */
	if (gb_edma_drm_ll_mode_prepare(pcie_info, dma_info, user_pages, gpu_phy)) {
		gb_err(pcie_info->gbdev->dev, "%s %d dma prepare failed!\n",
			__func__, __LINE__);
		return -1;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t2);
#else
	ktime_get_real_ts64(&t2);
#endif
	/* second, start DMA transferring */
	mutex_lock(&pcie_info->edma_para.dma_lock);
	dma_start_info.dma_dir = dma_info->dma_direction;
	dma_start_info.ll_table_addr = dma_info->ll_table_addr;
	dma_start_info.num_dma_entry = dma_info->num_dma_entry;

	if (gb_edma_ll_mode_start(pcie_info, &dma_start_info)) {
		mutex_unlock(&pcie_info->edma_para.dma_lock);
		gb_err(pcie_info->gbdev->dev, "%s %d dma transfer failed!\n",
			__func__, __LINE__);
		return -1;
	}
	mutex_unlock(&pcie_info->edma_para.dma_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
	getnstimeofday(&t3);
#else
	ktime_get_real_ts64(&t3);
#endif
#ifdef GB_TIME_DEBUG
	time_use1 = (t2.tv_sec - t1.tv_sec) * 1000000 +
		(t2.tv_nsec - t1.tv_nsec)/1000;
	time_use2 = (t3.tv_sec - t2.tv_sec) * 1000000 +
		(t3.tv_nsec - t2.tv_nsec)/1000;
	gb_printf(KERN_INFO, "%s gb_edma_drm_ll_mode_prepare:%ld gb_edma_ll_mode_start:%ld size %lld, %dx*%d\n",
                __func__, time_use1, time_use2, (long long)(dma_info->height * dma_info->width * 4), dma_info->width, dma_info->height);
#endif
	return 0;
}

struct page **gb_edma_get_user_cpu_pages(
	u64 cpu_va, enum gb_dma_data_direction dma_dir,
	unsigned long num_dma_pages)
{
	struct page **user_pages = NULL;
	unsigned long getted_num = 0;
	int i, cnt;

	user_pages = kvzalloc(sizeof(struct page *) * num_dma_pages,
			GFP_KERNEL);
	if (!user_pages) {
		gb_printf(KERN_ERR, "%s %d alloc dma %lu pages failed!\n",
			__func__, __LINE__, num_dma_pages);
		return NULL;
	}

	do {
		unsigned num_pages = num_dma_pages - getted_num;
		u64 userptr = cpu_va + getted_num * PAGE_SIZE;
		struct page **pages = user_pages + getted_num;

		cnt = get_user_pages_fast(userptr, num_pages,
				dma_dir == GB_DMA_FROM_DEVICE ? FOLL_WRITE : 0,
				pages);
		if (cnt < 0)
			goto err_release_pages;

		getted_num += cnt;

	} while (getted_num < num_dma_pages);

	return user_pages;

err_release_pages:
	for (i = 0; i < getted_num; i++)
		put_page(user_pages[i]);
	kvfree(user_pages);

	return NULL;
}

int pcie_hdma_init(struct gb_pcie_info *pcie_info)
{
	int result = 0;
	struct hdma_ctrl_info *dma_ctrl = NULL;
	struct pcie_edma_chan *chan = NULL;
	int hdmac_reg_bar_id = get_genbu_gpu_reg_bar_id(pcie_info->gb_dev_info);
	int hdma_reg_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);
	int i, ll_bitmap_sz;

	gb_printf(KERN_INFO, "Start to init PCIE HDMA:\n");
	dma_ctrl = (struct hdma_ctrl_info *)devm_kzalloc(&pcie_info->pdev->dev,
		sizeof(struct hdma_ctrl_info), GFP_KERNEL);
	if (!dma_ctrl) {
		gb_printf(KERN_ERR, "Error: %s, kmalloc failed\n", __func__);
		return -ENOMEM;
	}

	//init dma_channel_usgae
	pcie_info->edma_para.rd_channel_usage = 0;
	pcie_info->edma_para.wr_channel_usage = 0;

	dma_ctrl->pdev = pcie_info->pdev;
	dma_ctrl->hdmac_regs = pcie_info->pci_bars[hdmac_reg_bar_id].mmio + CS_S2_HDMAC_BASE;
	dma_ctrl->regs = pcie_info->pci_bars[hdma_reg_bar_id].mmio + EDMA_CFG_REG_BAR4_OFFSET;
	/* write channel */
	for (i = 0; i < GB_HDMA_MAX_CH; i++) {
		chan = &(dma_ctrl->wr_chan[i]);
		chan->id = i;
		chan->rd_chnl_flag = 0;
		chan->chan_regs = dma_ctrl->regs + COMMON_REG_LEN + i * CHAN_REG_LEN;
		chan->chip = dma_ctrl;
	}
	/* read channel */
	for (i = 0; i < GB_HDMA_MAX_CH; i++) {
		chan = &(dma_ctrl->rd_chan[i]);
		chan->id = i;
		chan->rd_chnl_flag = 1;
		chan->chan_regs = dma_ctrl->regs + COMMON_REG_LEN + i * CHAN_REG_LEN + WR_READ_CHAN_OFF;
		chan->chip = dma_ctrl;
	}

	dma_ctrl->ll_data_element_num_max = DWC_PCIE_DMA_LL_ELEMENT_NUM_MAX;
	dma_ctrl->ll_link_element_num = DWC_PCIE_DMA_LL_LINK_ELEMENT_NUM;

	pcie_info->edma_para.dma_ctrl = dma_ctrl;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
	result = dma_set_mask(&pcie_info->pdev->dev, DMA_BIT_MASK(64));
#else
	result = pci_set_dma_mask(pcie_info->pdev, DMA_BIT_MASK(64));
#endif
	if(result) {
		gb_printf(KERN_ERR, "%s:%d:No suitable PCI mapping available(%d).\n", __func__, __LINE__, result);
		goto err;
	}

	pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap_maxno =
		LL_TABLE_LEN >> GPU_PAGE_SHIFT;
	ll_bitmap_sz = BITS_TO_LONGS(LL_TABLE_LEN >> GPU_PAGE_SHIFT) *
		sizeof(long);
	pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap =
		devm_kzalloc(&pcie_info->pdev->dev, ll_bitmap_sz, GFP_KERNEL);

	if (!pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap) {
		gb_printf(KERN_ERR, "%s %d gb alloc 0x%lx dma_ll_mem_bitmap failed!\n",
			__func__, __LINE__,
			pcie_info->edma_para.ll_bitmap.dma_ll_mem_bitmap_maxno);
		goto err;
	}

	mutex_init(&pcie_info->edma_para.channel_rd_lock);
	mutex_init(&pcie_info->edma_para.channel_wr_lock);
	mutex_init(&pcie_info->edma_para.dma_lock);
	mutex_init(&pcie_info->edma_para.ll_bitmap.lock);

	return 0;
err:
	devm_kfree(&pcie_info->pdev->dev, dma_ctrl);
	return -1;
}
