/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include "gb_hdmac.h"
#include "gb_hdmac_regmap.h"
#include "common/gb_pcie_info.h"
#include "common/gb_common.h"

static DEFINE_MUTEX(hdmac_gpu_page_alloc_lock);
static DEFINE_MUTEX(hdmac_pgd_alloc_lock);
static DEFINE_MUTEX(hdmac_gpu_vpfn_alloc_lock);
static DEFINE_MUTEX(hdmac_cpu_vpfn_alloc_lock);
static DEFINE_MUTEX(hdmac_start_lock);
static DEFINE_MUTEX(hdmac_done_lock);
static DEFINE_MUTEX(hdmac_abort_clear_lock);
static DEFINE_MUTEX(hdmac_status_clear_lock);
static DEFINE_MUTEX(hdmac_int_status_lock);
u64 gb_hdmac_pgt_offset;
u64 gb_hdmac_gpu_page_offset;
u64 gb_hdmac_gpu_vpfn_offset;
u64 gb_hdmac_cpu_vpfn_offset;

extern struct hdmac_ctrl_info *dma_infos;
void gb_hdmac_reg_write(u32 offset, u32 value)
{
	iowrite32(value, dma_infos->regs + offset);
	hdmac_debug("writereg 0x%x 0x%x\n", offset, value);
	udelay(50);
}

u32 gb_hdmac_reg_read(u32 offset)
{
	u32 val = 0;

	val = ioread32(dma_infos->regs + offset);
	hdmac_debug("readreg 0x%x 0x%x\n", offset, val);
	return val;
}

void gb_hdmac_hdma_reg_write(u32 offset, u32 value)
{
	iowrite32(value, dma_infos->hdma_regs + offset);
	hdmac_debug("write hdmA reg:0x%x:0x%x\n", offset, value);
}

u32 gb_hdmac_hdma_reg_read(u32 offset)
{
	u32 val = 0;

	val = ioread32(dma_infos->hdma_regs + offset);
	hdmac_debug("read hdmA reg:0x%x:0x%x\n", offset, val);
	return val;
}

static inline u64 lock_region(u64 pfn,
		u32 num_pages)
{
	u64 region;

	/* can't lock a zero sized range */
	GB_DEBUG_ASSERT(num_pages);

	region = pfn << GPU_PAGE_SHIFT;
	/*
	 * fls returns (given the ASSERT above):
	 * 1 .. 32
	 *
	 * 10 + fls(num_pages)
	 * results in the range (11 .. 42)
	 */

	/* gracefully handle num_pages being zero */
	if (num_pages == 0)
		region |= 11;
	else {
		u8 region_width;

		region_width = 10 + fls(num_pages);
		if (num_pages != (1ul << (region_width - 11))) {
			/* not pow2, so must go up to the next pow2 */
			region_width += 1;
		}
		GB_DEBUG_ASSERT(region_width <= GB_LOCK_REGION_MAX_SIZE);
		GB_DEBUG_ASSERT(region_width >= GB_LOCK_REGION_MIN_SIZE);
		region |= region_width;
	}
	return region;
}

static int wait_ready(unsigned int as_nr)
{
	unsigned int max_loops = GB_AS_INACTIVE_MAX_LOOPS;
	u32 val = gb_hdmac_reg_read(MMU_AS_REG(as_nr, AS_STATUS));

	/* Wait for the MMU status to indicate there is no active command, in
	 * case one is pending. Do not log remaining register accesses.
	 */
	while (--max_loops && (val & AS_STATUS_AS_ACTIVE)) {
		val = gb_hdmac_reg_read(MMU_AS_REG(as_nr, AS_STATUS));
		msleep(500);
	}

	if (max_loops == 0)
		return -1;

	/* If waiting in loop was performed, log last read value. */
	if (GB_AS_INACTIVE_MAX_LOOPS - 1 > max_loops)
		gb_hdmac_reg_read(MMU_AS_REG(as_nr, AS_STATUS));

	return 0;
}

static int write_cmd(int as_nr, u32 cmd)
{
	int status;

	/* write AS_COMMAND when MMU is ready to accept another command */
	status = wait_ready(as_nr);
	if (status == 0)
		gb_hdmac_reg_write(MMU_AS_REG(as_nr, AS_COMMAND), cmd);
	else
		gb_printf(KERN_ERR, "%s %d BUG...\n", __func__, __LINE__);

	return status;
}

void gb_hdmac_mmu_hw_configure(struct gb_hdmac_as *as)
{
	struct gb_hdmac_mmu_setup *current_setup = &as->current_setup;
	u32 transcfg = 0;

	transcfg = current_setup->transcfg & 0xFFFFFFFFUL;

	/* Set flag AS_TRANSCFG_PTW_MEMATTR_WRITE_BACK */
	/* Clear PTW_MEMATTR bits */
	transcfg &= ~AS_TRANSCFG_PTW_MEMATTR_MASK;
	/* Enable correct PTW_MEMATTR bits */
	transcfg |= AS_TRANSCFG_PTW_MEMATTR_WRITE_BACK;

	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_TRANSCFG_LO),
			transcfg);
	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_TRANSCFG_HI),
			(current_setup->transcfg >> 32) & 0xFFFFFFFFUL);

	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_TRANSTAB_LO),
			current_setup->transtab & 0xFFFFFFFFUL);
	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_TRANSTAB_HI),
			(current_setup->transtab >> 32) & 0xFFFFFFFFUL);

	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_MEMATTR_LO),
			current_setup->memattr & 0xFFFFFFFFUL);
	gb_hdmac_reg_write(MMU_AS_REG(as->number, AS_MEMATTR_HI),
			(current_setup->memattr >> 32) & 0xFFFFFFFFUL);

	write_cmd(as->number, AS_COMMAND_UPDATE);
}

static int gb_hdmac_mmu_hw_do_operation(int as_nr, u64 vpfn, u32 nr,
	u32 op, unsigned int handling_irq)
{
	int ret;

	if (op == AS_COMMAND_UNLOCK)
		/* Unlock doesn't require a lock first */
		ret = write_cmd(as_nr, AS_COMMAND_UNLOCK);
	else {
		u64 lock_addr = lock_region(vpfn, nr);

		/* Lock the region that needs to be updated */
		gb_hdmac_reg_write(MMU_AS_REG(as_nr, AS_LOCKADDR_LO),
		 lock_addr & 0xFFFFFFFFUL);
		gb_hdmac_reg_write(MMU_AS_REG(as_nr, AS_LOCKADDR_HI),
		 (lock_addr >> 32) & 0xFFFFFFFFUL);
		write_cmd(as_nr, AS_COMMAND_LOCK);

		/* Run the MMU operation */
		write_cmd(as_nr, op);

		/* Wait for the flush to complete */
		ret = wait_ready(as_nr);
	}

	return ret;
}

u64 *gb_phys_to_virt(phys_addr_t phy_add)
{
	u64 *virt;

	virt = (u64 *)(dma_infos->vram_addr + phy_add);
	return virt;
}

phys_addr_t gb_hdmac_alloc_gpu_page(void)
{
	phys_addr_t target_page;

	mutex_lock(&hdmac_gpu_page_alloc_lock);
	target_page = HDMAC_PHY_OFFSET + gb_hdmac_gpu_page_offset;
	gb_hdmac_gpu_page_offset += (1 << GPU_PAGE_SHIFT);
	mutex_unlock(&hdmac_gpu_page_alloc_lock);

	return target_page;
}

void gb_hdmac_mmu_free_all_gpu_page(void)
{
	gb_hdmac_gpu_page_offset = 0;
}

phys_addr_t gb_hdmac_mmu_alloc_pgd(struct gb_hdmac_context *gbctx)
{
	u64 *page;
	phys_addr_t target_pgd;
	int i = 0;

	mutex_lock(&hdmac_pgd_alloc_lock);
	target_pgd = HDMAC_MMU_OFFSET + gb_hdmac_pgt_offset;
	gb_hdmac_pgt_offset += (1 << GPU_PAGE_SHIFT);

	page = gb_phys_to_virt(target_pgd);
	memset_io(page, 0, GPU_PAGE_SIZE);
	for (i = 0; i < GB_MMU_PAGE_ENTRIES; i++)
		gbctx->mmu_mode->entry_invalidate(&page[i]);
	mutex_unlock(&hdmac_pgd_alloc_lock);

	return target_pgd;
}

void gb_hdmac_mmu_free_all_pgd(void)
{
	gb_hdmac_pgt_offset = 0;
}

static int mmu_get_next_pgd(struct gb_hdmac_context *gbctx,
		phys_addr_t *pgd, u64 vpfn, int level)
{
	u64 *page;
	phys_addr_t target_pgd;

	/*
	 * Architecture spec defines level-0 as being the top-most.
	 * This is a bit unfortunate here, but we keep the same convention.
	 */
	vpfn >>= (3 - level) * 9;
	vpfn &= 0x1FF;

	page = gb_phys_to_virt(*pgd);
	if (page == NULL) {
		gb_printf(KERN_ERR, "%s: gb_phys_to_virt failure\n", __func__);
		return -EINVAL;
	}

	target_pgd = gbctx->mmu_mode->pte_to_phy_addr(page[vpfn]);

	if (!target_pgd) {
		target_pgd = gb_hdmac_mmu_alloc_pgd(gbctx);
		if (!target_pgd) {
			gb_printf(KERN_ERR, "%s: gb_hdmac_mmu_alloc_pgd failure\n",
			 __func__);
			return -ENOMEM;
		}

		gbctx->mmu_mode->entry_set_pte(&page[vpfn], target_pgd);
	}

	*pgd = target_pgd;
	return 0;
}

static int mmu_get_bottom_pgd(struct gb_hdmac_context *gbctx,
	u64 vpfn, phys_addr_t *out_pgd)
{
	phys_addr_t pgd;
	int l;

	pgd = gbctx->pgd;
	for (l = GENBU_MMU_TOPLEVEL; l < GENBU_MMU_BOTTOMLEVEL; l++) {
		int err = mmu_get_next_pgd(gbctx, &pgd, vpfn, l);

		/* Handle failure condition */
		if (err) {
			gb_printf(KERN_ERR, "%s: mmu_get_next_pgd failure\n", __func__);
			return err;
		}
	}

	*out_pgd = pgd;

	return 0;
}

int gb_hdmac_mmu_insert_pages_no_flush(struct gb_hdmac_context *gbctx, u64 vpfn,
	phys_addr_t *phys, size_t nr, unsigned long flags)
{
	phys_addr_t pgd;
	u64 *pgd_page;
	/*
	 * In case the insert_pages only partially completes we need to be able
	 * to recover
	 */
	size_t remain = nr;
	int err;

	GB_DEBUG_ASSERT(vpfn != 0);

	/* Early out if there is nothing to do */
	if (nr == 0)
		return 0;

	while (remain) {
		unsigned int i;
		unsigned int index = vpfn & 0x1FF;
		unsigned int count = GB_MMU_PAGE_ENTRIES - index;

		if (count > remain)
			count = remain;

		/*
		 * Repeatedly calling mmu_get_bottom_pte() is clearly
		 * suboptimal. We don't have to re-parse the whole tree
		 * each time (just cache the l0-l2 sequence).
		 * On the other hand, it's only a gain when we map more than
		 * 256 pages at once (on average). Do we really care?
		 */
		do {
			err = mmu_get_bottom_pgd(gbctx, vpfn, &pgd);
			if (err != -ENOMEM)
				break;
		} while (!err);
		if (err) {
			gb_printf(KERN_ERR, "gb_hdmac_mmu_insert_pages: mmu_get_bottom_pgd failure\n");
			goto fail_unlock;
		}

		pgd_page = gb_phys_to_virt(pgd);
		if (!pgd_page) {
			gb_printf(KERN_ERR, "gb_hdmac_mmu_insert_pages: gb_phys_to_virt failure\n");
			err = -ENOMEM;
			goto fail_unlock;
		}
		for (i = 0; i < count; i++) {
			unsigned int ofs = index + i;

			GB_DEBUG_ASSERT(0 == (pgd_page[ofs] & 1UL));
			gbctx->mmu_mode->entry_set_ate(&pgd_page[ofs],
			 phys[i], flags);
		}

		phys += count;
		vpfn += count;
		remain -= count;
	}

	return 0;

fail_unlock:
	return err;
}

int gb_hdmac_mmu_insert_pages(struct gb_hdmac_context *gbctx, u64 vpfn,
	phys_addr_t *phys, size_t nr, unsigned long flags)
{
	int err;

	err = gb_hdmac_mmu_insert_pages_no_flush(gbctx, vpfn, phys, nr, flags);
	err |= gb_hdmac_mmu_hw_do_operation(gbctx->as_nr,
	 vpfn, nr, AS_COMMAND_FLUSH_PT, 0);
	return err;
}

u64 gb_hdmac_mmu_alloc_gpu_vpfn(size_t nr_pages)
{
	u64 target_vpfn;

	mutex_lock(&hdmac_gpu_vpfn_alloc_lock);
	target_vpfn = (GPU_VIRT_BASE >> GPU_PAGE_SHIFT)
	 + gb_hdmac_gpu_vpfn_offset;
	gb_hdmac_gpu_vpfn_offset += nr_pages;
	mutex_unlock(&hdmac_gpu_vpfn_alloc_lock);

	return target_vpfn;
}

u64 gb_hdmac_mmu_alloc_cpu_vpfn(size_t nr_pages)
{
	u64 target_vpfn;

	mutex_lock(&hdmac_cpu_vpfn_alloc_lock);
	target_vpfn = (CPU_VIRT_BASE >> GPU_PAGE_SHIFT)
	 + gb_hdmac_cpu_vpfn_offset;
	gb_hdmac_cpu_vpfn_offset += nr_pages;
	mutex_unlock(&hdmac_cpu_vpfn_alloc_lock);

	return target_vpfn;
}

void gb_hdmac_mmu_free_all_vpfn(void)
{
	gb_hdmac_gpu_vpfn_offset = 0;
	gb_hdmac_cpu_vpfn_offset = 0;
}

void gb_hdmac_mmu_softreset(void)
{
	gb_hdmac_reg_write(DMA_CTRL, HDMAC_MMU_SOFT_RESET);
	msleep(10);
}

void gb_hdmac_softreset(void)
{
	gb_hdmac_reg_write(DMA_CTRL, HDMAC_SOFT_RESET);
	msleep(10);
}

int gb_hdmac_gpu_mmap(struct gb_hdmac_context *gbctx, size_t nr_pages,
	bool is_vram, phys_addr_t *phys, u64 *va)
{
	int err;
	u64 vpfn;

	if (is_vram)
		vpfn = gb_hdmac_mmu_alloc_gpu_vpfn(nr_pages);
	else
		vpfn = gb_hdmac_mmu_alloc_cpu_vpfn(nr_pages);

	*va = vpfn << GPU_PAGE_SHIFT;
	/* set pagetable in vram */
	err = gb_hdmac_mmu_insert_pages(gbctx, vpfn,
			phys,
			nr_pages,
			GB_REG_CPU_WR | GB_REG_GPU_WR |
			 GB_REG_CPU_RD | GB_REG_GPU_RD);
	if (err)
		gb_printf(KERN_ERR, "mmap fail vpfn:0x%llx, nr_page:0x%zx, phy_base:0x%llx",
		 vpfn, nr_pages, phys[0]);

	return err;
}

/* mmu set */
#define ENTRY_TYPE_MASK     3ULL
/* For valid ATEs bit 1 = (level == 3) ? 1 : 0.
 * The MMU is only ever configured by the driver so that ATEs
 * are at level 3, so bit 1 should always be set
 */
#define ENTRY_IS_ATE        3ULL
#define ENTRY_IS_INVAL      2ULL
#define ENTRY_IS_PTE        3ULL

#define ENTRY_ATTR_BITS (7ULL << 2)	/* bits 4:2 */
#define ENTRY_ACCESS_RW (1ULL << 6)     /* bits 6:7 */
#define ENTRY_ACCESS_RO (3ULL << 6)
#define ENTRY_SHARE_BITS (3ULL << 8)	/* bits 9:8 */
#define ENTRY_ACCESS_BIT (1ULL << 10)
#define ENTRY_NX_BIT (1ULL << 54)

static inline void page_table_entry_set(u64 *pte, u64 phy)
{
	*pte = phy;
}

static void mmu_get_as_setup(struct gb_hdmac_context *gbctx,
	  struct gb_hdmac_mmu_setup *setup)
{
	/* Set up the required caching policies at the correct indices
	 * in the memattr register.
	 */
	setup->memattr = (AS_MEMATTR_IMPL_DEF_CACHE_POLICY <<
	 (AS_MEMATTR_INDEX_IMPL_DEF_CACHE_POLICY * 8)) |
		(AS_MEMATTR_FORCE_TO_CACHE_ALL <<
		 (AS_MEMATTR_INDEX_FORCE_TO_CACHE_ALL * 8)) |
		(AS_MEMATTR_WRITE_ALLOC	<<
		 (AS_MEMATTR_INDEX_WRITE_ALLOC * 8)) |
		(AS_MEMATTR_AARCH64_OUTER_IMPL_DEF <<
		 (AS_MEMATTR_INDEX_OUTER_IMPL_DEF * 8)) |
		(AS_MEMATTR_AARCH64_OUTER_WA <<
		 (AS_MEMATTR_INDEX_OUTER_WA * 8));

	setup->transtab = (u64)gbctx->pgd & AS_TRANSTAB_BASE_MASK;
	hdmac_debug("HDMAC setup->transtab: 0x%llx, as_base_mask:0x%llx\n",
	 setup->transtab, AS_TRANSTAB_BASE_MASK);

	setup->transcfg = AS_TRANSCFG_ADRMODE_AARCH64_4K;
}

static void mmu_update(struct gb_hdmac_context *gbctx)
{
	struct gb_hdmac_as as;

	memset(&as, 0, sizeof(struct gb_hdmac_as));
	as.number = gbctx->as_nr;
	mmu_get_as_setup(gbctx, &as.current_setup);

	/* Apply the address space setting */
	gb_hdmac_mmu_hw_configure(&as);
}

static void mmu_disable_as(struct gb_hdmac_context *gbctx)
{
	struct gb_hdmac_as as;

	memset(&as, 0, sizeof(struct gb_hdmac_as));
	as.number = gbctx->as_nr;
	mmu_get_as_setup(gbctx, &as.current_setup);

	as.current_setup.transtab = 0ULL;
	as.current_setup.transcfg = AS_TRANSCFG_ADRMODE_UNMAPPED;

	/* Apply the address space setting */
	gb_hdmac_mmu_hw_configure(&as);
}

static phys_addr_t pte_to_phy_addr(u64 entry)
{
	if (!(entry & 1))
		return 0;

	return entry & ~(GPU_PAGE_SIZE - 1);
}

static int ate_is_valid(u64 ate)
{
	return ((ate & ENTRY_TYPE_MASK) == ENTRY_IS_ATE);
}

static int pte_is_valid(u64 pte)
{
	return ((pte & ENTRY_TYPE_MASK) == ENTRY_IS_PTE);
}

/*
 * Map GB_REG flags to MMU flags
 */
static u64 get_mmu_flags(unsigned long flags)
{
	u64 mmu_flags;

	/* store mem_attr index as 4:2 (macro called ensures 3 bits already) */
	mmu_flags = GB_REG_MEMATTR_VALUE(flags) << 2;

	/* Set access flags - note that AArch64 stage 1 does not support
	 * write-only access, so we use read/write instead
	 */
	if (flags & GB_REG_GPU_WR)
		mmu_flags |= ENTRY_ACCESS_RW;
	else if (flags & GB_REG_GPU_RD)
		mmu_flags |= ENTRY_ACCESS_RO;

	/* nx if requested */
	mmu_flags |= (flags & GB_REG_GPU_NX) ? ENTRY_NX_BIT : 0;

	if (flags & GB_REG_SHARE_BOTH)
		/* inner and outer shareable */
		mmu_flags |= SHARE_BOTH_BITS;
	else if (flags & GB_REG_SHARE_IN)
		/* inner shareable coherency */
		mmu_flags |= SHARE_INNER_BITS;

	return mmu_flags;
}

static void entry_set_ate(u64 *entry, phys_addr_t phy, unsigned long flags)
{
	page_table_entry_set(entry, (phy & ~0xFFF) | get_mmu_flags(flags) |
	 ENTRY_ACCESS_BIT | ENTRY_IS_ATE);
}

static void entry_set_pte(u64 *entry, phys_addr_t phy)
{
	page_table_entry_set(entry, (phy & ~0xFFF) |
	 ENTRY_ACCESS_BIT | ENTRY_IS_PTE);
}

static void entry_invalidate(u64 *entry)
{
	page_table_entry_set(entry, ENTRY_IS_INVAL);
}

static struct gb_hdmac_mmu_mode hdmac_mmu64_mode = {
	.update = mmu_update,
	.get_as_setup = mmu_get_as_setup,
	.disable_as = mmu_disable_as,
	.pte_to_phy_addr = pte_to_phy_addr,
	.ate_is_valid = ate_is_valid,
	.pte_is_valid = pte_is_valid,
	.entry_set_ate = entry_set_ate,
	.entry_set_pte = entry_set_pte,
	.entry_invalidate = entry_invalidate
};

struct gb_hdmac_mmu_mode *get_mmu_mode(void)
{
	return &hdmac_mmu64_mode;
}

void gb_hdmac_dump_pg_table(void)
{
#ifdef HDMAC_MMU_DUMP
	u64 *mmu_base = NULL;
	u64 mmu_phy;
	int i;

	mmu_base = gb_phys_to_virt(HDMAC_MMU_OFFSET);
	mmu_phy = HDMAC_MMU_OFFSET;
	gb_printf(KERN_ERR, "=========mmu table dump start========\n");
	for (i = 0; i < (0x80000 / 8); i++) {
		if ((*mmu_base != 2) && (*mmu_base != 0))
			gb_printf(KERN_ERR, "0x%llx:0x%llx\n", mmu_phy, *mmu_base);
		mmu_base++;
		mmu_phy += 8;
	}
	gb_printf(KERN_ERR, "=========mmu table dump end========\n");
#endif
}

int gb_hdmac_test(u64 size,
	phys_addr_t *src_phys, phys_addr_t *dst_phys,
	bool is_read, int channel_id, bool is_done, u64 abort_pg_num,
	u64 cpu_virt, u64 gpu_virt)
{
	int ret = 0, retry = 0;
	u64 pg_cnt = size >> GPU_PAGE_SHIFT;
	u32 src_phy_l, src_phy_h, dst_phy_l, dst_phy_h, hdmac_int_status;
	u64 src_phy, dst_phy;

	/* config start hdmac */
	mutex_lock(&hdmac_start_lock);
	if (is_read) {
		gb_hdmac_reg_write(DMA_SRC_ADDR_L,
		 (cpu_virt & 0xfffffff) | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_SRC_ADDR_H,
		 (cpu_virt >> 28)  | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_DST_ADDR_L,
		 (gpu_virt & 0xfffffff)  | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_DST_ADDR_H,
		 (gpu_virt >> 28)  | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_PG_CNT,
		 pg_cnt | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_PG_SIZE,
		 0 | ((2 * channel_id + 1) << 28));
		gb_hdmac_reg_write(DMA_START,
		 1 | ((2 * channel_id + 1) << 28));
	} else {
		gb_hdmac_reg_write(DMA_SRC_ADDR_L,
		 (gpu_virt & 0xfffffff)  | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_SRC_ADDR_H,
		 (gpu_virt >> 28)  | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_DST_ADDR_L,
		 (cpu_virt & 0xfffffff)  | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_DST_ADDR_H,
		 (cpu_virt >> 28)  | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_PG_CNT,
		 pg_cnt | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_PG_SIZE,
		 0 | ((2 * channel_id) << 28));
		gb_hdmac_reg_write(DMA_START,
		 1 | ((2 * channel_id) << 28));
	}
	mutex_unlock(&hdmac_start_lock);
	/* config end */

	mutex_lock(&hdmac_int_status_lock);
retry_get_status:
	if (is_done)
		hdmac_int_status = gb_hdmac_reg_read(DMA_INT_STATUS);
	else
		hdmac_int_status = gb_hdmac_reg_read(DMA_ABORT_STATUS);

	if (!is_read) {
		if (hdmac_int_status & (1 << (2 * channel_id))) {
			if (is_done) {
				gb_hdmac_reg_write(DMA_STATUS_CLEAR,
				 (1 << (2 * channel_id)));
				gb_hdmac_reg_write(DMA_STATUS_CLEAR, 0);
				gb_hdmac_reg_write(DMA_IRQ_CLEAR, 1);
				ret = 0;
				goto test_done;
			} else {
				src_phy_l =
				 gb_hdmac_reg_read(DMA_ABORT_SRC_L_i(2
				  * channel_id));
				src_phy_h =
				 gb_hdmac_reg_read(DMA_ABORT_SRC_H_i(2
				  * channel_id));
				dst_phy_l =
				 gb_hdmac_reg_read(DMA_ABORT_DST_L_i(2
				  * channel_id));
				dst_phy_h =
				 gb_hdmac_reg_read(DMA_ABORT_DST_H_i(2
				  * channel_id));
				src_phy = ((u64)src_phy_h << 32) | src_phy_l;
				dst_phy = ((u64)dst_phy_h << 32) | dst_phy_l;
				if ((src_phy == src_phys[abort_pg_num])
				 && (dst_phy == dst_phys[abort_pg_num]))
					ret = 0;
				else {
					gb_printf(KERN_ERR, "HDMAC ABORT ERROR!hdmac abort src_phy:0x%llx, user abort src_phy:0x%llx, dst:0x%llx, user dst:0x%llx\n",
					 src_phy, src_phys[abort_pg_num],
					  dst_phy, dst_phys[abort_pg_num]);
					ret = -1;
				}
				gb_hdmac_reg_write(DMA_ABORT_CLEAR,
				 (1 << (2 * channel_id)));
				gb_hdmac_reg_write(DMA_ABORT_CLEAR, 0);
				gb_hdmac_reg_write(DMA_IRQ_CLEAR, 1);
				goto test_done;
			}
		} else {
			if (retry < 10) {
				retry++;
				mdelay(1);
				goto retry_get_status;
			}
			gb_printf(KERN_ERR, "hdmac write test ERR!!!is_read:%d, is_done:%d, channel_id:%d, size:0x%llx, status:0x%x, abort_status:0x%x\n",
			 is_read, is_done, channel_id, size, hdmac_int_status,
			  gb_hdmac_reg_read(DMA_ABORT_STATUS));
			ret = -1;
			goto test_done;
		}
	} else {
		if (hdmac_int_status & (1 << (2 * channel_id + 1))) {
			if (is_done) {
				gb_hdmac_reg_write(DMA_STATUS_CLEAR,
				 (1 << (2 * channel_id + 1)));
				gb_hdmac_reg_write(DMA_STATUS_CLEAR, 0);
				gb_hdmac_reg_write(DMA_IRQ_CLEAR, 1);
				ret = 0;
				goto test_done;
			} else {
				src_phy_l =
				 gb_hdmac_reg_read(DMA_ABORT_SRC_L_i(2 *
				  channel_id + 1));
				src_phy_h =
				 gb_hdmac_reg_read(DMA_ABORT_SRC_H_i(2 *
				  channel_id + 1));
				dst_phy_l =
				 gb_hdmac_reg_read(DMA_ABORT_DST_L_i(2 *
				  channel_id + 1));
				dst_phy_h =
				 gb_hdmac_reg_read(DMA_ABORT_DST_H_i(2 *
				  channel_id + 1));
				src_phy =
				 ((u64)src_phy_h << 32) | src_phy_l;
				dst_phy =
				 ((u64)dst_phy_h << 32) | dst_phy_l;
				if ((src_phy == src_phys[abort_pg_num])
				 && (dst_phy == dst_phys[abort_pg_num]))
					ret = 0;
				else {
					gb_printf(KERN_ERR, "HDMAC ABORT ERROR!hdmac abort src_phy:0x%llx, user abort src_phy:0x%llx, dst:0x%llx, user dst:0x%llx\n",
					 src_phy, src_phys[abort_pg_num],
					  dst_phy, dst_phys[abort_pg_num]);
					ret = -1;
				}
				gb_hdmac_reg_write(DMA_ABORT_CLEAR,
				 (1 << (2 * channel_id + 1)));
				gb_hdmac_reg_write(DMA_ABORT_CLEAR, 0);
				gb_hdmac_reg_write(DMA_IRQ_CLEAR, 1);
				goto test_done;
			}
		} else {
			if (retry < 10) {
				retry++;
				mdelay(1);
				goto retry_get_status;
			}
			gb_printf(KERN_ERR, "hdmac read test ERR, is_read: %d, is_done: %d, channel_id: %d, size: 0x%llx, status: 0x%x, abort_status: 0x%x\n",
			 is_read, is_done, channel_id, size,
			  hdmac_int_status,
			   gb_hdmac_reg_read(DMA_ABORT_STATUS));
			ret = -1;
			goto test_done;
		}
	}
test_done:
	mutex_unlock(&hdmac_int_status_lock);
	return ret;
}

void hdmac_enable_mmu_irq(void)
{
	gb_hdmac_reg_write(MMU_IRQ_CLEAR, 0xffffffff);
	gb_hdmac_reg_write(MMU_IRQ_MASK, 0xffffffff);
}

u32 gb_hdmac_get_mmu_irq(void)
{
	return gb_hdmac_reg_read(MMU_IRQ_STATUS);
}
