/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "common/gb_common.h"
#ifndef __GB_HDMAC_REGMAP_H__
#define __GB_HDMAC_REGMAP_H__

#define HDMAC_PHY_OFFSET HDMAC_VRAM_START_ADDR
#define HDMAC_MMU_LEN (10*MB)
#define HDMAC_MMU_OFFSET (HDMAC_VRAM_START_ADDR + (HDMAC_VRAM_LEN - HDMAC_MMU_LEN))

#define CPU_VIRT_BASE 0x210000000
#define GPU_VIRT_BASE 0x110000000

#define GENBU_MMU_VA_BITS 48

#if GENBU_MMU_VA_BITS > 39
#define GENBU_MMU_TOPLEVEL    0
#else
#define GENBU_MMU_TOPLEVEL    1
#endif

#define GENBU_MMU_BOTTOMLEVEL 3


#define GB_MMU_PAGE_ENTRIES 512

#define GPU_PAGE_SHIFT (12)
#define GPU_PAGE_SIZE (1 << GPU_PAGE_SHIFT)
#define GB_LOCK_REGION_MAX_SIZE (63)
#define GB_LOCK_REGION_MIN_SIZE (11)

#define GB_DEBUG_SKIP_TRACE 0
#define GB_DEBUG_SKIP_FUNCTION_NAME 0

#if !GB_DEBUG_SKIP_TRACE
#define GB_DEBUG_PRINT_TRACE \
		"In file: " __FILE__
#if !GB_DEBUG_SKIP_FUNCTION_NAME
#define GB_DEBUG_PRINT_FUNCTION __func__
#else
#define GB_DEBUG_PRINT_FUNCTION ""
#endif
#else
#define GB_DEBUG_PRINT_TRACE ""
#endif

#define GB_DEBUG_ASSERT_OUT(trace, function, ...)\
		do { \
			gb_printf(KERN_ERR, "Botson<ASSERT>: %s function:%s ",\
			 trace, function);\
			gb_printf(KERN_ERR, __VA_ARGS__);\
			gb_printf(KERN_ERR, "\n");\
		} while (false)

#define GB_DEBUG_ASSERT(expr) \
	GB_DEBUG_ASSERT_MSG(expr, #expr)

#define GB_DEBUG_ASSERT_MSG(expr, ...) \
		do { \
			if (!(expr)) { \
				GB_DEBUG_ASSERT_OUT(GB_DEBUG_PRINT_TRACE,\
				 GB_DEBUG_PRINT_FUNCTION, __VA_ARGS__);\
				gb_printf(KERN_ERR, "Assert Error: inside %s %d\n",\
				 __func__, __LINE__); \
			} \
		} while (false)

enum gb_hdmac_mmu_fault_type {
	GB_MMU_FAULT_TYPE_UNKNOWN = 0,
	GB_MMU_FAULT_TYPE_PAGE,
	GB_MMU_FAULT_TYPE_BUS,
	GB_MMU_FAULT_TYPE_PAGE_UNEXPECTED,
	GB_MMU_FAULT_TYPE_BUS_UNEXPECTED
};

enum gb_hdmac_share_attr_bits {
	/* (1ULL << 8) bit is reserved */
	SHARE_BOTH_BITS = (2ULL << 8), /* inner and out shareable coherency */
	SHARE_INNER_BITS = (3ULL << 8) /* inner shareable coherency */
};

struct gb_hdmac_mmu_setup {
	u64	transtab;
	u64	memattr;
	u64	transcfg;
};

struct gb_hdmac_as {
	int number;
	enum gb_hdmac_mmu_fault_type fault_type;
	bool protected_mode;
	u32 fault_status;
	u64 fault_addr;
	u64 fault_extra_addr;
	struct gb_hdmac_mmu_setup current_setup;
};

struct gb_hdmac_context {
	pid_t tgid;
	pid_t pid;
	int as_nr;
	phys_addr_t pgd;
	struct gb_hdmac_mmu_mode *mmu_mode;
	u64 fault_cpu_addr;
	u64 fault_gpu_addr;
};

struct gb_hdmac_mmu_mode {
	void (*update)(struct gb_hdmac_context *gbctx);
	void (*get_as_setup)(struct gb_hdmac_context *gbctx,
			struct gb_hdmac_mmu_setup * const setup);
	void (*disable_as)(struct gb_hdmac_context *gbctx);
	phys_addr_t (*pte_to_phy_addr)(u64 entry);
	int (*ate_is_valid)(u64 ate);
	int (*pte_is_valid)(u64 pte);
	void (*entry_set_ate)(u64 *entry, phys_addr_t phy, unsigned long flags);
	void (*entry_set_pte)(u64 *entry, phys_addr_t phy);
	void (*entry_invalidate)(u64 *entry);
};

/*
 * Maximum number of loops polling the GPU for a cache flush before
 * we assume it must have completed
 */
#define GB_CLEAN_CACHE_MAX_LOOPS     100000
/*
 * Maximum number of loops polling the GPU for an AS command to complete
 * before we assume the GPU has hung
 */
#define GB_AS_INACTIVE_MAX_LOOPS     10


/* region flags */
#define GB_REG_FREE              (1ul << 0)
/* CPU write access */
#define GB_REG_CPU_WR            (1ul << 1)
/* GPU write access */
#define GB_REG_GPU_WR            (1ul << 2)
/* No eXecute flag */
#define GB_REG_GPU_NX            (1ul << 3)
/* Is CPU cached? */
#define GB_REG_CPU_CACHED        (1ul << 4)
/* Is GPU cached? */
#define GB_REG_GPU_CACHED        (1ul << 5)

#define GB_REG_GROWABLE          (1ul << 6)
/* Can grow on pf? */
#define GB_REG_PF_GROW           (1ul << 7)

/* VA managed by us */
#define GB_REG_CUSTOM_VA         (1ul << 8)

/* inner shareable coherency */
#define GB_REG_SHARE_IN          (1ul << 9)
/* inner & outer shareable coherency */
#define GB_REG_SHARE_BOTH        (1ul << 10)

/* Space for 4 different zones */
#define GB_REG_ZONE_MASK         (3ul << 11)
#define GB_REG_ZONE(x)           (((x) & 3) << 11)

/* GPU read access */
#define GB_REG_GPU_RD            (1ul<<13)
/* CPU read access */
#define GB_REG_CPU_RD            (1ul<<14)

/* Index of chosen MEMATTR for this region (0..7) */
#define GB_REG_MEMATTR_MASK      (7ul << 16)
#define GB_REG_MEMATTR_INDEX(x)  (((x) & 7) << 16)
#define GB_REG_MEMATTR_VALUE(x)  (((x) & GB_REG_MEMATTR_MASK) >> 16)

#define GB_REG_SECURE            (1ul << 19)

#define GB_REG_DONT_NEED         (1ul << 20)

/* hdmA reg, just for read to debug */
#define HDMA_WR_SRC_PHY_L		0x20
#define HDMA_WR_SRC_PHY_H		0x24
#define HDMA_WR_DST_PHY_L		0x28
#define HDMA_WR_DST_PHY_H		0x2c
#define HDMA_WR_PG_SIZE			0x1c
#define HDMA_WR_START			0x00

#define HDMA_RD_SRC_PHY_L		0x120
#define HDMA_RD_SRC_PHY_H		0x124
#define HDMA_RD_DST_PHY_L		0x128
#define HDMA_RD_DST_PHY_H		0x12c
#define HDMA_RD_PG_SIZE			0x11c
#define HDMA_RD_START			0x100

#define HDMA_WR_SRC_PHY_L_i(i)		(HDMA_WR_SRC_PHY_L + 0x200 * (i))
#define HDMA_WR_SRC_PHY_H_i(i)		(HDMA_WR_SRC_PHY_H + 0x200 * (i))
#define HDMA_WR_DST_PHY_L_i(i)		(HDMA_WR_DST_PHY_L + 0x200 * (i))
#define HDMA_WR_DST_PHY_H_i(i)		(HDMA_WR_DST_PHY_H + 0x200 * (i))
#define HDMA_WR_PG_SIZE_i(i)		(HDMA_WR_PG_SIZE + 0x200 * (i))
#define HDMA_WR_START_i(i)		(HDMA_WR_START + 0x200 * (i))

#define HDMA_RD_SRC_PHY_L_i(i)		(HDMA_RD_SRC_PHY_L + 0x200 * (i))
#define HDMA_RD_SRC_PHY_H_i(i)		(HDMA_RD_SRC_PHY_H + 0x200 * (i))
#define HDMA_RD_DST_PHY_L_i(i)		(HDMA_RD_DST_PHY_L + 0x200 * (i))
#define HDMA_RD_DST_PHY_H_i(i)		(HDMA_RD_DST_PHY_H + 0x200 * (i))
#define HDMA_RD_PG_SIZE_i(i)		(HDMA_RD_PG_SIZE + 0x200 * (i))
#define HDMA_RD_START_i(i)		(HDMA_RD_START + 0x200 * (i))

#if 1
#define HDMA_INT_STATUS_OFF_WRCH 0x84
#define HDMA_INT_STATUS_OFF_RDCH 0x184
#define HDMA_INT_STATUS_OFF_WRCH_i(i) (HDMA_INT_STATUS_OFF_WRCH + 0x200 * (i))
#define HDMA_INT_STATUS_OFF_RDCH_i(i) (HDMA_INT_STATUS_OFF_RDCH + 0x200 * (i))
#else
#define HDMA_INT_STATUS_OFF_WRCH 0x84
#define HDMA_INT_STATUS_OFF_RDCH 0x284
#define HDMA_INT_STATUS_OFF_WRCH_i(i) (HDMA_INT_STATUS_OFF_WRCH + 0x400 * (i))
#define HDMA_INT_STATUS_OFF_RDCH_i(i) (HDMA_INT_STATUS_OFF_RDCH + 0x400 * (i))
#endif

/* hdmA reg end */

#define HDMAC_CHANNEL_NUM		4
#define HDMAC_MMU_SOFT_RESET	0xfffffffe
#define HDMAC_SOFT_RESET		0xfffffffd

/* hdmaC reg map start */
#define HDMAC_BASE				0x2000
#define DMA_ID					(HDMAC_BASE + 0x20)
/* bit0:set 0 to softreset mmu; bit1:set 0 to softreset hdmac */
#define DMA_CTRL				(HDMAC_BASE + 0x24)
/* 0~7, totally 8 AS */
#define DMA_ASID				(HDMAC_BASE + 0x30)
//#define DMA_CHANNEL				(HDMAC_BASE + 0x5c)
#define DMA_SRC_ADDR_L			(HDMAC_BASE + 0x48)
#define DMA_SRC_ADDR_H			(HDMAC_BASE + 0x4c)
#define DMA_DST_ADDR_L			(HDMAC_BASE + 0x50)
#define DMA_DST_ADDR_H			(HDMAC_BASE + 0x54)
#define DMA_PG_CNT				(HDMAC_BASE + 0x58)
#define DMA_PG_SIZE				(HDMAC_BASE + 0x44)
/*
 * must be set at last... bit0 = 1:start dma;
 * bit31 = 1:dma read; bit31 = 0:dma_write
 */
#define DMA_START				(HDMAC_BASE + 0x40)

#define DMA_INT_STATUS			(HDMAC_BASE + 0x60)
#define DMA_ABORT_STATUS		(HDMAC_BASE + 0x64)
#define DMA_STATUS_CLEAR		(HDMAC_BASE + 0x68)
#define DMA_ABORT_CLEAR			(HDMAC_BASE + 0x6c)
#define DMA_TEST_INT			(HDMAC_BASE + 0x70)

#define DMA_ABORT_SRC_L_BASE	(HDMAC_BASE + 0x80)
#define DMA_ABORT_SRC_H_BASE	(HDMAC_BASE + 0x84)
#define DMA_ABORT_SRC_L_i(r)    (DMA_ABORT_SRC_L_BASE + (8 * (r)))
#define DMA_ABORT_SRC_H_i(r)    (DMA_ABORT_SRC_H_BASE + (8 * (r)))

#define DMA_ABORT_DST_L_BASE	(HDMAC_BASE + 0x100)
#define DMA_ABORT_DST_H_BASE	(HDMAC_BASE + 0x104)
#define DMA_ABORT_DST_L_i(r)    (DMA_ABORT_DST_L_BASE + (8 * (r)))
#define DMA_ABORT_DST_H_i(r)    (DMA_ABORT_DST_H_BASE + (8 * (r)))

#define DMA_ABORT_PNUM_BASE		(HDMAC_BASE + 0x180)
#define DMA_ABORT_PNUM_i(r)     (DMA_ABORT_PNUM_BASE + (4 * (r)))

#define DMA_CFG_PCIE_BASE_ADDR	(HDMAC_BASE + 0x1c0)

/* 0 dianping; 1maichong  */
#define DMA_IRQ_MODE		(HDMAC_BASE + 0x1c4)

#define DMA_IRQ_CLEAR		(HDMAC_BASE + 0x1c8)

/* enable hdmac to clear hdma irq */
#define DMA_CLEAR_HDMA_IRQ	(HDMAC_BASE + 0x1cc)

/* hdmaC reg map end */

/* mmu reg map */
#define MEMORY_MANAGEMENT_BASE  0x0
#define MMU_REG(r)              (MEMORY_MANAGEMENT_BASE + (r))

#define MMU_AS0     0x0	/* Configuration registers for address space 0 */

#define MMU_AS_REG(n, r)        (MMU_REG(MMU_AS0 + ((n) << 6)) + (r))

/* (RW) Translation Table Base Address for address space n, low word */
#define AS_TRANSTAB_LO			0x00
/* (RW) Translation Table Base Address for address space n, high word */
#define AS_TRANSTAB_HI			0x04
/* (RW) Memory attributes for address space n, low word. */
#define AS_MEMATTR_LO			0x08
/* (RW) Memory attributes for address space n, high word. */
#define AS_MEMATTR_HI			0x0C
/* (RW) Lock region address for address space n, low word */
#define AS_LOCKADDR_LO			0x10
/* (RW) Lock region address for address space n, high word */
#define AS_LOCKADDR_HI			0x14

#define AS_COMMAND	0x18 /* (WO) MMU command register for address space n */
/* (RO) MMU fault status register for address space n */
#define AS_FAULTSTATUS 0x1C
/* (RO) Fault Address for address space n, low word */
#define AS_FAULTADDRESS_LO		0x20
/* (RO) Fault Address for address space n, high word */
#define AS_FAULTADDRESS_HI		0x24
#define AS_STATUS	0x28	/*(RO) Status flags for address space n*/


/* (RW) Translation table configuration for address space n, low word */
#define AS_TRANSCFG_LO			0x30
/* (RW) Translation table configuration for address space n, high word */
#define AS_TRANSCFG_HI			0x34
/* (RO) Secondary fault address for address space n, low word */
#define AS_FAULTEXTRA_LO		0x38
/* (RO) Secondary fault address for address space n, high word */
#define AS_FAULTEXTRA_HI		0x3C

#define MMU_IRQ_RAWSTAT		0x400	/* (RW) Raw interrupt status register */
#define MMU_IRQ_CLEAR		0x404	/* (WO) Interrupt clear register */
#define MMU_IRQ_MASK		0x408	/* (RW) Interrupt mask register */
#define MMU_IRQ_STATUS		0x40C	/* (RO) Interrupt status register */

/* End Register Offsets */

/*
 * MMU_IRQ_RAWSTAT register values. Values are valid also for
   MMU_IRQ_CLEAR, MMU_IRQ_MASK, MMU_IRQ_STATUS registers.
 */

#define MMU_PAGE_FAULT_FLAGS   16

/* Macros returning a bitmask to retrieve page fault or bus error flags from
 * MMU registers
 */
#define MMU_PAGE_FAULT(n)      (1UL << (n))
#define MMU_BUS_ERROR(n)       (1UL << ((n) + MMU_PAGE_FAULT_FLAGS))

/*
 * Begin LPAE MMU TRANSTAB register values
 */
#define AS_TRANSTAB_LPAE_ADDR_SPACE_MASK   0xfffff000
#define AS_TRANSTAB_LPAE_ADRMODE_UNMAPPED  (0u << 0)
#define AS_TRANSTAB_LPAE_ADRMODE_IDENTITY  (1u << 1)
#define AS_TRANSTAB_LPAE_ADRMODE_TABLE     (3u << 0)
#define AS_TRANSTAB_LPAE_READ_INNER        (1u << 2)
#define AS_TRANSTAB_LPAE_SHARE_OUTER       (1u << 4)

#define AS_TRANSTAB_LPAE_ADRMODE_MASK      0x00000003

/*
 * Begin AARCH64 MMU TRANSTAB register values
 */
#define MMU_HW_OUTA_BITS 40
#define AS_TRANSTAB_BASE_MASK ((1ULL << MMU_HW_OUTA_BITS) - (1ULL << 4))

/*
 * Begin MMU STATUS register values
 */
#define AS_STATUS_AS_ACTIVE 0x01

#define AS_FAULTSTATUS_EXCEPTION_CODE_MASK                    (0x7<<3)
#define AS_FAULTSTATUS_EXCEPTION_CODE_TRANSLATION_FAULT       (0x0<<3)
#define AS_FAULTSTATUS_EXCEPTION_CODE_PERMISSION_FAULT        (0x1<<3)
#define AS_FAULTSTATUS_EXCEPTION_CODE_TRANSTAB_BUS_FAULT      (0x2<<3)
#define AS_FAULTSTATUS_EXCEPTION_CODE_ACCESS_FLAG             (0x3<<3)

#define AS_FAULTSTATUS_EXCEPTION_CODE_ADDRESS_SIZE_FAULT      (0x4<<3)
#define AS_FAULTSTATUS_EXCEPTION_CODE_MEMORY_ATTRIBUTES_FAULT (0x5<<3)

#define AS_FAULTSTATUS_ACCESS_TYPE_MASK                  (0x3<<8)
#define AS_FAULTSTATUS_ACCESS_TYPE_ATOMIC                (0x0<<8)
#define AS_FAULTSTATUS_ACCESS_TYPE_EX                    (0x1<<8)
#define AS_FAULTSTATUS_ACCESS_TYPE_READ                  (0x2<<8)
#define AS_FAULTSTATUS_ACCESS_TYPE_WRITE                 (0x3<<8)

/*
 * Begin MMU TRANSCFG register values
 */

#define AS_TRANSCFG_ADRMODE_LEGACY      0
#define AS_TRANSCFG_ADRMODE_UNMAPPED    1
#define AS_TRANSCFG_ADRMODE_IDENTITY    2
#define AS_TRANSCFG_ADRMODE_AARCH64_4K  6
#define AS_TRANSCFG_ADRMODE_AARCH64_64K 8

#define AS_TRANSCFG_ADRMODE_MASK        0xF


/*
 * Begin TRANSCFG register values
 */
#define AS_TRANSCFG_PTW_MEMATTR_MASK (3 << 24)
#define AS_TRANSCFG_PTW_MEMATTR_NON_CACHEABLE (1 << 24)
#define AS_TRANSCFG_PTW_MEMATTR_WRITE_BACK (2 << 24)

#define AS_TRANSCFG_PTW_SH_MASK ((3 << 28))
#define AS_TRANSCFG_PTW_SH_OS (2 << 28)
#define AS_TRANSCFG_PTW_SH_IS (3 << 28)

/* AS_COMMAND register commands */
#define AS_COMMAND_NOP         0x00	/* NOP Operation */
/* Broadcasts the values in AS_TRANSTAB and ASn_MEMATTR to all MMUs */
#define AS_COMMAND_UPDATE      0x01
#define AS_COMMAND_LOCK	0x02 /* Issue a lock region command to all MMUs */
#define AS_COMMAND_UNLOCK 0x03 /* Issue a flush region command to all MMUs */
/*
 * Flush all L2 caches then issue a flush region command to all MMUs
 * (deprecated - only for use with T60x)
 */
#define AS_COMMAND_FLUSH 0x04
/* Flush all L2 caches then issue a flush region command to all MMUs */
#define AS_COMMAND_FLUSH_PT    0x04
/*
 * Wait for memory accesses to complete, flush all the L1s cache then
 * flush all L2 caches then issue a flush region command to all MMUs
 */
#define AS_COMMAND_FLUSH_MEM   0x05

/* AS<n>_MEMATTR values: */
/* Use GPU implementation-defined caching policy. */
#define AS_MEMATTR_IMPL_DEF_CACHE_POLICY 0x88ull
/* The attribute set to force all resources to be cached. */
#define AS_MEMATTR_FORCE_TO_CACHE_ALL    0x8Full
/* Inner write-alloc cache setup, no outer caching */
#define AS_MEMATTR_WRITE_ALLOC           0x8Dull

/* Set to implementation defined, outer caching */
#define AS_MEMATTR_AARCH64_OUTER_IMPL_DEF 0x88ull
/* Set to write back memory, outer caching */
#define AS_MEMATTR_AARCH64_OUTER_WA       0x8Dull

/* HW implementation defined caching */
#define AS_MEMATTR_INDEX_IMPL_DEF_CACHE_POLICY 0
/* Force cache on */
#define AS_MEMATTR_INDEX_FORCE_TO_CACHE_ALL    1
/* Write-alloc */
#define AS_MEMATTR_INDEX_WRITE_ALLOC           2
/* Outer coherent, inner implementation defined policy */
#define AS_MEMATTR_INDEX_OUTER_IMPL_DEF        3
/* Outer coherent, write alloc inner */
#define AS_MEMATTR_INDEX_OUTER_WA              4
struct gb_hdmac_mmu_mode *get_mmu_mode(void);
phys_addr_t gb_hdmac_mmu_alloc_pgd(struct gb_hdmac_context *gbctx);
void gb_hdmac_mmu_free_all_pgd(void);
phys_addr_t gb_hdmac_alloc_gpu_page(void);
void gb_hdmac_mmu_free_all_gpu_page(void);
void gb_hdmac_mmu_free_all_vpfn(void);
int gb_hdmac_gpu_mmap(struct gb_hdmac_context *gbctx, size_t nr_pages,
	bool is_vram, phys_addr_t *phys, u64 *va);
int gb_hdmac_test(u64 size, phys_addr_t *src_phys, phys_addr_t *dst_phys,
	bool is_read, int channel_id, bool is_done,
	u64 abort_pg_num, u64 cpu_virt, u64 gpu_virt);
void gb_hdmac_dump_pg_table(void);
u32 gb_hdmac_reg_read(u32 offset);
void gb_hdmac_reg_write(u32 offset, u32 value);
void gb_hdmac_mmu_softreset(void);
void gb_hdmac_softreset(void);
void hdmac_enable_mmu_irq(void);
u32 gb_hdmac_get_mmu_irq(void);
#endif
