#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include "common/xt.h"
#include "common/gb_irq.h"
#include "common/gb_pcie_info.h"
#include "gb_v2vdma.h"

static struct v2vdma_ctrl_info *v2vdma_infos;

static inline void axi_dma_disable(struct v2vdma_ctrl_info *chip)
{
    u32 val;

    val = dma_reg_read32(chip, DMAC_CFG);
    val &= ~DMAC_EN_MASK;
    dma_reg_write32(chip, DMAC_CFG, val);
}

static inline void axi_dma_enable(struct v2vdma_ctrl_info *chip)
{
    u32 val;

    val = dma_reg_read32(chip, DMAC_CFG);
    val |= DMAC_EN_MASK;
    dma_reg_write32(chip, DMAC_CFG, val);
}

static inline void axi_dma_irq_disable(struct v2vdma_ctrl_info *chip)
{
    u32 val;

    val = dma_reg_read32(chip, DMAC_CFG);
    val &= ~INT_EN_MASK;
    dma_reg_write32(chip, DMAC_CFG, val);
}

static inline void axi_dma_irq_enable(struct v2vdma_ctrl_info *chip)
{
    u32 val;

    val = dma_reg_read32(chip, DMAC_CFG);
    val |= INT_EN_MASK;
    dma_reg_write32(chip, DMAC_CFG, val);
}

static inline void axi_chan_irq_disable(struct axi_dma_chan *chan, u32 irq_mask)
{
    u32 val;

    if (likely(irq_mask == DWAXIDMAC_IRQ_ALL)) {
        chan_reg_write32(chan, CH_INTSTATUS_ENA, DWAXIDMAC_IRQ_NONE);
    } else {
        val = chan_reg_read32(chan, CH_INTSTATUS_ENA);
        val &= ~irq_mask;
        chan_reg_write32(chan, CH_INTSTATUS_ENA, val);
    }
}

static inline void axi_chan_irq_set(struct axi_dma_chan *chan, u32 irq_mask)
{
    chan_reg_write32(chan, CH_INTSTATUS_ENA, irq_mask);
}

static inline void axi_chan_irq_sig_set(struct axi_dma_chan *chan, u32 irq_mask)
{
    chan_reg_write32(chan, CH_INTSIGNAL_ENA, irq_mask);
}

static inline void axi_chan_irq_clear(struct axi_dma_chan *chan, u32 irq_mask)
{
    chan_reg_write32(chan, CH_INTCLEAR, irq_mask);
}

static inline u32 axi_chan_irq_read(struct axi_dma_chan *chan)
{
    return chan_reg_read32(chan, CH_INTSTATUS);
}

static inline void axi_chan_disable(struct axi_dma_chan *chan)
{
    u32 val;

    val = dma_reg_read32(chan->chip, DMAC_CHEN);
    val &= ~(BIT(chan->id) << DMAC_CHAN_EN_SHIFT);
    val |=   BIT(chan->id) << DMAC_CHAN_EN_WE_SHIFT;
    dma_reg_write32(chan->chip, DMAC_CHEN, val);
}

static inline void axi_chan_enable(struct axi_dma_chan *chan)
{
    u32 val;

    val = dma_reg_read32(chan->chip, DMAC_CHEN);
    val |= BIT(chan->id) << DMAC_CHAN_EN_SHIFT |
           BIT(chan->id) << DMAC_CHAN_EN_WE_SHIFT;
    dma_reg_write32(chan->chip, DMAC_CHEN, val);
}

static int axi_dma_hw_init(struct v2vdma_ctrl_info *chip)
{
	u32 i = 0;

    for (i = 0; i < DW_AXI_DMA_CHANNELS; i++) {
		axi_chan_irq_disable(&chip->chan[i], DWAXIDMAC_IRQ_ALL);
		axi_chan_disable(&chip->chan[i]);
	}

	return 0;
}

static inline bool axi_chan_is_hw_enable(struct axi_dma_chan *chan)
{
	u32 val;

	val = dma_reg_read32(chan->chip, DMAC_CHEN);

	return !!(val & (BIT(chan->id) << DMAC_CHAN_EN_SHIFT));
}

static noinline void axi_chan_handle_err(struct axi_dma_chan *chan, u32 status)
{
#if 0
    int ret = 0, i = 0;
    struct dma_ctrl_info *chip = chan->chip;
    struct axi_dma_desc *desc;

	axi_chan_disable(chan);

    if (status & DWAXIDMAC_IRQ_INVALID_ERR) {
        if (chip->dma_mode_flag == DMA_SHADOW_MODE) {
            /* shadow mode */

            /* prepare for next block's transfer */
            desc = list_next_entry(chan->first_desc, xfer_list);
            for (i = 0;i < chan->shadow_blk_cnt;i++) {
                desc = list_next_entry(desc, xfer_list);
            }

            if (desc) {
                ret = write_chan_shadow(chan, desc);
            }

            chan_reg_write32(chan, CH_BLK_TFR_RESUMEREQ, 1);

            chan->shadow_blk_cnt++;
            axi_chan_enable(chan);
        }
    }
#endif
}

static void axi_chan_block_xfer_complete(struct axi_dma_chan *chan)
{
#if 0
	struct dma_ctrl_info *chip = chan->chip;
	long xfer_period_us = 0L;

    if (unlikely(axi_chan_is_hw_enable(chan))) {
        gb_printf(KERN_ERR, "BUG: caught DWAXIDMAC_IRQ_DMA_TRF, but channel not idle!\n");
        axi_chan_disable(chan);
    }

	if (chip->invalid_test1_flag) {
		chip->invalid_test1_flag = false;

        do_gettimeofday(&(chip->dma_end_tv));

        xfer_period_us = (chip->dma_end_tv.tv_sec * 1000 * 1000 + chip->dma_end_tv.tv_usec) -
                        (chip->dma_start_tv.tv_sec * 1000 * 1000 + chip->dma_start_tv.tv_usec);
        gb_printf(KERN_INFO, "invalid test1 transfer cost : %ld us.\n", xfer_period_us);
	}
#endif
}

irqreturn_t gb_v2vdma_irq_handler(int irq, void *arg)
{
    struct v2vdma_ctrl_info *chip = v2vdma_infos;
    struct axi_dma_chan *chan;
    int i = 0;
    u32 status = 0, val = 0;

	/* gb_printf(KERN_DEBUG, "%s:%d:enter.\n", __func__, __LINE__); */

    /* Disable DMAC inerrupts. We'll enable them after processing chanels */
    axi_dma_irq_disable(chip);

    status = dma_reg_read32(chip, DMAC_COMMON_INTSTATUS);
    if (status) {
        dma_reg_write32(chip, DMAC_COMMON_INTCLEAR, status);
        val = dma_reg_read32(chip, DMAC_COMMON_INTSTATUS_ENA);
        val &= ~status;
        dma_reg_write32(chip, DMAC_COMMON_INTSTATUS_ENA, val);
    }

    /* Poll, clear and process every chanel interrupt status */
    for (i = 0; i < DW_AXI_DMA_CHANNELS; i++) {
        chan = &chip->chan[i];
        status = axi_chan_irq_read(chan);
        axi_chan_irq_clear(chan, status);

        gb_printf(KERN_DEBUG, "%s %u IRQ status: 0x%08x\n",
            __func__, i, status);

        if (status & DWAXIDMAC_IRQ_ALL_ERR)
            axi_chan_handle_err(chan, status);
        else if (status & DWAXIDMAC_IRQ_DMA_TRF) {
            axi_chan_block_xfer_complete(chan);
            atomic_set(&chan->stop, 0);
        }
    }

    /* Re-enable interrupts */
    axi_dma_irq_enable(chip);

	/* TODO: send event to UMD */
    return IRQ_HANDLED;
}

/* single block mode */
int gb_v2vdma_start(phys_addr_t src_addr,
	phys_addr_t dst_addr, uint32_t size)
{
	struct axi_dma_chan *chan = NULL;
	u32 dma_chnl = 0, val = 0, irq_mask = 0;
	unsigned long long addr = 0LL;
	struct v2vdma_ctrl_info *dma_ctrl = v2vdma_infos;
	u64 timeout = 0;
	u32 status;

	if (dma_chnl >= DW_AXI_DMA_CHANNELS) {
		gb_printf(KERN_ERR, "%s:%d: can not get DMA channel(%d).\n",
			__func__, __LINE__, dma_chnl);
		return -1;
	}
	chan = &(dma_ctrl->chan[dma_chnl]);
	atomic_set(&chan->stop, 1);
	/* DMAC_CFGREG */
	axi_dma_enable(dma_ctrl);
	axi_dma_irq_enable(dma_ctrl);

	/* DMAC_COMMON_INTSTATUS_ENA */
	dma_reg_write32(dma_ctrl, DMAC_COMMON_INTSTATUS_ENA, GENMASK(20, 0));
	/* DMAC_COMMON_INTSIGNAL_ENA */
	dma_reg_write32(dma_ctrl, DMAC_COMMON_INTSIGNAL_ENA, GENMASK(20, 0));

	/* CHX_SAR */
	addr = V2VDMA_VRAM_OFFSET + src_addr;
	chan_reg_write64(chan, CH_SAR, cpu_to_le64(addr));
	/* CHX_DAR */
	addr = V2VDMA_VRAM_OFFSET + dst_addr;
	chan_reg_write64(chan, CH_DAR, cpu_to_le64(addr));
	/* CHX_BLOCK_TS */
	val = (size >> DWAXIDMAC_TRANS_WIDTH_256) - 1;
	chan_reg_write32(chan, CH_BLOCK_TS, cpu_to_le32(val));
	/* CHX_CTL */
	val = (DWAXIDMAC_BURST_TRANS_LEN_16 << CH_CTL_L_DST_MSIZE_POS |
            DWAXIDMAC_BURST_TRANS_LEN_16 << CH_CTL_L_SRC_MSIZE_POS |
            DW_AXI_DMA_DATA_WIDTH << CH_CTL_L_DST_WIDTH_POS |
            DW_AXI_DMA_DATA_WIDTH << CH_CTL_L_SRC_WIDTH_POS |
            DWAXIDMAC_CH_CTL_L_INC << CH_CTL_L_DST_INC_POS |
            DWAXIDMAC_CH_CTL_L_INC << CH_CTL_L_SRC_INC_POS);
	val &= ~(CH_CTL_L_SRC_MAST | CH_CTL_L_DST_MAST);
	chan_reg_write32(chan, CH_CTL_L, val);

	/* CHX_CFG */
	val = (DWAXIDMAC_MBLK_TYPE_CONTIGUOUS << CH_CFG_L_DST_MULTBLK_TYPE_POS |
           DWAXIDMAC_MBLK_TYPE_CONTIGUOUS << CH_CFG_L_SRC_MULTBLK_TYPE_POS);
	chan_reg_write32(chan, CH_CFG_L, val);
	val = (DWAXIDMAC_TT_FC_MEM_TO_MEM_DMAC << CH_CFG_H_TT_FC_POS |
           DW_AXI_DMA_TEST_CHAN_PRIORITY << CH_CFG_H_PRIORITY_POS |
           DWAXIDMAC_HS_SEL_SW << CH_CFG_H_HS_SEL_DST_POS |
           DWAXIDMAC_HS_SEL_SW << CH_CFG_H_HS_SEL_SRC_POS);
	chan_reg_write32(chan, CH_CFG_H, val);

	/* CH_INTSTATUS_ENA */
	irq_mask = DWAXIDMAC_IRQ_DMA_TRF | DWAXIDMAC_IRQ_ALL_ERR;
	axi_chan_irq_sig_set(chan, irq_mask);

	/* Generate 'suspend' status but don't generate interrupt */
	irq_mask |= DWAXIDMAC_IRQ_SUSPENDED;
	axi_chan_irq_set(chan, irq_mask);

	/* DMAC_CHENREG */
	axi_chan_enable(chan);
	/* wait v to v dma trans end */
	do {
		status = axi_chan_irq_read(chan);
		axi_chan_irq_clear(chan, status);
		if (status & DWAXIDMAC_IRQ_ALL_ERR) {
			axi_chan_handle_err(chan, status);
			break;
		} else if (status & DWAXIDMAC_IRQ_DMA_TRF) {
			axi_chan_block_xfer_complete(chan);
			return 0;
		}
		timeout++;
		udelay(2);
	} while (timeout < 30000);
	gb_printf(KERN_INFO, "%s IRQ status: 0x%08x timeout\n", __func__, atomic_read(&chan->stop));

	return -1;
}

int gb_v2vdma_init(struct gb_pcie_info *pcie_info)
{
	int result = 0;
	struct v2vdma_ctrl_info *dma_ctrl;
	int i = 0;
	int ddr_bar_id = get_genbu_ddr_bar_id(pcie_info->gb_dev_info);
	int reg_bar_id = get_genbu_gpu_reg_bar_id(pcie_info->gb_dev_info);

	dma_ctrl = (struct v2vdma_ctrl_info *)kzalloc(sizeof(struct v2vdma_ctrl_info), GFP_KERNEL);
	if (dma_ctrl == NULL) {
		gb_printf(KERN_ERR, "Error: %s, kmalloc failed\n", __func__);
		return -ENOMEM;
	}

	dma_ctrl->regs = pcie_info->pci_bars[reg_bar_id].mmio + CS_CNOC_S0_V2VDMA_BASE;
	dma_ctrl->vram_addr = pcie_info->pci_bars[ddr_bar_id].mmio;
	for (i = 0; i < DW_AXI_DMA_CHANNELS; i++) {
		struct axi_dma_chan *chan = &(dma_ctrl->chan[i]);
		chan->id = i;
		chan->chan_regs = dma_ctrl->regs + COMMON_REG_LEN + i * CHAN_REG_LEN;
    	chan->chip = dma_ctrl;
    	atomic_set(&chan->descs_allocated, 0);
	}

	result = axi_dma_hw_init(dma_ctrl);
	if (result)
		goto err_kfree;

	v2vdma_infos = dma_ctrl;

	return result;
err_kfree:
	kfree(dma_ctrl);
	gb_printf(KERN_ERR, "%s %d fail!\n", __func__, __LINE__);
	return result;
}

void gb_v2vdma_release(struct gb_pcie_info *pcie_info)
{
	gb_printf(KERN_ERR, "%s %d v2vdma_info:0x%pK\n",
		__func__, __LINE__, v2vdma_infos);

	if (v2vdma_infos) {
		kfree(v2vdma_infos);
		v2vdma_infos = NULL;
	}
}
