// SPDX-License-Identifier: GPL-2.0-only
/*
 * GB backlight control, board code has to setup
 */

#include <linux/version.h>
#include <linux/delay.h>
#include <linux/gpio/consumer.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/regulator/consumer.h>
#include <linux/slab.h>
#include "gb_backlight.h"
#include "ip/gb_dp_dptx.h"
#include "ip/gb_dp.h"

#define MCU_PWM_CONFIG_REG_BASE (0x0003FC08)
#define DPTX_AUX_I2C_ADDR 0x37

static struct backlight_data *bl_data[GB02_CONNECTER_NUM];
static u32 reserve_brightness[GB02_CONNECTER_NUM];

struct backlight_data *gb_get_backlight_data(int index)
{
	return bl_data[index];
}

static void gb_backlight_update_brightness(struct dptx *dptx_info, u32 brightness)
{
	int index = dptx_info->index;
	struct backlight_data *data = bl_data[index];
	char brightness_buf[7] = {0x51, 0x84, 0x3, 0x10, 0x0, 0x0, 0x0}; //DPCD brightness configure CMD
	char temp = 0;

	if (!data || data->last_brightness == brightness)
		return;

	if (dptx_info->is_edp) {
		base_reg_write(brightness, data->sysctl_iofunc_base + 0x1000000, MCU_PWM_CONFIG_REG_BASE + index * 4);
	} else {
		if (!data->by_dptx)
			dptx_aux_rw(dptx_info, false, true, true, true, DPTX_AUX_I2C_ADDR, &temp, 0x1);

		brightness_buf[5] = brightness;
		brightness_buf[6] = 0xa8 ^ brightness;
		dptx_aux_rw(dptx_info, false, true, true, false, DPTX_AUX_I2C_ADDR, brightness_buf, 0x7);

		if (!data->by_dptx)
			dptx_aux_rw(dptx_info, false, true, false, true, DPTX_AUX_I2C_ADDR, &temp, 0x1);
	}
	data->by_dptx = false;
	data->last_brightness = brightness;
	reserve_brightness[index] = brightness;
}

static int gb_backlight_update_status(struct backlight_device *bl)
{
	int brightness = bl->props.brightness;
	struct backlight_data *data = bl_get_data(bl);

	if (brightness == 0) {
		bl->props.brightness = data->last_brightness;
		return -EINVAL;
	}

	gb_backlight_update_brightness(data->dptx_info, brightness);
	return 0;
}

static const struct backlight_ops gb_backlight_ops = {
	.update_status = gb_backlight_update_status,
};

int gb_register_backlight(struct gb_pcie_info *pcie_info, struct dptx *dptx_info)
{
	struct backlight_properties props;
	struct backlight_data *data;
	char backlight_name[20];
	int index = dptx_info->index;

	if (bl_data[index])
		return 0;

	bl_data[dptx_info->index] = kzalloc(sizeof(struct backlight_data), GFP_KERNEL);
	if (!bl_data[dptx_info->index])
		return -ENOMEM;
	data = bl_data[index];

	data->sysctl_iofunc_base = pcie_info->pci_bars[0].mmio;
	data->dptx_info = dptx_info;

	memset(&props, 0, sizeof(props));
	props.type = BACKLIGHT_RAW;
	props.max_brightness = 100;
	sprintf(backlight_name, "gb_backlight%d", index);
	data->bl = backlight_device_register(backlight_name, dptx_info->dev, data,
					&gb_backlight_ops, &props);

	if (IS_ERR(data->bl)) {
		dev_err(dptx_info->dev, "failed to register backlight\n");
		data->bl = NULL;
		return -ENODEV;
	}
	if (reserve_brightness[index]) {
		data->bl->props.brightness = reserve_brightness[index];
		backlight_update_status(data->bl);
	}

	return 0;
}

void gb_unregister_backlight(struct dptx *dptx_info)
{
	int index = dptx_info->index;
	struct backlight_data *data = bl_data[index];

	if (!data)
		return;
	if (data->bl)
		backlight_device_unregister(data->bl);
	kfree(data);
	bl_data[index] = NULL;
}
