// SPDX-License-Identifier: GPL-2.0
/*
 *  Hwmon driver for fans connected to PWM lines.
 *
 */

#include "common/gb-peripherals-common.h"
#include <linux/hwmon-sysfs.h>
#include <linux/hwmon.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/pci.h>
#include <linux/gpio.h>
#include <linux/sysfs.h>
#include <linux/version.h>
#include <linux/thermal.h>
#include "gpu/gb_device.h"

#define LEVEL_NUM 5
//#define FAN_DEBUG
#ifdef FAN_DEBUG
#define GB02_FAN_DEBUG(arg...) gb_printf(KERN_INFO, "[" GB02_PWM_FAN_DEV_NAME "] " arg)
#else
#define GB02_FAN_DEBUG(arg...)
#endif

struct fan_ctx {
	struct mutex lock;
	unsigned int fan_state;
	unsigned int fan_max_state;
	unsigned int *fan_cooling_levels;
	void __iomem *sysctl_iofunc_base; // bar0
	unsigned int rpm;
	bool user_mode;
	unsigned int user_state;
	struct thermal_cooling_device *cdev;
};

static void fan_update_state(struct fan_ctx *ctx, unsigned long pwm)
{
	int i;

	for (i = 0; i < ctx->fan_max_state; ++i) {
		if (pwm < ctx->fan_cooling_levels[i + 1])
			break;
	}

	ctx->fan_state = i;
}

static ssize_t set_level(struct device *dev, struct device_attribute *attr, const char *buf,
					   size_t count)
{
	struct fan_ctx *ctx = dev_get_drvdata(dev);
	unsigned long		level;

	if (kstrtoul(buf, 10, &level) || level > LEVEL_NUM)
		return -EINVAL;

	if (level > 0) {
		base_reg_write(level - 1, ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC04);
		ctx->user_mode = true;
		ctx->user_state = level - 1;
	} else {
		ctx->user_mode = false;
		ctx->user_state = 0xf;
	}

	fan_update_state(ctx, ctx->fan_cooling_levels[level-1]);
	return count;
}

static ssize_t show_level(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct fan_ctx *ctx = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", ctx->fan_state + 1);
}


static ssize_t show_rpm(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct fan_ctx *ctx = dev_get_drvdata(dev);

	ctx->rpm = base_reg_read(ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC00);

	return sprintf(buf, "%u\n", ctx->rpm);
}

static SENSOR_DEVICE_ATTR(fan1_level, 0644, show_level, set_level, 0);
static SENSOR_DEVICE_ATTR(fan1_input, 0444, show_rpm, NULL, 0);

static struct attribute *fan_attrs[] = {
	&sensor_dev_attr_fan1_input.dev_attr.attr,
	&sensor_dev_attr_fan1_level.dev_attr.attr,
	NULL,
};

ATTRIBUTE_GROUPS(fan);

/* thermal cooling device callbacks */
static int fan_get_max_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
	struct fan_ctx *ctx = cdev->devdata;

	if (!ctx)
		return -EINVAL;

	*state = ctx->fan_max_state;

	return 0;
}

static int fan_get_cur_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
	struct fan_ctx *ctx = cdev->devdata;

	if (!ctx)
		return -EINVAL;

	*state = ctx->fan_state;

	return 0;
}

static int fan_set_cur_state(struct thermal_cooling_device *cdev, unsigned long state)
{
	struct fan_ctx *ctx = cdev->devdata;

	if (!ctx || (state > ctx->fan_max_state))
		return -EINVAL;

	if (state == ctx->fan_state)
		return 0;

	if (ctx->user_mode)
		return 0;

	base_reg_write(state, ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC04);
	ctx->fan_state = state;

	return 0;
}

static const struct thermal_cooling_device_ops fan_cooling_ops = {
	.get_max_state = fan_get_max_state,
	.get_cur_state = fan_get_cur_state,
	.set_cur_state = fan_set_cur_state,
};

static int pwm_fan_get_cooling_data(struct device *dev, struct fan_ctx *ctx)
{
	int num = LEVEL_NUM, i;

	ctx->fan_cooling_levels = devm_kcalloc(dev, LEVEL_NUM, sizeof(u32), GFP_KERNEL);
	if (!ctx->fan_cooling_levels)
		return -ENOMEM;
	for (i = 0; i < LEVEL_NUM; i++) {
		ctx->fan_cooling_levels[i] =  i + 1;
		if (ctx->fan_cooling_levels[i] > LEVEL_NUM) {
			dev_err(dev, "fan state[%d]:%d > %d\n", i, ctx->fan_cooling_levels[i], LEVEL_NUM);
			return -EINVAL;
		}
	}

	ctx->fan_max_state = num - 1;

	return 0;
}

int gb02_fan_drv_init(struct gb_device *gbdev, struct gb02_mcu_peripherals_info *peri_info)
{
	struct thermal_cooling_device *cdev;
	struct fan_ctx *ctx;
	struct device *hwmon;
	int ret;

	ctx = devm_kzalloc(gbdev->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	mutex_init(&ctx->lock);

	ctx->sysctl_iofunc_base = gbdev->gb_pcie->pci_bars[0].mmio; // bar0

	hwmon = devm_hwmon_device_register_with_groups(gbdev->dev, "pwmfan", ctx, fan_groups);
	if (IS_ERR(hwmon)) {
		dev_err(gbdev->dev, "Failed to register hwmon device\n");
		ret = PTR_ERR(hwmon);
		return ret;
	}

	ret = pwm_fan_get_cooling_data(gbdev->dev, ctx);
	if (ret)
		return ret;

	ctx->fan_state = ctx->fan_max_state;

	if (IS_ENABLED(CONFIG_THERMAL)) {
		cdev = thermal_of_cooling_device_register(gbdev->dev->of_node, "pwm-fan", ctx,
				&fan_cooling_ops);
		if (IS_ERR(cdev)) {
			dev_err(gbdev->dev, "Failed to register pwm-fan as cooling device");
			ret = PTR_ERR(cdev);
			return ret;
		}
		ctx->cdev = cdev;
	}
	peri_info->priv = ctx;

	return ret;
}

void gb02_fan_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	struct fan_ctx *ctx = (struct fan_ctx *)peri_info->priv;

	thermal_cooling_device_unregister(ctx->cdev);
}

int gb02_fan_drv_suspend(struct gb02_mcu_peripherals_info *peri_info)
{
	struct fan_ctx *ctx = (struct fan_ctx *)peri_info->priv;

	base_reg_write(0, ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC04);
	ctx->fan_state = 0;

	return 0;
}

int gb02_fan_drv_resume(struct gb02_mcu_peripherals_info *peri_info)
{
	struct fan_ctx *ctx = (struct fan_ctx *)peri_info->priv;

	if (ctx->user_mode)
		base_reg_write(ctx->user_state, ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC04);

	return 0;
}
