// SPDX-License-Identifier: GPL-2.0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <linux/io.h>
#include <linux/uaccess.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/pinctrl/consumer.h>
#include <linux/seq_file.h>
#include "gpu/gb_device.h"
#include "pinctrl-gb02-mcu.h"
#include "common/gb-peripherals-common.h"

//#define PINCTRL_DEBUG
#ifdef PINCTRL_DEBUG
#define GB02_PINCTRL_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_LGPIO_DEVICE_NAME "] " arg)
#else
#define GB02_PINCTRL_DEBUG(arg...)
#endif


static DEFINE_MUTEX(sysfs_lock);
static int gb_gpiod_export(unsigned int gpio);

struct gb02_pin_group {
	const char *name;
	unsigned int start_pin;
	unsigned int npins;
	u32 reg_mask;
	u32 val[MAX_FUNCS];
	const char *funcs[MAX_FUNCS];
	unsigned int *pins;
};

struct gb02_pinctrl_pin_data {
	int nr_pins;
	char *name;
	struct gb02_pin_group *groups;
	int ngroups;
};

struct gb02_pmx_func {
	const char *name;
	const char **groups;
	unsigned int ngroups;
};

struct gb02_gpio_bank {
	void __iomem *base;
	spinlock_t lock;
};

struct gb02_attr_data {
	unsigned int gpio;
	struct mutex mutex;
};

struct gb02_pinctrl {
	void __iomem *regs;
	void __iomem *sysctl_iofunc_base;			 // bar0
	void __iomem *sysctl_system_peripheral_base; // bar1
	void __iomem *sysctl_cfg_base;				 // bar4
	struct device *dev;
	struct gb02_pin_group *groups;
	unsigned int ngroups;
	struct gb02_pmx_func *funcs;
	unsigned int nfuncs;
	unsigned int nbanks;
	struct gb02_gpio_bank *banks;
	struct gb02_pinctrl_pin_data *data;
};

static struct gb02_pinctrl *g_pinctrl_infos;

static inline int gb02_gpio_pin(int gpio)
{
	return gpio % GB02_MCU_GPIO_PINS_PER_BANK;
}

/* GPIO functions */
static inline void __gb02_gpio_set(struct gb02_gpio_bank *bank,
	 unsigned int offset, int value)
{
	int dat = 0;

	dat = readl_relaxed(bank->base + NUCLEI_GPIO_OUTPUT_VAL);
	dat &= ~BIT(offset);
	GB02_PINCTRL_DEBUG("offset %u set %d\n", offset, value);
	if (value)
		dat |= BIT(offset);
	writel_relaxed(dat, bank->base + NUCLEI_GPIO_OUTPUT_VAL);
}

static inline u32 __gb02_gpio_get(struct gb02_gpio_bank *bank,
	unsigned int offset, int mode)
{
	u32 val = 0;

	if (mode == GB02_GPIO_LINE_DIRECTION_IN)
		val = readl_relaxed(bank->base + NUCLEI_GPIO_INPUT_VAL);
	else if (mode == GB02_GPIO_LINE_DIRECTION_OUT)
		val = readl_relaxed(bank->base + NUCLEI_GPIO_OUTPUT_VAL);
	GB02_PINCTRL_DEBUG("offset %d get %d\n", offset, val);

	return !!(val & BIT(offset));
}

static inline bool gpio_is_valid(unsigned int gpio)
{
	int ngpio;
	struct gb02_pinctrl *info = g_pinctrl_infos;

	ngpio = info->data->nr_pins;
	if (gpio > ngpio)
		return false;

	return true;
}

static ssize_t direction_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	ssize_t			status;
	int mode;
	unsigned int gpio;
	struct gb02_attr_data *data = dev_get_drvdata(dev);

	gpio = data->gpio;
	mode = gb02_gpio_get_direction(gpio);
	status = sprintf(buf, "%s\n", mode ? "out" : "in");

	return status;
}

static ssize_t direction_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	ssize_t			status;
	unsigned int gpio;
	struct gb02_attr_data *data = dev_get_drvdata(dev);

	gpio = data->gpio;
	if (sysfs_streq(buf, "out"))
		status = gb02_gpio_set_direction(gpio, GB02_GPIO_LINE_DIRECTION_OUT);
	else if (sysfs_streq(buf, "in"))
		status = gb02_gpio_set_direction(gpio, GB02_GPIO_LINE_DIRECTION_IN);
	else
		status = -EINVAL;

	return status ? : size;
}
static DEVICE_ATTR_RW(direction);

static ssize_t value_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	ssize_t			status;
	unsigned int gpio;
	struct gb02_attr_data *data = dev_get_drvdata(dev);

	gpio = data->gpio;
	status = gb02_gpio_get(gpio);
	if (status < 0)
		goto err;

	status = sprintf(buf, "%ld\n", status);
err:

	return status;
}

static ssize_t value_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	ssize_t status = 0;
	long		value;
	unsigned int gpio;
	struct gb02_attr_data *data = dev_get_drvdata(dev);

	gpio = data->gpio;
	if (size <= 2 && isdigit(buf[0]) &&
	    (size == 1 || buf[1] == '\n'))
		value = buf[0] - '0';
	else
		status = kstrtol(buf, 0, &value);

	if (status == 0) {
		gb02_gpio_set(gpio, value);
		status = size;
	}

	return status;
}
static DEVICE_ATTR_PREALLOC(value, S_IWUSR | S_IRUGO, value_show, value_store);

static struct attribute *gpio_attrs[] = {
	&dev_attr_direction.attr,
	&dev_attr_value.attr,
	NULL,
};

static const struct attribute_group gpio_group = {
	.attrs = gpio_attrs,
};

static const struct attribute_group *gpio_groups[] = {
	&gpio_group,
	NULL
};

static int gb02_device_match_name(struct device *dev, const void *name)
{
	return !strcmp(dev_name(dev), name);
}

/*
 * /sys/class/gpio/export ... write-only
 *	integer N ... number of GPIO to export (full access)
 * /sys/class/gpio/unexport ... write-only
 *	integer N ... number of GPIO to unexport
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
static ssize_t export_store(struct class *class,
				struct class_attribute *attr,
				const char *buf, size_t len)
#else
static ssize_t export_store(const struct class *class,
				const struct class_attribute *attr,
				const char *buf, size_t len)
#endif	
{
	long			gpio;
	int			status;

	status = kstrtol(buf, 0, &gpio);
	if (status < 0)
		goto done;

	status = gb02_gpio_request(gpio);
	if (status < 0) {
		if (status == -EPROBE_DEFER)
			status = -ENODEV;
		goto done;
	}

	status = gb_gpiod_export(gpio);
	if (status < 0)
		gb02_gpio_free(gpio);
done:
	if (status)
		GB02_PINCTRL_DEBUG("%s: status %d\n", __func__, status);
	return status ? : len;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
static CLASS_ATTR_WO(export);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
static ssize_t unexport_store(struct class *class,
				struct class_attribute *attr,
				const char *buf, size_t len)
#else
static ssize_t unexport_store(const struct class *class,
				const struct class_attribute *attr,
				const char *buf, size_t len)
#endif	
{
	long			gpio;
	int			status;
	struct device *dev;
	char dev_name[10];

	status = kstrtol(buf, 0, &gpio);
	if (status < 0)
		goto done;

	sprintf(dev_name, "gpio%ld", gpio);
	/* No extra locking here; FLAG_SYSFS just signifies that the
	 * request and export were done by on behalf of userspace, so
	 * they may be undone on its behalf too.
	 */
	dev = class_find_device(class, NULL, dev_name, gb02_device_match_name);
	if (!dev)
		gb_printf(KERN_ERR, "can't find dev\n");
	device_destroy(class, dev->devt);
	gb02_gpio_free(gpio);
done:
	if (status)
		GB02_PINCTRL_DEBUG("%s: status %d\n", __func__, status);
	return status ? : len;
}
static CLASS_ATTR_WO(unexport);

static struct attribute *gb02_gpio_class_attrs[] = {
	&class_attr_export.attr,
	&class_attr_unexport.attr,
	NULL,
};

ATTRIBUTE_GROUPS(gb02_gpio_class);


static struct class gb02_gpio_class = {
	.name =		"gb02_gpio",
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	.owner =	THIS_MODULE,
#endif
	.class_groups = gb02_gpio_class_groups,
};


static int gb_gpiod_export(unsigned int gpio)
{
	int			status;
	const char		*ioname = NULL;
	struct device		*dev;
	struct gb02_attr_data *data;

	mutex_lock(&sysfs_lock);

	data = kzalloc(sizeof(*data), GFP_KERNEL);
	data->gpio = gpio;
	mutex_init(&data->mutex);
	dev = device_create_with_groups(&gb02_gpio_class, g_pinctrl_infos->dev,
					MKDEV(0, 0), data, gpio_groups,
					ioname ? ioname : "gpio%u",
					gpio);
	if (IS_ERR(dev)) {
		status = PTR_ERR(dev);
		goto err_unlock;
	}

	mutex_unlock(&sysfs_lock);
	return 0;

err_unlock:
	mutex_unlock(&sysfs_lock);
	return status;
}

static int gb02_gpio_sysfs_init(void)
{
	int status = 0;

	status = class_register(&gb02_gpio_class);
	if (status < 0)
		return status;

	return status;
}

#ifdef TEST_MISC_GPIO
static int gb02_pinctrl_mcu_gpio_ioctl(struct gb02_pinctrl *info, u32 gpio)
{
	/* GPIO range judge */
	if ((gpio < GPIO_IRQ_SUPPORT_GPIO_NO_MIN) ||
	 (gpio > GPIO_IRQ_SUPPORT_GPIO_NO_MAX)) {
		gb_printf(KERN_ERR, "GPIO IRQ : unsupported gpio(%d)\n", gpio);
		return -1;
	}

	/* GPIO output mode */
	gb02_gpio_set(gpio, 1);
	udelay(10);
	gb02_gpio_set(gpio, 0);
	udelay(10);
	gb02_gpio_set(gpio, 1);

	return 0;
}

static int my_open(struct inode *inode, struct file *filp)
{
	filp->private_data = g_pinctrl_infos;
	return 0;
}

static int my_release(struct inode *inode, struct file *filp)
{
	filp->private_data = NULL;
	return 0;
}

static long gb02_test_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	struct gb02_pinctrl *info = filp->private_data;

	GB02_PINCTRL_DEBUG("enter ioctrl!\n");

	switch (cmd) {
	case PINCTRL_MCU_CMD_GPIO40:
		GB02_PINCTRL_DEBUG("=====MCU's GPIO40 start!======\n");
		ret = gb02_pinctrl_mcu_gpio_ioctl(info, 40);
		break;
	case PINCTRL_MCU_CMD_GPIO41:
		GB02_PINCTRL_DEBUG("=====MCU's GPIO41 start!======\n");
		ret = gb02_pinctrl_mcu_gpio_ioctl(info, 41);
		break;
	case PINCTRL_MCU_CMD_GPIO42:
		GB02_PINCTRL_DEBUG("=====MCU's GPIO42 start!======\n");
		ret = gb02_pinctrl_mcu_gpio_ioctl(info, 42);
		break;
	case PINCTRL_MCU_CMD_GPIO43:
		GB02_PINCTRL_DEBUG("=====MCU's GPIO43 start!======\n");
		ret = gb02_pinctrl_mcu_gpio_ioctl(info, 43);
		break;
	default:
		break;
	}

	return 0;
}

static const struct file_operations mmap_fops = {
	.owner = THIS_MODULE,
	.open = my_open,
	.release = my_release,
	.unlocked_ioctl = gb02_test_ioctl,
};

static struct miscdevice gbfpga_misc = {
	.fops = &mmap_fops,
	.name = "gbfpgagpio",
	.minor = MISC_DYNAMIC_MINOR,
};
#endif

static void gb02_pmx_set_mode(struct gb02_gpio_bank *bank,
	int pin, u32 mode)
{
	u32 val;
	unsigned long flags;

	spin_lock_irqsave(&bank->lock, flags);

	if (!mode) {
		/* GPIO input mode */
		val = readl_relaxed(bank->base + 0x4);
		val |= BIT(pin);
		writel_relaxed(val, bank->base + 0x4);
		val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE0);
		val &= ~BIT(pin);
		writel_relaxed(val, bank->base + NUCLEI_GPIO_MODE0);

		val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE1);
		val &= ~BIT(pin);
		writel_relaxed(val, bank->base + NUCLEI_GPIO_MODE1);
	} else {
		/* GPIO output mode */
		val = readl_relaxed(bank->base + 0x4);
		val &= ~BIT(pin);
		writel_relaxed(val, bank->base + 0x4);

		val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE0);
		val &= ~BIT(pin);
		val |= BIT(pin);
		writel_relaxed(val, bank->base + NUCLEI_GPIO_MODE0);

		val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE1);
		val &= ~BIT(pin);
		writel_relaxed(val, bank->base + NUCLEI_GPIO_MODE1);
	}

	spin_unlock_irqrestore(&bank->lock, flags);
}

static int gb02_pmx_set_by_name(struct gb02_pinctrl *info,
	const char *name,
	struct gb02_pin_group *grp)
{
	unsigned int mask = grp->reg_mask;
	int func = 0, val = 0, dat = 0;

	func = match_string(grp->funcs, MAX_FUNCS, name);
	if (func < 0)
		return -ENOTSUPP;

	val = grp->val[func];
	dat = base_reg_read(info->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
	dat &= ~mask;
	dat |= val;
	base_reg_write(dat, info->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);

	return 0;
}

static struct gb02_pin_group *gb02_find_next_grp_by_pin(
	struct gb02_pinctrl *info, int pin, int *grp)
{
	while (*grp < info->ngroups) {
		struct gb02_pin_group *group = &info->groups[*grp];
		int j;

		*grp = *grp + 1;
		for (j = 0; j < group->npins; j++)
			if (group->pins[j] == pin)
				return group;
	}
	return NULL;
}

static void gb02_pmx_get_mode(struct gb02_gpio_bank *bank, int pin, u32 *mode)
{
	u32 val = 0, mode0 = 0, mode1 = 0;
	unsigned long flags;

	spin_lock_irqsave(&bank->lock, flags);

	val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE0);
	val &= BIT(pin);
	mode0 = !!val;

	val = readl_relaxed(bank->base + NUCLEI_GPIO_MODE1);
	val &= BIT(pin);
	mode1 = !!val;

	*mode = (mode1 << 1) | mode0;

	spin_unlock_irqrestore(&bank->lock, flags);
}

static int gb02_pmx_set_default(struct gb02_pinctrl *info,
	struct gb02_pin_group *grp)
{
	unsigned int mask = grp->reg_mask;
	int func = 0, val = 0, dat = 0;

	if (strcmp(grp->name, "test") == 0)
		func = 1;

	GB02_PINCTRL_DEBUG("enable function %s group %s\n",
	  grp->funcs[func], grp->name);

	val = grp->val[func];
	dat = base_reg_read(info->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
	dat &= ~mask;
	dat |= val;
	GB02_PINCTRL_DEBUG("sysctl_iofunc_set default %x\n", dat);
	base_reg_write(dat, info->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);

	return 0;
}

static struct gb02_gpio_bank *gb02_gpio_to_bank(unsigned int gpio)
{
	int nbank;
	struct gb02_pinctrl *info = g_pinctrl_infos;

	nbank = gpio / GB02_MCU_GPIO_PINS_PER_BANK;

	return &info->banks[nbank];
}

void gb02_gpio_irq_enable(unsigned int gpio, bool enable)
{
	u32 val = 0;
	/* enable or disable LGPIO5_0---LGPIO5_3 sw interrupt */
	if (gpio < 40) {
		GB02_PINCTRL_DEBUG("unsupport gpio%d to irq\n", gpio);
		return;
	}
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, g_pinctrl_infos->sysctl_cfg_base, CFG_MCU_XIP_ADDR_OFFSET);
	base_reg_write(INT_CONTROLLER_BASE_ADDR, g_pinctrl_infos->sysctl_cfg_base, CFG_SYS_PERI_ADDR_OFFSET);
	val = base_reg_read(g_pinctrl_infos->sysctl_system_peripheral_base, INT_CONTROLLER_OFFSET + INTC_IRQ_INTEN_H_OFFSET);
	if (enable)
		val |= BIT(gpio - 40) << (45 - 32);
	else
		val &= ~BIT(gpio - 40) << (45 - 32);
	base_reg_write(val, g_pinctrl_infos->sysctl_system_peripheral_base, INT_CONTROLLER_OFFSET + INTC_IRQ_INTEN_H_OFFSET);
}

int gb02_gpio_request(unsigned int gpio)
{
	struct gb02_pinctrl *info = g_pinctrl_infos;
	struct gb02_pin_group *group;
	int grp = 0;
	bool valid;

	GB02_PINCTRL_DEBUG("requesting gpio %d\n", gpio);

	valid = gpio_is_valid(gpio);
	if (!valid) {
		gb_printf(KERN_ERR,"gpio%u is valid\n", gpio);
		return -1;
	}

	while ((group = gb02_find_next_grp_by_pin(info, gpio, &grp)))
		gb02_pmx_set_by_name(info, "gpio", group);

	return 0;
}

void gb02_gpio_free(unsigned int gpio)
{
	struct gb02_pinctrl *info = g_pinctrl_infos;
	struct gb02_pin_group *group;
	int grp = 0;

	GB02_PINCTRL_DEBUG("set pin %u as default\n", gpio);

	/* Set the pin to some default state */
	while ((group = gb02_find_next_grp_by_pin(info, gpio, &grp)))
		gb02_pmx_set_default(info, group);
}

int gb02_gpio_set_direction(unsigned int gpio, bool direction)
{
	struct gb02_gpio_bank *bank;
	int pin;
	bool valid;

	valid = gpio_is_valid(gpio);
	if (!valid) {
		gb_printf(KERN_ERR,"gpio%u is valid\n", gpio);
		return -1;
	}

	bank = gb02_gpio_to_bank(gpio);
	pin = gb02_gpio_pin(gpio);

	gb02_pmx_set_mode(bank, pin, direction);

	return 0;
}

int gb02_gpio_get_direction(unsigned int gpio)
{
	struct gb02_gpio_bank *bank = gb02_gpio_to_bank(gpio);
	int pin = gb02_gpio_pin(gpio);
	int ret = 0;
	u32 mode;

	gb02_pmx_get_mode(bank, pin, &mode);
	if (mode == GB02_GPIO_LINE_DIRECTION_IN)
		ret = 0;
	else if (mode == GB02_GPIO_LINE_DIRECTION_OUT)
		ret = 1;
	else if (mode == 2) {
		gb_printf(KERN_ERR, "pin %u is in open-drain mode\n", gpio);
		ret = -EINVAL;
	} else
		ret = -EINVAL;

	return ret;
}

int gb02_gpio_get(unsigned int gpio)
{
	struct gb02_gpio_bank *bank = gb02_gpio_to_bank(gpio);
	int pin = gb02_gpio_pin(gpio);
	int mode = gb02_gpio_get_direction(gpio);

	return __gb02_gpio_get(bank, pin, mode);
}

int gb02_gpio_set(unsigned int gpio, int value)
{
	int ret = 0;
	u32 mode;
	struct gb02_gpio_bank *bank = gb02_gpio_to_bank(gpio);
	int pin = gb02_gpio_pin(gpio);

	gb02_pmx_get_mode(bank, pin, &mode);
	if (mode != GB02_GPIO_LINE_DIRECTION_OUT) {
		ret = gb02_gpio_set_direction(gpio, GB02_GPIO_LINE_DIRECTION_OUT);
		if (ret)
			gb_printf(KERN_ERR,"gb02_gpio_set gpio%u set %d falile\n", gpio, value);
	}
	__gb02_gpio_set(bank, pin, value);

	return ret;
}

static int gb02_add_function(struct gb02_pmx_func *funcs,
	int *funcsize, const char *name)
{
	int i = 0;

	if (*funcsize <= 0)
		return -EOVERFLOW;

	while (funcs->ngroups) {
		/* function already there */
		if (strcmp(funcs->name, name) == 0) {
			funcs->ngroups++;

			return -EEXIST;
		}
		funcs++;
		i++;
	}

	/* append new unique function */
	funcs->name = name;
	funcs->ngroups = 1;
	(*funcsize)--;

	return 0;
}

static int gb02_fill_group(struct gb02_pinctrl *info)
{
	int n, num = 0, funcsize = info->data->nr_pins;

	for (n = 0; n < info->ngroups; n++) {
		struct gb02_pin_group *grp = &info->groups[n];
		int i, f;

		grp->pins = devm_kcalloc(info->dev,
			grp->npins, sizeof(*grp->pins),
			GFP_KERNEL);
		if (!grp->pins)
			return -ENOMEM;

		for (i = 0; i < grp->npins; i++)
			grp->pins[i] = grp->start_pin + i;

		for (f = 0; (f < MAX_FUNCS) && grp->funcs[f]; f++) {
			int ret;
			/* check for unique functions and count groups */
			ret = gb02_add_function(info->funcs, &funcsize,
			   grp->funcs[f]);
			if (ret == -EOVERFLOW)
				dev_err(info->dev,
				  "More functions than pins(%d)\n",
				  info->data->nr_pins);
			if (ret < 0)
				continue;
			num++;
		}
	}

	info->nfuncs = num;

	return 0;
}

static int gb02_fill_func(struct gb02_pinctrl *info)
{
	struct gb02_pmx_func *funcs = info->funcs;
	int n;

	for (n = 0; n < info->nfuncs; n++) {
		const char *name = funcs[n].name;
		const char **groups;
		int g;

		funcs[n].groups = devm_kcalloc(info->dev,
			funcs[n].ngroups,
			sizeof(*(funcs[n].groups)),
			GFP_KERNEL);
		if (!funcs[n].groups)
			return -ENOMEM;

		groups = funcs[n].groups;

		for (g = 0; g < info->ngroups; g++) {
			struct gb02_pin_group *gp = &info->groups[g];
			int f;

			f = match_string(gp->funcs, MAX_FUNCS, name);
			if (f < 0)
				continue;

			*groups = gp->name;
			groups++;
		}
	}
	return 0;
}

static int gb02_fill_bank(struct gb02_pinctrl *info)
{
	int ret = 0, j = 0;
	struct gb02_gpio_bank *bank;

	info->banks = devm_kcalloc(info->dev, GB02_MCU_GPIO_BANKS_NUM, sizeof(*info->banks),
	   GFP_KERNEL);
	if (!info->banks)
		return -ENOMEM;

	for (j = 0; j < GB02_MCU_GPIO_BANKS_NUM; j++) {
		bank = &info->banks[info->nbanks];
		bank->base = info->regs + info->nbanks * 0x1000;
		spin_lock_init(&bank->lock);
		info->nbanks++;
	}

	return ret;
}

static int gb02_pinctrl_register(struct device *dev,
	struct gb02_pinctrl *info)
{
	const struct gb02_pinctrl_pin_data *pin_data = info->data;
	int ret;

	info->groups = pin_data->groups;
	info->ngroups = pin_data->ngroups;

	/*
	 * we allocate functions for number of pins and hope there are
	 * fewer unique functions than pins available
	 */
	info->funcs = devm_kcalloc(dev,
	   pin_data->nr_pins,
	   sizeof(struct gb02_pmx_func),
	   GFP_KERNEL);
	if (!info->funcs)
		return -ENOMEM;

	ret = gb02_fill_bank(info);
	if(ret)
		return ret;

	ret = gb02_fill_group(info);
	if (ret)
		return ret;

	ret = gb02_fill_func(info);
	if (ret)
		return ret;

	return 0;
}

static int gb02_pinctl_init(struct gb02_pinctrl *info)
{
#ifdef TEST_MISC_GPIO
	int ret = 0;
	int i = 0;
	u32 gpio = 0;

	for (i = 0; i < 4; i++) {
		gpio = 40 + i;

		/* GPIO request */
		ret = gb02_gpio_request(gpio);
		if (ret) {
			gb_printf(KERN_ERR, "Failed to request gpio(%d)\n", gpio);
			return ret;
		}

		/* GPIO input mode */
		gb02_gpio_set_direction(gpio, GB02_GPIO_LINE_DIRECTION_IN);
	}
#endif
	return 0;
}

#define PIN_GRP_GPIO(_name, _start, _nr, _mask, _func1) \
	{	  \
		.name = _name,	\
		.start_pin = _start,	\
		.npins = _nr,	 \
		.reg_mask = _mask,   \
		.val = {0, _mask},   \
		.funcs = { _func1,   \
				   "gpio" }  \
	}

#define PIN_GRP_GPIO_2(_name, _start, _nr, _mask, _val1, _val2, _func1) \
	{	   \
		.name = _name,	 \
		.start_pin = _start,	 \
		.npins = _nr,	  \
		.reg_mask = _mask,	\
		.val = {_val1, _val2},   \
		.funcs = { _func1,	\
				   "gpio" }   \
	}

static struct gb02_pin_group gb02_mcu_pin_groups[] = {
	PIN_GRP_GPIO_2("test", 0, 8, BIT(11), BIT(11), 0, "plltest"),
	PIN_GRP_GPIO("spis", 8, 4, BIT(0), "spis"),
	PIN_GRP_GPIO("uart0", 12, 2, BIT(1), "uart"),
	PIN_GRP_GPIO("uart1", 14, 2, BIT(2), "uart"),
	PIN_GRP_GPIO("qspi0", 16, 6, BIT(3), "qspi"),
	PIN_GRP_GPIO("qspi1", 22, 6, BIT(4), "qspi"),
	PIN_GRP_GPIO("i2c0", 28, 2, BIT(5), "i2c"),
	PIN_GRP_GPIO("i2c1", 30, 2, BIT(6), "i2c"),
	PIN_GRP_GPIO("i2c2", 32, 2, BIT(7), "i2c"),
	PIN_GRP_GPIO("i2c3", 34, 2, BIT(8), "i2c"),
	PIN_GRP_GPIO("timer0", 36, 4, BIT(9), "timer"),
	PIN_GRP_GPIO("timer1", 40, 4, BIT(10), "timer"),
};

static const struct gb02_pinctrl_pin_data gb02_pin_data = {
	.nr_pins = 44,
	.name = "GB02-GPIO",
	.groups = gb02_mcu_pin_groups,
	.ngroups = ARRAY_SIZE(gb02_mcu_pin_groups),
};

static int gb02_gpio_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_pinctrl *gb02_pinctrl;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id, peri_base_bar_id;
	int id, ret = 0;

	GB02_PINCTRL_DEBUG("Start to probe GB02 pinctrl MCU:\n");
	pcie_info = gb_dev->gb_pcie;
	id = device_id;

	gb02_pinctrl = devm_kzalloc(gb_dev->dev, sizeof(struct gb02_pinctrl), GFP_KERNEL);
	if (gb02_pinctrl == NULL)
		return -ENOMEM;

	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	peri_base_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);

	gb02_pinctrl->dev = gb_dev->dev;
	gb02_pinctrl->data = (struct gb02_pinctrl_pin_data *)&gb02_pin_data;
	gb02_pinctrl->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	gb02_pinctrl->sysctl_cfg_base =  gb_dev->gb_pcie->pci_bars[peri_base_bar_id].mmio; //bar4
	gb02_pinctrl->sysctl_system_peripheral_base = gb_dev->gb_pcie->pci_bars[GB02_SYSTEM_PERIPHERAL_BAR].mmio; //bar1
	gb02_pinctrl->regs = gb02_pinctrl->sysctl_iofunc_base + MCU_PERI_ADDR_IN_BAR0 + MCU_LGPIO_PERI_ADDR_BASE;

	ret = gb02_pinctrl_register(gb_dev->dev, gb02_pinctrl);
	if (ret)
		return ret;

	ret = gb02_pinctl_init(gb02_pinctrl);
	if (ret)
		return ret;
	g_pinctrl_infos = gb02_pinctrl;
	peri_info->priv = gb02_pinctrl;
	gb02_gpio_sysfs_init();
	dev_info(gb02_pinctrl->dev, "Pinctrl GB02 initialized\n");

	return ret;
}

static void gb02_mcu_gpio_remove(struct gb02_mcu_peripherals_info *peri_info)
{
#ifdef TEST_MISC_GPIO
	int i = 0;

	for (i = 0; i < 4; i++)
		gb02_gpio_free(40 + i);
#endif
	g_pinctrl_infos = NULL;
}

int gb02_mcu_gpio_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;

#ifdef TEST_MISC_GPIO
	ret = misc_register(&gbfpga_misc);
	if (ret)
		dev_err(gb_dev->dev, "unable to register gpio misc device(%d)\n", ret);
#endif
	GB02_PINCTRL_DEBUG("platform gpio register\n");
	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_gpio_init(gb_dev, peri_info, device_id);
	return ret;
}

void gb02_mcu_gpio_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{

	gb02_mcu_gpio_remove(peri_info);
#ifdef TEST_MISC_GPIO
	misc_deregister(&gbfpga_misc);
#endif
}
