// SPDX-License-Identifier: GPL-2.0
/*
 *  Devfreq driver for freqs.
 *
 */

#include "common/gb-peripherals-common.h"
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/sysfs.h>
#include <linux/version.h>
#include <linux/random.h>
#include <linux/delay.h>
#include "gpu/gb_device.h"
#include "gb_devfreq.h"
#include "gpu_freq.h"
#include "gb_regs.h"

#define GB_PERF_FREQ 900
#define GB_HIGH_FREQ 800
#define GB_MID_FREQ 700
#define GB_LOW_FREQ 600
#define GB_LOWEST_FREQ 400
#define POLLING_TIME 50
#define GB_USE_CNT_UPTHRESHOLD 70
#define GPUPLL_CFG0_OFF 0x130
#define GB02_FREQ_DEV_NAME "gpu_freq"
#define OPP_TABLE_CNT (sizeof(gb_opp_table) / sizeof(struct gb_opp_table))

//#define FREQ_DEBUG
#ifdef FREQ_DEBUG
#define GB02_FREQ_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_FREQ_DEVICE_NAME "] " arg)
#else
#define GB02_FREQ_DEBUG(arg...)
#endif

union gpupll_cfg {
	uint32_t val;
	struct {
		uint32_t gpu_pll_pllen:1;
		uint32_t gpu_pll_bypass:1;
		uint32_t gpu_pll_refdiv:6;
		uint32_t gpu_pll_postdiv1:3;
		uint32_t gpu_pll_postdiv2:3;
		uint32_t gpu_pll_fbdiv:12;
		uint32_t gpu_pll_dsmen:1;
		uint32_t reserve:5;
	} type;
};

struct gb02_freq_ctx {
	struct gb_device *gbdev;
	void __iomem *sysctl_cfg_base;    // bar4
	void __iomem *sysctl_iofunc_base; // bar0
	union gpupll_cfg cfg;
	spinlock_t lock;
	struct gb_devfreq *devfreq;
	unsigned long cur_freq;
	unsigned long cur_volt;
	unsigned long last_freq;
	struct notifier_block nb;
	bool limit_freq;
};

BLOCKING_NOTIFIER_HEAD(limit_freq_notifier_chain);
static struct gb02_freq_ctx *g_ctx;
static struct mutex slot_lock[NUM_JOB_SLOTS];
static struct gb_opp_table gb_opp_table[] = {
	{
		.available = false,
		.freq = GB_LOWEST_FREQ,
		.volt = 0,
	},
	{
		.available = true,
		.freq = GB_LOW_FREQ,
		.volt = 0,
	},
	{
		.available = true,
		.freq = GB_MID_FREQ,
		.volt = 0,
	},
	{
		.available = true,
		.freq = GB_HIGH_FREQ,
		.volt = 0,
	},
	{
		.available = false,
		.freq = GB_PERF_FREQ,
		.volt = 0,
	},
};

u32 get_cur_gpu_freq(void)
{
	return ((g_ctx->cur_freq) << 20);
}

static void gb_devfreq_update_utilization(struct gb_device *gb_dev, int slot)
{
	struct gb_devfreq_slot *devfreq_slot = &gb_dev->devfreq.slot[slot];
	ktime_t now;
	ktime_t last;

	if (!gb_dev->devfreq.devfreq)
		return;

	now = ktime_get();
	last = gb_dev->devfreq.slot[slot].time_last_update;

	/* If we last recorded a transition to busy, we have been idle since */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
	if (devfreq_slot->busy)
		gb_dev->devfreq.slot[slot].busy_time += ktime_sub(now, last);
	else
		gb_dev->devfreq.slot[slot].idle_time += ktime_sub(now, last);
#else
	if (devfreq_slot->busy)
		gb_dev->devfreq.slot[slot].busy_time.tv64 += ktime_sub(now, last).tv64;
	else
		gb_dev->devfreq.slot[slot].idle_time.tv64 += ktime_sub(now, last).tv64;
#endif
	gb_dev->devfreq.slot[slot].time_last_update = now;
}

/* The job scheduler is expected to call this at every transition busy <-> idle */
void gb_devfreq_record_transition(struct gb_device *gb_dev, int slot)
{
	struct gb_devfreq_slot *devfreq_slot = &gb_dev->devfreq.slot[slot];

	mutex_lock(&slot_lock[slot]);
	gb_devfreq_update_utilization(gb_dev, slot);
	devfreq_slot->busy = !devfreq_slot->busy;
	mutex_unlock(&slot_lock[slot]);
}

static void freq_apply_state(struct gb02_freq_ctx *ctx, enum freq_e freq)
{
	int locked = 1;
	if (ctx->cur_freq == freq)
		return;

	locked = spin_trylock(&ctx->gbdev->hw_irq_lock);
	if (!locked)
		return;

	ctx->cfg.type.gpu_pll_pllen = 0x1;
	ctx->cfg.type.gpu_pll_dsmen = 0x0;
	ctx->cfg.type.gpu_pll_refdiv = 0x1;
	ctx->cfg.type.gpu_pll_bypass = 0x0;
	ctx->cfg.type.gpu_pll_postdiv2 = 0x1;
	ctx->cfg.type.gpu_pll_postdiv1 = 0x1;
	switch (freq) {
	case GB_MID_FREQ:
		ctx->cfg.type.gpu_pll_fbdiv = 0x1C;
		GB02_FREQ_DEBUG("set freq 700\n");
		break;
	case GB_LOW_FREQ:
		ctx->cfg.type.gpu_pll_fbdiv = 0x18;
		GB02_FREQ_DEBUG("set freq 600\n");
		break;
	case GB_PERF_FREQ:
		ctx->cfg.type.gpu_pll_fbdiv = 0x24;
		GB02_FREQ_DEBUG("set freq 900\n");
		break;
	case GB_LOWEST_FREQ:
		ctx->cfg.type.gpu_pll_fbdiv = 0x10;
		GB02_FREQ_DEBUG("set freq 400\n");
		break;
	case GB_HIGH_FREQ:
	default:
		ctx->cfg.type.gpu_pll_fbdiv = 0x20;
		GB02_FREQ_DEBUG("set freq 800\n");
		break;
	}
	//enable pll and config
	base_reg_write(ctx->cfg.val, ctx->sysctl_iofunc_base, GPUPLL_CFG0_OFF);
	ctx->cur_freq = freq;
	spin_unlock(&ctx->gbdev->hw_irq_lock);
}

static ssize_t perf_test_store(struct device *dev, struct device_attribute *attr, const char *buf,
			size_t count)
{
	int ret;
	int value;
	struct gb_opp_table *freq_table = gb_devfreq_opp_get_table(dev->parent, GB_PERF_FREQ);

	ret = sscanf(buf, "%d", &value);
	if (ret != 1)
		return -EINVAL;

	if (value == 1)
		gb_devfreq_opp_enable(freq_table, true);
	else if (value == 0)
		gb_devfreq_opp_enable(freq_table, false);

	return count;
}

static int gb_open_tpu_core(unsigned long open_bit)
{
	unsigned long val;
	int size, limit_size, i, bit;
	unsigned int on_bit = 0, off_bit = 0;
	unsigned int change_mask = 0xFFFFFFFF; //分bin方案：后续通过接口获取可操作核心bitmap mask

	limit_size = hweight32(change_mask);
	if (open_bit > limit_size)
		open_bit = limit_size;
	else if (open_bit < 1)
		open_bit = 1;

	val = gpu_read(g_ctx->gbdev, TPU_READY_LO);
	size = hweight32(val);
	if (size > limit_size)
		return -EINVAL;

	if (size < open_bit) {
		on_bit = val;
		for (i = 0; i < open_bit - size; i++){
			bit = find_first_zero_bit(&val, 32);
			val |= BIT(bit);
			if (change_mask & BIT(bit))
				on_bit |= BIT(bit);
		}
		gpu_write(g_ctx->gbdev, SHADER_PWRON_LO, on_bit);
	} else if (size > open_bit){
		for (i = 0; i < size - open_bit; i++) {
			bit = find_first_bit(&val, 32);
			val &= ~BIT(bit);
			if (change_mask & BIT(bit))
				off_bit |= BIT(bit);
		}
		gpu_write(g_ctx->gbdev, SHADER_PWROFF_LO, off_bit);
	}

	return 0;
}

static ssize_t gb02_tpu_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	u32 val;
	int size;

	val = gpu_read(g_ctx->gbdev, TPU_READY_LO);
	size = hweight32(val);

	return sprintf(buf, "TPU_READY:0x%x, open_core:%d\n", val, size);
}

static ssize_t gb02_tpu_store(struct device *dev, struct device_attribute *attr, const char *buf,
			size_t count)
{
	unsigned long open_bit;
	int ret;

	if (kstrtoul(buf, 10, &open_bit))
		return -EINVAL;

	ret = gb_open_tpu_core(open_bit);
	if (ret)
		return ret;

	return count;
}

static DEVICE_ATTR_RW(gb02_tpu);
static DEVICE_ATTR_WO(perf_test);

static struct attribute *gb_gpufreq_attrs[] = {
	&dev_attr_gb02_tpu.attr,
	&dev_attr_perf_test.attr,
	NULL,
};

static const struct attribute_group gb_gpufreq_attr_group = {
	.attrs = gb_gpufreq_attrs,
};

static void gb_devfreq_reset(struct gb_device *gb_dev)
{
	ktime_t now = ktime_get();
	int i;

	for (i = 0; i < NUM_JOB_SLOTS; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
		gb_dev->devfreq.slot[i].busy_time = 0;
		gb_dev->devfreq.slot[i].idle_time = 0;
#else
		gb_dev->devfreq.slot[i].busy_time.tv64 = 0;
		gb_dev->devfreq.slot[i].idle_time.tv64 = 0;
#endif
		gb_dev->devfreq.slot[i].time_last_update = now;
	}
}

static int get_cur_gpufreq_level(void)
{
	int i = 0;
	int cur_freq = g_ctx->cur_freq;

	for (i = 0; i < OPP_TABLE_CNT; i++) {
		if (gb_opp_table[i].available && (cur_freq == gb_opp_table[i].freq))
			return i;
	}
	return -1;
}

static int gb_devfreq_target(struct device *dev, unsigned long *freq)
{
	struct gb02_freq_ctx *ctx = g_ctx;
	struct gb_devfreq *df = ctx->devfreq;
	unsigned long target_rate;
	struct gb_opp_table *freq_table;
	struct userspace_data *user_data = NULL;

	freq_table = gb_devfreq_opp_get_table(dev, *freq);
	if (freq_table->available)
		target_rate = freq_table->freq;
	else
		target_rate = ctx->last_freq;

	*freq = target_rate;

	if (ctx->limit_freq && ctx->cur_freq < target_rate)
		return 0;

	if ((*freq == GB_HIGH_FREQ) || (*freq == GB_PERF_FREQ))
		df->profile->polling_ms = 1000;
	else if (*freq == GB_LOW_FREQ)
		df->profile->polling_ms = 50;

	spin_lock(&ctx->lock);
	freq_apply_state(ctx, target_rate);
	spin_unlock(&ctx->lock);

	if (df->custom_mode.is_custom) {
		if (df->data)
			user_data = (struct userspace_data *)df->data;
		gb_open_tpu_core(user_data->tpu_cnt);
	}

	return 0;
}

static int gb_devfreq_get_dev_status(struct device *dev,
		struct gb_devfreq_dev_status *status)
{
	struct gb02_freq_ctx *ctx = g_ctx;
	struct gb_device *gb_dev = ctx->gbdev;
	int i;

	for (i = 0; i < NUM_JOB_SLOTS; i++) {
		gb_devfreq_update_utilization(gb_dev, i);
	}

	status->current_frequency = ctx->cur_freq;
	status->total_time = ktime_to_ns(ktime_add(gb_dev->devfreq.slot[0].busy_time,
						   gb_dev->devfreq.slot[0].idle_time));

	status->busy_time = 0;
	for (i = 0; i < NUM_JOB_SLOTS; i++) {
		status->busy_time += ktime_to_ns(gb_dev->devfreq.slot[i].busy_time);
	}

	//make sure faster frequency rise
	if (status->busy_time / (status->total_time / 100) > GB_USE_CNT_UPTHRESHOLD) {
		spin_lock(&ctx->lock);
		freq_apply_state(ctx, GB_HIGH_FREQ);
		spin_unlock(&ctx->lock);
	}

	/* We're scheduling only to one core atm, so don't divide for now */
	/* status->busy_time /= NUM_JOB_SLOTS; */
	gb_devfreq_reset(gb_dev);

	/* dev_dbg(gb_dev->dev, "busy %lu total %lu %lu %% freq %lu MHz\n", status->busy_time,
		status->total_time,
		status->busy_time / (status->total_time / 100),
		status->current_frequency); */

	return 0;
}

static int gb_devfreq_get_cur_freq(struct device *dev, unsigned long *freq)
{
	struct gb02_freq_ctx *ctx = g_ctx;

	if (!ctx)
		return -EINVAL;

	*freq = ctx->cur_freq;

	return 0;
}

static void gb02_freq_init_pre(struct gb02_freq_ctx *ctx)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, ctx->sysctl_cfg_base, CFG_MCU_XIP_ADDR_OFFSET);
}

static int limit_freq_notifier_callback(struct notifier_block *nb,
						unsigned long action, void *data)
{
	int level = get_cur_gpufreq_level();

	switch (action) {
	case TEMP_NORMAL_EVENT:
		g_ctx->limit_freq = false;
		level += 1;
		freq_apply_state(g_ctx, gb_opp_table[level].freq);
		break;
	case TEMP_HOT_EVENT:
		g_ctx->limit_freq = true;
		level -= 1;
		if (level < 0)
			level = 0;
		freq_apply_state(g_ctx, gb_opp_table[level].freq);
		break;
	default:
		g_ctx->limit_freq = false;
		break;
	}

	return NOTIFY_OK;
}

static struct gb_devfreq_dev_profile gb_devfreq_profile = {
	.polling_ms = POLLING_TIME,
	.target = gb_devfreq_target,
	.get_dev_status = gb_devfreq_get_dev_status,
	.get_cur_freq = gb_devfreq_get_cur_freq,
	.freq_table = gb_opp_table,
	.max_state = OPP_TABLE_CNT,
};

static int gb02_freq_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_freq_ctx *ctx;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id, peri_base_bar_id;
	int i = 0, ret = 0;

	pcie_info = gb_dev->gb_pcie;
	ctx = devm_kzalloc(gb_dev->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	peri_base_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);
	ctx->gbdev = gb_dev;
	ctx->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	ctx->sysctl_cfg_base = gb_dev->gb_pcie->pci_bars[peri_base_bar_id].mmio; //bar4
	ctx->cur_freq = GB_HIGH_FREQ;

	gb02_freq_init_pre(ctx);
	spin_lock_init(&ctx->lock);
	for (i = 0; i < NUM_JOB_SLOTS; i++)
		mutex_init(&slot_lock[i]);

	gb_devfreq_reset(gb_dev);
	gb_devfreq_profile.initial_freq = GB_HIGH_FREQ;
	ctx->devfreq = devm_gb_devfreq_add_device(gb_dev->dev,
			&gb_devfreq_profile, GB_DEVFREQ_GOV_USERSPACE,
			NULL);
	if (IS_ERR(ctx->devfreq)) {
		DRM_DEV_ERROR(gb_dev->dev, "Couldn't initialize GPU devfreq\n");
		ret = PTR_ERR(ctx->devfreq);
		ctx->devfreq = NULL;
		return ret;
	}
	gb_dev->devfreq.devfreq = ctx->devfreq;

	ctx->nb.notifier_call = limit_freq_notifier_callback;
	blocking_notifier_chain_register(&limit_freq_notifier_chain, &ctx->nb);

	g_ctx = ctx;
	peri_info->priv = ctx;

	ret = sysfs_create_group(&ctx->devfreq->dev.kobj, &gb_gpufreq_attr_group);
	if (ret)
		dev_err(&ctx->devfreq->dev, "failed to register sysfs\n");

	return ret;
}

static void gb02_freq_deinit(void)
{
	blocking_notifier_chain_unregister(&limit_freq_notifier_chain, &g_ctx->nb);
}

static void gb02_freq_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	gb_devfreq_governor_deinit();
	gb_devfreq_deinit();
	gb02_freq_deinit();
	return;
}

int gb02_freq_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	ret = gb_devfreq_init();
	ret = gb_devfreq_governor_init();
	ret = gb02_freq_init(gb_dev, peri_info);
	return ret;
}

void gb02_freq_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	gb02_freq_remove(peri_info);
}
