// SPDX-License-Identifier: GPL-2.0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <linux/io.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/regmap.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/seq_file.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include "common/gb-peripherals-common.h"
#include "gpu/gb_device.h"
#include "i2c-gb02-mcu.h"
//#define I2C_DEBUG
#ifdef I2C_DEBUG
#define GB02_I2C_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_I2C_DEVICE_NAME "] " arg)
#else
#define GB02_I2C_DEBUG(arg...)
#endif
//#define I2C_TEST
static struct gb02_mcu_i2c gb02_i2c_infos[4];

static const struct gb02_i2c_pdata gb02_i2c_extra[] = {
	{.frequency = 100},
	{.frequency = 100},
	{.frequency = 100},
	{.frequency = 100}
};

static inline void gb02_i2c_write_reg(unsigned int val,
		struct gb02_i2c_adapter *i2c, unsigned int reg)
{
	writel(val, i2c->regs + reg);
	//GB02_I2C_DEBUG("addr %x reg %x\n",i2c->regs + reg,val);
}

static inline unsigned long gb02_i2c_read_reg(struct gb02_i2c_adapter *i2c,
		unsigned int reg)
{
	//unsigned int val;
	return readl(i2c->regs + reg);
	//val=readl(i2c->regs + reg);
	//GB02_I2C_DEBUG("addr %x reg %x\n",i2c->regs + reg,val);
	//return val;
}

static inline void gb02_i2c_disable_irq(struct gb02_i2c_adapter *i2c)
{
	gb02_i2c_write_reg(0, i2c, NUCLEI_I2C_INT_IE);
}

static inline void gb02_i2c_disable_ack(struct gb02_i2c_adapter *i2c)
{
	unsigned long tmp;

	tmp = readl(i2c->regs + NUCLEI_I2C_SETUP);
	writel(tmp | NUCLEI_I2C_SETUP_ACK, i2c->regs + NUCLEI_I2C_SETUP);
}

static inline void gb02_i2c_enable_ack(struct gb02_i2c_adapter *i2c)
{
	unsigned long tmp;

	tmp = readl(i2c->regs + NUCLEI_I2C_SETUP);
	writel(tmp & ~NUCLEI_I2C_SETUP_ACK, i2c->regs + NUCLEI_I2C_SETUP);
}

static inline void gb02_i2c_start(struct gb02_i2c_adapter *i2c)
{
	unsigned long tmp;

	tmp = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(tmp | NUCLEI_I2C_SETUP_START, i2c, NUCLEI_I2C_SETUP);
}

struct i2c_client *
gb02_i2c_new_device(struct i2c_adapter *adap, struct i2c_board_info const *info)
{
	struct i2c_client	*client;
	int	status = -1;

	client = kzalloc(sizeof *client, GFP_KERNEL);
	if (!client)
		return NULL;

	client->adapter = adap;

	client->dev.platform_data = info->platform_data;

	client->flags = info->flags;
	client->addr = info->addr;

	strlcpy(client->name, info->type, sizeof(client->name));

	if (client->addr == 0x00 || client->addr > 0x7f)
		goto out_err;

	client->dev.parent = &client->adapter->dev;
	client->dev.bus = &i2c_bus_type;
	//client->dev.type = &i2c_client_type;
	client->dev.of_node = info->of_node;
	client->dev.fwnode = info->fwnode;

	dev_set_name(&client->dev, "%d-%04x", i2c_adapter_id(adap), client->addr);

	status = device_register(&client->dev);
	if (status)
		goto out_err;

	dev_dbg(&adap->dev, "client [%s] registered with bus id %s\n",
		client->name, dev_name(&client->dev));

	return client;

out_err:
	dev_err(&adap->dev,
		"Failed to register i2c client %s at 0x%02x (%d)\n",
		client->name, client->addr, status);
	kfree(client);
	return NULL;
}

static bool is_ack(struct gb02_i2c_adapter *i2c)
{
	int tries;

	for (tries = 5; tries; --tries) {
		if (!(readl(i2c->regs + NUCLEI_I2C_STATUS) &
			  NUCLEI_I2C_STATUS_ACK))
			return true;
		usleep_range(50, 100);
	}
	dev_err(i2c->dev, "ack was not received\n");
	return false;
}

static int gb02_i2c_xfer_ongoing_check(struct gb02_i2c_adapter *i2c)
{
	unsigned long iicstat;
	int timeout = 400;

	while (timeout-- > 0) {
		iicstat = gb02_i2c_read_reg(i2c, NUCLEI_I2C_STATUS);

		if (!(iicstat & NUCLEI_I2C_STATUS_BYTE_ON_GOING))
			return 0;

		usleep_range(50, 100);
	}

	return -ETIMEDOUT;
}

#ifdef I2C_TEST
/*
 * get the i2c bus for a master transaction
 */
static int gb02_i2c_set_master(struct gb02_i2c_adapter *i2c)
{
	unsigned long iicstat;
	int timeout = 400;

	while (timeout-- > 0) {
		iicstat = gb02_i2c_read_reg(i2c, NUCLEI_I2C_STATUS);

		if (!(iicstat & NUCLEI_I2C_STATUS_BUSBUSY))
			return 0;

		usleep_range(50, 100);
	}

	return -ETIMEDOUT;
}
#endif

static int i2c_gb02_bus_busy(struct gb02_i2c_adapter *i2c, int for_busy)
{
	unsigned long orig_jiffies = jiffies;
	unsigned int temp;

	while (1) {
		temp = gb02_i2c_read_reg(i2c, NUCLEI_I2C_STATUS);

		/* check for arbitration lost */
		if (temp & NUCLEI_I2C_STATUS_ARB_LOST) {
			dev_err(&i2c->adap.dev,
					"<%s> arbitration lost\n", __func__);
			return -EAGAIN;
		}

		if (for_busy && !(temp & NUCLEI_I2C_STATUS_BUSBUSY))
			break;

		if (time_after(jiffies, orig_jiffies + msecs_to_jiffies(500))) {
			dev_dbg(&i2c->adap.dev,
					"<%s> I2C bus is busy\n", __func__);
			return -ETIMEDOUT;
		}
		schedule();
	}

	return 0;
}

static int i2c_gb02_start(struct gb02_i2c_adapter *i2c)
{
	int ret = 0;

	gb02_i2c_disable_ack(i2c);
	ret = i2c_gb02_bus_busy(i2c, 1);
	if (ret)
		return ret;

	gb02_i2c_start(i2c);

	return ret;
}

static int i2c_gb02_stop(struct gb02_i2c_adapter *i2c)
{
	unsigned int val = 0;
	int ret = 0;

	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_STOP, i2c, NUCLEI_I2C_SETUP);

#ifdef I2C_TEST
	int i = 0;
	for (i = 0; i < GB02_I2C_STOP_TRY_MAX; i++) {
		ret = gb02_i2c_set_master(i2c);
		if (ret != 0)
			ret = -1;
		gb02_i2c_disable_ack(i2c);
		gb02_i2c_start(i2c);

		/* write slave address */
		gb02_i2c_write_reg(i2c->slave_addr, i2c, NUCLEI_I2C_TXDATA);

		val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
		gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_WRITE, i2c, NUCLEI_I2C_SETUP);
		ret = gb02_i2c_xfer_ongoing_check(i2c);
		if (ret < 0)
			/* gb_printf(KERN_ERR, "I2C%d:transfer is on going\n", i2c->busno); */
			ret = -2;
		ret = is_ack(i2c);
		if (!ret)
			/* gb_printf(KERN_ERR, "I2C%d:no ACK occurs\n", i2c->busno); */
			ret = -3;
		else {
			ret = 0;
			break;
		}

		/* STOP */
		val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
		gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_STOP, i2c, NUCLEI_I2C_SETUP);

		mdelay(1);
	}

	if (ret < 0)
		gb_printf(KERN_ERR, "%s:%d:stop error(%d)\n", __func__, __LINE__, ret);

	/* STOP */
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_STOP, i2c, NUCLEI_I2C_SETUP);
#endif

	return ret;
}

static u8 gb_i2c_8bit_addr_from_msg(struct i2c_msg *msg)
{
	return (msg->addr << 1) | (msg->flags & I2C_M_RD ? 1 : 0);
}

static int i2c_gb02_write(struct gb02_i2c_adapter *i2c, struct i2c_msg *msgs)
{
	int i = 0, ret = 0;
	unsigned long val;

	dev_dbg(&i2c->adap.dev, "<%s> write slave address: addr=0x%x\n",
			__func__, gb_i2c_8bit_addr_from_msg(msgs));

	/* write slave address */
	gb02_i2c_write_reg(gb_i2c_8bit_addr_from_msg(msgs), i2c, NUCLEI_I2C_TXDATA);

	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_WRITE, i2c, NUCLEI_I2C_SETUP);
	ret = gb02_i2c_xfer_ongoing_check(i2c);
	if (ret < 0) {
		gb_printf(KERN_ERR, "I2C%d:transfer is on going\n", i2c->bus_no);
		return ret;
	}
	ret = is_ack(i2c);
	if (!ret) {
		gb_printf(KERN_ERR, "I2C%d:no ACK occurs\n", i2c->bus_no);
		ret = -1;
		return ret;
	}

	dev_dbg(&i2c->adap.dev, "<%s> write data\n", __func__);

	/* write data */
	for (i = 0; i < msgs->len; i++) {
		//		GB02_I2C_DEBUG("<%s> write byte: B%d=0x%X\n",__func__, i, msgs->buf[i]);
		gb02_i2c_write_reg(msgs->buf[i], i2c, NUCLEI_I2C_TXDATA);

		val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
		gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_WRITE, i2c, NUCLEI_I2C_SETUP);

		ret = gb02_i2c_xfer_ongoing_check(i2c);
		if (ret < 0) {
			gb_printf(KERN_ERR, "I2C%d:transfer is on going\n", i2c->bus_no);
			return ret;
		}
		ret = is_ack(i2c);
		if (!ret) {
			gb_printf(KERN_ERR, "I2C%d:no ACK occurs\n", i2c->bus_no);
			ret = -1;
			return ret;
		}

	}

	return 0;
}

static int i2c_gb02_read(struct gb02_i2c_adapter *i2c, struct i2c_msg *msgs, bool is_lastmsg)
{
	int i = 0, ret = 0;
	unsigned long val;

	//GB02_I2C_DEBUG("<%s> write slave address: addr=0x%x\n",__func__, gb_i2c_8bit_addr_from_msg(msgs));

	/* write slave address */
	gb02_i2c_write_reg(gb_i2c_8bit_addr_from_msg(msgs), i2c, NUCLEI_I2C_TXDATA);

	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_WRITE, i2c, NUCLEI_I2C_SETUP);
	ret = gb02_i2c_xfer_ongoing_check(i2c);
	if (ret < 0) {
		gb_printf(KERN_ERR, "I2C%d:transfer is on going\n", i2c->bus_no);
		return ret;
	}
	ret = is_ack(i2c);
	if (!ret) {
		gb_printf(KERN_ERR, "I2C%d:no ACK occurs\n", i2c->bus_no);
		ret = -1;
		return ret;
	}

	dev_dbg(&i2c->adap.dev, "<%s> setup bus\n", __func__);

	/* setup bus to read data */
	gb02_i2c_enable_ack(i2c);

	dev_dbg(&i2c->adap.dev, "<%s> read data\n", __func__);
	/* read data */
	for (i = 0; i < msgs->len; i++) {
		if (i == (msgs->len - 1)) {
			if (is_lastmsg) {
				dev_dbg(&i2c->adap.dev,
						"<%s> set NOACK\n", __func__);

				gb02_i2c_disable_ack(i2c);
			}
		}

		val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
		gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_READ, i2c, NUCLEI_I2C_SETUP);

		ret = gb02_i2c_xfer_ongoing_check(i2c);
		if (ret < 0) {
			gb_printf(KERN_ERR, "I2C%d:transfer is on going\n", i2c->bus_no);
			return ret;
		}

		msgs->buf[i] = gb02_i2c_read_reg(i2c, NUCLEI_I2C_RXDATA);
		dev_dbg(&i2c->adap.dev,
				"<%s> read byte: B%d=0x%X\n",
				__func__, i, msgs->buf[i]);
		//GB02_I2C_DEBUG("<%s> read byte: B%d=0x%X\n",
		//__func__, i, msgs->buf[i]);
	}
	return 0;
}

/*
 * this starts an i2c transfer
 */
static int gb02_i2c_doxfer(struct gb02_i2c_adapter *i2c,
						   struct i2c_msg *msgs, int num)
{
	int ret = 0;
	unsigned int i = 0;
	bool is_lastmsg = false;

	/* Start I2C transfer */
	ret = i2c_gb02_start(i2c);
	if (ret) {
		if (i2c->adap.bus_recovery_info) {
			i2c_recover_bus(&i2c->adap);
			ret = i2c_gb02_start(i2c);
		}
	}

	if (ret)
		goto fail0;

	i2c->slave_addr = (msgs[0].addr & 0x7f) << 1;

	/* read/write data */
	for (i = 0; i < num; i++) {
		if (i == num - 1)
			is_lastmsg = true;

		if (i) {
			dev_dbg(&i2c->adap.dev,
					"<%s> repeated start\n", __func__);
			gb02_i2c_start(i2c);
		}

		dev_dbg(&i2c->adap.dev,
				"<%s> transfer message: %d\n", __func__, i);
		/* write/read data */
		if (msgs[i].flags & I2C_M_RD)
			ret = i2c_gb02_read(i2c, &msgs[i], is_lastmsg);
		else
			ret = i2c_gb02_write(i2c, &msgs[i]);
		if (ret)
			goto fail0;
	}
	//return 0;
fail0:
	/* Stop I2C transfer */
	ret = i2c_gb02_stop(i2c);
	return (ret < 0) ? ret : num;
}

/*
 * first port of call from the i2c bus code when an message needs
 * transferring across the i2c bus.
 */
static int gb02_i2c_xfer(struct i2c_adapter *adap,
						 struct i2c_msg *msgs, int num)
{
	struct gb02_i2c_adapter *i2c = (struct gb02_i2c_adapter *)adap->algo_data;
	int retry;
	int ret;

	for (retry = 0; retry < adap->retries; retry++) {
		ret = gb02_i2c_doxfer(i2c, msgs, num);
		if (ret != -EAGAIN)
			return ret;
		dev_dbg(i2c->dev, "Retrying transmission (%d)\n", retry);
		udelay(100);
	}

	return -EREMOTEIO;
}

/* declare our i2c functionality */
static u32 gb02_i2c_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL | I2C_FUNC_PROTOCOL_MANGLING;
}

/* i2c bus registration info */
static const struct i2c_algorithm gb02_i2c_algorithm = {
	.master_xfer = gb02_i2c_xfer,
	.functionality = gb02_i2c_func,
};

void gb02_i2c_set_sclduty(struct gb02_i2c_adapter *i2c, enum I2C_DutyTypedef duty)
{
	unsigned long val;

	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	val &= ~(I2C_SETUP_SCL_DUTY_SEL_MASK);
	gb02_i2c_write_reg(val, i2c, NUCLEI_I2C_SETUP);
	val |= duty << 18;
	gb02_i2c_write_reg(val, i2c, NUCLEI_I2C_SETUP);
}

/*
 * work out a divisor for the user requested frequency setting,
 * either by the requested frequency, or scanning the acceptable
 * range of frequencies until something is found
 */
static int gb02_i2c_clockrate(struct gb02_i2c_adapter *i2c, unsigned int freq, enum I2C_DutyTypedef duty)
{
	unsigned int div, periph_freq_div, periph_freq;
	struct gb02_mcu_i2c *gb02_i2c = i2c->driver_data;

	dev_dbg(i2c->dev, "i2c%d desired frequency %d KHz\n", i2c->bus_no, freq);
	periph_freq_div = base_reg_read(gb02_i2c->sysctl_iofunc_base, SYSCTL_CLK_DIV_OFFSET) & 0x7;
	periph_freq = GB02_MCU_SYS_CLK / (periph_freq_div + 1);

	//gb02_i2c_set_sclduty(i2c, duty);
	//switch (duty) {
	//case I2C_30_DUTY:
	//case I2C_50_DUTY:
	//	div = ((periph_freq / freq) / 6) - 1;
	//	break;
	//case I2C_40_DUTY:
	//	div = ((periph_freq / freq) / 5) - 1;
	//	break;
	//default:
	//	break;
	//}

	div = ((periph_freq / freq) / 5) - 1;
	/* set I2C frequency */
	gb02_i2c_write_reg(div, i2c, NUCLEI_I2C_DIV_NUM);

	return 0;
}

static int gb02_i2c_unit_init(struct gb02_i2c_adapter *i2c)
{
	struct gb02_mcu_i2c *gb02_i2c = i2c->driver_data;
	unsigned int freq = gb02_i2c->pdata[i2c->bus_no].frequency;
	unsigned long val = 0;

	gb02_i2c_disable_irq(i2c);
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val & ~NUCLEI_I2C_SETUP_ENABLE, i2c, NUCLEI_I2C_SETUP);

	/* soft reset*/
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_SOFT_RESET, i2c, NUCLEI_I2C_SETUP);
	udelay(100);
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val & ~NUCLEI_I2C_SETUP_SOFT_RESET, i2c, NUCLEI_I2C_SETUP);

	/* I2C/I3C mode */
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val & ~NUCLEI_I2C_SETUP_I3C_MODE, i2c, NUCLEI_I2C_SETUP);

	/* clock divider */
	if (gb02_i2c_clockrate(i2c, freq, I2C_40_DUTY) != 0) {
		dev_err(i2c->dev, "cannot meet bus frequency required\n");
		return -EINVAL;
	}
	GB02_I2C_DEBUG("bus frequency set to %d KHz\n", freq);

	/* I2C master mode */
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val & ~NUCLEI_I2C_SETUP_ROLE_MODE, i2c, NUCLEI_I2C_SETUP);

	/* write slave address */

	/* PULLUP enable */
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	val |= (NUCLEI_I2C_SETUP_SCL_PULLUP | NUCLEI_I2C_SETUP_SDA_PULLUP);
	gb02_i2c_write_reg(val, i2c, NUCLEI_I2C_SETUP);

	/* I2C enable */
	val = gb02_i2c_read_reg(i2c, NUCLEI_I2C_SETUP);
	gb02_i2c_write_reg(val | NUCLEI_I2C_SETUP_ENABLE, i2c, NUCLEI_I2C_SETUP);

	return 0;
}

static void gb02_i2c_init_pre(struct gb02_mcu_i2c *gb02_i2c)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, gb02_i2c->sysctl_cfg_base, CFG_MCU_XIP_ADDR_OFFSET);
	gb02_i2c->pdata = gb02_i2c_extra;
}

static int gb02_i2c_init(struct gb02_mcu_i2c *gb02_i2c)
{
	int ret = 0;
	struct gb02_i2c_adapter *i2c = &gb02_i2c->adapter;
	char label[10];

	i2c->driver_data = gb02_i2c;
	snprintf(label, 11, "gb02-i2c%d", gb02_i2c->i2c_bus_no);
	strscpy(i2c->adap.name, label, sizeof(i2c->adap.name));
	i2c->adap.owner = THIS_MODULE;
	i2c->adap.algo = &gb02_i2c_algorithm;
	i2c->adap.retries = 1;
	i2c->adap.class = I2C_CLASS_DEPRECATED;
	i2c->tx_setup = 50;
	i2c->bus_no = gb02_i2c->i2c_bus_no; // -1 for non static bus_num

	init_waitqueue_head(&i2c->wait);
	i2c->dev = gb02_i2c->dev;

	i2c->regs = gb02_i2c->regs;
	dev_dbg(gb02_i2c->dev, "registers %p (%d)\n",
			i2c->regs, i2c->bus_no);

	/* setup info block for the i2c core */
	i2c->adap.algo_data = i2c;
	i2c->adap.dev.parent = gb02_i2c->dev;

	ret = gb02_i2c_unit_init(i2c);
	if (ret != 0) {
		dev_err(gb02_i2c->dev, "I2C controller init failed\n");
		return ret;
	}

	ret = i2c_add_adapter(&i2c->adap);
	if (ret < 0) {
		dev_warn(gb02_i2c->dev, "cannot add %s I2C adapter\n", i2c->adap.name);
		return ret;
	}

	GB02_I2C_DEBUG("%s: GB02 I2C adapter(%d)\n", dev_name(&i2c->adap.dev), gb02_i2c->i2c_bus_no);
	return 0;
}

static int gb02_i2c_mcu_ioctl_read(struct gb02_i2c_adapter *i2c, unsigned int addr,
								   unsigned int len, u8 *data)
{
	int ret = 0, i = 0;
	u8 out_buf[ADDR_LEN];
	u8 in_buf[I2C_MCU_TEST_DATA_LENGTH];
	struct i2c_msg msgs[] = {
		{
			.addr = I2C_MCU_TEST_DEVICE_ADDR,
			.flags = 0,
			.len = ADDR_LEN,
			.buf = out_buf,
		},
		{
			.addr = I2C_MCU_TEST_DEVICE_ADDR,
			.flags = I2C_M_RD,
			.len = len,
			.buf = in_buf,
		}
	};

	out_buf[0] = addr & 0xff;
	out_buf[1] = addr >> 8;

	ret = i2c_transfer(&i2c->adap, msgs, 2);
	if (ret != 2) {
		gb_printf(KERN_ERR, "%s:%d:I2C transfer error(%d)\n", __func__, __LINE__, ret);
	} else {
		memcpy(data, in_buf, len);
		for (i = 0; i < len; i++)
			GB02_I2C_DEBUG("read data:0x%x ", in_buf[i]);
		ret = 0;
		GB02_I2C_DEBUG("%s:%d:I2C read finished\n", __func__, __LINE__);
	}

	return ret;
}

static int gb02_i2c_mcu_ioctl_write(struct gb02_i2c_adapter *i2c, unsigned int addr,
									unsigned int len, u8 *data)
{
	int ret = 0, i = 0;
	u8 in_buf[ADDR_LEN + I2C_MCU_TEST_DATA_LENGTH];
	struct i2c_msg msgs[] = {
		{
			.addr = I2C_MCU_TEST_DEVICE_ADDR,
			.flags = 0,
			.len = ADDR_LEN + len,
			.buf = in_buf,
		}
	};

	in_buf[0] = addr & 0xff;
	in_buf[1] = addr >> 8;
	memcpy(in_buf + ADDR_LEN, data, len);

	for (i = 0; i < len + ADDR_LEN; i++)
		GB02_I2C_DEBUG("write data 0x%x ", in_buf[i]);
	ret = i2c_transfer(&i2c->adap, msgs, 1);
	if (ret != 1)
		gb_printf(KERN_ERR, "%s:%d:I2C transfer error(%d)\n", __func__, __LINE__, ret);
	else
		GB02_I2C_DEBUG("%s:%d:I2C write finished\n", __func__, __LINE__);

	return ret;
}

static int gb02_i2c_mcu_ioctl(struct gb02_i2c_adapter *i2c)
{
	int ret = 0;
	unsigned int start_addr = 0, len = I2C_MCU_TEST_DATA_LENGTH;
	u8 rdata[I2C_MCU_TEST_DATA_LENGTH] = {0},
										 wdata[I2C_MCU_TEST_DATA_LENGTH] = {0};
	int i = 0;

	/* read */
	ret = gb02_i2c_mcu_ioctl_read(i2c, start_addr, len, rdata);
	if (ret) {
		gb_printf(KERN_ERR, "%s:%d:I2C read failed(%d)\n", __func__, __LINE__, ret);
		return ret;
	}

	/* write */
	for (i = 0; i < len; i++)
		wdata[i] = 0x5a + i;
	ret = gb02_i2c_mcu_ioctl_write(i2c, start_addr, len, wdata);
	if (ret) {
		gb_printf(KERN_ERR, "%s:%d:I2C write failed(%d)\n", __func__, __LINE__, ret);
		return ret;
	}
	/* read again */
	ret = gb02_i2c_mcu_ioctl_read(i2c, start_addr, len, rdata);
	if (ret) {
		gb_printf(KERN_ERR, "%s:%d:I2C read failed(%d)\n", __func__, __LINE__, ret);
		return ret;
	}
	return 0;
}

static int gb02_i2c_mcu_read_ioctl(struct gb02_i2c_adapter *i2c)
{
	int ret = 0;
	//int i = 0;
	unsigned int start_addr = 0, len = I2C_MCU_TEST_DATA_LENGTH;
	u8 rdata[I2C_MCU_TEST_DATA_LENGTH] = {0};

	/* read */
	ret = gb02_i2c_mcu_ioctl_read(i2c, start_addr, len, rdata);

	if (ret) {
		gb_printf(KERN_ERR, "%s:%d:I2C read failed(%d)\n", __func__, __LINE__, ret);
		//for (i = 0; i < len; i++)
		//	GB02_I2C_DEBUG("read data 0x%x\n", rdata[i]);
		return ret;
	}

	return 0;
}

static int gb02_i2c_mcu_write_ioctl(struct gb02_i2c_adapter *i2c)
{
	int ret = 0;
	unsigned int start_addr = 0, len = I2C_MCU_TEST_DATA_LENGTH;
	u8 wdata[I2C_MCU_TEST_DATA_LENGTH] = {0};
	int i = 0;

	/* write */
	for (i = 0; i < len; i++)
		wdata[i] = 0x5a + i;

	ret = gb02_i2c_mcu_ioctl_write(i2c, start_addr, len, wdata);
	if (ret) {
		gb_printf(KERN_ERR, "%s:%d:I2C write failed(%d)\n", __func__, __LINE__, ret);
		return ret;
	}

	return 0;
}

static int my_open(struct inode *inode, struct file *filp)
{
	filp->private_data = gb02_i2c_infos;
	return 0;
}

static int my_release(struct inode *inode, struct file *filp)
{
	filp->private_data = NULL;
	return 0;
}

static long gb02_test_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	struct gb02_mcu_i2c *mcu_i2c = filp->private_data;
	struct gb02_i2c_adapter *i2c;

	GB02_I2C_DEBUG("enter ioctrl!\n");

	switch (cmd) {
	case I2C_MCU_CMD_I2C0:
		GB02_I2C_DEBUG("=====MCU's I2C0 start!======\n");
		i2c = &mcu_i2c[0].adapter;
		ret = gb02_i2c_mcu_ioctl(i2c);
		break;
	case I2C_MCU_CMD_I2C1:
		GB02_I2C_DEBUG("=====MCU's I2C1 start!======\n");
		i2c = &mcu_i2c[1].adapter;
		ret = gb02_i2c_mcu_ioctl(i2c);
		break;
	case I2C_MCU_CMD_I2C2:
		GB02_I2C_DEBUG("=====MCU's I2C2 start!======\n");
		i2c = &mcu_i2c[2].adapter;
		ret = gb02_i2c_mcu_ioctl(i2c);
		break;
	case I2C_MCU_CMD_I2C3:
		GB02_I2C_DEBUG("=====MCU's I2C3 start!======\n");
		i2c = &mcu_i2c[3].adapter;
		ret = gb02_i2c_mcu_ioctl(i2c);
		break;
	case I2C_MCU_CMD_I2C0_RD:
		GB02_I2C_DEBUG("=====MCU's I2C0 read start!======\n");
		i2c = &mcu_i2c[0].adapter;
		ret = gb02_i2c_mcu_read_ioctl(i2c);
		break;
	case I2C_MCU_CMD_I2C0_WR:
		GB02_I2C_DEBUG("=====MCU's I2C0 write start!======\n");
		i2c = &mcu_i2c[0].adapter;
		ret = gb02_i2c_mcu_write_ioctl(i2c);
		break;
	default:
		break;
	}

	return 0;
}

static const struct file_operations mmap_fops = {
	.owner = THIS_MODULE,
	.open = my_open,
	.release = my_release,
	.unlocked_ioctl = gb02_test_ioctl,
};
static struct miscdevice gbfpga_misc = {
	.fops = &mmap_fops,
	.name = "gbfpga_i2c",
	.minor = MISC_DYNAMIC_MINOR,
};

static int gb02_mcu_i2c_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_mcu_i2c *gb02_i2c;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id, peri_base_bar_id;
	int id, ret = 0;

	GB02_I2C_DEBUG("Start to probe GB02 MCU's I2C:%d\n", device_id);
	pcie_info = gb_dev->gb_pcie;
	id = device_id;
	gb02_i2c = devm_kzalloc(gb_dev->dev, sizeof(struct gb02_mcu_i2c), GFP_KERNEL);
	if (gb02_i2c == NULL)
		return -ENOMEM;
	gb02_i2c->dev = gb_dev->dev;
	gb02_i2c->i2c_bus_no = id;

	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	peri_base_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);
	gb02_i2c->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	gb02_i2c->sysctl_cfg_base = gb_dev->gb_pcie->pci_bars[peri_base_bar_id].mmio; //bar4
	gb02_i2c->regs = gb02_i2c->sysctl_iofunc_base + MCU_PERI_ADDR_IN_BAR0 + MCU_I2C0_PERI_ADDR_BASE + MCU_I2C_PERI_ADDR_OFST * id;
	gb02_i2c_init_pre(gb02_i2c);
	ret = gb02_i2c_init(gb02_i2c);
	if (ret)
		return ret;
	memcpy(&gb02_i2c_infos[id], gb02_i2c, sizeof(struct gb02_mcu_i2c));
	peri_info->priv = gb02_i2c;
	GB02_I2C_DEBUG("GB02 MCU's I2C initialized\n");

	return ret;
}

static void gb02_mcu_i2c_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_mcu_i2c *gb02_i2c = (struct gb02_mcu_i2c *)peri_info->priv;

	i2c_del_adapter(&(gb02_i2c->adapter.adap));
	memset(gb02_i2c_infos, 0, sizeof(gb02_i2c_infos));
}

int gb02_mcu_i2c_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;
	static bool i2c_misc_inited;

	GB02_I2C_DEBUG("mcu_i2c_drv_init\n");
	if (!i2c_misc_inited) {
		ret = misc_register(&gbfpga_misc);
		if (ret)
			dev_err(gb_dev->dev, "unable to register i2c misc device(%d)\n", ret);
		else
			i2c_misc_inited = true;
	}
	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_mcu_i2c_init(gb_dev, peri_info, device_id);
	return ret;
}

void gb02_mcu_i2c_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	gb02_mcu_i2c_remove(peri_info);
	misc_deregister(&gbfpga_misc);
	GB02_I2C_DEBUG("Bye!\n");
}
