// SPDX-License-Identifier: GPL-2.0
#include <linux/clk.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/iopoll.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/thermal.h>
#include <linux/reboot.h>
#include <linux/workqueue.h>
#include <linux/pci.h>
#include "gpu/gb_device.h"
#include "gb02-pvt.h"
#include "mcu_peripherals/gpu_freq/gpu_freq.h"

//#define PVT_TEST
#ifdef PVT_DEBUG
#define GB02_PVT_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_PVT_DEVICE_NAME "] " arg)
#else
#define GB02_PVT_DEBUG(arg...)
#endif

#define GB02_SYSCTRL_SYSCKEN0_OFST 0x150
#define GB02_SYSCTRL_SYSCKEN1_OFST 0x154
#define GB02_SYSCTRL_SWRST0_OFST 0x158
#define GB02_SYSCTRL_SWRST1_OFST 0x15C

#define GB02_PVT_THERMAL_ASIC_TEMP_COOL	30000	/* 30C */
#define GB02_PVT_THERMAL_ASIC_TEMP_NORM	50000	/* 50C */
#define GB02_PVT_THERMAL_ASIC_TEMP_HIGH	70000	/* 70C */
#define GB02_PVT_THERMAL_ASIC_TEMP_HOT	90000	/* 90C */
#define GB02_PVT_THERMAL_ASIC_TEMP_CRIT	110000	/* 110C */
#define GB02_PVT_THERMAL_HYSTERESIS_TEMP	0	/* 0C */

#define CQ2040_PCIe_C21 (0x0005)
#define CQ2040_MXM_M60 (0x0008)
#define CQ2040_PCIe_P42 (0x000B)
#define CQ2040_PUB (0x000C)

#define GB02_PVT_THERMAL_MAX_STATE	4
#define POWEROFF_DELAY_MS 2000

static bool power_off_triggered;
static DEFINE_MUTEX(poweroff_lock);

enum gb02_pvt_thermal_trips {
	GB02_PVT_THERMAL_TEMP_TRIP_COOL,
	GB02_PVT_THERMAL_TEMP_TRIP_NORM,
	GB02_PVT_THERMAL_TEMP_TRIP_HIGH,
	GB02_PVT_THERMAL_TEMP_TRIP_HOT,
	GB02_PVT_THERMAL_TEMP_TRIP_CRIT,
};

struct gb02_pvt_thermal_trip {
	int type;
	int temp;
	int hyst;
	int min_state;
	int max_state;
};
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
static const struct gb02_pvt_thermal_trip default_thermal_trips[] = {
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temp		= GB02_PVT_THERMAL_ASIC_TEMP_COOL,
		.hyst		= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
		.min_state	= 0,
		.max_state	= 1,
	},
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temp		= GB02_PVT_THERMAL_ASIC_TEMP_NORM,
		.hyst		= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
		.min_state	= 1,
		.max_state	= 2,
	},
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temp		= GB02_PVT_THERMAL_ASIC_TEMP_HIGH,
		.hyst		= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
		.min_state	= 2,
		.max_state	= 3,
	},
	{	/* Warning */
		.type		= THERMAL_TRIP_ACTIVE,
		.temp		= GB02_PVT_THERMAL_ASIC_TEMP_HOT,
		.hyst		= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
		.min_state	= GB02_PVT_THERMAL_MAX_STATE,
		.max_state	= GB02_PVT_THERMAL_MAX_STATE,
	},
	{	/* Critical - soft poweroff */
		.type		= THERMAL_TRIP_HOT,
		.temp		= GB02_PVT_THERMAL_ASIC_TEMP_CRIT,
		.min_state	= GB02_PVT_THERMAL_MAX_STATE,
		.max_state	= GB02_PVT_THERMAL_MAX_STATE,
	}
};
#else
static const struct thermal_trip default_thermal_trips[] = {
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temperature	= GB02_PVT_THERMAL_ASIC_TEMP_COOL,
		.hysteresis	= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
	},
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temperature	= GB02_PVT_THERMAL_ASIC_TEMP_NORM,
		.hysteresis	= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
	},
	{
		.type		= THERMAL_TRIP_ACTIVE,
		.temperature	= GB02_PVT_THERMAL_ASIC_TEMP_HIGH,
		.hysteresis	= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
	},
	{	/* Warning */
		.type		= THERMAL_TRIP_ACTIVE,
		.temperature	= GB02_PVT_THERMAL_ASIC_TEMP_HOT,
		.hysteresis	= GB02_PVT_THERMAL_HYSTERESIS_TEMP,
	},
	{	/* Critical - soft poweroff */
		.type		= THERMAL_TRIP_HOT,
		.temperature	= GB02_PVT_THERMAL_ASIC_TEMP_CRIT,
	}
};
#endif

#define GB02_PVT_THERMAL_NUM_TRIPS	ARRAY_SIZE(default_thermal_trips)
#define GB02_PVT_THERMAL_TRIP_MASK	(BIT(GB02_PVT_THERMAL_NUM_TRIPS) - 1)

static struct thermal_zone_params gb02_pvt_zone_params = {
	.governor_name = "step_wise",
};
/**
 * struct gb02_pvt_thermal_priv - gb02_pvt thermal sensor private structure
 * @pvt_tzd: struct thermal_zone_device where the sensor is registered
 * @lock: prevents read sensor in parallel
 * @regs: pointer to base address of the thermal sensor
 * @dev: struct device pointer
 */

struct gb02_pvt_thermal_priv {
	struct thermal_zone_device *pvt_tzd;
	struct mutex lock;
	void __iomem *regs;
	void __iomem *sysctl_iofunc_base; // bar0
	void __iomem *gb_comp_irq_reg_bar_base;
	struct delayed_work work;
	int shutdown_temp;
	struct device *dev;
	struct gb02_pvt_thermal_trip trips[GB02_PVT_THERMAL_NUM_TRIPS];
};

static int gb02_pvt_init(struct gb02_pvt_thermal_priv *priv, unsigned char mode);
static u16 gb02_pvt_get_dout_data(struct gb02_pvt_thermal_priv *priv);

static unsigned long long gb02_pvt_get_dout_tempcal_data(unsigned short dout_act_in)
{
	unsigned int Kcal = 0, Tcal = 0;
	unsigned long long ret_cal = 0;
	struct gb02_pvt_thermal_priv priv;
	struct gb_pcie_info *gb_pcie = get_pcie_gl_info();
	struct pci_dev *pdev = gb_pcie->pdev;

	if (gb_pcie == NULL)
		return 0;

	priv.regs = gb_pcie->pci_bars[PVT_DEFAULT_BAR].mmio + (CS_CNOC_S1_PVT_OFFSET - GB02_DEFAULT_CFG_SYS_PERI_ADDR);

	/* Reinitialize pvt when get register data 0, Fix bug GB2EN-1431 */
	if (dout_act_in == 0) {
		gb02_pvt_init(&priv, TEMP_MODE);
		dout_act_in = gb02_pvt_get_dout_data(&priv);
	}

	if (D1 < dout_act_in && 4095 >= dout_act_in) {
		GB02_PVT_DEBUG("D1-4096 (%lld-4096): %d\n", D1, dout_act_in);
		Kcal = 198;
		Tcal = -16;
	} else if ((DOUT75act < dout_act_in) && (dout_act_in <= D1)) {
		GB02_PVT_DEBUG("DOUTact75-D1(%lld-%lld: %d)\n", DOUT75act, D1, dout_act_in);
		Kcal = 41;
		Tcal = -3;
	} else if ((dout_act_in > D2) && (dout_act_in <= DOUT75act)) {
		GB02_PVT_DEBUG("D2-DOUTact75(%lld-%lld: %d)\n", D2, DOUT75act, dout_act_in);
		Kcal = -111;
		Tcal = 7;
	} else if ((DOUT25act < dout_act_in) && (dout_act_in <= D2)) {
		GB02_PVT_DEBUG("DOUTact25-D2(%lld-%lld: %d)\n", DOUT25act, D2, dout_act_in);
		Kcal = 111;
		Tcal = -2;
	} else if ((dout_act_in >= D3) && (DOUT25act >= dout_act_in)) {
		GB02_PVT_DEBUG("D3-DOUTact25(%lld-%lld: %d)\n", D3, DOUT25act, dout_act_in);
		Kcal = -233;
		Tcal = 5;
	} else if (D4 <= dout_act_in && D3 > dout_act_in) {
		GB02_PVT_DEBUG("D4-D3(%lld-%lld: %d)\n", D4, D3, dout_act_in);
		Kcal = -250;
		Tcal = 5;
	} else if (0 <= dout_act_in && D4 > dout_act_in) {
		GB02_PVT_DEBUG("0-D4(0-%lld: %d)\n", D4, dout_act_in);
		Kcal = -267;
		Tcal = 5;
	}
	ret_cal = (DOUT75std - DOUT25std) * (dout_act_in * 8192 - (8192 + Tcal) * (3 * DOUT25act - DOUT75act) / 2) / ((8192 + Kcal) * (DOUT75act - DOUT25act)) + (3 * DOUT25std - DOUT75std) / 2;
	if ((pdev->subsystem_device == CQ2040_PCIe_C21) || (pdev->subsystem_device == CQ2040_MXM_M60)
			|| (pdev->subsystem_device == CQ2040_PUB) || (pdev->subsystem_device == CQ2040_PCIe_P42))
		/* 
		 * DOUT = (T / 44 + 27049) / 10
		 * thermal resistance compensation 8°C (8000 / 44 /10) = 18.1
		 */
		ret_cal += 18;
	GB02_PVT_DEBUG("RET_CAL:%lld", ret_cal);
	return ret_cal;
}

static int gb02_pvt_set_power(struct gb02_pvt_thermal_priv *priv, unsigned char power)
{
	void __iomem *regs;
	u32 val;
	int ret;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_00);
	if (power == PVT_POWER_DOWN) {
		val &= ~PU_VTDC_MASK; // power down
		ret = PVT_POWER_DOWN;
	} else {
		val |= PU_VTDC_MASK; // power on
		ret = PVT_POWER_UP;
	}
	writel_relaxed(val, regs + PVT_REG_00);
	val = readl_relaxed(regs + PVT_REG_00);
	GB02_PVT_DEBUG("REG_00:%x\n", val);
	return ret;
}
#ifdef PVT_TEST
static int gb02_pvt_set_mode(struct gb02_pvt_thermal_priv *priv, unsigned char mode)
{
	void __iomem *regs;
	u32 val = 0;
	int ret;

	regs = priv->regs;
	if (mode == VOLT_MODE) {
		val &= ~TDC_MODE_MASK;
		ret = VOLT_MODE;
	} else if (mode == TEMP_MODE) {
		val |= TDC_MODE_MASK;
		ret = TEMP_MODE;
	}
	writel_relaxed(val, regs + PVT_REG_00);
	return ret;
}
#endif
static void gb02_pvt_reset(struct gb02_pvt_thermal_priv *priv)
{

	void __iomem *regs;
	u32 val;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_00);
	val &= ~PVT_RESETn_MASK;
	writel_relaxed(val, regs + PVT_REG_00);
}

static void gb02_pvt_dereset(struct gb02_pvt_thermal_priv *priv)
{

	void __iomem *regs;
	u32 val;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_00);
	val |= PVT_RESETn_MASK;
	writel_relaxed(val, regs + PVT_REG_00);
}
static u16 gb02_pvt_get_dout_data(struct gb02_pvt_thermal_priv *priv)
{
	void __iomem *regs;
	static u32 old_dout;
	u32 val;
	u16 ret;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_14);
	ret = DOUT(val);
	if (ret == DOUT(0xffffffff)) {
		gb_printf(KERN_ERR, "%s PVT_REG_14:%x\n", __func__, ret);
		return old_dout;
	}
	old_dout = ret;
	return ret;
}
#ifdef VOLTAGE
static u8 gb02_pvt_cal_trim(struct gb02_pvt_thermal_priv *priv)
{
#ifdef PVT_TEST
	void __iomem *regs;
	u32 val;
	u8 trimval;
	u16 dout;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_00);
	// by following steps ,we can get a bigger DOUTacl , in case overvoltage
	while (gb02_pvt_get_dout_data(priv) > DOUTstd) {
		val++;
		writel_relaxed(val, regs + PVT_REG_00);
		trimval = BAND_GAP_TRIM(val);
		msleep(20);
	}
	while (gb02_pvt_get_dout_data(priv) < DOUTstd) {
		val--;
		writel_relaxed(val, regs + PVT_REG_00);
		trimval = BAND_GAP_TRIM(val);
		msleep(20);
	}
out:
#endif
	return trimval;
}
#endif

#ifdef PVT_TEST
static int gb02_pvt_set_thresh_normal(struct gb02_pvt_thermal_priv *priv, unsigned int low, unsigned int high)
{
	void __iomem *regs;
	u32 val;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_00);
	val |= (HIGH_TEMP_THER_CFG_NORMAL(high) | LOW_TEMP_THER_CFG_NORMAL(low));
	writel_relaxed(val, regs + PVT_REG_00);

	return 0;
}
#endif

#ifdef PVT_TEST
static void gb02_pvt_set_thresh_fatal(struct gb02_pvt_thermal_priv *priv, unsigned int low, unsigned int high)
{
	void __iomem *regs;
	u32 val;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_04);
	val |= (HIGH_TEMP_THER_CFG_FATAL(high) | LOW_TEMP_THER_CFG_FATAL(low));
	writel_relaxed(val, regs + PVT_REG_04);
}
#endif

#ifdef HAS_PVT_IRQ
static void gb02_pvt_clear_irq_flag(struct gb02_pvt_thermal_priv *priv)
{
	void __iomem *regs;
	u32 val;

	regs = priv->regs;
	val = readl_relaxed(regs + PVT_REG_10);
	val |= (TEMP_CLR_NORMAL_CLEAR | TEMP_CLR_FATAL_CLEAR);
	writel_relaxed(val, regs + PVT_REG_10);
}
#endif
static int gb02_pvt_init(struct gb02_pvt_thermal_priv *priv, unsigned char mode)
{
	void __iomem *regs;
	int ret = 0;

	regs = priv->regs;
	gb02_pvt_set_power(priv, PVT_POWER_UP);
//	gb02_pvt_set_mode(priv, mode);
//	gb02_pvt_set_thresh_normal(priv, DOUT_MIN, DOUT_MAX);
//	gb02_pvt_set_thresh_fatal(priv, DOUT_MIN, DOUT_MAX);
#ifdef VOLTAGE
	gb02_pvt_cal_trim(priv);
#endif

	gb02_pvt_dereset(priv);
	usleep_range(10, 20);
	gb02_pvt_reset(priv);
	usleep_range(10, 20);
	gb02_pvt_dereset(priv);
	return ret;
}

int get_temp_info(void)
{
	long long tempval;
	u16 outdata;
	struct gb02_pvt_thermal_priv priv;
	struct gb_pcie_info *gb_pcie = get_pcie_gl_info();

	if (gb_pcie == NULL)
		return 0;

	priv.regs = gb_pcie->pci_bars[PVT_DEFAULT_BAR].mmio + (CS_CNOC_S1_PVT_OFFSET - GB02_DEFAULT_CFG_SYS_PERI_ADDR);
	gb02_pvt_init(&priv, TEMP_MODE);
	outdata = gb02_pvt_get_dout_data(&priv);
	tempval = gb02_pvt_get_dout_tempcal_data(outdata);
	return ((tempval * 10 - 27049) * 44) / 1000;
}

static int gb02_get_cooling_device(struct gb02_pvt_thermal_priv *thermal,
								   struct thermal_cooling_device *cdev)
{
	int ret;

	ret = strcmp(cdev->type, "pwm-fan");
	if (ret)
		ret = strcmp(cdev->type, "gb02-freq");

	return ret;
}

static int gb02_pvt_thermal_bind(struct thermal_zone_device *tzdev,
								 struct thermal_cooling_device *cdev)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;
	struct device *dev = thermal->dev;
	int i, err;

	/* If the cooling device is one of ours bind it */
	if (gb02_get_cooling_device(thermal, cdev))
		return 0;

	for (i = 0; i < GB02_PVT_THERMAL_NUM_TRIPS; i++) {
		const struct gb02_pvt_thermal_trip *trip = &thermal->trips[i];

		err = thermal_zone_bind_cooling_device(tzdev, i, cdev,
											   trip->max_state,
											   trip->min_state,
											   THERMAL_WEIGHT_DEFAULT);
		if (err < 0) {
			dev_err(dev, "Failed to bind cooling device to trip %d\n", i);
			return err;
		}
	}
	return 0;
}

static int gb02_pvt_thermal_unbind(struct thermal_zone_device *tzdev,
								   struct thermal_cooling_device *cdev)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;
	struct device *dev = thermal->dev;
	int i;
	int err;

	/* If the cooling device is our one unbind it */
	if (gb02_get_cooling_device(thermal, cdev))
		return 0;

	for (i = 0; i < GB02_PVT_THERMAL_NUM_TRIPS; i++) {
		err = thermal_zone_unbind_cooling_device(tzdev, i, cdev);
		if (err < 0) {
			dev_err(dev, "Failed to unbind cooling device\n");
			return err;
		}
	}
	return 0;
}

/**
 * thermal_emergency_poweroff_func - emergency poweroff work after a known delay
 * @work: work_struct associated with the emergency poweroff function
 *
 */
static void thermal_emergency_poweroff(void)
{
	/*
	 * We have reached here after the emergency thermal shutdown
	 * Waiting period has expired. This means orderly_poweroff has
	 * not been able to shut off the system for some reason.
	 * Try to shut down the system immediately using kernel_power_off
	 * if populated
	 */
	WARN(1, "Attempting kernel_power_off: Temperature too high\n");
	kernel_power_off();

	/*
	 * Worst of the worst case trigger emergency restart
	 */
	WARN(1, "Attempting emergency_restart: Temperature too high\n");
	emergency_restart();
}

static void gb02_pvt_thermal_notify(struct work_struct *work)
{
	struct gb02_pvt_thermal_priv *priv =
			container_of(work, struct gb02_pvt_thermal_priv, work.work);

	if (priv->shutdown_temp > GB02_PVT_THERMAL_ASIC_TEMP_CRIT)
	{
		//shutdown
		dev_emerg(&priv->pvt_tzd->device,
			"critical temperature reached (%d C), shutting down\n",
			priv->shutdown_temp / 1000);
		mutex_lock(&poweroff_lock);
		if (!power_off_triggered) {
			/*
			 * Queue a backup emergency shutdown in the event of
			 * orderly_poweroff failure
			 */
			thermal_emergency_poweroff();
			orderly_poweroff(true);
			power_off_triggered = true;
		}
		mutex_unlock(&poweroff_lock);
	}
}

static int gb02_pvt_thermal_get_temp(struct thermal_zone_device *thermal, int *temp)
{
	struct gb02_pvt_thermal_priv *priv = thermal->devdata;
	long long tempval;
	int ret = 0;
	u16 outdata;
	bool temp_exceed_threshold;
	static bool trip_hot = false;

	if (thermal->trips_disabled)
		thermal->trips_disabled = 0;

	if (!priv->pvt_tzd)
		return -EAGAIN;
	msleep(20);
	mutex_lock(&priv->lock);
//	gb02_pvt_set_mode(priv, TEMP_MODE);
	outdata = gb02_pvt_get_dout_data(priv);

	GB02_PVT_DEBUG("DOUTact: %d\n", outdata);

	tempval = gb02_pvt_get_dout_tempcal_data(outdata);

	GB02_PVT_DEBUG("DOUTcal: %lld\n", tempval);
	/*
	 * Temperature calulation:
	 * Dout = 2.2632 * T + 2704.9
	 * To avoid float-poing arithmetic,increase
	 * the value of Temperature by a factor of
	 * 1000:
	 * T = (10*DOUT - 27049)*44
	 */
	if (thermal->emul_temperature)
		*temp = thermal->emul_temperature;
	else
		*temp = (tempval * 10 - 27049) * 44; // 441 -> 1000/2.2632

	GB02_PVT_DEBUG("caculate temp: %d\n", *temp);
	priv->shutdown_temp = *temp;
	mutex_unlock(&priv->lock);

	temp_exceed_threshold = (*temp > GB02_PVT_THERMAL_ASIC_TEMP_HOT);
	if (temp_exceed_threshold && !trip_hot) {
		trip_hot = true;
		blocking_notifier_call_chain(&limit_freq_notifier_chain, TEMP_HOT_EVENT, NULL);
	} else if (!temp_exceed_threshold && trip_hot) {
		trip_hot = false;
		blocking_notifier_call_chain(&limit_freq_notifier_chain, TEMP_NORMAL_EVENT, NULL);
	}

	if (*temp > GB02_PVT_THERMAL_ASIC_TEMP_CRIT)
		schedule_delayed_work(&priv->work, msecs_to_jiffies(POWEROFF_DELAY_MS));

	return ret;
}

static int gb02_pvt_thermal_get_trip_type(struct thermal_zone_device *tzdev,
		int trip,
		enum thermal_trip_type *p_type)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;

	if (trip < 0 || trip >= GB02_PVT_THERMAL_NUM_TRIPS)
		return -EINVAL;

	*p_type = thermal->trips[trip].type;
	return 0;
}

static int gb02_pvt_thermal_get_trip_temp(struct thermal_zone_device *tzdev,
		int trip, int *p_temp)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;

	if (trip < 0 || trip >= GB02_PVT_THERMAL_NUM_TRIPS)
		return -EINVAL;

	*p_temp = thermal->trips[trip].temp;
	return 0;
}

static int gb02_pvt_thermal_set_trip_temp(struct thermal_zone_device *tzdev,
		int trip, int temp)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;

	if (trip < 0 || trip >= GB02_PVT_THERMAL_NUM_TRIPS)
		return -EINVAL;

	thermal->trips[trip].temp = temp;
	return 0;
}

static int gb02_pvt_thermal_get_trip_hyst(struct thermal_zone_device *tzdev,
		int trip, int *p_hyst)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;

	*p_hyst = thermal->trips[trip].hyst;
	return 0;
}

static int gb02_pvt_thermal_set_trip_hyst(struct thermal_zone_device *tzdev,
		int trip, int hyst)
{
	struct gb02_pvt_thermal_priv *thermal = tzdev->devdata;

	thermal->trips[trip].hyst = hyst;
	return 0;
}

#if (!IS_ENABLED(CONFIG_THERMAL_EMULATION))
static ssize_t
emul_temp_store(struct device *dev, struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct thermal_zone_device *tz =
				container_of(dev, struct thermal_zone_device, device);
	int temperature;

	if (kstrtoint(buf, 10, &temperature))
		return -EINVAL;

	mutex_lock(&tz->lock);
	tz->emul_temperature = temperature;
	mutex_unlock(&tz->lock);

	return count;
}
static DEVICE_ATTR_WO(emul_temp);
#endif

static struct thermal_zone_device_ops gb02_pvt_thermal_ops = {
	.bind = gb02_pvt_thermal_bind,
	.unbind = gb02_pvt_thermal_unbind,
	.get_temp = gb02_pvt_thermal_get_temp,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)	
	.get_trip_type	= gb02_pvt_thermal_get_trip_type,
	.get_trip_temp	= gb02_pvt_thermal_get_trip_temp,
#endif	
	.set_trip_temp	= gb02_pvt_thermal_set_trip_temp,
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)	
	.get_trip_hyst	= gb02_pvt_thermal_get_trip_hyst,
#endif	
	.set_trip_hyst	= gb02_pvt_thermal_set_trip_hyst,
};

static int gb02_pvt_thermal_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_pvt_thermal_priv *priv;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id;
	int id, ret = 0;

	GB02_PVT_DEBUG("Start to probe GB02 PVT\n");

	pcie_info = gb_dev->gb_pcie;
	id = device_id;
	priv = devm_kzalloc(gb_dev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	priv->regs = gb_dev->gb_pcie->pci_bars[PVT_DEFAULT_BAR].mmio + (CS_CNOC_S1_PVT_OFFSET - GB02_DEFAULT_CFG_SYS_PERI_ADDR);
	priv->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	priv->gb_comp_irq_reg_bar_base = gb_dev->gb_pcie->pci_bars[GB02_SYSTEM_PERIPHERAL_BAR].mmio; //bar1

	if (IS_ERR(priv->regs))
		return PTR_ERR(priv->regs);

	mutex_init(&priv->lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)	
	priv->pvt_tzd = thermal_zone_device_register("gb02_pvt-thermal.0", GB02_PVT_THERMAL_NUM_TRIPS, GB02_PVT_THERMAL_TRIP_MASK, priv, &gb02_pvt_thermal_ops, &gb02_pvt_zone_params, 0, 1000);
#else
	priv->pvt_tzd = thermal_zone_device_register_with_trips("gb02_pvt-thermal.0", default_thermal_trips, GB02_PVT_THERMAL_NUM_TRIPS, GB02_PVT_THERMAL_TRIP_MASK, priv, &gb02_pvt_thermal_ops, &gb02_pvt_zone_params, 0, 1000);
#endif
	if (IS_ERR(priv->pvt_tzd)) {
		ret = PTR_ERR(priv->pvt_tzd);
		dev_err(gb_dev->dev, "failed to register sensor: %d\n", ret);
		return ret;
	}

	GB02_PVT_DEBUG("regist thermal: %s\n", priv->pvt_tzd->type);
	memcpy(priv->trips, default_thermal_trips, sizeof(priv->trips));
	priv->dev = gb_dev->dev;
	peri_info->priv = priv;
	gb02_pvt_init(priv, TEMP_MODE);
	INIT_DELAYED_WORK(&priv->work, gb02_pvt_thermal_notify);
#if (!IS_ENABLED(CONFIG_THERMAL_EMULATION))
	ret = sysfs_create_file(&priv->pvt_tzd->device.kobj, &dev_attr_emul_temp.attr);
	if (ret)
		dev_err(gb_dev->dev, "Failed to register emul_temp node\n");
#endif
	return ret;
}

static int gb02_pvt_thermal_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_pvt_thermal_priv *priv = (struct gb02_pvt_thermal_priv *)peri_info->priv;

	gb02_pvt_reset(priv);
	gb02_pvt_set_power(priv, PVT_POWER_DOWN);
	thermal_zone_device_unregister(priv->pvt_tzd);
	return 0;
}

int gb02_pvt_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;

	GB02_PVT_DEBUG("pvt sensor register\n");
	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_pvt_thermal_init(gb_dev, peri_info, device_id);

	return ret;
}

void gb02_pvt_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	gb02_pvt_thermal_exit(peri_info);
}
