// SPDX-License-Identifier: GPL-2.0
/*
 *  Hwmon driver for fans connected to PWM lines.
 *
 */

#include "common/gb-peripherals-common.h"
#include <linux/hwmon-sysfs.h>
#include <linux/hwmon.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/of.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/pwm.h>
#include <linux/gpio.h>
#include <linux/sysfs.h>
#include <linux/version.h>
#include <linux/thermal.h>
#include "../gb_pwm.h"
#include "mcu_peripherals/gpio/pinctrl-gb02-mcu.h"

#define MAX_PWM 30000
#define LEVEL_NUM 5
//PERIOD use ns
#define FAN_PWM_PERIOD 30000
#define FAN_REG_READ
#define FAN_FG_GPIO 41
//#define FAN_DEBUG
#ifdef FAN_DEBUG
#define GB02_FAN_DEBUG(arg...) gb_printf(KERN_INFO, "[" GB02_PWM_FAN_DEV_NAME "] " arg)
#else
#define GB02_FAN_DEBUG(arg...)
#endif

struct pwm_fan_ctx {
	struct mutex lock;
	struct gb_pwm_device *pwm;
	unsigned int pwm_value;
	unsigned int pwm_fan_state;
	unsigned int pwm_fan_max_state;
	unsigned int *pwm_fan_cooling_levels;
	void __iomem *sysctl_iofunc_base; // bar0
	atomic_t pulses;
	unsigned int rpm;
	ktime_t sample_start;
	u8 pulses_per_revolution;
	struct timer_list rpm_timer;
	bool user_mode;
	int irq;
	struct thermal_cooling_device *cdev;
};
static struct pwm_fan_ctx *g_ctx;

static int __set_pwm(struct pwm_fan_ctx *ctx, unsigned long pwm)
{
	unsigned long	 period;
	int				 ret   = 0;
	struct gb_pwm_state state = {};

	GB02_FAN_DEBUG("pwm:%ld\n", pwm);
	mutex_lock(&ctx->lock);
	if (ctx->pwm_value == pwm)
		goto exit_set_pwm_err;

	period	= ctx->pwm->period;
	state.duty_cycle = DIV_ROUND_UP(pwm * (period - 1), MAX_PWM);
	state.period	 = FAN_PWM_PERIOD;
	state.enable	 = pwm ? true : false;

	ret = gb_pwm_apply(ctx->pwm, &state);
	if (!ret)
		ctx->pwm_value = pwm;
exit_set_pwm_err:
	mutex_unlock(&ctx->lock);
	return ret;
}

static void pwm_fan_update_state(struct pwm_fan_ctx *ctx, unsigned long pwm)
{
	int i;

	for (i = 0; i < ctx->pwm_fan_max_state; ++i) {
		if (pwm < ctx->pwm_fan_cooling_levels[i + 1])
			break;
	}

	ctx->pwm_fan_state = i;
}

static ssize_t set_level(struct device *dev, struct device_attribute *attr, const char *buf,
					   size_t count)
{
	struct pwm_fan_ctx *ctx = dev_get_drvdata(dev);
	unsigned long		level;
	int					ret;

	if (kstrtoul(buf, 10, &level) || level > LEVEL_NUM)
		return -EINVAL;

	if (level > 0) {
		ret = __set_pwm(ctx, ctx->pwm_fan_cooling_levels[level-1]);
		if (ret)
			return ret;
		ctx->user_mode = true;
	} else {
		ctx->user_mode = false;
	}

	pwm_fan_update_state(ctx, ctx->pwm_fan_cooling_levels[level-1]);
	return count;
}

static ssize_t show_level(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct pwm_fan_ctx *ctx = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", ctx->pwm_fan_state + 1);
}

static ssize_t set_pwm(struct device *dev, struct device_attribute *attr, const char *buf,
					   size_t count)
{
	struct pwm_fan_ctx *ctx = dev_get_drvdata(dev);
	unsigned long		pwm;
	int					ret;

	if (kstrtoul(buf, 10, &pwm) || pwm > MAX_PWM)
		return -EINVAL;

	if (pwm > 0) {
		ret = __set_pwm(ctx, pwm);
		if (ret)
			return ret;
		ctx->user_mode = true;
	} else {
		ctx->user_mode = false;
	}

	pwm_fan_update_state(ctx, pwm);
	return count;
}

static ssize_t show_pwm(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct pwm_fan_ctx *ctx = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", ctx->pwm_value);
}

static ssize_t show_rpm(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct pwm_fan_ctx *ctx = dev_get_drvdata(dev);
#ifdef FAN_REG_READ
	ctx->rpm = base_reg_read(ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC00);
#endif
	return sprintf(buf, "%u\n", ctx->rpm);
}

static SENSOR_DEVICE_ATTR(fan1_level, 0644, show_level, set_level, 0);
static SENSOR_DEVICE_ATTR(pwm1, 0644, show_pwm, set_pwm, 0);
static SENSOR_DEVICE_ATTR(fan1_input, 0444, show_rpm, NULL, 0);

static struct attribute *pwm_fan_attrs[] = {
	&sensor_dev_attr_pwm1.dev_attr.attr,
	&sensor_dev_attr_fan1_input.dev_attr.attr,
	&sensor_dev_attr_fan1_level.dev_attr.attr,
	NULL,
};

ATTRIBUTE_GROUPS(pwm_fan);

/* thermal cooling device callbacks */
static int pwm_fan_get_max_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
	struct pwm_fan_ctx *ctx = cdev->devdata;

	if (!ctx)
		return -EINVAL;

	*state = ctx->pwm_fan_max_state;

	return 0;
}

static int pwm_fan_get_cur_state(struct thermal_cooling_device *cdev, unsigned long *state)
{
	struct pwm_fan_ctx *ctx = cdev->devdata;

	if (!ctx)
		return -EINVAL;

	*state = ctx->pwm_fan_state;

	return 0;
}

static int pwm_fan_set_cur_state(struct thermal_cooling_device *cdev, unsigned long state)
{
	struct pwm_fan_ctx *ctx = cdev->devdata;
	int					ret;

	if (!ctx || (state > ctx->pwm_fan_max_state))
		return -EINVAL;

	if (state == ctx->pwm_fan_state)
		return 0;

	if (ctx->user_mode)
		return 0;
	ret = __set_pwm(ctx, ctx->pwm_fan_cooling_levels[state]);
	if (ret) {
		dev_err(&cdev->device, "Cannot set pwm!\n");
		return ret;
	}

	ctx->pwm_fan_state = state;

	return ret;
}

static const struct thermal_cooling_device_ops pwm_fan_cooling_ops = {
	.get_max_state = pwm_fan_get_max_state,
	.get_cur_state = pwm_fan_get_cur_state,
	.set_cur_state = pwm_fan_set_cur_state,
};

static int pwm_fan_get_cooling_data(struct device *dev, struct pwm_fan_ctx *ctx)
{
	int num = LEVEL_NUM, i;

	ctx->pwm_fan_cooling_levels = devm_kcalloc(dev, LEVEL_NUM, sizeof(u32), GFP_KERNEL);
	if (!ctx->pwm_fan_cooling_levels)
		return -ENOMEM;
	for (i = 0; i < LEVEL_NUM; i++) {
		ctx->pwm_fan_cooling_levels[i] = 6000 * (i + 1);
		if (ctx->pwm_fan_cooling_levels[i] > MAX_PWM) {
			dev_err(dev, "PWM fan state[%d]:%d > %d\n", i, ctx->pwm_fan_cooling_levels[i], MAX_PWM);
			return -EINVAL;
		}
	}

	ctx->pwm_fan_max_state = num - 1;

	return 0;
}

#ifndef FAN_REG_READ
static void sample_timer(struct timer_list *t)
{
	struct pwm_fan_ctx *ctx = from_timer(ctx, t, rpm_timer);
	int pulses;
	u64 tmp;

	pulses = atomic_read(&ctx->pulses);
	atomic_sub(pulses, &ctx->pulses);
	tmp = (u64)pulses * ktime_ms_delta(ktime_get(), ctx->sample_start) * 60;
	do_div(tmp, ctx->pulses_per_revolution * 1000);
	if (tmp == 0)
		tmp = base_reg_read(ctx->sysctl_iofunc_base + 0x1000000, 0x0003FC00);
	ctx->rpm = tmp;

	ctx->sample_start = ktime_get();
	mod_timer(&ctx->rpm_timer, jiffies + HZ);
}

static irqreturn_t gb02_pulse_handler(int irq, void *data)
{
	struct pwm_fan_ctx *ctx = data;

	atomic_inc(&ctx->pulses);

	return IRQ_HANDLED;

}
#endif

int gb02_pwm_fan_init(struct gb_device *gbdev,
		struct gb02_mcu_peripherals_info *peri_info, struct gb02_pwm_device *pwm_dev)
{
	struct thermal_cooling_device *cdev;
	struct pwm_fan_ctx *ctx;
	struct device *hwmon;
	int ret;
	struct gb_pwm_state state = {};
	struct gb_pwm_chip *gb02_fan_pwm;

	gb02_fan_pwm = (struct gb_pwm_chip *)peri_info->priv;
	ctx = devm_kzalloc(gbdev->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	mutex_init(&ctx->lock);

	ctx->pwm = (struct gb_pwm_device *)&gb02_fan_pwm->pwms[FAN_PWM_CHAN_NO];
	if (IS_ERR(ctx->pwm)) {
		ret = PTR_ERR(ctx->pwm);
		if (ret != -EPROBE_DEFER)
			dev_err(gbdev->dev, "Could not get PWM: %d\n", ret);
		return ret;
	}

	ctx->pwm->period = FAN_PWM_PERIOD;
	ctx->pwm_value = MAX_PWM;
	ctx->sysctl_iofunc_base = gbdev->gb_pcie->pci_bars[0].mmio; // bar0
	/* Set duty cycle to maximum allowed and enable PWM output */
	state.duty_cycle = FAN_PWM_PERIOD - 1;
	state.enable	 = true;
	state.period	 = FAN_PWM_PERIOD;
	ret = gb_pwm_apply(ctx->pwm, &state);
	if (!ctx->pwm)
		pr_err("%s pwm is NULL\n", __func__);
	if (ret)
		return ret;

	hwmon = devm_hwmon_device_register_with_groups(gbdev->dev, "pwmfan", ctx, pwm_fan_groups);
	if (IS_ERR(hwmon)) {
		dev_err(gbdev->dev, "Failed to register hwmon device\n");
		ret = PTR_ERR(hwmon);
		goto err_pwm_disable;
	}

	ret = pwm_fan_get_cooling_data(gbdev->dev, ctx);
	if (ret)
		goto err_pwm_disable;

	ctx->pwm_fan_state = ctx->pwm_fan_max_state;
	ctx->pulses_per_revolution = 2;

#ifndef FAN_REG_READ
	ret = gb02_gpio_request(FAN_FG_GPIO);
	if (ret) {
		dev_err(gbdev->dev, "Failed to request  gpio\n");
		return ret;
	}
	ret = gb02_gpio_set_direction(FAN_FG_GPIO, GB02_GPIO_LINE_DIRECTION_IN);
	if (ret) {
		dev_err(gbdev->dev, "Failed to set  gpio direction\n");
		return ret;
	}

	gb02_gpio_irq_enable(FAN_FG_GPIO, true);
	//register irq
	ctx->irq = pci_irq_vector(gbdev->gb_pcie->pdev, 6);
	if (ctx->irq > 0) {
		ret = devm_request_threaded_irq(gbdev->dev, ctx->irq, NULL, gb02_pulse_handler, IRQF_SHARED | IRQF_TRIGGER_RISING,
				       "fan_fg", ctx);
		if (ret) {
			dev_err(gbdev->dev, "Failed to request interrupt: %d\n", ret);
			return ret;
		}
		ctx->sample_start = ktime_get();
		timer_setup(&ctx->rpm_timer, sample_timer, 0);
		mod_timer(&ctx->rpm_timer, jiffies + HZ);
	}
#endif

	if (IS_ENABLED(CONFIG_THERMAL)) {
		cdev = thermal_of_cooling_device_register(gbdev->dev->of_node, "pwm-fan", ctx,
				&pwm_fan_cooling_ops);
		if (IS_ERR(cdev)) {
			dev_err(gbdev->dev, "Failed to register pwm-fan as cooling device");
			ret = PTR_ERR(cdev);
			goto err_pwm_disable;
		}
		ctx->cdev = cdev;
	}

	g_ctx = ctx;

	return 0;

err_pwm_disable:
	state.enable = false;
	gb_pwm_apply(ctx->pwm, &state);

	return ret;
}

void gb02_pwm_fan_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	struct pwm_fan_ctx *ctx = g_ctx;

	thermal_cooling_device_unregister(ctx->cdev);
	if (ctx->pwm_value)
		gb_pwm_disable(ctx->pwm);
	del_timer_sync(&ctx->rpm_timer);
}

int pwm_fan_suspend(struct gb02_mcu_peripherals_info *peri_info)
{
	struct pwm_fan_ctx *ctx;
	int ret;

	ctx = g_ctx;

	ret = __set_pwm(ctx, ctx->pwm_fan_cooling_levels[0]);
	if (ret)
		return ret;
	ctx->pwm_fan_state = 0;

	return 0;
}

int pwm_fan_resume(struct gb_device *gbdev, struct gb02_mcu_peripherals_info *peri_info)
{
	struct pwm_fan_ctx *ctx;
	int ret;

	ctx = g_ctx;

	if (ctx->pwm_value == 0)
		return 0;
#ifndef FAN_REG_READ
	//restore gpio status
	ret = gb02_gpio_request(FAN_FG_GPIO);
	if (ret) {
		dev_err(gbdev->dev, "Failed to request  gpio\n");
		return ret;
	}
	ret = gb02_gpio_set_direction(FAN_FG_GPIO, GB02_GPIO_LINE_DIRECTION_IN);
	if (ret) {
		dev_err(gbdev->dev, "Failed to set  gpio direction\n");
		return ret;
	}

	gb02_gpio_irq_enable(FAN_FG_GPIO, true);
#endif

	return ret;
}
