// SPDX-License-Identifier: GPL-2.0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <linux/io.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/bitfield.h>
#include <linux/dmaengine.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/seq_file.h>
#include <linux/platform_device.h>
#include <linux/pwm.h>
#include "gb_pwm.h"
#include "common/gb-peripherals-common.h"

#ifdef PWM_DEBUG
#define GB02_PWM_DEBUG(arg...)		gb_printf(KERN_INFO, "["PLATFORM_PWM_DEVICE_NAME"] " arg)
#else
#define GB02_PWM_DEBUG(arg...)
#endif

#define PWM_MCU_TEST_CMD_BASE 'W'
#define PWM_MCU_CMD_1 _IOW(PWM_MCU_TEST_CMD_BASE, 1u, int)

#define CCMR_CHANNEL_SHIFT 8
#define CCMR_CHANNEL_MASK  0xFF
#define MAX_BREAKINPUT 2

static inline void gb02_pwm_write(unsigned int val,
				struct mcu_pwm_chip *mcu_pwm_chip, unsigned int reg)
{
	writel(val, mcu_pwm_chip->res + reg);
}

static inline unsigned long gb02_pwm_read(struct mcu_pwm_chip *mcu_pwm_chip,
		unsigned int reg)
{
	return readl(mcu_pwm_chip->res + reg);
}

static inline void gb02_pwm_mask_write(unsigned int dat,
				struct mcu_pwm_chip *mcu_pwm_chip, unsigned int reg, unsigned int mask)
{
	unsigned long val = 0;

	val = gb02_pwm_read(mcu_pwm_chip, reg);
	val &= ~mask;
	val |= dat;

	gb02_pwm_write(val, mcu_pwm_chip, reg);
}

static inline struct mcu_pwm_chip *to_mcu_pwm_chip(struct gb_pwm_chip *chip)
{
	return container_of(chip, struct mcu_pwm_chip, chip);
}

static void gb02_pwm_set_mux(struct mcu_pwm_chip *mcu_pwm_chip, int id)
{
	unsigned int val;

	val = base_reg_read(mcu_pwm_chip->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
	if (id == DEVICE_ID0)
		val &= (~BIT(9));
	else if (id == DEVICE_ID1)
		val &= (~BIT(10));
	GB02_PWM_DEBUG("pwm set mux %x\n", val);
	base_reg_write(val, mcu_pwm_chip->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
}

static u32 active_channels(struct mcu_pwm_chip *mcu_pwm_chip)
{
	u32 ccer;

	ccer = gb02_pwm_read(mcu_pwm_chip, TIM_CCER);
	return ccer & TIM_CCER_CCXE;
}

static void write_ccrx(struct mcu_pwm_chip *pwm, int ch, u32 value)
{
	switch (ch) {
	case 0:
		gb02_pwm_write(value, pwm, TIM_CCR1);
		break;
	case 1:
		gb02_pwm_write(value, pwm, TIM_CCR2);
		break;
	case 2:
		gb02_pwm_write(value, pwm, TIM_CCR3);
		break;
	case 3:
		gb02_pwm_write(value, pwm, TIM_CCR4);
		break;
	default:
		break;
	}
}

#define TIM_CCER_CC12P (TIM_CCER_CC1P | TIM_CCER_CC2P)
#define TIM_CCER_CC12E (TIM_CCER_CC1E | TIM_CCER_CC2E)
#define TIM_CCER_CC34P (TIM_CCER_CC3P | TIM_CCER_CC4P)
#define TIM_CCER_CC34E (TIM_CCER_CC3E | TIM_CCER_CC4E)
/* DIER register DMA enable bits */
static const u32 gb02_timers_dier_dmaen[GB02_TIMERS_MAX_DMAS] = {
	TIM_DIER_CC1DE,
	TIM_DIER_CC2DE,
	TIM_DIER_CC3DE,
	TIM_DIER_CC4DE,
	TIM_DIER_UIE,
	TIM_DIER_TDE,
	TIM_DIER_COMDE
};

static void gb02_timers_dma_done(void *p)
{
	struct gb02_timers_dma *dma = p;
	struct dma_tx_state state;
	enum dma_status status;

	status = dmaengine_tx_status(dma->chan, dma->chan->cookie, &state);
	if (status == DMA_COMPLETE)
		complete(&dma->completion);
}

/**
 * gb02_timers_dma_burst_read - Read from timers registers using DMA.
 *
 * Read from GB02 timers registers using DMA on a single event.
 * @dev: reference to gb02_timers_dma MFD device
 * @buf: DMA'able destination buffer
 * @id: gb02_timers_dmas event identifier (ch[1..4], up, trig or com)
 * @reg: registers start offset for DMA to read from (like CCRx for capture)
 * @num_reg: number of registers to read upon each DMA request, starting @reg.
 * @bursts: number of bursts to read (e.g. like two for pwm period capture)
 * @tmo_ms: timeout (milliseconds)
 */
static int gb02_timers_dma_burst_read(struct mcu_pwm_chip *mcu_pwm_chip, u32 *buf,
				enum gb02_timers_dmas id, u32 reg,
				unsigned int num_reg, unsigned int bursts,
				unsigned long tmo_ms)
{
	unsigned long timeout = msecs_to_jiffies(tmo_ms);
	struct gb02_timers_dma *dma = &mcu_pwm_chip->dma;
	struct device *dev = mcu_pwm_chip->chip.dev;
	size_t len = num_reg * bursts * sizeof(u32);
	struct dma_async_tx_descriptor *desc;
	struct dma_slave_config config;
	dma_cookie_t cookie;
	dma_addr_t dma_buf;
	u32 dbl, dba;
	long err;
	int ret;

	/* Sanity check */
	if (id < GB02_TIMERS_DMA_CH1 || id >= GB02_TIMERS_MAX_DMAS)
		return -EINVAL;

	if (!num_reg || !bursts || reg > GB02_TIMERS_MAX_REGISTERS ||
		(reg + num_reg * sizeof(u32)) > GB02_TIMERS_MAX_REGISTERS)
		return -EINVAL;

	if (!dma->chans[id])
		return -ENODEV;
	mutex_lock(&dma->lock);

	/* Select DMA channel in use */
	dma->chan = dma->chans[id];
	dma_buf = dma_map_single(dev, buf, len, DMA_FROM_DEVICE);
	if (dma_mapping_error(dev, dma_buf)) {
		ret = -ENOMEM;
		goto unlock;
	}

	/* Prepare DMA read from timer registers, using DMA burst mode */
	memset(&config, 0, sizeof(config));
	config.src_addr = (dma_addr_t)dma->phys_base + TIM_DMAR;
	config.src_addr_width = DMA_SLAVE_BUSWIDTH_4_BYTES;
	ret = dmaengine_slave_config(dma->chan, &config);
	if (ret)
		goto unmap;

	desc = dmaengine_prep_slave_single(dma->chan, dma_buf, len,
									   DMA_DEV_TO_MEM, DMA_PREP_INTERRUPT);
	if (!desc) {
		ret = -EBUSY;
		goto unmap;
	}

	desc->callback = gb02_timers_dma_done;
	desc->callback_param = dma;
	cookie = dmaengine_submit(desc);
	ret = dma_submit_error(cookie);
	if (ret)
		goto dma_term;

	reinit_completion(&dma->completion);
	dma_async_issue_pending(dma->chan);

	/* Setup and enable timer DMA burst mode */
	dbl = FIELD_PREP(TIM_DCR_DBL, bursts - 1);
	dba = FIELD_PREP(TIM_DCR_DBA, reg >> 2);
	gb02_pwm_write(dbl | dba, mcu_pwm_chip, TIM_DCR);

	/* Clear pending flags before enabling DMA request */
	gb02_pwm_write(0, mcu_pwm_chip, TIM_SR);

	gb02_pwm_mask_write(gb02_timers_dier_dmaen[id], mcu_pwm_chip, TIM_DIER, gb02_timers_dier_dmaen[id]);

	err = wait_for_completion_interruptible_timeout(&dma->completion,
			timeout);
	if (err == 0)
		ret = -ETIMEDOUT;
	else if (err < 0)
		ret = err;

	gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_DIER, gb02_timers_dier_dmaen[id]);
	gb02_pwm_write(0, mcu_pwm_chip, TIM_SR);

	gb02_pwm_write(0, mcu_pwm_chip, TIM_DCR);
dma_term:
	dmaengine_terminate_all(dma->chan);
unmap:
	dma_unmap_single(dev, dma_buf, len, DMA_FROM_DEVICE);
unlock:
	dma->chan = NULL;
	mutex_unlock(&dma->lock);

	return ret;
}

static void gb02_timers_dma_probe(struct mcu_pwm_chip *mcu_pwm_chip)
{
	int i;
	char name[4];
	struct gb02_timers_dma *dma = &mcu_pwm_chip->dma;
	struct device *dev = mcu_pwm_chip->chip.dev;

	init_completion(&dma->completion);
	mutex_init(&dma->lock);
	dma->phys_base = (phys_addr_t)mcu_pwm_chip->res;
	/* Optional DMA support: get valid DMA channel(s) or NULL */
	for (i = GB02_TIMERS_DMA_CH1; i <= GB02_TIMERS_DMA_CH4; i++) {
		snprintf(name, ARRAY_SIZE(name), "ch%1d", i + 1);
		dma->chans[i] = dma_request_slave_channel(dev, name);
	}
	dma->chans[GB02_TIMERS_DMA_UP] =
		dma_request_slave_channel(dev, "up");
	dma->chans[GB02_TIMERS_DMA_TRIG] =
		dma_request_slave_channel(dev, "trig");
	dma->chans[GB02_TIMERS_DMA_COM] =
		dma_request_slave_channel(dev, "com");
}

static void gb02_timers_dma_remove(struct mcu_pwm_chip *mcu_pwm_chip)
{
	int i;
	struct gb02_timers_dma *dma = &mcu_pwm_chip->dma;

	for (i = GB02_TIMERS_DMA_CH1; i < GB02_TIMERS_MAX_DMAS; i++)
		if (dma->chans[i])
			dma_release_channel(dma->chans[i]);
}

/*
 * Capture using PWM input mode:
 *							  ___		  ___
 * TI[1, 2, 3 or 4]: ........._|   |________|
 *							 ^0  ^1	   ^2
 *							  .   .		.
 *							  .   .		XXXXX
 *							  .   .   XXXXX	 |
 *							  .  XXXXX	 .	|
 *							XXXXX .		.	|
 * COUNTER:		______XXXXX  .   .		.	|_XXX
 *				 start^	   .   .		.		^stop
 *					  .	   .   .		.
 *					  v	   v   .		v
 *								  v
 * CCR1/CCR3:	   tx..........t0...........t2
 * CCR2/CCR4:	   tx..............t1.........
 *
 * DMA burst transfer:		  |			|
 *							  v			v
 * DMA buffer:				  { t0, tx }   { t2, t1 }
 * DMA done:								 ^
 *
 * 0: IC1/3 snapchot on rising edge: counter value -> CCR1/CCR3
 *	+ DMA transfer CCR[1/3] & CCR[2/4] values (t0, tx: doesn't care)
 * 1: IC2/4 snapchot on falling edge: counter value -> CCR2/CCR4
 * 2: IC1/3 snapchot on rising edge: counter value -> CCR1/CCR3
 *	+ DMA transfer CCR[1/3] & CCR[2/4] values (t2, t1)
 *
 * DMA done, compute:
 * - Period	 = t2 - t0
 * - Duty cycle = t1 - t0
 */

static int gb02_pwm_raw_capture(struct mcu_pwm_chip *mcu_pwm_chip, struct gb_pwm_device *pwm,
								unsigned long tmo_ms, u32 *raw_prd,
								u32 *raw_dty)
{
	//struct device *parent = mcu_pwm_chip->chip.dev->parent;
	enum gb02_timers_dmas dma_id;
	u32 ccen, ccr;
	int ret;

	/* Ensure registers have been updated, enable counter and capture */
	gb02_pwm_mask_write(TIM_EGR_UG, mcu_pwm_chip, TIM_EGR, TIM_EGR_UG);
	gb02_pwm_mask_write(TIM_CR1_CEN, mcu_pwm_chip, TIM_CR1, TIM_CR1_CEN);

	/* Use cc1 or cc3 DMA resp for PWM input channels 1 & 2 or 3 & 4 */
	dma_id = pwm->hwpwm < 2 ? GB02_TIMERS_DMA_CH1 : GB02_TIMERS_DMA_CH3;
	ccen = pwm->hwpwm < 2 ? TIM_CCER_CC12E : TIM_CCER_CC34E;
	ccr = pwm->hwpwm < 2 ? TIM_CCR1 : TIM_CCR3;
	gb02_pwm_mask_write(ccen, mcu_pwm_chip, TIM_CCER, ccen);

	/*
	 * Timer DMA burst mode. Request 2 registers, 2 bursts, to get both
	 * CCR1 & CCR2 (or CCR3 & CCR4) on each capture event.
	 * We'll get two capture snapchots: { CCR1, CCR2 }, { CCR1, CCR2 }
	 * or { CCR3, CCR4 }, { CCR3, CCR4 }
	 */
	ret = gb02_timers_dma_burst_read(mcu_pwm_chip, mcu_pwm_chip->capture, dma_id, ccr, 2,
									 2, tmo_ms);
	if (ret)
		goto stop;

	/* Period: t2 - t0 (take care of counter overflow) */
	if (mcu_pwm_chip->capture[0] <= mcu_pwm_chip->capture[2])
		*raw_prd = mcu_pwm_chip->capture[2] - mcu_pwm_chip->capture[0];
	else
		*raw_prd = mcu_pwm_chip->max_arr - mcu_pwm_chip->capture[0] + mcu_pwm_chip->capture[2];

	/* Duty cycle capture requires at least two capture units */
	if (pwm->chip->npwm < 2)
		*raw_dty = 0;
	else if (mcu_pwm_chip->capture[0] <= mcu_pwm_chip->capture[3])
		*raw_dty = mcu_pwm_chip->capture[3] - mcu_pwm_chip->capture[0];
	else
		*raw_dty = mcu_pwm_chip->max_arr - mcu_pwm_chip->capture[0] + mcu_pwm_chip->capture[3];

	if (*raw_dty > *raw_prd) {
		/*
		 * Race beetween PWM input and DMA: it may happen
		 * falling edge triggers new capture on TI2/4 before DMA
		 * had a chance to read CCR2/4. It means capture[1]
		 * contains period + duty_cycle. So, subtract period.
		 */
		*raw_dty -= *raw_prd;
	}

stop:
	gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_CCER, ccen);
	gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_CR1, TIM_CR1_CEN);

	return ret;
}

static int gb02_pwm_capture(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
							struct gb_pwm_capture *result, unsigned long tmo_ms)
{
	struct mcu_pwm_chip *mcu_pwm_chip = to_mcu_pwm_chip(chip);
	unsigned long long prd, div, dty;
	unsigned long rate;
	unsigned int psc = 0, icpsc, scale;
	u32 raw_prd = 0, raw_dty = 0;
	int ret = 0;

	mutex_lock(&mcu_pwm_chip->chip.lock);

	if (active_channels(mcu_pwm_chip)) {
		ret = -EBUSY;
		goto unlock;
	}

	rate = mcu_pwm_chip->clk_rate;

	/* prescaler: fit timeout window provided by upper layer */
	div = (unsigned long long)rate * (unsigned long long)tmo_ms;
	do_div(div, MSEC_PER_SEC);
	prd = div;
	while ((div > mcu_pwm_chip->max_arr) && (psc < MAX_TIM_PSC)) {
		psc++;
		div = prd;
		do_div(div, psc + 1);
	}
	gb02_pwm_write(mcu_pwm_chip->max_arr, mcu_pwm_chip, TIM_ARR);
	gb02_pwm_write(psc, mcu_pwm_chip, TIM_PSC);

	/* Map TI1 or TI2 PWM input to IC1 & IC2 (or TI3/4 to IC3 & IC4) */
	gb02_pwm_mask_write(
		TIM_CCMR_CC1S | TIM_CCMR_CC2S, mcu_pwm_chip, pwm->hwpwm & 0x1 ?
		TIM_CCMR_CC1S_TI2 | TIM_CCMR_CC2S_TI2 :
		TIM_CCMR_CC1S_TI1 | TIM_CCMR_CC2S_TI1,
		pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2);

	/* Capture period on IC1/3 rising edge, duty cycle on IC2/4 falling. */
	gb02_pwm_mask_write(
		pwm->hwpwm < 2 ?
		TIM_CCER_CC2P : TIM_CCER_CC4P, mcu_pwm_chip, TIM_CCER, pwm->hwpwm < 2 ?
		TIM_CCER_CC12P : TIM_CCER_CC34P);

	ret = gb02_pwm_raw_capture(mcu_pwm_chip, pwm, tmo_ms, &raw_prd, &raw_dty);
	if (ret)
		goto stop;

	/*
	 * Got a capture. Try to improve accuracy at high rates:
	 * - decrease counter clock prescaler, scale up to max rate.
	 * - use input prescaler, capture once every /2 /4 or /8 edges.
	 */
	if (raw_prd) {
		u32 max_arr = mcu_pwm_chip->max_arr - 0x1000; /* arbitrary margin */

		scale = max_arr / min(max_arr, raw_prd);
	} else {
		scale = mcu_pwm_chip->max_arr; /* bellow resolution, use max scale */
	}

	if (psc && scale > 1) {
		/* 2nd measure with new scale */
		psc /= scale;

		gb02_pwm_write(psc, mcu_pwm_chip, TIM_PSC);
		ret = gb02_pwm_raw_capture(mcu_pwm_chip, pwm, tmo_ms, &raw_prd,
								   &raw_dty);
		if (ret)
			goto stop;
	}

	/* Compute intermediate period not to exceed timeout at low rates */
	prd = (unsigned long long)raw_prd * (psc + 1) * NSEC_PER_SEC;
	do_div(prd, rate);

	for (icpsc = 0; icpsc < MAX_TIM_ICPSC ; icpsc++) {
		/* input prescaler: also keep arbitrary margin */
		if (raw_prd >= (mcu_pwm_chip->max_arr - 0x1000) >> (icpsc + 1))
			break;
		if (prd >= (tmo_ms * NSEC_PER_MSEC) >> (icpsc + 2))
			break;
	}

	if (!icpsc)
		goto done;

	/* Last chance to improve period accuracy, using input prescaler */
	gb02_pwm_mask_write(FIELD_PREP(TIM_CCMR_IC1PSC, icpsc) |
						FIELD_PREP(TIM_CCMR_IC2PSC, icpsc), mcu_pwm_chip,
						pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2,
						TIM_CCMR_IC1PSC | TIM_CCMR_IC2PSC
					   );

	ret = gb02_pwm_raw_capture(mcu_pwm_chip, pwm, tmo_ms, &raw_prd, &raw_dty);
	if (ret)
		goto stop;

	if (raw_dty >= (raw_prd >> icpsc)) {
		/*
		 * We may fall here using input prescaler, when input
		 * capture starts on high side (before falling edge).
		 * Example with icpsc to capture on each 4 events:
		 *
		 *	   start   1st capture					 2nd capture
		 *		 v	 v							   v
		 *		 ___   _____   _____   _____   _____   ____
		 * TI1..4	 |__|	|__|	|__|	|__|	|__|
		 *			v  v	.  .	.  .	.	   v  v
		 * icpsc1/3:  .  0	.  1	.  2	.  3	.  0
		 * icpsc2/4:  0	   1	   2	   3	   0
		 *			v  v							v  v
		 * CCR1/3  ......t0..............................t2
		 * CCR2/4  ..t1..............................t1'...
		 *			   .							.  .
		 * Capture0:	 .<----------------------------->.
		 * Capture1:	 .<-------------------------->.  .
		 *			   .							.  .
		 * Period:	   .<------>					.  .
		 * Low side:								  .<>.
		 *
		 * Result:
		 * - Period = Capture0 / icpsc
		 * - Duty = Period - Low side = Period - (Capture0 - Capture1)
		 */
		raw_dty = (raw_prd >> icpsc) - (raw_prd - raw_dty);
	}

done:
	prd = (unsigned long long)raw_prd * (psc + 1) * NSEC_PER_SEC;
	result->period = DIV_ROUND_UP_ULL(prd, rate << icpsc);
	dty = (unsigned long long)raw_dty * (psc + 1) * NSEC_PER_SEC;
	result->duty_cycle = DIV_ROUND_UP_ULL(dty, rate);
stop:
	gb02_pwm_write(0, mcu_pwm_chip, TIM_CCER);
	gb02_pwm_write(0, mcu_pwm_chip, pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2);
	gb02_pwm_write(0, mcu_pwm_chip, TIM_PSC);
unlock:
	mutex_unlock(&mcu_pwm_chip->chip.lock);

	return ret;
}

static int gb02_pwm_config(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
						   int duty_ns, int period_ns)
{
	unsigned long long prd, div, dty;
	unsigned int prescaler = 0;
	u32 ccmr, mask, shift;
	int ch = pwm->hwpwm;
	struct mcu_pwm_chip *mcu_pwm_chip = to_mcu_pwm_chip(chip);

	gb02_pwm_set_mux(mcu_pwm_chip, mcu_pwm_chip->chip.chip_id);
	/* Period and prescaler values depends on clock rate */
	div = (unsigned long long)(mcu_pwm_chip->clk_rate) * period_ns;
	do_div(div, NSEC_PER_SEC);
	prd = div;

	while (div > mcu_pwm_chip->max_arr) {
		prescaler++;
		div = prd;
		do_div(div, prescaler + 1);
	}

	prd = div;

	if (prescaler > MAX_TIM_PSC)
		return -EINVAL;

	/*
	 * All channels share the same prescaler and counter so when two
	 * channels are active at the same time we can't change them
	 */
	if (active_channels(mcu_pwm_chip) & ~(1 << ch * 4)) {
		u32 psc, arr;

		psc = gb02_pwm_read(mcu_pwm_chip, TIM_PSC);
		arr = gb02_pwm_read(mcu_pwm_chip, TIM_ARR);
		if ((psc != prescaler) || (arr != prd - 1))
		{
			gb02_pwm_write(period_ns, mcu_pwm_chip, TIM_ARR);
			write_ccrx(mcu_pwm_chip, ch, duty_ns);
			pwm->state.duty_cycle = duty_ns;
			pwm->state.period = period_ns;
			return 0;
		}
	}

	gb02_pwm_write(prescaler, mcu_pwm_chip, TIM_PSC);
	gb02_pwm_write(prd - 1, mcu_pwm_chip, TIM_ARR);
	gb02_pwm_mask_write(TIM_CR1_ARPE, mcu_pwm_chip, TIM_CR1, TIM_CR1_ARPE);

	/* Calculate the duty cycles */
	dty = prd * duty_ns;
	do_div(dty, period_ns);
	write_ccrx(mcu_pwm_chip, ch, dty);
	/* Configure output mode */
	shift = (ch & 0x1) * CCMR_CHANNEL_SHIFT;
	ccmr = (TIM_CCMR_PE | TIM_CCMR_M1) << shift;
	mask = CCMR_CHANNEL_MASK << shift;

	if (ch < 2)
		gb02_pwm_mask_write(ccmr, mcu_pwm_chip, TIM_CCMR1, mask);
	else
		gb02_pwm_mask_write(ccmr, mcu_pwm_chip, TIM_CCMR2, mask);

	gb02_pwm_mask_write(TIM_BDTR_MOE | TIM_BDTR_AOE, mcu_pwm_chip,
						TIM_BDTR, TIM_BDTR_MOE | TIM_BDTR_AOE);
	pwm->state.duty_cycle = duty_ns;
	pwm->state.period = period_ns;
	return 0;
}

static int gb02_pwm_set_polarity(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
								 enum pwm_polarity polarity)
{
	u32 mask;
	int ch = pwm->hwpwm;
	struct mcu_pwm_chip *mcu_pwm_chip = to_mcu_pwm_chip(chip);
	return 0;

	mask = TIM_CCER_CC1P << (ch * 4);
	if (mcu_pwm_chip->have_complementary_output)
		mask |= TIM_CCER_CC1NP << (ch * 4);

	gb02_pwm_mask_write(polarity == PWM_POLARITY_NORMAL ? 0 : mask,
						mcu_pwm_chip, TIM_CCER, mask);
	pwm->state.polarity = polarity;
	return 0;
}

static int gb02_pwm_enable(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm)
{
	u32 mask;
	int ch = pwm->hwpwm;
	struct mcu_pwm_chip *mcu_pwm_chip = to_mcu_pwm_chip(chip);

	/* Enable channel */
	mask = TIM_CCER_CC1E << (ch * 4);
	if (mcu_pwm_chip->have_complementary_output)
		mask |= TIM_CCER_CC1NE << (ch * 4);

	gb02_pwm_mask_write(mask, mcu_pwm_chip, TIM_CCER, mask);

	/* Make sure that registers are updated */
	gb02_pwm_mask_write(TIM_EGR_UG, mcu_pwm_chip, TIM_EGR, TIM_EGR_UG);

	/* Enable controller */
	gb02_pwm_mask_write(TIM_CR1_CEN, mcu_pwm_chip, TIM_CR1, TIM_CR1_CEN);
	pwm->state.enable = true;

	return 0;
}

static void gb02_pwm_disable(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm)
{
	u32 mask;
	int ch = pwm->hwpwm;
	struct mcu_pwm_chip *mcu_pwm_chip = to_mcu_pwm_chip(chip);
	/* Disable channel */
	mask = TIM_CCER_CC1E << (ch * 4);
	if (mcu_pwm_chip->have_complementary_output)
		mask |= TIM_CCER_CC1NE << (ch * 4);

	gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_CCER, mask);

	/* When all channels are disabled, we can disable the controller */
	if (!active_channels(mcu_pwm_chip))
		gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_CR1, TIM_CR1_CEN);

	pwm->state.enable = false;
}

static int gb02_pwm_apply(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
                                                 struct gb_pwm_state *state)
{
       bool enabled;
       int ret;

       enabled = pwm->state.enable;
       if (enabled && !state->enable) {
               gb02_pwm_disable(chip, pwm);
               return 0;
       }

       if (state->polarity != pwm->state.polarity)
               gb02_pwm_set_polarity(chip, pwm, state->polarity);

       ret = gb02_pwm_config(chip, pwm,
                                                 state->duty_cycle, state->period);
       if (ret)
               return ret;

       if (!enabled && state->enable)
               ret = gb02_pwm_enable(chip, pwm);

       return ret;
}


static struct gb_pwm_ops gb02_pwm_ops = {
	.config = gb02_pwm_config,
	.set_polarity = gb02_pwm_set_polarity,
	.enable = gb02_pwm_enable,
	.disable = gb02_pwm_disable,
	.apply = gb02_pwm_apply,
	.capture = gb02_pwm_capture,
};

static void gb02_pwm_detect_complementary(struct mcu_pwm_chip *mcu_pwm_chip)
{
	u32 ccer;

	/*
	 * If complementary bit doesn't exist writing 1 will have no
	 * effect so we can detect it.
	 */
	gb02_pwm_mask_write(TIM_CCER_CC1NE, mcu_pwm_chip, TIM_CCER,
						TIM_CCER_CC1NE);
	ccer = gb02_pwm_read(mcu_pwm_chip, TIM_CCER);
	gb02_pwm_mask_write(0, mcu_pwm_chip, TIM_CCER, TIM_CCER_CC1NE);

	mcu_pwm_chip->have_complementary_output = (ccer != 0);
}

#ifdef PWM_MISCDEVICE_DEBUG
static int my_open(struct inode *inode, struct file *filp)
{
	filp->private_data = gb02_pwm_infos;
	return 0;
}

static int my_release(struct inode *inode, struct file *filp)
{
	filp->private_data = NULL;
	return 0;
}

static long gb02_pwm_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	GB02_PWM_DEBUG("enter ioctrl!\n");

	switch (cmd) {
	default:
		break;
	}

	return ret;
}

static const struct file_operations mmap_fops = {
	.owner = THIS_MODULE,
	.open = my_open,
	.release = my_release,
	.unlocked_ioctl = gb02_pwm_ioctl,
};
static struct miscdevice gb02_pwm_misc = {
	.fops	   = &mmap_fops,
	.name	   = "mcu_pwm_chip",
	.minor	  = MISC_DYNAMIC_MINOR,
};
#endif

static void gb02_pwm_init_pre(struct mcu_pwm_chip *mcu_pwm_chip)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, mcu_pwm_chip->pci_mmio_bar4, CFG_MCU_XIP_ADDR_OFFSET);
}

static unsigned long gb02_pwm_clockrate(struct mcu_pwm_chip *mcu_pwm_chip)
{
	unsigned long periph_freq_div, periph_freq;

	periph_freq_div = base_reg_read(mcu_pwm_chip->sysctl_iofunc_base, SYSCTL_CLK_DIV_OFFSET) & 0x7;
	periph_freq = GB02_MCU_SYS_CLK * 1000 / (periph_freq_div + 1);

	return periph_freq;
}

static int gb02_pwm_init(struct mcu_pwm_chip *mcu_pwm_chip, unsigned int chip_id)
{
	int ret = 0;
	int i;
	struct gb_pwm_device *pwm;

	mutex_init(&mcu_pwm_chip->chip.lock);
	if (mcu_pwm_chip->res == NULL)
		return -EINVAL;

	//mcu_pwm_chip->max_arr = gb02_pwm_get_arr_size(mcu_pwm_chip);
	mcu_pwm_chip->max_arr = 0xffffffff;
	mcu_pwm_chip->clk_rate = gb02_pwm_clockrate(mcu_pwm_chip);
	mcu_pwm_chip->chip.chip_id = chip_id;
	GB02_PWM_DEBUG(" pwm_ctrl:%x max_arr %x clk_rate %lx\n", mcu_pwm_chip->ctrl_no, mcu_pwm_chip->max_arr, mcu_pwm_chip->clk_rate);
	gb02_pwm_detect_complementary(mcu_pwm_chip);
	mcu_pwm_chip->chip.ops = &gb02_pwm_ops;
	mcu_pwm_chip->chip.npwm = 4;
	mcu_pwm_chip->chip.pwm_type = GB_MCU_PWM;
	mcu_pwm_chip->chip.driver_data = mcu_pwm_chip;

	mcu_pwm_chip->chip.pwms = kzalloc(sizeof(struct gb_pwm_device) * mcu_pwm_chip->chip.npwm, GFP_KERNEL);
	if (!mcu_pwm_chip->chip.pwms)
		return -ENOMEM;
	for (i = 0; i < mcu_pwm_chip->chip.npwm; i++) {
		pwm = &mcu_pwm_chip->chip.pwms[i];
		pwm->chip = &mcu_pwm_chip->chip;
		pwm->hwpwm = i;
		pwm->state.polarity = PWM_POLARITY_NORMAL;
	}

	dev_dbg(mcu_pwm_chip->chip.dev, "registers %p (PWM%d)\n",
			mcu_pwm_chip->res, mcu_pwm_chip->chip.chip_id);

	return ret;
}

static int gb02_pwm_chip_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct mcu_pwm_chip *mcu_pwm_chip;
	struct gb_pcie_info *pcie_info;
	void __iomem *pcie_mmio_bar;
	int id, ret = 0;

	pcie_info = gb_dev->gb_pcie;

	id = device_id;
	if (id > MCU_PWM_PERI_NUM) {
		dev_err(gb_dev->dev, "invalid device num=%d\n", id);
		ret = -ENOENT;
		goto err_return;
	}

	mcu_pwm_chip = (struct mcu_pwm_chip *)devm_kzalloc(gb_dev->dev, sizeof(struct mcu_pwm_chip), GFP_KERNEL);
	if (mcu_pwm_chip == NULL)
		return -ENOMEM;
	mcu_pwm_chip->chip.dev = gb_dev->dev;
	/* PCIE's BAR0 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_MCU_ACCESS_BAR].mmio; //bar0
	if (!pcie_mmio_bar) {
		dev_err(gb_dev->dev, "cannot get PCIE's BAR0 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		mcu_pwm_chip->pci_mmio_bar0 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR0 : %p", __func__, __LINE__, mcu_pwm_chip->pci_mmio_bar0);
	}
	mcu_pwm_chip->res = pcie_mmio_bar + MCU_PERI_ADDR_IN_BAR0 + MCU_PWM_PERI_ADDR_BASE + id * MCU_PWM_PERI_OFFSET;

	/* PCIE's BAR1 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_SYSTEM_PERIPHERAL_BAR].mmio; //bar1
	if (!pcie_mmio_bar) {
		dev_err(gb_dev->dev, "cannot get PCIE's BAR1 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		mcu_pwm_chip->pci_mmio_bar1 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR1 : %p", __func__, __LINE__, mcu_pwm_chip->pci_mmio_bar1);
	}

	/* PCIE's BAR4 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_SYS_CFG_REG_BAR].mmio; //bar4
	if (!pcie_mmio_bar) {
		dev_err(gb_dev->dev, "cannot get PCIE's BAR4 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		mcu_pwm_chip->pci_mmio_bar4 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR4 : %p", __func__, __LINE__, mcu_pwm_chip->pci_mmio_bar4);
	}

	/* MCU's SYSCTL */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_MCU_ACCESS_BAR].mmio; //bar0
	if (!pcie_mmio_bar) {
		dev_err(gb_dev->dev, "cannot get MCU's SYSCTL info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		mcu_pwm_chip->sysctl_iofunc_base = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:SYSCTL : %p", __func__, __LINE__, mcu_pwm_chip->sysctl_iofunc_base);
	}

	gb02_pwm_init_pre(mcu_pwm_chip);
	ret = gb02_pwm_init(mcu_pwm_chip, id);
	if (ret) {
		dev_err(gb_dev->dev, "PWM init failed(%d)\n", ret);
		goto err_return;
	}

	peri_info->priv = &mcu_pwm_chip->chip;
	GB02_PWM_DEBUG("GB02 MCU's PWM initialized\n");
	if (id == FAN_PWM_CTRL_NO)
		gb02_timers_dma_probe(mcu_pwm_chip);
	return ret;
err_return:
	return ret;
}

static int gb02_pwm_chip_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb_pwm_chip *gb_pwm_chip = (struct gb_pwm_chip *)peri_info->priv;
	struct mcu_pwm_chip *mcu_pwm_chip = (struct mcu_pwm_chip *)gb_pwm_chip->driver_data;

	unsigned int i = 0;
	u32 id = gb_pwm_chip->chip_id;

	for (i = 0; i < gb_pwm_chip->npwm; i++)
		gb02_pwm_disable(gb_pwm_chip, &gb_pwm_chip->pwms[i]);

	if (id == FAN_PWM_CTRL_NO)
		gb02_timers_dma_remove(mcu_pwm_chip);

	kfree(gb_pwm_chip->pwms);
	gb_pwm_chip->pwms = NULL;

	return 0;
}

int gb02_mcu_pwm_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;
#ifdef PWM_MISCDEVICE_DEBUG
	static bool pwm_misc_inited;

	if (!pwm_misc_inited) {
		ret = misc_register(&gb02_pwm_misc);
		if (ret)
			dev_err(gb_dev->dev, "unable to register pwm misc device(%d)\n", ret);
		else
			pwm_misc_inited = true;
	}
#endif
	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_pwm_chip_init(gb_dev, peri_info, device_id);
	return ret;
}

void gb02_mcu_pwm_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
#ifdef PWM_MISCDEVICE_DEBUG
	misc_deregister(&gb02_pwm_misc);
#endif
	gb02_pwm_chip_remove(peri_info);
}
