// SPDX-License-Identifier: GPL-2.0-only
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/gpio/consumer.h>
#include <linux/hrtimer.h>
#include <linux/ktime.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/platform_device.h>
#include <linux/property.h>
#include <linux/pwm.h>
#include <linux/timekeeping.h>
#include "gb_pwm.h"
#include "mcu_peripherals/gpio/pinctrl-gb02-mcu.h"

//#define GPIO_PWM_DEBUG
#ifdef GPIO_PWM_DEBUG
#define GB02_GPIO_PWM_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_GPIO_PWM_DEVICE_NAME "] " arg)
#else
#define GB02_GPIO_PWM_DEBUG(arg...)
#endif

static inline struct gpio_pwm_chip *to_gpio_pwm_chip(struct gb_pwm_chip *c)
{
	return container_of(c, struct gpio_pwm_chip, chip);
}

static void gpio_pwm_off(struct gpio_pwm_chip *gpio_pwm_chip)
{
	gb02_gpio_set(gpio_pwm_chip->gpio, 0);
}

static void gpio_pwm_on(struct gpio_pwm_chip *gpio_pwm_chip)
{
	gb02_gpio_set(gpio_pwm_chip->gpio, 1);
}

static int gpio_pwm_enable(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm)
{
	struct gpio_pwm_chip *gpio_pwm_chip = to_gpio_pwm_chip(chip);

	if (gpio_pwm_chip->off_time)
		hrtimer_start(&gpio_pwm_chip->timer, ktime_set(0, 0), HRTIMER_MODE_REL);
	else
		gpio_pwm_on(gpio_pwm_chip);

	return 0;
}

static void gpio_pwm_disable(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm)
{
	struct gpio_pwm_chip *gpio_pwm_chip = to_gpio_pwm_chip(chip);

	if (!gpio_pwm_chip->off_time)
		gpio_pwm_off(gpio_pwm_chip);
}

static enum hrtimer_restart gpio_pwm_timer(struct hrtimer *timer)
{
	struct gpio_pwm_chip *gpio_pwm_chip = container_of(timer,
					struct gpio_pwm_chip, timer);

	if (!gpio_pwm_chip->pin_on) {
		hrtimer_forward_now(&gpio_pwm_chip->timer, ns_to_ktime(gpio_pwm_chip->on_time));

		if (gpio_pwm_chip->on_time) {
			GB02_GPIO_PWM_DEBUG("--%lld us--", ktime_to_ns(ktime_get()));
			gpio_pwm_on(gpio_pwm_chip);
			gpio_pwm_chip->pin_on = true;
		}

	} else {
		hrtimer_forward_now(&gpio_pwm_chip->timer, ns_to_ktime(gpio_pwm_chip->off_time));

		if (gpio_pwm_chip->off_time) {
			GB02_GPIO_PWM_DEBUG("***%lld us***", ktime_to_ns(ktime_get()));
			gpio_pwm_off(gpio_pwm_chip);
			gpio_pwm_chip->pin_on = false;
		}
	}

	return HRTIMER_RESTART;
}

static int gpio_pwm_config(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
						   int duty_ns, int period_ns)
{
	struct gpio_pwm_chip *gpio_pwm_chip = to_gpio_pwm_chip(chip);

	gpio_pwm_chip->on_time = duty_ns;
	gpio_pwm_chip->off_time = period_ns - duty_ns;
	GB02_GPIO_PWM_DEBUG("config duty:%d, period:%d\n", duty_ns, period_ns);
	return 0;
}

static int gpio_pwm_apply(struct gb_pwm_chip *chip, struct gb_pwm_device *pwm,
				struct gb_pwm_state *state)
{
	bool enabled;
	int ret;

	enabled = pwm->state.enable;

	if (enabled && !state->enable) {
		gpio_pwm_disable(chip, pwm);
		return 0;
	}

	ret = gpio_pwm_config(chip, pwm,
				   state->duty_cycle, state->period);
	if (ret)
		return ret;

	if (!enabled && state->enable)
		ret = gpio_pwm_enable(chip, pwm);

	return ret;
}

static struct gb_pwm_ops gpio_pwm_ops = {
	.config = gpio_pwm_config,
	.enable = gpio_pwm_enable,
	.disable = gpio_pwm_disable,
	.apply = gpio_pwm_apply,
};

static int gb_gpio_pwm_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, int device_id)
{
	int ret, i;
	struct gpio_pwm_chip *gpio_pwm_chip;
	struct gb_pwm_device *pwm;

	gpio_pwm_chip = devm_kzalloc(gb_dev->dev, sizeof(*gpio_pwm_chip), GFP_KERNEL);
	if (!gpio_pwm_chip)
		return -ENOMEM;

	gpio_pwm_chip->chip.dev = gb_dev->dev;
	gpio_pwm_chip->chip.ops = &gpio_pwm_ops;
	gpio_pwm_chip->chip.npwm = 1;
	gpio_pwm_chip->gpio = peri_info->gpio_num;
	gpio_pwm_chip->chip.pwm_type = GB_GPIO_PWM;
	gpio_pwm_chip->chip.driver_data = gpio_pwm_chip;
	ret = gb02_gpio_request(gpio_pwm_chip->gpio);
	if (ret) {
		GB02_GPIO_PWM_DEBUG("gpio%u request failed\n", gpio_pwm_chip->gpio);
		return ret;
	}
	ret = gb02_gpio_set_direction(gpio_pwm_chip->gpio, GB02_GPIO_LINE_DIRECTION_OUT);
	if (ret) {
		GB02_GPIO_PWM_DEBUG("gpio%u set_direction failed\n", gpio_pwm_chip->gpio);
		return ret;
	}
	hrtimer_init(&gpio_pwm_chip->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	gpio_pwm_chip->timer.function = &gpio_pwm_timer;
	gpio_pwm_chip->pin_on = false;

	gpio_pwm_chip->chip.pwms = kzalloc(sizeof(struct gb_pwm_device) * gpio_pwm_chip->chip.npwm, GFP_KERNEL);
	if (!gpio_pwm_chip->chip.pwms)
		return -ENOMEM;

	for (i = 0; i < gpio_pwm_chip->chip.npwm; i++) {
		pwm = &gpio_pwm_chip->chip.pwms[i];
		pwm->chip = &gpio_pwm_chip->chip;
		pwm->hwpwm = i;
		pwm->state.polarity = PWM_POLARITY_NORMAL;
	}

	if (!hrtimer_is_hres_active(&gpio_pwm_chip->timer))
		dev_warn(gb_dev->dev, "HR timer unavailable, restricting to low resolution\n");


	peri_info->priv = &gpio_pwm_chip->chip;
	return 0;
}

static void gb_gpio_pwm_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb_pwm_chip *gb_pwm_chip = (struct gb_pwm_chip *)peri_info->priv;
	struct gpio_pwm_chip *gpio_pwm_chip = (struct gpio_pwm_chip *)gb_pwm_chip->driver_data;

	gb02_gpio_free(peri_info->gpio_num);
	hrtimer_cancel(&gpio_pwm_chip->timer);

	kfree(gb_pwm_chip->pwms);
	gb_pwm_chip->pwms = NULL;
}

int gb02_mcu_gpio_pwm_drv_resume(struct gb02_mcu_peripherals_info *peri_info)
{
	unsigned gpio = peri_info->gpio_num;
	int ret = -1;

	ret = gb02_gpio_request(gpio);
	if (ret)
		GB02_GPIO_PWM_DEBUG("gpio%u request failed\n", gpio);
	ret = gb02_gpio_set_direction(gpio, GB02_GPIO_LINE_DIRECTION_OUT);
	if (ret)
		GB02_GPIO_PWM_DEBUG("gpio%u set_direction failed\n", gpio);
	return ret;
}

int gb02_mcu_gpio_pwm_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;

	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb_gpio_pwm_init(gb_dev, peri_info, device_id);

	return ret;
}

void gb02_mcu_gpio_pwm_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	gb_gpio_pwm_remove(peri_info);
}
