// SPDX-License-Identifier: GPL-2.0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <linux/io.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/cdev.h>
#include <linux/bitfield.h>
#include <linux/dmaengine.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/seq_file.h>
#include <linux/platform_device.h>
#include <linux/pwm.h>
#include "pwm-gb02-mcu.h"
#include "common/gb-peripherals-common.h"

#ifdef PWM_DEBUG
#define GB02_PWM_DEBUG(arg...)		gb_printf(KERN_INFO, "["PLATFORM_PWM_DEVICE_NAME"] " arg)
#else
#define GB02_PWM_DEBUG(arg...)
#endif

#define PWM_MCU_TEST_CMD_BASE 'W'
#define PWM_MCU_CMD_1 _IOW(PWM_MCU_TEST_CMD_BASE, 1u, int)

#define CCMR_CHANNEL_SHIFT 8
#define CCMR_CHANNEL_MASK  0xFF
#define MAX_BREAKINPUT 2

static struct gb02_mcu_pwm *gb02_pwm_infos[2];

static inline void gb02_pwm_write(unsigned int val,
				struct gb02_pwm *pwm, unsigned int reg)
{
	writel(val, pwm->regs + reg);
}

static inline unsigned long gb02_pwm_read(struct gb02_pwm *pwm,
		unsigned int reg)
{
	return readl(pwm->regs + reg);
}

static inline void gb02_pwm_mask_write(unsigned int dat,
				struct gb02_pwm *pwm, unsigned int reg, unsigned int mask)
{
	unsigned long val = 0;

	val = gb02_pwm_read(pwm, reg);
	val &= ~mask;
	val |= dat;

	gb02_pwm_write(val, pwm, reg);
}

static inline struct gb02_pwm *to_gb02_pwm_dev(struct pwm_chip *chip)
{
	return container_of(chip, struct gb02_pwm, chip);
}

static u32 active_channels(struct gb02_pwm *pwm)
{
	u32 ccer;

	ccer = gb02_pwm_read(pwm, TIM_CCER);
	return ccer & TIM_CCER_CCXE;
}

static void write_ccrx(struct gb02_pwm *pwm, int ch, u32 value)
{
	switch (ch) {
	case 0:
		gb02_pwm_write(value, pwm, TIM_CCR1);
		break;
	case 1:
		gb02_pwm_write(value, pwm, TIM_CCR2);
		break;
	case 2:
		gb02_pwm_write(value, pwm, TIM_CCR3);
		break;
	case 3:
		gb02_pwm_write(value, pwm, TIM_CCR4);
		break;
	default:
		break;
	}
}

#define TIM_CCER_CC12P (TIM_CCER_CC1P | TIM_CCER_CC2P)
#define TIM_CCER_CC12E (TIM_CCER_CC1E | TIM_CCER_CC2E)
#define TIM_CCER_CC34P (TIM_CCER_CC3P | TIM_CCER_CC4P)
#define TIM_CCER_CC34E (TIM_CCER_CC3E | TIM_CCER_CC4E)
/* DIER register DMA enable bits */
static const u32 gb02_timers_dier_dmaen[GB02_TIMERS_MAX_DMAS] = {
	TIM_DIER_CC1DE,
	TIM_DIER_CC2DE,
	TIM_DIER_CC3DE,
	TIM_DIER_CC4DE,
	TIM_DIER_UIE,
	TIM_DIER_TDE,
	TIM_DIER_COMDE
};

static void gb02_timers_dma_done(void *p)
{
	struct gb02_timers_dma *dma = p;
	struct dma_tx_state state;
	enum dma_status status;

	status = dmaengine_tx_status(dma->chan, dma->chan->cookie, &state);
	if (status == DMA_COMPLETE)
		complete(&dma->completion);
}

/**
 * gb02_timers_dma_burst_read - Read from timers registers using DMA.
 *
 * Read from GB02 timers registers using DMA on a single event.
 * @dev: reference to gb02_timers_dma MFD device
 * @buf: DMA'able destination buffer
 * @id: gb02_timers_dmas event identifier (ch[1..4], up, trig or com)
 * @reg: registers start offset for DMA to read from (like CCRx for capture)
 * @num_reg: number of registers to read upon each DMA request, starting @reg.
 * @bursts: number of bursts to read (e.g. like two for pwm period capture)
 * @tmo_ms: timeout (milliseconds)
 */
static int gb02_timers_dma_burst_read(struct gb02_pwm *priv, u32 *buf,
				enum gb02_timers_dmas id, u32 reg,
				unsigned int num_reg, unsigned int bursts,
				unsigned long tmo_ms)
{
	unsigned long timeout = msecs_to_jiffies(tmo_ms);
	struct gb02_timers_dma *dma = &priv->dma;
	struct device *dev = priv->dev;
	size_t len = num_reg * bursts * sizeof(u32);
	struct dma_async_tx_descriptor *desc;
	struct dma_slave_config config;
	dma_cookie_t cookie;
	dma_addr_t dma_buf;
	u32 dbl, dba;
	long err;
	int ret;

	/* Sanity check */
	if (id < GB02_TIMERS_DMA_CH1 || id >= GB02_TIMERS_MAX_DMAS)
		return -EINVAL;

	if (!num_reg || !bursts || reg > GB02_TIMERS_MAX_REGISTERS ||
		(reg + num_reg * sizeof(u32)) > GB02_TIMERS_MAX_REGISTERS)
		return -EINVAL;

	if (!dma->chans[id])
		return -ENODEV;
	mutex_lock(&dma->lock);

	/* Select DMA channel in use */
	dma->chan = dma->chans[id];
	dma_buf = dma_map_single(dev, buf, len, DMA_FROM_DEVICE);
	if (dma_mapping_error(dev, dma_buf)) {
		ret = -ENOMEM;
		goto unlock;
	}

	/* Prepare DMA read from timer registers, using DMA burst mode */
	memset(&config, 0, sizeof(config));
	config.src_addr = (dma_addr_t)dma->phys_base + TIM_DMAR;
	config.src_addr_width = DMA_SLAVE_BUSWIDTH_4_BYTES;
	ret = dmaengine_slave_config(dma->chan, &config);
	if (ret)
		goto unmap;

	desc = dmaengine_prep_slave_single(dma->chan, dma_buf, len,
									   DMA_DEV_TO_MEM, DMA_PREP_INTERRUPT);
	if (!desc) {
		ret = -EBUSY;
		goto unmap;
	}

	desc->callback = gb02_timers_dma_done;
	desc->callback_param = dma;
	cookie = dmaengine_submit(desc);
	ret = dma_submit_error(cookie);
	if (ret)
		goto dma_term;

	reinit_completion(&dma->completion);
	dma_async_issue_pending(dma->chan);

	/* Setup and enable timer DMA burst mode */
	dbl = FIELD_PREP(TIM_DCR_DBL, bursts - 1);
	dba = FIELD_PREP(TIM_DCR_DBA, reg >> 2);
	gb02_pwm_write(dbl | dba, priv, TIM_DCR);

	/* Clear pending flags before enabling DMA request */
	gb02_pwm_write(0, priv, TIM_SR);

	gb02_pwm_mask_write(gb02_timers_dier_dmaen[id], priv, TIM_DIER, gb02_timers_dier_dmaen[id]);

	err = wait_for_completion_interruptible_timeout(&dma->completion,
			timeout);
	if (err == 0)
		ret = -ETIMEDOUT;
	else if (err < 0)
		ret = err;

	gb02_pwm_mask_write(0, priv, TIM_DIER, gb02_timers_dier_dmaen[id]);
	gb02_pwm_write(0, priv, TIM_SR);

	gb02_pwm_write(0, priv, TIM_DCR);
dma_term:
	dmaengine_terminate_all(dma->chan);
unmap:
	dma_unmap_single(dev, dma_buf, len, DMA_FROM_DEVICE);
unlock:
	dma->chan = NULL;
	mutex_unlock(&dma->lock);

	return ret;
}

static void gb02_timers_dma_probe(struct gb02_pwm *priv)
{
	int i;
	char name[4];
	struct gb02_timers_dma *dma = &priv->dma;
	struct device *dev = priv->dev;

	init_completion(&dma->completion);
	mutex_init(&dma->lock);
	dma->phys_base = (phys_addr_t)priv->regs;
	/* Optional DMA support: get valid DMA channel(s) or NULL */
	for (i = GB02_TIMERS_DMA_CH1; i <= GB02_TIMERS_DMA_CH4; i++) {
		snprintf(name, ARRAY_SIZE(name), "ch%1d", i + 1);
		dma->chans[i] = dma_request_slave_channel(dev, name);
	}
	dma->chans[GB02_TIMERS_DMA_UP] =
		dma_request_slave_channel(dev, "up");
	dma->chans[GB02_TIMERS_DMA_TRIG] =
		dma_request_slave_channel(dev, "trig");
	dma->chans[GB02_TIMERS_DMA_COM] =
		dma_request_slave_channel(dev, "com");
}

static void gb02_timers_dma_remove(struct gb02_pwm *priv)
{
	int i;
	struct gb02_timers_dma *dma = &priv->dma;

	for (i = GB02_TIMERS_DMA_CH1; i < GB02_TIMERS_MAX_DMAS; i++)
		if (dma->chans[i])
			dma_release_channel(dma->chans[i]);
}

/*
 * Capture using PWM input mode:
 *							  ___		  ___
 * TI[1, 2, 3 or 4]: ........._|   |________|
 *							 ^0  ^1	   ^2
 *							  .   .		.
 *							  .   .		XXXXX
 *							  .   .   XXXXX	 |
 *							  .  XXXXX	 .	|
 *							XXXXX .		.	|
 * COUNTER:		______XXXXX  .   .		.	|_XXX
 *				 start^	   .   .		.		^stop
 *					  .	   .   .		.
 *					  v	   v   .		v
 *								  v
 * CCR1/CCR3:	   tx..........t0...........t2
 * CCR2/CCR4:	   tx..............t1.........
 *
 * DMA burst transfer:		  |			|
 *							  v			v
 * DMA buffer:				  { t0, tx }   { t2, t1 }
 * DMA done:								 ^
 *
 * 0: IC1/3 snapchot on rising edge: counter value -> CCR1/CCR3
 *	+ DMA transfer CCR[1/3] & CCR[2/4] values (t0, tx: doesn't care)
 * 1: IC2/4 snapchot on falling edge: counter value -> CCR2/CCR4
 * 2: IC1/3 snapchot on rising edge: counter value -> CCR1/CCR3
 *	+ DMA transfer CCR[1/3] & CCR[2/4] values (t2, t1)
 *
 * DMA done, compute:
 * - Period	 = t2 - t0
 * - Duty cycle = t1 - t0
 */

static int gb02_pwm_raw_capture(struct gb02_pwm *priv, struct pwm_device *pwm,
								unsigned long tmo_ms, u32 *raw_prd,
								u32 *raw_dty)
{
	//struct device *parent = priv->chip.dev->parent;
	enum gb02_timers_dmas dma_id;
	u32 ccen, ccr;
	int ret;

	/* Ensure registers have been updated, enable counter and capture */
	gb02_pwm_mask_write(TIM_EGR_UG, priv, TIM_EGR, TIM_EGR_UG);
	gb02_pwm_mask_write(TIM_CR1_CEN, priv, TIM_CR1, TIM_CR1_CEN);

	/* Use cc1 or cc3 DMA resp for PWM input channels 1 & 2 or 3 & 4 */
	dma_id = pwm->hwpwm < 2 ? GB02_TIMERS_DMA_CH1 : GB02_TIMERS_DMA_CH3;
	ccen = pwm->hwpwm < 2 ? TIM_CCER_CC12E : TIM_CCER_CC34E;
	ccr = pwm->hwpwm < 2 ? TIM_CCR1 : TIM_CCR3;
	gb02_pwm_mask_write(ccen, priv, TIM_CCER, ccen);

	/*
	 * Timer DMA burst mode. Request 2 registers, 2 bursts, to get both
	 * CCR1 & CCR2 (or CCR3 & CCR4) on each capture event.
	 * We'll get two capture snapchots: { CCR1, CCR2 }, { CCR1, CCR2 }
	 * or { CCR3, CCR4 }, { CCR3, CCR4 }
	 */
	ret = gb02_timers_dma_burst_read(priv, priv->capture, dma_id, ccr, 2,
									 2, tmo_ms);
	if (ret)
		goto stop;

	/* Period: t2 - t0 (take care of counter overflow) */
	if (priv->capture[0] <= priv->capture[2])
		*raw_prd = priv->capture[2] - priv->capture[0];
	else
		*raw_prd = priv->max_arr - priv->capture[0] + priv->capture[2];

	/* Duty cycle capture requires at least two capture units */
	if (pwm->chip->npwm < 2)
		*raw_dty = 0;
	else if (priv->capture[0] <= priv->capture[3])
		*raw_dty = priv->capture[3] - priv->capture[0];
	else
		*raw_dty = priv->max_arr - priv->capture[0] + priv->capture[3];

	if (*raw_dty > *raw_prd) {
		/*
		 * Race beetween PWM input and DMA: it may happen
		 * falling edge triggers new capture on TI2/4 before DMA
		 * had a chance to read CCR2/4. It means capture[1]
		 * contains period + duty_cycle. So, subtract period.
		 */
		*raw_dty -= *raw_prd;
	}

stop:
	gb02_pwm_mask_write(0, priv, TIM_CCER, ccen);
	gb02_pwm_mask_write(0, priv, TIM_CR1, TIM_CR1_CEN);

	return ret;
}

static int gb02_pwm_capture(struct pwm_chip *chip, struct pwm_device *pwm,
							struct pwm_capture *result, unsigned long tmo_ms)
{
	struct gb02_pwm *priv = to_gb02_pwm_dev(chip);
	unsigned long long prd, div, dty;
	unsigned long rate;
	unsigned int psc = 0, icpsc, scale;
	u32 raw_prd = 0, raw_dty = 0;
	int ret = 0;

	mutex_lock(&priv->lock);

	if (active_channels(priv)) {
		ret = -EBUSY;
		goto unlock;
	}

	rate = priv->clk_rate;

	/* prescaler: fit timeout window provided by upper layer */
	div = (unsigned long long)rate * (unsigned long long)tmo_ms;
	do_div(div, MSEC_PER_SEC);
	prd = div;
	while ((div > priv->max_arr) && (psc < MAX_TIM_PSC)) {
		psc++;
		div = prd;
		do_div(div, psc + 1);
	}
	gb02_pwm_write(priv->max_arr, priv, TIM_ARR);
	gb02_pwm_write(psc, priv, TIM_PSC);

	/* Map TI1 or TI2 PWM input to IC1 & IC2 (or TI3/4 to IC3 & IC4) */
	gb02_pwm_mask_write(
		TIM_CCMR_CC1S | TIM_CCMR_CC2S, priv, pwm->hwpwm & 0x1 ?
		TIM_CCMR_CC1S_TI2 | TIM_CCMR_CC2S_TI2 :
		TIM_CCMR_CC1S_TI1 | TIM_CCMR_CC2S_TI1,
		pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2);

	/* Capture period on IC1/3 rising edge, duty cycle on IC2/4 falling. */
	gb02_pwm_mask_write(
		pwm->hwpwm < 2 ?
		TIM_CCER_CC2P : TIM_CCER_CC4P, priv, TIM_CCER, pwm->hwpwm < 2 ?
		TIM_CCER_CC12P : TIM_CCER_CC34P);

	ret = gb02_pwm_raw_capture(priv, pwm, tmo_ms, &raw_prd, &raw_dty);
	if (ret)
		goto stop;

	/*
	 * Got a capture. Try to improve accuracy at high rates:
	 * - decrease counter clock prescaler, scale up to max rate.
	 * - use input prescaler, capture once every /2 /4 or /8 edges.
	 */
	if (raw_prd) {
		u32 max_arr = priv->max_arr - 0x1000; /* arbitrary margin */

		scale = max_arr / min(max_arr, raw_prd);
	} else {
		scale = priv->max_arr; /* bellow resolution, use max scale */
	}

	if (psc && scale > 1) {
		/* 2nd measure with new scale */
		psc /= scale;

		gb02_pwm_write(psc, priv, TIM_PSC);
		ret = gb02_pwm_raw_capture(priv, pwm, tmo_ms, &raw_prd,
								   &raw_dty);
		if (ret)
			goto stop;
	}

	/* Compute intermediate period not to exceed timeout at low rates */
	prd = (unsigned long long)raw_prd * (psc + 1) * NSEC_PER_SEC;
	do_div(prd, rate);

	for (icpsc = 0; icpsc < MAX_TIM_ICPSC ; icpsc++) {
		/* input prescaler: also keep arbitrary margin */
		if (raw_prd >= (priv->max_arr - 0x1000) >> (icpsc + 1))
			break;
		if (prd >= (tmo_ms * NSEC_PER_MSEC) >> (icpsc + 2))
			break;
	}

	if (!icpsc)
		goto done;

	/* Last chance to improve period accuracy, using input prescaler */
	gb02_pwm_mask_write(FIELD_PREP(TIM_CCMR_IC1PSC, icpsc) |
						FIELD_PREP(TIM_CCMR_IC2PSC, icpsc), priv,
						pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2,
						TIM_CCMR_IC1PSC | TIM_CCMR_IC2PSC
					   );

	ret = gb02_pwm_raw_capture(priv, pwm, tmo_ms, &raw_prd, &raw_dty);
	if (ret)
		goto stop;

	if (raw_dty >= (raw_prd >> icpsc)) {
		/*
		 * We may fall here using input prescaler, when input
		 * capture starts on high side (before falling edge).
		 * Example with icpsc to capture on each 4 events:
		 *
		 *	   start   1st capture					 2nd capture
		 *		 v	 v							   v
		 *		 ___   _____   _____   _____   _____   ____
		 * TI1..4	 |__|	|__|	|__|	|__|	|__|
		 *			v  v	.  .	.  .	.	   v  v
		 * icpsc1/3:  .  0	.  1	.  2	.  3	.  0
		 * icpsc2/4:  0	   1	   2	   3	   0
		 *			v  v							v  v
		 * CCR1/3  ......t0..............................t2
		 * CCR2/4  ..t1..............................t1'...
		 *			   .							.  .
		 * Capture0:	 .<----------------------------->.
		 * Capture1:	 .<-------------------------->.  .
		 *			   .							.  .
		 * Period:	   .<------>					.  .
		 * Low side:								  .<>.
		 *
		 * Result:
		 * - Period = Capture0 / icpsc
		 * - Duty = Period - Low side = Period - (Capture0 - Capture1)
		 */
		raw_dty = (raw_prd >> icpsc) - (raw_prd - raw_dty);
	}

done:
	prd = (unsigned long long)raw_prd * (psc + 1) * NSEC_PER_SEC;
	result->period = DIV_ROUND_UP_ULL(prd, rate << icpsc);
	dty = (unsigned long long)raw_dty * (psc + 1) * NSEC_PER_SEC;
	result->duty_cycle = DIV_ROUND_UP_ULL(dty, rate);
stop:
	gb02_pwm_write(0, priv, TIM_CCER);
	gb02_pwm_write(0, priv, pwm->hwpwm < 2 ? TIM_CCMR1 : TIM_CCMR2);
	gb02_pwm_write(0, priv, TIM_PSC);
unlock:
	mutex_unlock(&priv->lock);

	return ret;
}

static int gb02_pwm_config(struct gb02_pwm *priv, int ch,
						   int duty_ns, int period_ns)
{
	unsigned long long prd, div, dty;
	unsigned int prescaler = 0;
	u32 ccmr, mask, shift;

	/* Period and prescaler values depends on clock rate */
	div = (unsigned long long)(priv->clk_rate) * period_ns;

	do_div(div, NSEC_PER_SEC);
	prd = div;

	while (div > priv->max_arr) {
		prescaler++;
		div = prd;
		do_div(div, prescaler + 1);
	}

	prd = div;

	if (prescaler > MAX_TIM_PSC)
		return -EINVAL;

	/*
	 * All channels share the same prescaler and counter so when two
	 * channels are active at the same time we can't change them
	 */
	if (active_channels(priv) & ~(1 << ch * 4)) {
		u32 psc, arr;

		psc = gb02_pwm_read(priv, TIM_PSC);
		arr = gb02_pwm_read(priv, TIM_ARR);

		if ((psc != prescaler) || (arr != prd - 1))
			return -EBUSY;
	}

	gb02_pwm_write(prescaler, priv, TIM_PSC);
	gb02_pwm_write(prd - 1, priv, TIM_ARR);
	gb02_pwm_mask_write(TIM_CR1_ARPE, priv, TIM_CR1, TIM_CR1_ARPE);

	/* Calculate the duty cycles */
	dty = prd * duty_ns;
	do_div(dty, period_ns);

	write_ccrx(priv, ch, dty);

	/* Configure output mode */
	shift = (ch & 0x1) * CCMR_CHANNEL_SHIFT;
	ccmr = (TIM_CCMR_PE | TIM_CCMR_M1) << shift;
	mask = CCMR_CHANNEL_MASK << shift;

	if (ch < 2)
		gb02_pwm_mask_write(ccmr, priv, TIM_CCMR1, mask);
	else
		gb02_pwm_mask_write(ccmr, priv, TIM_CCMR2, mask);

	gb02_pwm_mask_write(TIM_BDTR_MOE | TIM_BDTR_AOE, priv,
						TIM_BDTR, TIM_BDTR_MOE | TIM_BDTR_AOE);

	return 0;
}

static int gb02_pwm_set_polarity(struct gb02_pwm *priv, int ch,
								 enum pwm_polarity polarity)
{
	u32 mask;

	mask = TIM_CCER_CC1P << (ch * 4);
	if (priv->have_complementary_output)
		mask |= TIM_CCER_CC1NP << (ch * 4);

	gb02_pwm_mask_write(polarity == PWM_POLARITY_NORMAL ? 0 : mask,
						priv, TIM_CCER, mask);

	return 0;
}

static int gb02_pwm_enable(struct gb02_pwm *priv, int ch)
{
	u32 mask;

	/* Enable channel */
	mask = TIM_CCER_CC1E << (ch * 4);
	if (priv->have_complementary_output)
		mask |= TIM_CCER_CC1NE << (ch * 4);

	gb02_pwm_mask_write(mask, priv, TIM_CCER, mask);

	/* Make sure that registers are updated */
	gb02_pwm_mask_write(TIM_EGR_UG, priv, TIM_EGR, TIM_EGR_UG);

	/* Enable controller */
	gb02_pwm_mask_write(TIM_CR1_CEN, priv, TIM_CR1, TIM_CR1_CEN);

	return 0;
}

static void gb02_pwm_disable(struct gb02_pwm *priv, int ch)
{
	u32 mask;

	/* Disable channel */
	mask = TIM_CCER_CC1E << (ch * 4);
	if (priv->have_complementary_output)
		mask |= TIM_CCER_CC1NE << (ch * 4);

	gb02_pwm_mask_write(0, priv, TIM_CCER, mask);

	/* When all channels are disabled, we can disable the controller */
	if (!active_channels(priv))
		gb02_pwm_mask_write(0, priv, TIM_CR1, TIM_CR1_CEN);
}

static int gb02_pwm_apply(struct pwm_chip *chip, struct pwm_device *pwm,
						  const struct pwm_state *state)
{
	bool enabled;
	struct gb02_pwm *priv = to_gb02_pwm_dev(chip);
	int ret;

	enabled = pwm->state.enabled;

	if (enabled && !state->enabled) {
		gb02_pwm_disable(priv, pwm->hwpwm);
		return 0;
	}

	if (state->polarity != pwm->state.polarity)
		gb02_pwm_set_polarity(priv, pwm->hwpwm, state->polarity);

	ret = gb02_pwm_config(priv, pwm->hwpwm,
						  state->duty_cycle, state->period);
	if (ret)
		return ret;

	if (!enabled && state->enabled)
		ret = gb02_pwm_enable(priv, pwm->hwpwm);

	return ret;
}
#if KERNEL_VERSION(5, 3, 18) <= LINUX_VERSION_CODE
static int gb02_pwm_apply_locked(struct pwm_chip *chip, struct pwm_device *pwm,
								 const struct pwm_state *state)
#else
static int gb02_pwm_apply_locked(struct pwm_chip *chip, struct pwm_device *pwm,
								 struct pwm_state *state)
#endif
{
	struct gb02_pwm *priv = to_gb02_pwm_dev(chip);
	int ret;

	/* protect common prescaler for all active channels */
	mutex_lock(&priv->lock);
	GB02_PWM_DEBUG("%s %d %d %d %d\n", __func__, chip->base, chip->npwm, pwm->hwpwm, pwm->pwm);
	ret = gb02_pwm_apply(chip, pwm, state);
	mutex_unlock(&priv->lock);

	return ret;
}

static void gb02_pwm_get_state(struct pwm_chip *chip,
							   struct pwm_device *pwm,
							   struct pwm_state *state)
{
	struct gb02_pwm *priv = to_gb02_pwm_dev(chip);

	mutex_lock(&priv->lock);

	//xxxxxx;
	GB02_PWM_DEBUG("%s %d %llx %llx %d\n", __func__, state->polarity, state->period, state->duty_cycle, state->enabled);
	mutex_unlock(&priv->lock);
}

static const struct pwm_ops gb02_pwm_ops = {
	.owner = THIS_MODULE,
	.get_state = gb02_pwm_get_state,
	.apply = gb02_pwm_apply_locked,
	.capture = IS_ENABLED(CONFIG_DMA_ENGINE) ? gb02_pwm_capture : NULL,
};
static void gb02_pwm_detect_complementary(struct gb02_pwm *priv)
{
	u32 ccer;

	/*
	 * If complementary bit doesn't exist writing 1 will have no
	 * effect so we can detect it.
	 */
	gb02_pwm_mask_write(TIM_CCER_CC1NE, priv, TIM_CCER,
						TIM_CCER_CC1NE);
	ccer = gb02_pwm_read(priv, TIM_CCER);
	gb02_pwm_mask_write(0, priv, TIM_CCER, TIM_CCER_CC1NE);

	priv->have_complementary_output = (ccer != 0);
}

static int gb02_pwm_detect_channels(struct gb02_pwm *priv)
{
	u32 ccer;
	int npwm = 0;

	/*
	 * If channels enable bits don't exist writing 1 will have no
	 * effect so we can detect and count them.
	 */
	gb02_pwm_mask_write(TIM_CCER_CCXE, priv, TIM_CCER,
						TIM_CCER_CCXE);
	ccer = gb02_pwm_read(priv, TIM_CCER);
	gb02_pwm_mask_write(0, priv, TIM_CCER, TIM_CCER_CCXE);

	if (ccer & TIM_CCER_CC1E)
		npwm++;

	if (ccer & TIM_CCER_CC2E)
		npwm++;

	if (ccer & TIM_CCER_CC3E)
		npwm++;

	if (ccer & TIM_CCER_CC4E)
		npwm++;

	return npwm;
}

#ifdef PWM_MISCDEVICE_DEBUG
static int my_open(struct inode *inode, struct file *filp)
{
	filp->private_data = gb02_pwm_infos;
	return 0;
}

static int my_release(struct inode *inode, struct file *filp)
{
	filp->private_data = NULL;
	return 0;
}

static long gb02_pwm_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	GB02_PWM_DEBUG("enter ioctrl!\n");

	switch (cmd) {
	default:
		break;
	}

	return ret;
}

static const struct file_operations mmap_fops = {
	.owner = THIS_MODULE,
	.open = my_open,
	.release = my_release,
	.unlocked_ioctl = gb02_pwm_ioctl,
};
static struct miscdevice gb02_pwm_misc = {
	.fops	   = &mmap_fops,
	.name	   = "gb02_pwm",
	.minor	  = MISC_DYNAMIC_MINOR,
};
#endif

static void gb02_pwm_init_pre(struct gb02_mcu_pwm *gb02_mcu_pwm)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, gb02_mcu_pwm->pci_mmio_bar4, CFG_MCU_XIP_ADDR_OFFSET);
}

static void gb02_pwm_set_mux(struct gb02_mcu_pwm *gb02_mcu_pwm)
{
	unsigned int val;

	val = base_reg_read(gb02_mcu_pwm->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
	val &= (~BIT(10));
	val &= (~BIT(9));
	GB02_PWM_DEBUG("pwm set mux %x\n", val);
	base_reg_write(val, gb02_mcu_pwm->sysctl_iofunc_base, SYSCTL_IOFUNC_OFFSET);
}

static unsigned long gb02_pwm_clockrate(struct gb02_mcu_pwm *gb02_mcu_pwm)
{
	unsigned long periph_freq_div, periph_freq;

	periph_freq_div = base_reg_read(gb02_mcu_pwm->sysctl_iofunc_base, SYSCTL_CLK_DIV_OFFSET)
					  & 0x7;
	periph_freq = GB02_MCU_SYS_CLK * 1000 / (periph_freq_div + 1);

	return periph_freq;
}

static u32 gb02_pwm_get_arr_size(struct gb02_pwm *priv)
{
	u32 val = 0;

	/*
	 * Only the available bits will be written so when readback
	 * we get the maximum value of auto reload register
	 */
	gb02_pwm_write(~0L, priv, TIM_ARR);
	val = gb02_pwm_read(priv, TIM_ARR);
	gb02_pwm_write(0x0, priv, TIM_ARR);

	return val;
}

static int gb02_pwm_init(struct gb02_mcu_pwm *gb02_mcu_pwm, u32 pwm_ctrl_no)
{
	struct gb02_pwm *priv = &(gb02_mcu_pwm->pwm);
	int ret = 0;

	mutex_init(&priv->lock);
	priv->regs = gb02_mcu_pwm->res;
	if (priv->regs == NULL) {
		dev_err(gb02_mcu_pwm->dev, "invalid address\n");
		ret = -EINVAL;
		return ret;
	}
	priv->ctrl_no = pwm_ctrl_no;
	priv->max_arr = gb02_pwm_get_arr_size(priv);
	priv->clk_rate = gb02_pwm_clockrate(gb02_mcu_pwm);
	GB02_PWM_DEBUG(" pwm_ctrl:%x max_arr %x clk_rate %lx\n", priv->ctrl_no, priv->max_arr, priv->clk_rate);
	gb02_pwm_detect_complementary(priv);

	priv->chip.base = -1;
	priv->chip.dev = gb02_mcu_pwm->dev;
	priv->chip.ops = &gb02_pwm_ops;
	priv->chip.npwm = gb02_pwm_detect_channels(priv);

	ret = pwmchip_add(&priv->chip);
	if (ret < 0)
		return ret;

	priv->driver_data = gb02_mcu_pwm;

	dev_dbg(gb02_mcu_pwm->dev, "registers %p (PWM%d)\n",
			priv->regs, priv->ctrl_no);

	return 0;
}

static int gb02_mcu_pwm_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_mcu_pwm *gb02_mcu_pwm;
	struct gb_pcie_info *pcie_info;
	void __iomem *pcie_mmio_bar;
	int id, ret = 0;

	GB02_PWM_DEBUG("Start to init GB02 MCU's PWM%d:\n", device_id);
	pcie_info = gb_dev->gb_pcie;
	id = device_id;
	gb02_mcu_pwm = devm_kzalloc(gb_dev->dev, sizeof(struct gb02_mcu_pwm), GFP_KERNEL);
	if (gb02_mcu_pwm == NULL)
		return -ENOMEM;
	gb02_mcu_pwm->dev = gb_dev->dev;
	if (id > MCU_PWM_PERI_NUM) {
		dev_err(gb02_mcu_pwm->dev, "invalid device num=%d\n", id);
		ret = -ENOENT;
		goto err_return;
	}
	gb02_mcu_pwm->pwm.dev = gb02_mcu_pwm->dev;

	/* PCIE's BAR0 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_MCU_ACCESS_BAR].mmio; //bar0
	if (!pcie_mmio_bar) {
		dev_err(gb02_mcu_pwm->dev, "cannot get PCIE's BAR0 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		gb02_mcu_pwm->pci_mmio_bar0 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR0 : %p", __func__, __LINE__, gb02_mcu_pwm->pci_mmio_bar0);
	}
	gb02_mcu_pwm->res = pcie_mmio_bar + MCU_PERI_ADDR_IN_BAR0 + MCU_PWM_PERI_ADDR_BASE + id * MCU_PWM_PERI_OFFSET;

	/* PCIE's BAR1 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_SYSTEM_PERIPHERAL_BAR].mmio; //bar1
	if (!pcie_mmio_bar) {
		dev_err(gb02_mcu_pwm->dev, "cannot get PCIE's BAR1 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		gb02_mcu_pwm->pci_mmio_bar1 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR1 : %p", __func__, __LINE__, gb02_mcu_pwm->pci_mmio_bar1);
	}

	/* PCIE's BAR4 */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_SYS_CFG_REG_BAR].mmio; //bar4
	if (!pcie_mmio_bar) {
		dev_err(gb02_mcu_pwm->dev, "cannot get PCIE's BAR4 info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		gb02_mcu_pwm->pci_mmio_bar4 = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:BAR4 : %p", __func__, __LINE__, gb02_mcu_pwm->pci_mmio_bar4);
	}

	/* MCU's SYSCTL */
	pcie_mmio_bar = gb_dev->gb_pcie->pci_bars[GB02_MCU_ACCESS_BAR].mmio; //bar0
	if (!pcie_mmio_bar) {
		dev_err(gb02_mcu_pwm->dev, "cannot get MCU's SYSCTL info\n");
		ret = -ENOMEM;
		goto err_return;
	} else {
		gb02_mcu_pwm->sysctl_iofunc_base = pcie_mmio_bar;
		GB02_PWM_DEBUG("%s:%d:SYSCTL : %p", __func__, __LINE__, gb02_mcu_pwm->sysctl_iofunc_base);
	}

	gb02_pwm_init_pre(gb02_mcu_pwm);
	gb02_pwm_set_mux(gb02_mcu_pwm);
	ret = gb02_pwm_init(gb02_mcu_pwm, id);
	if (ret) {
		dev_err(gb02_mcu_pwm->dev, "PWM init failed(%d)\n", ret);
		goto err_return;
	}

	peri_info->priv = gb02_mcu_pwm;
	gb02_pwm_infos[id] = gb02_mcu_pwm;
	GB02_PWM_DEBUG("GB02 MCU's PWM initialized\n");
	if (id == FAN_PWM_CTRL_NO)
		gb02_timers_dma_probe(&gb02_mcu_pwm->pwm);

	return ret;
err_return:
	return ret;
}

static int gb02_pwm_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_mcu_pwm *gb02_mcu_pwm = (struct gb02_mcu_pwm *)peri_info->priv;
	struct gb02_pwm *priv = &(gb02_mcu_pwm->pwm);
	unsigned int i = 0;
	u32 id = priv->ctrl_no;

	for (i = 0; i < priv->chip.npwm; i++)
		pwm_disable(&priv->chip.pwms[i]);

	pwmchip_remove(&priv->chip);
	if (id == FAN_PWM_CTRL_NO)
		gb02_timers_dma_remove(priv);

	gb02_pwm_infos[id] = NULL;
	return 0;
}

int gb02_mcu_pwm_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;
#ifdef PWM_MISCDEVICE_DEBUG
	static bool pwm_misc_inited;

	if (!pwm_misc_inited) {
		ret = misc_register(&gb02_pwm_misc);
		if (ret)
			dev_err(gb_dev->dev, "unable to register pwm misc device(%d)\n", ret);
		else
			pwm_misc_inited = true;
	}
#endif
	GB02_PWM_DEBUG("gb02_pwm_drv_init\n");
	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_mcu_pwm_init(gb_dev, peri_info, device_id);
	return ret;
}

void gb02_mcu_pwm_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
#ifdef PWM_MISCDEVICE_DEBUG
	misc_deregister(&gb02_pwm_misc);
#endif
	gb02_pwm_remove(peri_info);
}
