// SPDX-License-Identifier: GPL-2.0-or-later
//
// GB02 SPI controller driver (master mode only)
//
// Based on code from SiFive SPI Driver
//

#include <linux/clk.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/io.h>
#include <linux/log2.h>
#include "gpu/gb_device.h"
#include "spi-gb02-mcu.h"

#ifdef SPI_DEBUG
#define GB02_SPI_DEBUG(arg...) gb_printf(KERN_INFO, "[" PLATFORM_SPI_DEVICE_NAME "] " arg)
#else
#define GB02_SPI_DEBUG(arg...)
#endif

struct gb02_mcu_spi {
	void __iomem *regs;				  /* virt. address of control registers */
	void __iomem *sysctl_iofunc_base; // bar0
	void __iomem *sysctl_cfg_base;	  // bar4
	struct spi_device *spidev;
	u32 version;
	u32 feature;
	unsigned long clkrate;	 /* bus clock */
	unsigned int fifo_depth; /* fifo depth in words */
	u32 cs_inactive;		 /* level of the CS pins when inactive */
	struct completion done;	 /* wake-up from interrupt */
};

static void gb02_mcu_spi_write(struct gb02_mcu_spi *spi, int offset, u32 value)
{
	GB02_SPI_DEBUG("write offset 0x%x value 0x%08x\n", offset, value);
	iowrite32(value, spi->regs + offset);
}

static u32 gb02_mcu_spi_read(struct gb02_mcu_spi *spi, int offset)
{
	u32 tmp;

	tmp = ioread32(spi->regs + offset);
	GB02_SPI_DEBUG("read offset 0x%x value 0x%08x\n", offset, tmp);
	return tmp;
}

static void gb02_mcu_spi_init_pre(struct gb02_mcu_spi *spi)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, spi->sysctl_cfg_base, CFG_MCU_XIP_ADDR_OFFSET);
}

static unsigned long gb02_spi_clockrate(struct gb02_mcu_spi *spi)
{
	unsigned long periph_freq_div, periph_freq;

	periph_freq_div = base_reg_read(spi->sysctl_iofunc_base, SYSCTL_CLK_DIV_OFFSET) & 0x7;
	periph_freq = GB02_MCU_SYS_CLK * 1000 / (periph_freq_div + 1);
	GB02_SPI_DEBUG("------freq:%ld----------\n", periph_freq);
	return periph_freq;
}

static void gb02_mcu_spi_init(struct gb02_mcu_spi *spi)
{
	/* Watermark interrupts are disabled by default */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_IE, 0);

	if ((spi->feature & GB02_SPI_FEATURE_32B_DATA) == GB02_SPI_FEATURE_32B_DATA) {
		/* Set spi cr reg: master mode, uDMA disabled, ddr disabled, cs output enable, hdsmode disabled */
		gb02_mcu_spi_write(spi, GB02_SPI_REG_CR, 0x11);
		/* Set FORCE register to 0x1, force enable, write protect disable */
		gb02_mcu_spi_write(spi, GB02_SPI_REG_FORCE, 0x1);
	}

	/* Default watermark FIFO threshold values */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_TXMARK, 1);
	gb02_mcu_spi_write(spi, GB02_SPI_REG_RXMARK, 0);

	/* Set CS/SCK Delays and Inactive Time to defaults */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_DELAY0,
					   GB02_SPI_DELAY0_CSSCK(1) |
					   GB02_SPI_DELAY0_SCKCS(1));
	gb02_mcu_spi_write(spi, GB02_SPI_REG_DELAY1,
					   GB02_SPI_DELAY1_INTERCS(1) |
					   GB02_SPI_DELAY1_INTERXFR(0));

	/* Exit specialized memory-mapped SPI flash mode */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_FCTRL, 0);
}

static int
gb02_mcu_spi_prepare_message(struct spi_master *master, struct spi_message *msg)
{
	struct gb02_mcu_spi *spi = spi_master_get_devdata(master);
	struct spi_device *device = msg->spi;

	/* Update the chip select polarity */
	if (device->mode & SPI_CS_HIGH)
		spi->cs_inactive &= ~BIT(device->chip_select);
	else
		spi->cs_inactive |= BIT(device->chip_select);
	GB02_SPI_DEBUG("spi @0x%p device mode 0x%x, cs_inactive 0x%x, chip_select 0x%x\n",
				   spi->regs, device->mode, spi->cs_inactive, device->chip_select);
	gb02_mcu_spi_write(spi, GB02_SPI_REG_CSDEF, spi->cs_inactive);

	/* Select the correct device */
	if ((spi->feature & GB02_SPI_FEATURE_32B_DATA) == GB02_SPI_FEATURE_32B_DATA) {
		/* select device 0 using cs 1 for nuspi */
		gb02_mcu_spi_write(spi, GB02_SPI_REG_CSID, device->chip_select + 1);
	} else {
		gb02_mcu_spi_write(spi, GB02_SPI_REG_CSID, device->chip_select);
	}

	/* Set clock mode */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_SCKMODE,
					   device->mode & GB02_SPI_SCKMODE_MODE_MASK);

	return 0;
}

static void gb02_mcu_spi_set_cs(struct spi_device *device, bool is_high)
{
	struct gb02_mcu_spi *spi = spi_master_get_devdata(device->master);

	/* Reverse polarity is handled by SCMR/CPOL. Not inverted CS. */
	if (device->mode & SPI_CS_HIGH)
		is_high = !is_high;

	gb02_mcu_spi_write(spi, GB02_SPI_REG_CSMODE, is_high ? 0x03 : 0x02);
}

static int
gb02_mcu_spi_prep_transfer(struct gb02_mcu_spi *spi, struct spi_device *device,
						   struct spi_transfer *t)
{
	u32 cr;
	unsigned long rate;
	unsigned int mode;

	rate = min_t(unsigned long, spi->clkrate >> 1, t->speed_hz);
	/* Calculate and program the clock rate */
	// cr =0x07;
	cr = DIV_ROUND_UP(spi->clkrate >> 1, rate) - 1;
	GB02_SPI_DEBUG("----------cr:0x%x clkrate: %ld speed %d rate %ld-----------\n", cr, spi->clkrate, t->speed_hz, rate);

	cr = 0x07;
	cr &= GB02_SPI_SCKDIV_DIV_MASK;

	gb02_mcu_spi_write(spi, GB02_SPI_REG_SCKDIV, cr);

	mode = max_t(unsigned int, t->rx_nbits, t->tx_nbits);

	/* Set frame format */
	cr = GB02_SPI_FMT_LEN(t->bits_per_word);
	switch (mode) {
	case SPI_NBITS_QUAD:
		cr |= GB02_SPI_FMT_PROTO_QUAD;
		break;
	case SPI_NBITS_DUAL:
		cr |= GB02_SPI_FMT_PROTO_DUAL;
		break;
	default:
		cr |= GB02_SPI_FMT_PROTO_SINGLE;
		break;
	}
	if (device->mode & SPI_LSB_FIRST)
		cr |= GB02_SPI_FMT_ENDIAN;
	if (!t->rx_buf)
		cr |= GB02_SPI_FMT_DIR;
	gb02_mcu_spi_write(spi, GB02_SPI_REG_FMT, cr);

	/* We will want to poll if the time we need to wait is
	 * less than the context switching time.
	 * Let's call that threshold 5us. The operation will take:
	 *	(8/mode) * fifo_depth / hz <= 5 * 10^-6
	 *	1600000 * fifo_depth <= hz * mode
	 */

	return 1600000 * spi->fifo_depth <= t->speed_hz * mode;
}

static void gb02_mcu_spi_prope_feature(struct gb02_mcu_spi *spi)
{
	u32 data;

	data = gb02_mcu_spi_read(spi, GB02_SPI_REG_VERSION);
	spi->version = data;
	if (data >= GB02_SPI_VERSION_110)
		spi->feature |= GB02_SPI_FEATURE_32B_DATA;
	else
		spi->feature &= ~GB02_SPI_FEATURE_32B_DATA;
}
#ifdef GB02_SPI_USE_IRQ
static irqreturn_t gb02_mcu_spi_irq(int irq, void *dev_id)
{
	struct gb02_mcu_spi *spi = dev_id;
	u32 ip = gb02_mcu_spi_read(spi, GB02_SPI_REG_IP);

	if (ip & (GB02_SPI_IP_TXWM | GB02_SPI_IP_RXWM)) {
		/* Disable interrupts until next transfer */
		gb02_mcu_spi_write(spi, GB02_SPI_REG_IE, 0);
		complete(&spi->done);
		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}
#endif
static void gb02_mcu_spi_wait(struct gb02_mcu_spi *spi, u32 bit, int poll)
{
	GB02_SPI_DEBUG("spi @0x%p wait poll %d , bits %x start\n",
				   spi->regs, poll, bit);
	if (poll) {
		u32 cr;
		u32 tmp = 10;

		do {
			cr = gb02_mcu_spi_read(spi, GB02_SPI_REG_IP);
			tmp--;
		} while ((!(cr & bit)) & tmp);
		/* wait for busy flag is clear */
		if ((spi->feature & GB02_SPI_FEATURE_32B_DATA) != 0) {
			while ((gb02_mcu_spi_read(spi, GB02_SPI_REG_STATUS) & GB02_SPI_STATUS_BUSY_FLAG) != 0)
				;
		}
	} else {
		reinit_completion(&spi->done);
		gb02_mcu_spi_write(spi, GB02_SPI_REG_IE, bit);
		wait_for_completion(&spi->done);
	}
	GB02_SPI_DEBUG("spi @0x%p wait poll %d end\n",
				   spi->regs, poll);
}

static void gb02_mcu_spi_tx(struct gb02_mcu_spi *spi, const u8 *tx_ptr)
{
	if ((spi->feature & GB02_SPI_FEATURE_32B_DATA) == 0) {
		WARN_ON_ONCE((gb02_mcu_spi_read(spi, GB02_SPI_REG_TXDATA) & GB02_SPI_TXDATA_FULL) != 0);
		gb02_mcu_spi_write(spi, GB02_SPI_REG_TXDATA,
						   *tx_ptr & GB02_SPI_TXDATA_DATA_MASK);
	} else {
		WARN_ON_ONCE((gb02_mcu_spi_read(spi, GB02_SPI_REG_STATUS) & GB02_SPI_STATUS_TX_FULL) != 0);
		gb02_mcu_spi_write(spi, GB02_SPI_REG_TXDATA,
						   *tx_ptr & GB02_SPI_TXDATA_DATA_MASK);
	}
}

static void gb02_mcu_spi_rx(struct gb02_mcu_spi *spi, u8 *rx_ptr)
{
	u32 data;

	if ((spi->feature & GB02_SPI_FEATURE_32B_DATA) == 0) {
		data = gb02_mcu_spi_read(spi, GB02_SPI_REG_RXDATA);

		WARN_ON_ONCE((data & GB02_SPI_RXDATA_EMPTY) != 0);
		*rx_ptr = data & GB02_SPI_RXDATA_DATA_MASK;
	} else {
		u32 status = gb02_mcu_spi_read(spi, GB02_SPI_REG_STATUS);
		// TODO fix in newer IP, which can use spi_status rx empty flag
		WARN_ON_ONCE((status & GB02_SPI_STATUS_RX_EMPTY) != 0);
		data = gb02_mcu_spi_read(spi, GB02_SPI_REG_RXDATA);
		*rx_ptr = data & GB02_SPI_RXDATA_DATA_MASK;
	}
}

static int
gb02_mcu_spi_transfer_one(struct spi_master *master, struct spi_device *device,
						  struct spi_transfer *t)
{
	struct gb02_mcu_spi *spi = spi_master_get_devdata(master);
	int poll = gb02_mcu_spi_prep_transfer(spi, device, t);
	const u8 *tx_ptr = t->tx_buf;
	u8 *rx_ptr = t->rx_buf;
	u8 *rx_ptr2;
	unsigned int remaining_words = t->len;
	int j;

	GB02_SPI_DEBUG("spi @0x%p %s %d, %d\n", spi->regs, __func__, remaining_words, spi->fifo_depth);
	GB02_SPI_DEBUG("tx ");

	for (j = 0; j < remaining_words; j++)
		GB02_SPI_DEBUG("0x%x ", tx_ptr[j]);
	GB02_SPI_DEBUG("\n");
	while (remaining_words) {
		unsigned int n_words = min(remaining_words, spi->fifo_depth);
		unsigned int i;

		/* Enqueue n_words for transmission */
		for (i = 0; i < n_words; i++)
			gb02_mcu_spi_tx(spi, tx_ptr++);

		if (rx_ptr) {
			GB02_SPI_DEBUG("rx_ptr ");
			/* Wait for transmission + reception to complete */
			gb02_mcu_spi_write(spi, GB02_SPI_REG_RXMARK,
							   n_words - 1);
			gb02_mcu_spi_wait(spi, GB02_SPI_IP_RXWM, poll);

			/* Read out all the data from the RX FIFO */
			GB02_SPI_DEBUG("rx ");
			for (i = 0; i < n_words; i++) {
				rx_ptr2 = rx_ptr;
				gb02_mcu_spi_rx(spi, rx_ptr++);
				GB02_SPI_DEBUG("0x%x ", rx_ptr2[0]);
			}
			GB02_SPI_DEBUG("\n");
		} else {
			GB02_SPI_DEBUG("else ");
			/* Wait for transmission to complete */
			gb02_mcu_spi_wait(spi, GB02_SPI_IP_TXWM, poll);
		}

		remaining_words -= n_words;
	}
	GB02_SPI_DEBUG("spi @0x%p %s end\n", spi->regs, __func__);

	return 0;
}
static struct spi_board_info board_info = {
	.modalias = "spidev",
	.bus_num = 0,
	.chip_select = 0,
	.max_speed_hz = 0,
};
/*TODO add spidev board_info in controler*/
static int gb02_spi_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_mcu_spi *spi;
	int num_cs;
	u32 cs_bits;
	// u32 cs_bits, max_bits_per_word=8;//Only 8bit SPI words supported by the driver
	struct spi_master *master;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id, peri_base_bar_id;
	unsigned long clkrate;
	int id, ret = 0;

	pcie_info = gb_dev->gb_pcie;
	id = device_id;
	master = spi_alloc_master(gb_dev->dev, sizeof(struct gb02_mcu_spi));
	if (!master) {
		dev_err(gb_dev->dev, "out of memory\n");
		return -ENOMEM;
	}

	spi = spi_master_get_devdata(master);

	init_completion(&spi->done);
	peri_info->priv = spi;
	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	peri_base_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);
	spi->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	spi->sysctl_cfg_base = gb_dev->gb_pcie->pci_bars[peri_base_bar_id].mmio; //bar4
	spi->regs = spi->sysctl_iofunc_base + MCU_PERI_ADDR_IN_BAR0 + MCU_SPI_PERI_ADDR_BASE; // get from platfrom_devdata
	GB02_SPI_DEBUG("mapped at %p\n", spi->regs);
	gb02_mcu_spi_init_pre(spi);
	clkrate = gb02_spi_clockrate(spi);
	spi->clkrate = clkrate;
	spi->fifo_depth = GB02_SPI_DEFAULT_DEPTH;
	/* probe the number of CS lines */
	spi->cs_inactive = gb02_mcu_spi_read(spi, GB02_SPI_REG_CSDEF);
	gb02_mcu_spi_write(spi, GB02_SPI_REG_CSDEF, 0xffffffffU);
	cs_bits = gb02_mcu_spi_read(spi, GB02_SPI_REG_CSDEF);
	gb02_mcu_spi_write(spi, GB02_SPI_REG_CSDEF, spi->cs_inactive);
	if (!cs_bits) {
		dev_err(gb_dev->dev, "Could not auto probe CS lines\n");
		ret = -EINVAL;
		goto err_out;
	}

	num_cs = ilog2(cs_bits) + 1;
	if (num_cs > GB02_SPI_MAX_CS) {
		dev_err(gb_dev->dev, "Invalid number of spi slaves\n");
		ret = -EINVAL;
		goto err_out;
	}

	/* Define our master */
	master->bus_num = -1; // non static bus_num
	master->num_chipselect = num_cs;
	master->mode_bits = SPI_CPHA | SPI_CPOL | SPI_CS_HIGH | SPI_LSB_FIRST | SPI_TX_DUAL | SPI_TX_QUAD | SPI_RX_DUAL | SPI_RX_QUAD;
	/* TODO: add driver support for bits_per_word < 8
	 * we need to "left-align" the bits (unless SPI_LSB_FIRST)
	 */
	master->bits_per_word_mask = SPI_BPW_MASK(8);
	master->flags = SPI_CONTROLLER_MUST_TX | SPI_MASTER_GPIO_SS;
	master->prepare_message = gb02_mcu_spi_prepare_message;
	master->set_cs = gb02_mcu_spi_set_cs;
	master->transfer_one = gb02_mcu_spi_transfer_one;

	/* probe gb02_mcu spi features */
	gb02_mcu_spi_prope_feature(spi);
	/* Configure the SPI master hardware */
	gb02_mcu_spi_init(spi);
	ret = devm_spi_register_master(gb_dev->dev, master);
	if (ret < 0) {
		dev_err(gb_dev->dev, "spi_register_master failed\n");
		goto err_out;
	}
	board_info.bus_num = master->bus_num;
	board_info.max_speed_hz = clkrate;
	spi->spidev = spi_new_device(master, &board_info);

	return 0;

err_out:
	dev_err(gb_dev->dev, "err_out failed\n");
	spi_master_put(master);

	return ret;
}

static int gb02_mcu_spi_remove(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_mcu_spi *spi = (struct gb02_mcu_spi *)peri_info->priv;

	/* Disable all the interrupts just in case */
	gb02_mcu_spi_write(spi, GB02_SPI_REG_IE, 0);
	spi_unregister_device(spi->spidev);
	return 0;
}

int gb02_mcu_spi_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	int ret;
	int device_id = 0;

	GB02_SPI_DEBUG("gb02_pci_spi_init\n");

	device_id = peri_info->mcu_peripherals_device_id;
	ret = gb02_spi_init(gb_dev, peri_info, device_id);
	return ret;
}

void gb02_mcu_spi_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	gb02_mcu_spi_remove(peri_info);
	GB02_SPI_DEBUG("gb02_pci_spi_exit!\n");
}
