/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __SPI_GB02_MCU_H
#define __SPI_GB02_MCU_H
#include "gb-peripherals-common.h"
#define GB02_SPI_DRIVER_NAME "gb02_mcu_spi"

#define GB02_SPI_MAX_CS 32
#define GB02_SPI_DEFAULT_DEPTH 4
#define GB02_SPI_DEFAULT_MAX_BITS 8

/* register offsets */
#define GB02_SPI_REG_SCKDIV 0x00    /* Serial clock divisor */
#define GB02_SPI_REG_SCKMODE 0x04   /* Serial clock mode */
#define GB02_SPI_REG_SCKSAMPLE 0x08 /* SPI data sampling divisor */
#define GB02_SPI_REG_FORCE 0x0C     /* SPI oe ctrl when not use other pad */
#define GB02_SPI_REG_CSID 0x10      /* Chip select ID */
#define GB02_SPI_REG_CSDEF 0x14     /* Chip select default */
#define GB02_SPI_REG_CSMODE 0x18    /* Chip select mode */
#define GB02_SPI_REG_VERSION 0x1C   /* SPI version */
#define GB02_SPI_REG_DELAY0 0x28    /* Delay control 0 */
#define GB02_SPI_REG_DELAY1 0x2c    /* Delay control 1 */
#define GB02_SPI_REG_FMT 0x40       /* Frame format */
#define GB02_SPI_REG_TXDATA 0x48    /* Tx FIFO data */
#define GB02_SPI_REG_RXDATA 0x4c    /* Rx FIFO data */
#define GB02_SPI_REG_TXMARK 0x50    /* Tx FIFO watermark */
#define GB02_SPI_REG_RXMARK 0x54    /* Rx FIFO watermark */
#define GB02_SPI_REG_FCTRL 0x60     /* SPI flash interface control */
#define GB02_SPI_REG_FFMT 0x64      /* SPI flash instruction format */
#define GB02_SPI_REG_IE 0x70        /* Interrupt Enable Register */
#define GB02_SPI_REG_IP 0x74        /* Interrupt Pendings Register */
#define GB02_SPI_REG_FFMT1 0x78     /* SPI flash instruction format 1 */
#define GB02_SPI_REG_STATUS 0x7C    /* SPI busy status */
#define GB02_SPI_REG_RXEDGE 0x80    /* SPI RX sample edge ctrl */
#define GB02_SPI_REG_CR 0x84        /* SPI control register */

/* spi version tags */
#define GB02_SPI_VERSION_110 0x00010100

/* sckdiv bits */
#define GB02_SPI_SCKDIV_DIV_MASK 0xfffU

/* sckmode bits */
#define GB02_SPI_SCKMODE_PHA BIT(0)
#define GB02_SPI_SCKMODE_POL BIT(1)
#define GB02_SPI_SCKMODE_MODE_MASK (GB02_SPI_SCKMODE_PHA | GB02_SPI_SCKMODE_POL)

/* csmode bits */
#define GB02_SPI_CSMODE_MODE_AUTO 0U
#define GB02_SPI_CSMODE_MODE_HOLD 2U
#define GB02_SPI_CSMODE_MODE_OFF 3U

/* delay0 bits */
#define GB02_SPI_DELAY0_CSSCK(x) ((u32)(x))
#define GB02_SPI_DELAY0_CSSCK_MASK 0xffU
#define GB02_SPI_DELAY0_SCKCS(x) ((u32)(x) << 16)
#define GB02_SPI_DELAY0_SCKCS_MASK (0xffU << 16)

/* delay1 bits */
#define GB02_SPI_DELAY1_INTERCS(x) ((u32)(x))
#define GB02_SPI_DELAY1_INTERCS_MASK 0xffU
#define GB02_SPI_DELAY1_INTERXFR(x) ((u32)(x) << 16)
#define GB02_SPI_DELAY1_INTERXFR_MASK (0xffU << 16)

/* fmt bits */
#define GB02_SPI_FMT_PROTO_SINGLE 0U
#define GB02_SPI_FMT_PROTO_DUAL 1U
#define GB02_SPI_FMT_PROTO_QUAD 2U
#define GB02_SPI_FMT_PROTO_MASK 3U
#define GB02_SPI_FMT_ENDIAN BIT(2)
#define GB02_SPI_FMT_DIR BIT(3)
#define GB02_SPI_FMT_LEN(x) ((u32)(x) << 16)
#define GB02_SPI_FMT_LEN_MASK (0xfU << 16)

/* txdata bits */
#define GB02_SPI_TXDATA_DATA_MASK 0xffU
#define GB02_SPI_TXDATA_FULL BIT(31)

/* rxdata bits */
#define GB02_SPI_RXDATA_DATA_MASK 0xffU
#define GB02_SPI_RXDATA_EMPTY BIT(31)

/* ie and ip bits */
#define GB02_SPI_IP_TXWM BIT(0)
#define GB02_SPI_IP_RXWM BIT(1)

/* status bits */
#define GB02_SPI_STATUS_BUSY_FLAG BIT(0)
#define GB02_SPI_STATUS_OVERRUN BIT(2)
#define GB02_SPI_STATUS_UNDERRUN BIT(3)
#define GB02_SPI_STATUS_TX_FULL BIT(4)
#define GB02_SPI_STATUS_RX_EMPTY BIT(5)

#define GB02_SPI_FEATURE_32B_DATA BIT(0)

#ifdef CONFIG_MCU_SPI
int gb02_mcu_spi_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info);
void gb02_mcu_spi_drv_exit(struct gb02_mcu_peripherals_info *peri_info);
void  spidev_exit(void);
int spidev_init(void);
#else
static inline int gb02_mcu_spi_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	return 0;
}
static inline void gb02_mcu_spi_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	return;
}
static inline void  spidev_exit(void)
{
	return;
}
static inline int spidev_init(void)
{
	return 0;
}
#endif
#endif /* __SPI_GB02_MCU_H */
