// SPDX-License-Identifier: GPL-2.0

#include <linux/clk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_irq.h>
#include <linux/platform_device.h>
#include <linux/serial_core.h>
#include <linux/serial_reg.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include "gpu/gb_device.h"
#include "gb02_usart.h"

#define PORT_GB02 200

/*
 * Register offsets
 */

/* TXDATA */
#define GB02_SERIAL_TXDATA_OFFS		0x0
#define GB02_SERIAL_TXDATA_FULL_SHIFT		31
#define GB02_SERIAL_TXDATA_FULL_MASK		(1 << GB02_SERIAL_TXDATA_FULL_SHIFT)
#define GB02_SERIAL_TXDATA_DATA_SHIFT		0
#define GB02_SERIAL_TXDATA_DATA_MASK		(0xff << GB02_SERIAL_TXDATA_DATA_SHIFT)

/* RXDATA */
#define GB02_SERIAL_RXDATA_OFFS		0x4
#define GB02_SERIAL_RXDATA_EMPTY_SHIFT	31
#define GB02_SERIAL_RXDATA_EMPTY_MASK		(1 << GB02_SERIAL_RXDATA_EMPTY_SHIFT)
#define GB02_SERIAL_RXDATA_DATA_SHIFT		0
#define GB02_SERIAL_RXDATA_DATA_MASK		(0xff << GB02_SERIAL_RXDATA_DATA_SHIFT)

/* TXCTRL */
#define GB02_SERIAL_TXCTRL_OFFS		0x8
#define GB02_SERIAL_TXCTRL_TXCNT_SHIFT	16
#define GB02_SERIAL_TXCTRL_TXCNT_MASK		(0x7 << GB02_SERIAL_TXCTRL_TXCNT_SHIFT)
#define GB02_SERIAL_TXCTRL_NSTOP_SHIFT	1
#define GB02_SERIAL_TXCTRL_NSTOP_MASK		(1 << GB02_SERIAL_TXCTRL_NSTOP_SHIFT)
#define GB02_SERIAL_TXCTRL_TXEN_SHIFT		0
#define GB02_SERIAL_TXCTRL_TXEN_MASK		(1 << GB02_SERIAL_TXCTRL_TXEN_SHIFT)

/* RXCTRL */
#define GB02_SERIAL_RXCTRL_OFFS		0xC
#define GB02_SERIAL_RXCTRL_RXCNT_SHIFT	16
#define GB02_SERIAL_RXCTRL_RXCNT_MASK		(0x7 << GB02_SERIAL_TXCTRL_TXCNT_SHIFT)
#define GB02_SERIAL_RXCTRL_RXEN_SHIFT		0
#define GB02_SERIAL_RXCTRL_RXEN_MASK		(1 << GB02_SERIAL_RXCTRL_RXEN_SHIFT)

/* IE */
#define GB02_SERIAL_IE_OFFS			0x10
#define GB02_SERIAL_IE_RXWM_SHIFT		1
#define GB02_SERIAL_IE_RXWM_MASK		(1 << GB02_SERIAL_IE_RXWM_SHIFT)
#define GB02_SERIAL_IE_TXWM_SHIFT		0
#define GB02_SERIAL_IE_TXWM_MASK		(1 << GB02_SERIAL_IE_TXWM_SHIFT)

/* IP */
#define GB02_SERIAL_IP_OFFS			0x14
#define GB02_SERIAL_IP_RXWM_SHIFT		1
#define GB02_SERIAL_IP_RXWM_MASK		(1 << GB02_SERIAL_IP_RXWM_SHIFT)
#define GB02_SERIAL_IP_TXWM_SHIFT		0
#define GB02_SERIAL_IP_TXWM_MASK		(1 << GB02_SERIAL_IP_TXWM_SHIFT)

/* DIV */
#define GB02_SERIAL_DIV_OFFS			0x18
#define GB02_SERIAL_DIV_DIV_SHIFT		0
#define GB02_SERIAL_DIV_DIV_MASK		(0xffff << GB02_SERIAL_IP_DIV_SHIFT)

/* SETUP */
#define GB02_SERIAL_SETUP_OFFS		0x1c

/*
 * Config macros
 */

/*
 * GB02_SERIAL_MAX_PORTS: maximum number of UARTs on a device that can
 *						  host a serial console
 */
#define GB02_SERIAL_MAX_PORTS			2

/*
 * GB02_DEFAULT_BAUD_RATE: default baud rate that the driver should
 *						   configure itself to use
 */
#define GB02_DEFAULT_BAUD_RATE		115200

/* GB02_SERIAL_NAME: our driver's name that we pass to the operating system */
#define GB02_SERIAL_NAME			"gb02-mcu-uart"

/* GB02_TTY_PREFIX: tty name prefix for GB02 serial ports */
#define GB02_TTY_PREFIX			"ttyGB"

/* GB02_TX_FIFO_DEPTH: depth of the TX FIFO (in bytes) */
#define GB02_TX_FIFO_DEPTH			8

/* GB02_RX_FIFO_DEPTH: depth of the TX FIFO (in bytes) */
#define GB02_RX_FIFO_DEPTH			8

#if (GB02_TX_FIFO_DEPTH != GB02_RX_FIFO_DEPTH)
#error Driver does not support configurations with different TX, RX FIFO sizes
#endif

/*
 *
 */

/**
 * gb02_serial_port - driver-specific data extension to struct uart_port
 * @port: struct uart_port embedded in this struct
 * @dev: struct device *
 * @ier: shadowed copy of the interrupt enable register
 * @clkin_rate: input clock to the UART IP block.
 * @baud_rate: UART serial line rate (e.g., 115200 baud)
 * @clk_notifier: clock rate change notifier for upstream clock changes
 *
 * Configuration data specific to this GB02 UART.
 */
struct gb02_serial_port {
	void __iomem *sysctl_iofunc_base; // bar0
	void __iomem *sysctl_cfg_base;	  // bar4
	struct uart_port	port;
	struct device		*dev;
	unsigned char		ier;
	unsigned long		clkin_rate;
	unsigned long		baud_rate;
	bool task_run;
	struct task_struct *uart_task;
};

/*
 * Structure container-of macros
 */

#define port_to_gb02_serial_port(p) (container_of((p), \
							struct gb02_serial_port, \
							port))

/*
 * Internal functions
 */
static void gb02_serial_stop_tx(struct uart_port *port);

/**
 * gsp_early_writel() - write to a GB02 serial port register (early)
 * @port: pointer to a struct uart_port record
 * @offs: register address offset from the IP block base address
 * @v: value to write to the register
 *
 * Given a pointer @port to a struct uart_port record, write the value
 * @v to the IP block register address offset @offs.  This function is
 * intended for early console use.
 *
 * Context: Intended to be used only by the earlyconsole code.
 */
static void gsp_early_writel(u32 v, u16 offs, struct uart_port *port)
{
	writel_relaxed(v, port->membase + offs);
}

/**
 * gsp_early_readl() - read from a GB02 serial port register (early)
 * @port: pointer to a struct uart_port record
 * @offs: register address offset from the IP block base address
 *
 * Given a pointer @port to a struct uart_port record, read the
 * contents of the IP block register located at offset @offs from the
 * IP block base and return it.  This function is intended for early
 * console use.
 *
 * Context: Intended to be called only by the earlyconsole code or by
 *		  gsp_readl() or gsp_writel() (in this driver)
 *
 * Returns: the register value read from the UART.
 */
static u32 gsp_early_readl(struct uart_port *port, u16 offs)
{
	return readl_relaxed(port->membase + offs);
}

/**
 * gsp_writel() - write to a GB02 serial port register
 * @v: value to write to the register
 * @offs: register address offset from the IP block base address
 * @gsp: pointer to a struct gb02_serial_port record
 *
 * Write the value @v to the IP block register located at offset @offs from the
 * IP block base, given a pointer @gsp to a struct gb02_serial_port record.
 *
 * Context: Any context.
 */
static void gsp_writel(u32 v, u16 offs, struct gb02_serial_port *gsp)
{
	gsp_early_writel(v, offs, &gsp->port);
}

/**
 * gsp_readl() - read from a GB02 serial port register
 * @gsp: pointer to a struct gb02_serial_port record
 * @offs: register address offset from the IP block base address
 *
 * Read the contents of the IP block register located at offset @offs from the
 * IP block base, given a pointer @gsp to a struct gb02_serial_port record.
 *
 * Context: Any context.
 *
 * Returns: the value of the UART register
 */
static u32 gsp_readl(struct gb02_serial_port *gsp, u16 offs)
{
	return gsp_early_readl(&gsp->port, offs);
}

/**
 * gb02_serial_is_txfifo_full() - is the TXFIFO full?
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Read the transmit FIFO "full" bit, returning a non-zero value if the
 * TX FIFO is full, or zero if space remains.  Intended to be used to prevent
 * writes to the TX FIFO when it's full.
 *
 * Returns: GB02_SERIAL_TXDATA_FULL_MASK (non-zero) if the transmit FIFO
 * is full, or 0 if space remains.
 */
static int gb02_serial_is_txfifo_full(struct gb02_serial_port *gsp)
{
	return gsp_readl(gsp, GB02_SERIAL_TXDATA_OFFS) &
		   GB02_SERIAL_TXDATA_FULL_MASK;
}

/**
 * gsp_transmit_char() - enqueue a byte to transmit onto the TX FIFO
 * @gsp: pointer to a struct gb02_serial_port
 * @ch: character to transmit
 *
 * Enqueue a byte @ch onto the transmit FIFO, given a pointer @gsp to the
 * struct gb02_serial_port * to transmit on.  Caller should first check to
 * ensure that the TXFIFO has space; see gb02_serial_is_txfifo_full().
 *
 * Context: Any context.
 */
static void gsp_transmit_char(struct gb02_serial_port *gsp, int ch)
{
	gsp_writel(ch, GB02_SERIAL_TXDATA_OFFS, gsp);
}

/**
 * gsp_transmit_chars() - enqueue multiple bytes onto the TX FIFO
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Transfer up to a TX FIFO size's worth of characters from the Linux serial
 * transmit buffer to the GB02 UART TX FIFO.
 *
 * Context: Any context.  Expects @gsp->port.lock to be held by caller.
 */
static void gsp_transmit_chars(struct gb02_serial_port *gsp)
{
	struct circ_buf *xmit = &gsp->port.state->xmit;
	int count;

	if (gsp->port.x_char) {
		gsp_transmit_char(gsp, gsp->port.x_char);
		gsp->port.icount.tx++;
		gsp->port.x_char = 0;
		return;
	}
	if (uart_circ_empty(xmit) || uart_tx_stopped(&gsp->port)) {
		gb02_serial_stop_tx(&gsp->port);
		return;
	}
	count = GB02_TX_FIFO_DEPTH;
	do {
		gsp_transmit_char(gsp, xmit->buf[xmit->tail]);
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		gsp->port.icount.tx++;
		if (uart_circ_empty(xmit))
			break;
	} while (--count > 0);

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&gsp->port);

	if (uart_circ_empty(xmit))
		gb02_serial_stop_tx(&gsp->port);
}

/**
 * gsp_enable_txwm() - enable transmit watermark interrupts
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Enable interrupt generation when the transmit FIFO watermark is reached
 * on the GB02 UART referred to by @gsp.
 */
static void gsp_enable_txwm(struct gb02_serial_port *gsp)
{
	if (gsp->ier & GB02_SERIAL_IE_TXWM_MASK)
		return;

	gsp->ier |= GB02_SERIAL_IE_TXWM_MASK;
	gsp_writel(gsp->ier, GB02_SERIAL_IE_OFFS, gsp);
}

/**
 * gsp_enable_rxwm() - enable receive watermark interrupts
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Enable interrupt generation when the receive FIFO watermark is reached
 * on the GB02 UART referred to by @gsp.
 */
static void gsp_enable_rxwm(struct gb02_serial_port *gsp)
{
	if (gsp->ier & GB02_SERIAL_IE_RXWM_MASK)
		return;

	gsp->ier |= GB02_SERIAL_IE_RXWM_MASK;
	gsp_writel(gsp->ier, GB02_SERIAL_IE_OFFS, gsp);
}

/**
 * gsp_disable_txwm() - disable transmit watermark interrupts
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Disable interrupt generation when the transmit FIFO watermark is reached
 * on the UART referred to by @gsp.
 */
static void gsp_disable_txwm(struct gb02_serial_port *gsp)
{
	if (!(gsp->ier & GB02_SERIAL_IE_TXWM_MASK))
		return;

	gsp->ier &= ~GB02_SERIAL_IE_TXWM_MASK;
	gsp_writel(gsp->ier, GB02_SERIAL_IE_OFFS, gsp);
}

/**
 * gsp_disable_rxwm() - disable receive watermark interrupts
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Disable interrupt generation when the receive FIFO watermark is reached
 * on the UART referred to by @gsp.
 */
static void gsp_disable_rxwm(struct gb02_serial_port *gsp)
{
	if (!(gsp->ier & GB02_SERIAL_IE_RXWM_MASK))
		return;

	gsp->ier &= ~GB02_SERIAL_IE_RXWM_MASK;
	gsp_writel(gsp->ier, GB02_SERIAL_IE_OFFS, gsp);
}

/**
 * gsp_receive_char() - receive a byte from the UART
 * @gsp: pointer to a struct gb02_serial_port
 * @is_empty: char pointer to return whether the RX FIFO is empty
 *
 * Try to read a byte from the GB02 UART RX FIFO, referenced by
 * @gsp, and to return it.  Also returns the RX FIFO empty bit in
 * the char pointed to by @ch.  The caller must pass the byte back to the
 * Linux serial layer if needed.
 *
 * Returns: the byte read from the UART RX FIFO.
 */
static char gsp_receive_char(struct gb02_serial_port *gsp, char *is_empty)
{
	u32 v;
	u8 ch;

	v = gsp_readl(gsp, GB02_SERIAL_RXDATA_OFFS);

	if (!is_empty)
		WARN_ON(1);
	else
		*is_empty = (v & GB02_SERIAL_RXDATA_EMPTY_MASK) >>
					GB02_SERIAL_RXDATA_EMPTY_SHIFT;

	ch = (v & GB02_SERIAL_RXDATA_DATA_MASK) >>
		 GB02_SERIAL_RXDATA_DATA_SHIFT;

	return ch;
}

/**
 * gsp_receive_chars() - receive multiple bytes from the UART
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Receive up to an RX FIFO's worth of bytes from the GB02 UART referred
 * to by @gsp and pass them up to the Linux serial layer.
 *
 * Context: Expects gsp->port.lock to be held by caller.
 */
static void gsp_receive_chars(struct gb02_serial_port *gsp)
{
	unsigned char ch;
	char is_empty;
	int c;

	for (c = GB02_RX_FIFO_DEPTH; c > 0; --c) {
		ch = gsp_receive_char(gsp, &is_empty);
		if (is_empty)
			break;

		gsp->port.icount.rx++;
		uart_insert_char(&gsp->port, 0, 0, ch, TTY_NORMAL);
	}

	//spin_unlock(&gsp->port.lock);
	tty_flip_buffer_push(&gsp->port.state->port);
	//spin_lock(&gsp->port.lock);
}

/**
 * gsp_update_div() - calculate the divisor setting by the line rate
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Calculate the appropriate value of the clock divisor for the UART
 * and target line rate referred to by @gsp and write it into the
 * hardware.
 */
static void gsp_update_div(struct gb02_serial_port *gsp)
{
	u16 div;

	div = DIV_ROUND_UP(gsp->clkin_rate, gsp->baud_rate) - 1;

	gsp_writel(div, GB02_SERIAL_DIV_OFFS, gsp);
}

/**
 * gsp_update_baud_rate() - set the UART "baud rate"
 * @gsp: pointer to a struct gb02_serial_port
 * @rate: new target bit rate
 *
 * Calculate the UART divisor value for the target bit rate @rate for the
 * GB02 UART described by @gsp and program it into the UART.  There may
 * be some error between the target bit rate and the actual bit rate implemented
 * by the UART due to clock ratio granularity.
 */
static void gsp_update_baud_rate(struct gb02_serial_port *gsp,
								 unsigned int rate)
{
	if (gsp->baud_rate == rate)
		return;

	gsp->baud_rate = rate;
	gsp_update_div(gsp);
}

/**
 * gsp_set_stop_bits() - set the number of stop bits
 * @gsp: pointer to a struct gb02_serial_port
 * @nstop: 1 or 2 (stop bits)
 *
 * Program the GB02 UART referred to by @gsp to use @nstop stop bits.
 */
static void gsp_set_stop_bits(struct gb02_serial_port *gsp, char nstop)
{
	u32 v;

	if (nstop < 1 || nstop > 2) {
		WARN_ON(1);
		return;
	}

	v = gsp_readl(gsp, GB02_SERIAL_TXCTRL_OFFS);
	v &= ~GB02_SERIAL_TXCTRL_NSTOP_MASK;
	v |= (nstop - 1) << GB02_SERIAL_TXCTRL_NSTOP_SHIFT;
	gsp_writel(v, GB02_SERIAL_TXCTRL_OFFS, gsp);
}

/**
 * gsp_wait_for_xmitr() - wait for an empty slot on the TX FIFO
 * @gsp: pointer to a struct gb02_serial_port
 *
 * Delay while the UART TX FIFO referred to by @gsp is marked as full.
 *
 * Context: Any context.
 */
static void __maybe_unused gsp_wait_for_xmitr(struct gb02_serial_port *gsp)
{
	while (gb02_serial_is_txfifo_full(gsp))
		udelay(1); /* XXX Could probably be more intelligent here */
}

/*
 * Linux serial API functions
 */

static void gb02_serial_stop_tx(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_disable_txwm(gsp);
}

static void gb02_serial_stop_rx(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_disable_rxwm(gsp);
}

static void gb02_serial_start_tx(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_enable_txwm(gsp);
	if (!gsp->task_run) {
		wake_up_process(gsp->uart_task);
		gsp->task_run = true;
	}
}

static int gb02_serial_irq(void *dev_id)
{
	struct gb02_serial_port *gsp = dev_id;
	u32 ip, rx_num;

	while (!kthread_should_stop()) {
		ip = gsp_readl(gsp, GB02_SERIAL_IP_OFFS);
		rx_num = gsp_readl(gsp, 0x4c);
		if (ip & GB02_SERIAL_IP_RXWM_MASK)
			gsp_receive_chars(gsp);
		if (ip & GB02_SERIAL_IP_TXWM_MASK)
			gsp_transmit_chars(gsp);
		if (!gsp->task_run) {
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule();
		}
	}

	return 0;
}

static unsigned int gb02_serial_tx_empty(struct uart_port *port)
{
	return TIOCSER_TEMT;
}

static unsigned int gb02_serial_get_mctrl(struct uart_port *port)
{
	return TIOCM_CAR | TIOCM_CTS | TIOCM_DSR;
}

static void gb02_serial_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
	/* IP block does not support these signals */
}

static void gb02_serial_break_ctl(struct uart_port *port, int break_state)
{
	/* IP block does not support sending a break */
}

static int gb02_serial_startup(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_enable_rxwm(gsp);
	if (!gsp->task_run) {
		wake_up_process(gsp->uart_task);
		gsp->task_run = true;
	}

	return 0;
}

static void gb02_serial_shutdown(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_disable_rxwm(gsp);
	gsp_disable_txwm(gsp);
	gsp->task_run = false;
}

static void gb02_serial_set_termios(struct uart_port *port,
									struct ktermios *termios,
									struct ktermios *old)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);
	unsigned long flags;
	u32 v, old_v;
	int rate;
	char nstop;

	if ((termios->c_cflag & CSIZE) != CS8)
		dev_err_once(gsp->port.dev, "only 8-bit words supported\n");
	if (termios->c_iflag & (INPCK | PARMRK))
		dev_err_once(gsp->port.dev, "parity checking not supported\n");
	if (termios->c_iflag & BRKINT)
		dev_err_once(gsp->port.dev, "BREAK detection not supported\n");

	/* Set number of stop bits */
	nstop = (termios->c_cflag & CSTOPB) ? 2 : 1;
	gsp_set_stop_bits(gsp, nstop);

	/* Set line rate */
	rate = uart_get_baud_rate(port, termios, old, 0, gsp->clkin_rate / 16);
	gsp_update_baud_rate(gsp, rate);

	spin_lock_irqsave(&gsp->port.lock, flags);

	/* Update the per-port timeout */
	uart_update_timeout(port, termios->c_cflag, rate);

	gsp->port.read_status_mask = 0;

	/* Ignore all characters if CREAD is not set */
	v = gsp_readl(gsp, GB02_SERIAL_RXCTRL_OFFS);
	old_v = v;
	if ((termios->c_cflag & CREAD) == 0)
		v &= GB02_SERIAL_RXCTRL_RXEN_MASK;
	else
		v |= GB02_SERIAL_RXCTRL_RXEN_MASK;
	if (v != old_v)
		gsp_writel(v, GB02_SERIAL_RXCTRL_OFFS, gsp);

	spin_unlock_irqrestore(&gsp->port.lock, flags);
}

static void gb02_serial_release_port(struct uart_port *port)
{
}

static int gb02_serial_request_port(struct uart_port *port)
{
	return 0;
}

static void gb02_serial_config_port(struct uart_port *port, int flags)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp->port.type = PORT_GB02;
}

static int gb02_serial_verify_port(struct uart_port *port,
								   struct serial_struct *ser)
{
	return -EINVAL;
}

static const char *gb02_serial_type(struct uart_port *port)
{
	return port->type == PORT_GB02 ? "GB02 UART/USART" : NULL;
}

#ifdef CONFIG_CONSOLE_POLL
static int gb02_serial_poll_get_char(struct uart_port *port)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);
	char is_empty, ch;

	ch = gsp_receive_char(gsp, &is_empty);
	if (is_empty)
		return NO_POLL_CHAR;

	return ch;
}

static void gb02_serial_poll_put_char(struct uart_port *port,
									  unsigned char c)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_wait_for_xmitr(gsp);
	gsp_transmit_char(gsp, c);
}
#endif /* CONFIG_CONSOLE_POLL */

/*
 * Early console support
 */

#ifdef CONFIG_SERIAL_EARLYCON
static void early_gb02_serial_putc(struct uart_port *port, int c)
{
	while (gsp_early_readl(port, GB02_SERIAL_TXDATA_OFFS) &
		   GB02_SERIAL_TXDATA_FULL_MASK)
		cpu_relax();

	gsp_early_writel(c, GB02_SERIAL_TXDATA_OFFS, port);
}

static void early_gb02_serial_write(struct console *con, const char *s,
									unsigned int n)
{
	struct earlycon_device *dev = con->data;
	struct uart_port *port = &dev->port;

	uart_console_write(port, s, n, early_gb02_serial_putc);
}

static int __init early_gb02_serial_setup(struct earlycon_device *dev,
		const char *options)
{
	struct uart_port *port = &dev->port;

	if (!port->membase)
		return -ENODEV;

	dev->con->write = early_gb02_serial_write;

	return 0;
}

OF_EARLYCON_DECLARE(gb02, "gb02,uart0", early_gb02_serial_setup);
#endif /* CONFIG_SERIAL_EARLYCON */

/*
 * Linux console interface
 */

#ifdef CONFIG_SERIAL_GB02_CONSOLE

static struct gb02_serial_port *gb02_serial_console_ports[GB02_SERIAL_MAX_PORTS];

static void gb02_serial_console_putchar(struct uart_port *port, int ch)
{
	struct gb02_serial_port *gsp = port_to_gb02_serial_port(port);

	gsp_wait_for_xmitr(gsp);
	gsp_transmit_char(gsp, ch);
}

static void gb02_serial_console_write(struct console *co, const char *s,
									  unsigned int count)
{
	struct gb02_serial_port *gsp = gb02_serial_console_ports[co->index];
	unsigned long flags;
	unsigned int ier;
	int locked = 1;

	if (!gsp)
		return;

	local_irq_save(flags);
	if (gsp->port.sysrq)
		locked = 0;
	else if (oops_in_progress)
		locked = spin_trylock(&gsp->port.lock);
	else
		spin_lock(&gsp->port.lock);

	ier = gsp_readl(gsp, GB02_SERIAL_IE_OFFS);
	gsp_writel(0, GB02_SERIAL_IE_OFFS, gsp);

	uart_console_write(&gsp->port, s, count, gb02_serial_console_putchar);

	gsp_writel(ier, GB02_SERIAL_IE_OFFS, gsp);

	if (locked)
		spin_unlock(&gsp->port.lock);
	local_irq_restore(flags);
}

static int __init gb02_serial_console_setup(struct console *co, char *options)
{
	struct gb02_serial_port *gsp;
	int baud = GB02_DEFAULT_BAUD_RATE;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	if (co->index < 0 || co->index >= GB02_SERIAL_MAX_PORTS)
		return -ENODEV;

	gsp = gb02_serial_console_ports[co->index];
	if (!gsp)
		return -ENODEV;

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);

	return uart_set_options(&gsp->port, co, baud, parity, bits, flow);
}

static struct uart_driver gb02_serial_uart_driver;

static struct console gb02_serial_console = {
	.name		= GB02_TTY_PREFIX,
	.write		= gb02_serial_console_write,
	.device		= uart_console_device,
	.setup		= gb02_serial_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &gb02_serial_uart_driver,
};

static int __init gb02_console_init(void)
{
	register_console(&gb02_serial_console);
	return 0;
}

console_initcall(gb02_console_init);

static void gsp_add_console_port(struct gb02_serial_port *gsp)
{
	gb02_serial_console_ports[gsp->port.line] = gsp;
}

static void gsp_remove_console_port(struct gb02_serial_port *gsp)
{
	gb02_serial_console_ports[gsp->port.line] = 0;
}

#define GB02_SERIAL_CONSOLE	(&gb02_serial_console)

#else

#define GB02_SERIAL_CONSOLE	NULL

static void gsp_add_console_port(struct gb02_serial_port *gsp)
{}
static void gsp_remove_console_port(struct gb02_serial_port *gsp)
{}

#endif

static unsigned long gb02_uart_clockrate(struct gb02_serial_port *gsp)
{
	unsigned long periph_freq_div, periph_freq;

	periph_freq_div = base_reg_read(gsp->sysctl_iofunc_base, SYSCTL_CLK_DIV_OFFSET) & 0x7;
	periph_freq = GB02_MCU_SYS_CLK * 1000 / (periph_freq_div + 1);
	return periph_freq;
}

static void gb02_uart_init_pre(struct gb02_serial_port *gsp)
{
	base_reg_write(MCU_SYSCTRL_BASE_ADDR, gsp->sysctl_cfg_base, CFG_MCU_XIP_ADDR_OFFSET);
}

static const struct uart_ops gb02_serial_uops = {
	.tx_empty	= gb02_serial_tx_empty,
	.set_mctrl	= gb02_serial_set_mctrl,
	.get_mctrl	= gb02_serial_get_mctrl,
	.stop_tx	= gb02_serial_stop_tx,
	.start_tx	= gb02_serial_start_tx,
	.stop_rx	= gb02_serial_stop_rx,
	.break_ctl	= gb02_serial_break_ctl,
	.startup	= gb02_serial_startup,
	.shutdown	= gb02_serial_shutdown,
	.set_termios	= gb02_serial_set_termios,
	.type		= gb02_serial_type,
	.release_port	= gb02_serial_release_port,
	.request_port	= gb02_serial_request_port,
	.config_port	= gb02_serial_config_port,
	.verify_port	= gb02_serial_verify_port,
#ifdef CONFIG_CONSOLE_POLL
	.poll_get_char	= gb02_serial_poll_get_char,
	.poll_put_char	= gb02_serial_poll_put_char,
#endif
};

static struct uart_driver gb02_serial_uart_driver = {
	.owner		= THIS_MODULE,
	.driver_name	= GB02_SERIAL_NAME,
	.dev_name	= GB02_TTY_PREFIX,
	.nr		= GB02_SERIAL_MAX_PORTS,
	.cons		= GB02_SERIAL_CONSOLE,
};

static int gb02_serial_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info, unsigned int device_id)
{
	struct gb02_serial_port *gsp;
	struct gb_pcie_info *pcie_info;
	int mcu_peri_bar_id, peri_base_bar_id;
	int id, ret = 0;

	pcie_info = gb_dev->gb_pcie;
	id = device_id;
	if (id > GB02_SERIAL_MAX_PORTS) {
		dev_err(gb_dev->dev, "too many UARTs (%d)\n", id);
		return -EINVAL;
	}

	gsp = devm_kzalloc(gb_dev->dev, sizeof(*gsp), GFP_KERNEL);
	if (!gsp)
		return -ENOMEM;
	mcu_peri_bar_id = get_genbu_mcu_peri_bar_id(pcie_info->gb_dev_info);
	peri_base_bar_id = get_genbu_peri_base_bar_id(pcie_info->gb_dev_info);
	gsp->sysctl_iofunc_base = gb_dev->gb_pcie->pci_bars[mcu_peri_bar_id].mmio; // bar0
	gsp->sysctl_cfg_base = gb_dev->gb_pcie->pci_bars[peri_base_bar_id].mmio; //bar4
	gb02_uart_init_pre(gsp);
	gsp->port.dev = gb_dev->dev;
	gsp->port.type = PORT_GB02;
	gsp->port.iotype = UPIO_MEM;
	gsp->port.fifosize = GB02_TX_FIFO_DEPTH;
	gsp->port.ops = &gb02_serial_uops;
	gsp->port.line = id;
	gsp->port.membase = gsp->sysctl_iofunc_base + MCU_PERI_ADDR_IN_BAR0 + MCU_UART0_PERI_ADDR_BASE + MCU_UART_PERI_ADDR_OFST * id;
	gsp->dev = gb_dev->dev;
	gsp->task_run = false;

	gsp->uart_task = kthread_create(gb02_serial_irq, gsp, "gb02_uart%d", id);
	if (IS_ERR(gsp->uart_task)) {
		dev_err(gb_dev->dev, "gb02 uart_task creat failed\n");
		goto init_out;
	}

	/* Set up clock divider */
	gsp->clkin_rate = gb02_uart_clockrate(gsp);
	gsp->baud_rate = GB02_DEFAULT_BAUD_RATE;
	gsp->port.uartclk = gsp->clkin_rate;
	peri_info->priv = gsp;
	gsp_update_div(gsp);

	/* Enable transmits and set the watermark level to 1 */
	gsp_writel((1 << GB02_SERIAL_TXCTRL_TXCNT_SHIFT) |
			   GB02_SERIAL_TXCTRL_TXEN_MASK,
			   GB02_SERIAL_TXCTRL_OFFS, gsp);

	/* Enable receives and set the watermark level to 0 */
	gsp_writel((0 << GB02_SERIAL_RXCTRL_RXCNT_SHIFT) |
			   GB02_SERIAL_RXCTRL_RXEN_MASK,
			   GB02_SERIAL_RXCTRL_OFFS, gsp);

	/* TODO: adjust using driver data for UART_SETUP reg */
	/* No partity check, 8 bit len, cts/rts disable, dma disable */
	gsp_writel(0x3<<4, GB02_SERIAL_SETUP_OFFS, gsp);

	gsp_add_console_port(gsp);

	ret = uart_add_one_port(&gb02_serial_uart_driver, &gsp->port);
	if (ret != 0) {
		dev_err(gb_dev->dev, "could not add uart: %d\n", ret);
		goto init_out;
	}
	wake_up_process(gsp->uart_task);//TODO

	return 0;

init_out:
	gsp_remove_console_port(gsp);
	return ret;
}

static void gb02_serial_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	struct gb02_serial_port *gsp = (struct gb02_serial_port *)peri_info->priv;

	kthread_stop(gsp->uart_task);
	gsp_remove_console_port(gsp);
	uart_remove_one_port(&gb02_serial_uart_driver, &gsp->port);
	uart_unregister_driver(&gb02_serial_uart_driver);
}

int gb02_mcu_uart_drv_init(struct gb_device *gb_dev, struct gb02_mcu_peripherals_info *peri_info)
{
	static u8 uart_complete;
	int device_id = 0;
	int ret;

	device_id = peri_info->mcu_peripherals_device_id;
	//uart driver only register once
	if (!uart_complete) {
		ret = uart_register_driver(&gb02_serial_uart_driver);
		if (ret)
			return ret;
	}

	uart_complete = 1;
	ret = gb02_serial_init(gb_dev, peri_info, device_id);

	return ret;
}

void gb02_mcu_uart_drv_exit(struct gb02_mcu_peripherals_info *peri_info)
{
	uart_unregister_driver(&gb02_serial_uart_driver);
	gb02_serial_exit(peri_info);
}
