/* SPDX-License-Identifier: GPL-3.0-or-later */
/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * jessy 20221019 Sietium
 */

#ifndef __VPU_COMM_H__
#define __VPU_COMM_H__
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/spinlock.h>
#include <linux/dma-mapping.h>

#include "vpu_list.h"
#include "vpu/vpu_comm/gb_vpu.h"
#include "vpu_vcmd_registers.h"
#include "vpu_basetype.h"

enum vcmd_module_type {
	VCMD_TYPE_ENCODER = 0,
	VCMD_TYPE_CUTREE,
	VCMD_TYPE_DECODER,
	VCMD_TYPE_JPEG_ENCODER,
	VCMD_TYPE_JPEG_DECODER,
	MAX_VCMD_TYPE
};

#define EXECUTING_CMDBUF_ID_ADDR    26
#define VCMD_REGISTER_SIZE			(128 * 4)
#define WORKING_STATE_IDLE          0
#define WORKING_STATE_WORKING       1
#define CMDBUF_EXE_STATUS_OK        0
#define CMDBUF_EXE_STATUS_CMDERR    1
#define CMDBUF_EXE_STATUS_BUSERR    2
#define CMDBUF_PRIORITY_NORMAL      0
#define CMDBUF_PRIORITY_HIGH        1
#define MAX_VCMD_NUMBER				(MAX_VCMD_TYPE * 4)

#define VCMD_HW_ID                  0x0012
#define INT_MIN_SUM_OF_IMAGE_SIZE   (4096 * 2160 * 4 * 1)
#define VCMD_EXE_CMDBUF_COUNT       3
/* variables declaration related with race condition */
#define CMDBUF_MAX_SIZE			(512 * 4 * 4)

/* approximately=128x(320x240)=128x2k=128x8kbyte=1Mbytes */
#define CMDBUF_POOL_TOTAL_SIZE		(2 * 1024 * 1024)
#define TOTAL_DISCRETE_CMDBUF_NUM (CMDBUF_POOL_TOTAL_SIZE / CMDBUF_MAX_SIZE / 2)
#define CMDBUF_VCMD_REGISTER_TOTAL_SIZE \
	(9 * 1024 * 1024 - CMDBUF_POOL_TOTAL_SIZE * 2)

#define PROCESS_MAX_VIDEO_SIZE (4096 * 2160 * 4 * MAX_PROCESS_CORE_NUMBER)
#define PROCESS_MAX_JPEG_SIZE           (2147483648U) // 32K*32K*2
#define PROCESS_MAX_SUM_OF_IMAGE_SIZE\
	(PROCESS_MAX_VIDEO_SIZE > PROCESS_MAX_JPEG_SIZE ?\
	PROCESS_MAX_VIDEO_SIZE:PROCESS_MAX_JPEG_SIZE)

#ifdef EMU
#define MTLB_PCIE_START_ADDRESS  0x04100000
#define PAGE_PCIE_START_ADDRESS  0x04200000
#define STLB_PCIE_START_ADDRESS  0x04300000
#else
#define MTLB_PCIE_START_ADDRESS  0x00100000
#define PAGE_PCIE_START_ADDRESS  0x00200000
#define STLB_PCIE_START_ADDRESS  0x00300000
#endif

#define MMU_REG_OFFSET              0
#define MMU_REG_HW_ID               (MMU_REG_OFFSET + 6 * 4)
#define MMU_REG_FLUSH               (MMU_REG_OFFSET + 97 * 4)
#define MMU_REG_PAGE_TABLE_ID       (MMU_REG_OFFSET + 107 * 4)
#define MMU_REG_CONTROL             (MMU_REG_OFFSET + 226 * 4)
#define MMU_REG_ADDRESS             (MMU_REG_OFFSET + 227 * 4)
#define MMU_REG_ADDRESS_MSB         (MMU_REG_OFFSET + 228 * 4)

/* VCMD */
#define OPCODE_WREG					(0x01 << 27)
#define OPCODE_END					(0x02 << 27)
#define OPCODE_NOP					(0x03 << 27)
#define OPCODE_RREG					(0x16 << 27)
#define OPCODE_INT					(0x18 << 27)
#define OPCODE_JMP					(0x19 << 27)
#define OPCODE_STALL				(0x09 << 27)
#define OPCODE_CLRINT				(0x1a << 27)
#define OPCODE_JMP_RDY0				(0x19 << 27)
#define OPCODE_JMP_RDY1				((0x19 << 27) | (1 << 26))
#define JMP_IE_1					(1 << 25)
#define JMP_RDY_1					(1 << 26)

#define HW_ID_1_0_C					0x00120fff
#define HW_ID_1_1_2					0x43421102

#define MAX_REG_COUNT 768
/* these size need to be modified according to hw config */
#define VCMD_ENCODER_REGISTER_SIZE              (479 * 4)
#define VCMD_DECODER_REGISTER_SIZE              (MAX_REG_COUNT * 4)
#define VCMD_IM_REGISTER_SIZE                   (479 * 4)
#define VCMD_JPEG_ENCODER_REGISTER_SIZE         (479 * 4)
#define VCMD_JPEG_DECODER_REGISTER_SIZE         (MAX_REG_COUNT * 4)

#define DWL_CLIENT_TYPE_H264_DEC        1U
#define DWL_CLIENT_TYPE_MPEG4_DEC       2U
#define DWL_CLIENT_TYPE_JPEG_DEC        3U
#define DWL_CLIENT_TYPE_PP              4U
#define DWL_CLIENT_TYPE_VC1_DEC         5U
#define DWL_CLIENT_TYPE_MPEG2_DEC       6U
#define DWL_CLIENT_TYPE_VP6_DEC         7U
#define DWL_CLIENT_TYPE_AVS_DEC         8U
#define DWL_CLIENT_TYPE_RV_DEC          9U
#define DWL_CLIENT_TYPE_VP8_DEC         10U
#define DWL_CLIENT_TYPE_VP9_DEC         11U
#define DWL_CLIENT_TYPE_HEVC_DEC        12U
#define DWL_CLIENT_TYPE_ST_PP           14U
#define DWL_CLIENT_TYPE_H264_MAIN10     15U
#define DWL_CLIENT_TYPE_AVS2_DEC        16U
#define DWL_CLIENT_TYPE_AV1_DEC         17U
#define DWL_CLIENT_TYPE_BO_AV1_DEC      31U

/* hantro G2 reg config */
#define HANTRO_G2_DEC_REGS              337 /* G2 total regs */
#define HANTRO_G1_DEC_REGS              155 /* G1 total regs */
#define HANTRO_VC8000D_REGS             MAX_REG_COUNT /* VC8000D total regs */

/* Logic module IRQs */
#define HXDEC_NO_IRQ                    -1
#define MAX(a, b)                       (((a) > (b)) ? (a) : (b))
#define DEC_IO_SIZE_MAX\
	(MAX(MAX(HANTRO_G2_DEC_REGS, HANTRO_G1_DEC_REGS),\
		HANTRO_VC8000D_REGS) * 4)

/* Base address DDR register */
#define PCI_DDR_BAR 0

/* Base address got control register */
#define PCI_CONTROL_BAR                 4

/* PCIe hantro driver offset in control register */
#define HANTRO_REG_OFFSET0               0x600000
#define HANTRO_REG_OFFSET1               0x700000

#define DEC_IO_SIZE_0                   DEC_IO_SIZE_MAX /* bytes */
#define DEC_IO_SIZE_1                   DEC_IO_SIZE_MAX /* bytes */

#define DEC_IRQ_0                       HXDEC_NO_IRQ
#define DEC_IRQ_1                       HXDEC_NO_IRQ
#define IS_G1(hw_id)                    (((hw_id) == 0x6731) ? 1 : 0)
#define IS_G2(hw_id)                    (((hw_id) == 0x6732) ? 1 : 0)
#define IS_VC8000D(hw_id)               (((hw_id) == 0x8001) ? 1 : 0)
#define IS_BIGOCEAN(hw_id)              (((hw_id) == 0xB16D) ? 1 : 0)

#define VC8000D_NUM_MASK_REG            336
#define VC8000D_NUM_MODE                4
#define VC8000D_MASK_REG_OFFSET         4096
#define VC8000D_MASK_BITS_PER_REG       1

#define AXIFE_NUM_MASK_REG              144
#define AXIFE_NUM_MODE                  1
#define AXIFE_MASK_REG_OFFSET           4096
#define AXIFE_MASK_BITS_PER_REG         1

#define CORE_TYPE_STR_CASE(ct)          case (ct): return(#ct + 3)

#define ANY_CMDBUF_ID 0xFFFF


/* Use 'v' as magic number for vcmd */
#define HANTRO_VCMD_IOC_MAGIC  'v'
/*
 * S means "Set" through a ptr,
 * T means "Tell" directly with the argument value
 * G means "Get": reply by setting through a pointer
 * Q means "Query": response is on the return value
 * X means "eXchange": G and S atomically
 * H means "sHift": T and Q atomically
 */
#define HANTRO_VCMD_IOCH_GET_CMDBUF_PARAMETER\
	_IOWR(HANTRO_VCMD_IOC_MAGIC, 20, struct cmdbuf_mem_parameter *)
#define HANTRO_VCMD_IOCH_GET_HWINFO_FROM_VCMD\
	_IOWR(HANTRO_VCMD_IOC_MAGIC, 21, unsigned long)
#define HANTRO_VCMD_IOCH_SET_CMDBUF_POOL_BASE\
	_IOWR(HANTRO_VCMD_IOC_MAGIC, 22, unsigned long)
#define HANTRO_VCMD_IOCH_GET_VCMD_PARAMETER\
	_IOWR(HANTRO_VCMD_IOC_MAGIC, 24, struct config_parameter *)

#define HANTRO_VCMD_IOCH_LINK_RUN_CMDBUF _IOR(HANTRO_VCMD_IOC_MAGIC, 26, u16 *)
#define HANTRO_VCMD_IOCH_WAIT_CMDBUF    _IOR(HANTRO_VCMD_IOC_MAGIC, 27, u16 *)
#define HANTRO_VCMD_IOCH_RELEASE_CMDBUF _IOR(HANTRO_VCMD_IOC_MAGIC, 28, u16 *)

#define HANTRO_VCMD_IOCH_POLLING_CMDBUF _IOR(HANTRO_VCMD_IOC_MAGIC, 40, u16 *)

#define HANTRO_VCMD_IOC_MAXNR		50

/* priority support */
#define MAX_CMDBUF_PRIORITY_TYPE    2 // 0:normal priority,1:high priority

// video decoder vcmd configuration
#define VCMD_DEC_IO_ADDR_0					0x600000
#define VCMD_DEC_IO_SIZE_0                  (ASIC_VCMD_SWREG_AMOUNT * 4)
#define VCMD_DEC_INT_PIN_0                  -1
#define VCMD_DEC_MODULE_TYPE_0              2
#define VCMD_DEC_MODULE_MAIN_ADDR_0         0x1000
#define VCMD_DEC_MODULE_DEC400_ADDR_0       0XFFFF
#define VCMD_DEC_MODULE_L2CACHE_ADDR_0      0X2000
#define VCMD_DEC_MODULE_MMU_ADDR_0          0XFFFF

#define VCMD_DEC_IO_ADDR_1                  0x700000
#define VCMD_DEC_IO_SIZE_1                  (ASIC_VCMD_SWREG_AMOUNT * 4)
#define VCMD_DEC_INT_PIN_1                  -1
#define VCMD_DEC_MODULE_TYPE_1              2
#define VCMD_DEC_MODULE_MAIN_ADDR_1         0x1000
#define VCMD_DEC_MODULE_DEC400_ADDR_1       0XFFFF
#define VCMD_DEC_MODULE_L2CACHE_ADDR_1      0X2000
#define VCMD_DEC_MODULE_MMU_ADDR_1          0XFFFF

#define MAX_PROCESS_CORE_NUMBER		(4 * 8)

struct noncache_mem {
	u32 *virtual_address;
	dma_addr_t  bus_address;
	size_t mmu_bus_address;  /* buffer physical address in MMU */
	u32 size;
	u16 cmdbuf_id;
};

struct exchange_parameter {
	u64 executing_time;
	u16 module_type; /* input vc8000e=0,IM=1,vc8000d=2,jpege=3,jpegd=4 */
	u16 cmdbuf_size; /* input, reserve not used; link and run is input */
	u16 priority;    /* input,normal=0, high/live=1 */
	u16 cmdbuf_id;   /* output, it is unique in driver */
	u16 core_id;     /* just used for polling */
};
struct cmdbuf_mem_parameter {
	u32 *virt_cmdbuf_addr;
	ptr_t phy_cmdbuf_addr;
	u32 mmu_phy_cmdbuf_addr;
	u32 cmdbuf_total_size;
	u16 cmdbuf_unit_size;
	u32 *virt_status_cmdbuf_addr;
	ptr_t phy_status_cmdbuf_addr;
	u32 mmu_phy_status_cmdbuf_addr;
	u32 status_cmdbuf_total_size;
	u16 status_cmdbuf_unit_size;
	ptr_t base_ddr_addr;
};


struct process_manager_obj {
	struct drm_file *filp;
	u64 total_exe_time;
	spinlock_t spinlock;
	wait_queue_head_t wait_queue;
};

struct cmdbuf_obj {
	u32 module_type; /* input vc8000e=0,IM=1,vc8000d=2,jpege=3, jpegd=4 */
	u32 priority;    /* current CMDBUFpriority: normal=0, high=1 */
	u64 executing_time;
	u32 cmdbuf_size;
	u32 *cmdbuf_virtual_address;
	size_t cmdbuf_bus_address;
	size_t mmu_cmdbuf_bus_address;
	u32 *status_virtual_address;
	size_t status_bus_address;
	size_t mmu_status_bus_address;
	u32 status_size;
	u32 executing_status;
	struct drm_file *filp;
	u16 core_id;
	u16 cmdbuf_id;
	u8 cmdbuf_data_loaded;
	u8 cmdbuf_data_linked;
	u8 cmdbuf_run_done;
	u8 cmdbuf_need_remove;
	u32 waited;
	u8 has_end_cmdbuf;
	u8 no_normal_int_cmdbuf;
	struct process_manager_obj *process_manager_obj;
};

struct vpu_pcie_info {
	u32 *vpu_vaddr_base;
	dma_addr_t  vpu_paddr_base;
	u32 *vram_vaddr_base;
	dma_addr_t  vram_paddr_base;
	u32 size;
};

struct subsys_config {
	unsigned long base_addr;
	int irq;
	u32 subsys_type;
	u32 submodule_offset[HW_CORE_MAX];
	u32 submodule_iosize[HW_CORE_MAX];
	volatile u8 *submodule_hwregs[HW_CORE_MAX];
	int has_apbfilter[HW_CORE_MAX];
};

struct apbfilter_cfg {
	u32 nbr_mask_regs;
	u32 mask_reg_offset;
	u32 page_sel_addr;
	u8  num_mode;
	u8  mask_bits_per_reg;
	u32 id; /* id of the subsystem */
	u32 type; /* type of core to be written */
	u32 has_apbfilter;
};

struct axife_cfg {
	u8 axi_rd_chn_num;
	u8 axi_wr_chn_num;
	u8 axi_rd_burst_length;
	u8 axi_wr_burst_length;
	u8 fe_mode;
	u32 id;
};

struct regsize_desc {
	u32 slice; /* id of the slice */
	u32 id; /* id of the subsystem */
	u32 type; /* type of core to be written */
	u32 size; /* iosize of the core */
};

struct core_desc {
	u32 id; /* id of the subsystem */
	u32 type; /* type of core to be written */
	u32 *regs; /* pointer to user registers */
	u32 size; /* size of register space */
	u32 reg_id; /* id of reigster to be read/written */
};

struct core_param {
	u32 slice; /* id of the slice */
	u32 id; /* id of the subsystem */
	u32 type; /* type of core to be written */
	u32 size; /* iosize of the core */
	u32 asic_id; /* asic id of the core */
};

struct subsys_desc {
	u32 subsys_num;   /* total subsystems count */
	u32 subsys_vcmd_num;  /* subsystems with vcmd */
};

typedef struct {
	u32 cfg[HXDEC_MAX_CORES];  /* indicate the supported format */
	u32 cfg_backup[HXDEC_MAX_CORES];       /* back up of cfg */
	int its_main_core_id[HXDEC_MAX_CORES]; /* indicate if main core exist */
	int its_aux_core_id[HXDEC_MAX_CORES];  /* indicate if aux core exist */
} core_cfg;

struct DEC_PRIVATE_DATA {
	int dec_irq;
	int pp_irq;
	u32 dec_regs[HXDEC_MAX_CORES][DEC_IO_SIZE_MAX / 4];
	u32 apbfilter_regs[HXDEC_MAX_CORES][DEC_IO_SIZE_MAX / 4 + 1];
	u32 shadow_dec_regs[HXDEC_MAX_CORES][DEC_IO_SIZE_MAX / 4];
	int reg_access_opt;
	atomic_t *irq_rx;
	atomic_t *irq_tx;

	struct vpu_pcie_info dec_pcie;

	core_cfg config;
	struct subsys_config vpu_subsys[MAX_SUBSYS_NUM];
	struct apbfilter_cfg apbfilter_cfg[MAX_SUBSYS_NUM][HW_CORE_MAX];
	struct axife_cfg axife_cfg[MAX_SUBSYS_NUM];
	struct semaphore dec_core_sem;
	struct drm_file *dec_owner[HXDEC_MAX_CORES];
	struct file *pp_owner[HXDEC_MAX_CORES];
};

enum VPU_MM_Status {
	MM_STATUS_OK                          =    0,
	MM_STATUS_FALSE                       =   -1,
	MM_STATUS_INVALID_ARGUMENT            =   -2,
	MM_STATUS_INVALID_OBJECT              =   -3,
	MM_STATUS_OUT_OF_MEMORY               =   -4,
	MM_STATUS_NOT_FOUND                   =   -19,
};

struct kernel_addr_desc {
	size_t	bus_address;  /* buffer virtual address */
	size_t	mmu_bus_address;  /* buffer physical address in MMU */
	unsigned int size;	/* physical size */
};

struct addr_desc {
	void *virtual_address;  /* buffer virtual address */
	size_t bus_address;  /* buffer physical address */
	unsigned int size;  /* physical size */
};

/* Used in vcmd initialization in hantro_vcmd_xxx.c. */
/* May be unified in next step. */
struct vcmd_config {
	unsigned long vcmd_base_addr;
	u32 vcmd_iosize;
	int vcmd_irq;
	u32 sub_module_type;        // input vc8000e=0,IM=1,vc8000d=2,jpege=3,jpegd=4
	u16 submodule_main_addr;    // in byte
	u16 submodule_dec400_addr;
	u16 submodule_L2Cache_addr; // in byte
	u16 submodule_MMU_addr;     // in byte
	u16 submodule_MMUWrite_addr;// in byte
	u16 submodule_axife_addr;   // in byte
};

struct hantrovcmd_dev {
	struct vcmd_config vcmd_core_cfg;
	u32 core_id; /* vcmd core id for driver and sw internal use */
	u32 sw_cmdbuf_rdy_num;
	spinlock_t *spinlock;
	wait_queue_head_t *wait_queue;
	wait_queue_head_t *wait_abort_queue;
	bi_list list_manager;
	volatile u8 *hwregs; /* IO mem base */
	u32 reg_mirror[ASIC_VCMD_SWREG_AMOUNT];
	u32 duration_without_int;  /* number of cmdbufs without interrupt */
	volatile u8 working_state;
	u64 total_exe_time;
	u16 status_cmdbuf_id; /* used for analyse configuration in cwl */
	u32 hw_version_id; /* megvii 0x43421001, later 0x43421102 */
	u32 *vcmd_reg_mem_virtual_address;
	size_t vcmd_reg_mem_bus_address;
	unsigned int mmu_vcmd_reg_mem_bus_address;
	u32  vcmd_reg_mem_size;
};

struct dec_hw_info {
	u32 hw_id;
	u32 build_id;
	u32 synth_cfg;
	u32 synth_cfg_2;
	u32 synth_cfg_3;
	u32 pp_synth_cfg;
	u32 fuse_cfg;
	u32 pp_cfg_stat;
};
struct DEC_VCMD_PRIVATE_DATA {
	int vcmd_type_core_num[MAX_VCMD_TYPE];
	u16 vcmd_position[MAX_VCMD_TYPE];
	u16 total_vcmd_core_num;
	u16 cmdbuf_used_pos;
	u16 cmdbuf_used_residual;
	u16 cmdbuf_used[TOTAL_DISCRETE_CMDBUF_NUM];
	unsigned long gBaseDDRHw;
	size_t base_ddr_addr;

	bi_list_node * global_cmdbuf_node[TOTAL_DISCRETE_CMDBUF_NUM];
	bi_list global_process_manager;
	struct semaphore vcmd_reserve_cmdbuf_sem[MAX_VCMD_TYPE];
	struct noncache_mem dec_vcmd_mem_pool[3];
	spinlock_t owner_lock_vcmd[MAX_VCMD_NUMBER];
	wait_queue_head_t wait_queue_vcmd[MAX_VCMD_NUMBER];
	wait_queue_head_t abort_queue_vcmd[MAX_VCMD_NUMBER];
	wait_queue_head_t mc_wait_queue;

	struct vpu_vcmd_buff vcmd_buff;
	struct dec_hw_info hw_info;
	struct vpu_vcmd_buff *vcmd_pool;
	struct vpu_vcmd_buff *vcmd_reg;
};

struct config_parameter {
	u16 module_type;
	u16 vcmd_core_num;
	u16 submodule_main_addr;
	u16 submodule_dec400_addr;
	u16 submodule_L2Cache_addr;
	u16 submodule_MMU_addr;
	u16 submodule_MMUWrite_addr;
	u16 submodule_axife_addr;
	u16 config_status_cmdbuf_id;
	u32 vcmd_hw_version_id;
};

/* SubsysDesc & CoreDesc are used for configuration */
struct SubsysDesc {
	int slice_index;    /* slice this subsys belongs to */
	int index;   /* subsystem index */
	long base;
};

struct CoreDesc {
	int slice;
	int subsys;     /* subsys this core belongs to */
	enum core_type core_type;
	int offset;     /* offset to subsystem base */
	int iosize;
	int irq;
	int has_apb;
};

bi_list_node *find_last_linked_cmdbuf(bi_list_node *current_node);
unsigned long long GetMMUAddress(void);
void free_process_manager_node(bi_list_node *process_node);
bi_list_node *remove_cmdbuf_node_from_list(bi_list *list,
					   bi_list_node *cmdbuf_node);
long release_process_node_cleanup(bi_list *list);
u64 calculate_executing_time_after_node(bi_list_node *exe_cmdbuf_node);
void create_kernel_process_manager(bi_list *global_process_manager);
bi_list_node *create_process_manager_node(void);
int wait_process_resource_rdy(struct process_manager_obj *manager_obj);
void free_cmdbuf_obj(struct cmdbuf_obj *cmdbuf_obj);
struct cmdbuf_obj *create_cmdbuf_obj(void);

#endif

