// SPDX-License-Identifier: GPL-3.0-or-later
/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * jessy 20221019 Sietium
 */
#include <linux/kernel.h>
#include <linux/cpumask.h>
#include <asm/irq.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>

#include "vpu_comm.h"
#include "vpu_list.h"
#include "vpu/vpu_dec/vpu_vcmd.h"

/* process manager object management*/
static struct process_manager_obj *create_process_manager_obj(void)
{
	struct process_manager_obj *process_manager_obj = NULL;
	process_manager_obj = vmalloc(sizeof(struct process_manager_obj));
	if (process_manager_obj == NULL) {
		gb_printf(KERN_ERR, "%s: vmalloc for process_manager_obj fail!\n",
		 __func__);
		return process_manager_obj;
	}
	memset(process_manager_obj, 0, sizeof(struct process_manager_obj));
	return process_manager_obj;
}

static void free_process_manager_obj(struct process_manager_obj *process_manager_obj)
{
	if (process_manager_obj == NULL) {
		gb_printf(KERN_ERR, "%s: process_manager_obj = NULL, return\n", __func__);
		return;
	}
	vfree(process_manager_obj);	//free current cmdbuf_obj
}

bi_list_node *create_process_manager_node(void)
{
	bi_list_node *current_node = NULL;
	struct process_manager_obj *process_manager_obj = NULL;

	process_manager_obj = create_process_manager_obj();
	if (process_manager_obj == NULL) {
		gb_printf(KERN_ERR, "%s: error process_manager_obj == NULL!\n", __func__);
		return NULL;
	}
	process_manager_obj->total_exe_time = 0;
	spin_lock_init(&process_manager_obj->spinlock);
	init_waitqueue_head(&process_manager_obj->wait_queue);
	current_node = bi_list_create_node();
	if (current_node == NULL) {
		gb_printf(KERN_ERR, "%s: bi_list_create_node fail!", __func__);
		free_process_manager_obj(process_manager_obj);
		return NULL;
	}
	current_node->data = (void *)process_manager_obj;
	return current_node;
}

void create_kernel_process_manager(bi_list *global_process_manager)
{
	bi_list_node *node;
	struct process_manager_obj *process_manager_obj = NULL;

	node = create_process_manager_node();
	process_manager_obj = (struct process_manager_obj *)node->data;
	process_manager_obj->filp = NULL;
	bi_list_insert_node_tail(global_process_manager, node);
}

/* cmdbuf object management */
struct cmdbuf_obj *create_cmdbuf_obj(void)
{
	struct cmdbuf_obj *cmdbuf_obj = NULL;

	cmdbuf_obj = vmalloc(sizeof(struct cmdbuf_obj));
	if (cmdbuf_obj == NULL) {
		gb_printf(KERN_ERR, "%s: vmalloc cmdbuf_obj fail cmdbuf_obj = NULL\n",
		 __func__);
		return cmdbuf_obj;
	}
	memset(cmdbuf_obj, 0, sizeof(struct cmdbuf_obj));
	return cmdbuf_obj;
}

void free_cmdbuf_obj(struct cmdbuf_obj *cmdbuf_obj)
{
	if (cmdbuf_obj == NULL) {
		gb_printf(KERN_ERR, "%s: remove_cmdbuf_obj cmdbuf_obj == NULL\n", __func__);
		return;
	}
	vfree(cmdbuf_obj);	/* free current cmdbuf_obj */
}

int wait_process_resource_rdy(struct process_manager_obj *manager_obj)
{
	return manager_obj->total_exe_time <= PROCESS_MAX_SUM_OF_IMAGE_SIZE;
}

bi_list_node *find_last_linked_cmdbuf(bi_list_node *current_node)
{
	bi_list_node *new_cmdbuf_node = current_node;
	bi_list_node *last_cmdbuf_node;
	struct cmdbuf_obj *cmdbuf_obj = NULL;

	if (current_node == NULL)
		return NULL;
	last_cmdbuf_node = new_cmdbuf_node;
	new_cmdbuf_node = new_cmdbuf_node->previous;
	while (1) {
		if (new_cmdbuf_node == NULL)
			return last_cmdbuf_node;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if (cmdbuf_obj->cmdbuf_data_linked)
			return new_cmdbuf_node;

		last_cmdbuf_node = new_cmdbuf_node;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	return NULL;
}

unsigned long long GetMMUAddress(void)
{
	unsigned long long address = 0;

	address = PAGE_PCIE_START_ADDRESS;
	return address;
}

void free_process_manager_node(bi_list_node *process_node)
{
	struct process_manager_obj *process_manager_obj = NULL;

	if (process_node == NULL) {
		gb_printf(KERN_ERR, "%s: process_node = NULL, so return\n", __func__);
		return;
	}
	process_manager_obj = (struct process_manager_obj *)process_node->data;
	/* free struct process_manager_obj */
	free_process_manager_obj(process_manager_obj);
	/* free current process_manager_obj entity. */
	bi_list_free_node(process_node);
}

/* just remove, not free the node. */
bi_list_node *remove_cmdbuf_node_from_list(bi_list *list,
	bi_list_node *cmdbuf_node)
{
	struct cmdbuf_obj *cmdbuf_obj = NULL;

	if (cmdbuf_node == NULL) {
		gb_printf(KERN_ERR, "%s: error remove_cmdbuf = NULL, return\n", __func__);
		return NULL;
	}
	if (cmdbuf_node->next) {
		cmdbuf_obj = (struct cmdbuf_obj *)cmdbuf_node->next->data;
		if (cmdbuf_obj->cmdbuf_need_remove == 1) {
			bi_list_remove_node(list, cmdbuf_node);
			return cmdbuf_node;
		} else {
			return NULL; /* cmdbuf_node not last one, should not be removed. */
		}
	} else
		return NULL; /* the last one, should not be removed. */
}

/* calculate executing_time of each vcmd */
u64 calculate_executing_time_after_node(bi_list_node *exe_cmdbuf_node)
{
	u64 time_run_all = 0;
	struct cmdbuf_obj *cmdbuf_obj_temp = NULL;

	while (1) {
		if (exe_cmdbuf_node == NULL)
			break;
		cmdbuf_obj_temp = (struct cmdbuf_obj *)exe_cmdbuf_node->data;
		time_run_all += cmdbuf_obj_temp->executing_time;
		exe_cmdbuf_node = exe_cmdbuf_node->next;
	}
	return time_run_all;
}

long release_process_node_cleanup(bi_list *list)
{
	bi_list_node *new_process_node = NULL;

	while (1) {
		new_process_node = list->head;
		if (new_process_node == NULL)
			break;
		/* remove node from list */
		bi_list_remove_node(list, new_process_node);
		/* remove node from list */
		free_process_manager_node(new_process_node);
	}
	return 0;
}


