// SPDX-License-Identifier: GPL-3.0-or-later
/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * jessy 20221019 Sietium
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#include "common/gb_common.h"
#include "vpu_vc8000E_irq.h"
#include "vpu_vc8000E_vcmd.h"
#include "vpu/vpu_comm/vpu_list.h"
#include "vpu/vpu_comm/vpu_vcmd_registers.h"
#include "vpu/vpu_comm/vpu_comm.h"
#include "vpu/vpu_comm/vpu_basetype.h"

static void enc_vcmd_irq_reset_handle(struct enc_hantrovcmd_dev *dev)
{
	bi_list_node *new_cmdbuf_node = NULL;
	bi_list_node *base_cmdbuf_node = NULL;
	struct cmdbuf_obj *cmdbuf_obj = NULL;

	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0))
			break;
		new_cmdbuf_node = new_cmdbuf_node->next;
	}
	base_cmdbuf_node = new_cmdbuf_node;
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
	if (dev->sw_cmdbuf_rdy_num != 0)
		enc_vcmd_start(dev, base_cmdbuf_node); /*restart new command*/
}

static int enc_vcmd_irq_abort_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	size_t base_ddr_addr = 0;
	size_t exe_cmdbuf_busAddress;
	bi_list_node *base_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();
	base_ddr_addr = enc_data->enc_base_ddr_addr;

	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	if (dev->hw_version_id > HW_ID_1_0_C) {
		new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
		if (new_cmdbuf_node == NULL) {
			gb_printf(KERN_ERR, "%s: error, new_cmdbuf_node == NULL\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	} else {
		exe_cmdbuf_busAddress =
			VCMDGetAddrRegisterValue((const void *)dev->hwregs,
			dev->reg_mirror, HWIF_VCMD_EXECUTING_CMD_ADDR);
		/* find the cmderror cmdbuf */
		while (1) {
			if (new_cmdbuf_node == NULL) {
				spin_unlock_irqrestore(dev->spinlock, flags);
				return -1;
			}
			cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
			if ((((cmdbuf_obj->cmdbuf_bus_address -
			 base_ddr_addr) <= exe_cmdbuf_busAddress) &&
			  (((cmdbuf_obj->cmdbuf_bus_address -
			   base_ddr_addr + cmdbuf_obj->cmdbuf_size)
			    > exe_cmdbuf_busAddress)))
				&& (cmdbuf_obj->cmdbuf_run_done == 0))
				break;
			new_cmdbuf_node = new_cmdbuf_node->next;
		}
	}
	base_cmdbuf_node = new_cmdbuf_node;

	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	base_cmdbuf_node = base_cmdbuf_node->next;
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	if (enc_data->software_triger_abort == 0) {
		enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
		if (dev->sw_cmdbuf_rdy_num != 0)
			enc_vcmd_start(dev, base_cmdbuf_node);
	}
	spin_unlock_irqrestore(dev->spinlock, flags);
	if (*cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	wake_up_interruptible_all(dev->wait_abort_queue);
	return 0;
}

static int enc_vcmd_irq_buserr_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	size_t base_ddr_addr = 0;
	size_t exe_cmdbuf_busAddress;
	bi_list_node *base_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();
	base_ddr_addr = enc_data->enc_base_ddr_addr;

	/* bus error, don't need to reset where to record status? */
	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	if (dev->hw_version_id > HW_ID_1_0_C) {
		new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
		if (new_cmdbuf_node == NULL) {
			gb_printf(KERN_ERR, "%s: error, new_cmdbuf_node == NULL\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	} else {
		exe_cmdbuf_busAddress =
			VCMDGetAddrRegisterValue((const void *)dev->hwregs,
				dev->reg_mirror, HWIF_VCMD_EXECUTING_CMD_ADDR);
		/* find the cmderror cmdbuf */
		while (1) {
			if (new_cmdbuf_node == NULL) {
				spin_unlock_irqrestore(dev->spinlock, flags);
				return -1;
			}
			cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
			if ((((cmdbuf_obj->cmdbuf_bus_address -
			 base_ddr_addr) <= exe_cmdbuf_busAddress) &&
			  (((cmdbuf_obj->cmdbuf_bus_address -
			   base_ddr_addr + cmdbuf_obj->cmdbuf_size)
			    > exe_cmdbuf_busAddress)))
				&& (cmdbuf_obj->cmdbuf_run_done == 0))
				break;
			new_cmdbuf_node = new_cmdbuf_node->next;
		}
	}
	base_cmdbuf_node = new_cmdbuf_node;

	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	new_cmdbuf_node = base_cmdbuf_node;
	if (new_cmdbuf_node != NULL) {
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_BUSERR;
	}
	base_cmdbuf_node = base_cmdbuf_node->next;
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
	if (dev->sw_cmdbuf_rdy_num != 0)
		enc_vcmd_start(dev, base_cmdbuf_node); /*restart new command*/
	spin_unlock_irqrestore(dev->spinlock, flags);
	if (*cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	return 0;
}

static void enc_vcmd_reset_current_asic(struct enc_hantrovcmd_dev *dev)
{
	u32 result;

	if (dev->hwregs != NULL) {
		/* disable interrupt at first */
		vcmd_write_reg((const void *)dev->hwregs,
			VCMD_REGISTER_INT_CTL_OFFSET, 0x0000);
		/* reset all	*/
		vcmd_write_reg((const void *)dev->hwregs,
			VCMD_REGISTER_CONTROL_OFFSET, 0x0002);
		/* read status register */
		result = vcmd_read_reg((const void *)dev->hwregs,
			VCMD_REGISTER_INT_STATUS_OFFSET);
		/* clean status register */
		vcmd_write_reg((const void *)dev->hwregs,
			VCMD_REGISTER_INT_STATUS_OFFSET, result);
	}
}

static int enc_vcmd_irq_timeout_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	size_t base_ddr_addr = 0;
	size_t exe_cmdbuf_busAddress;
	bi_list_node *base_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();
	base_ddr_addr = enc_data->enc_base_ddr_addr;

	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	if (dev->hw_version_id > HW_ID_1_0_C) {
		new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
		if (new_cmdbuf_node == NULL) {
			gb_printf(KERN_ERR, "%s: error, new_cmdbuf_node == NULL\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	} else {
		exe_cmdbuf_busAddress =
			VCMDGetAddrRegisterValue((const void *)dev->hwregs,
				dev->reg_mirror, HWIF_VCMD_EXECUTING_CMD_ADDR);
		/* find the timeout cmdbuf */
		while (1) {
			if (new_cmdbuf_node == NULL) {
				spin_unlock_irqrestore(dev->spinlock, flags);
				return -1;
			}
			cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
			if ((((cmdbuf_obj->cmdbuf_bus_address -
			 base_ddr_addr) <= exe_cmdbuf_busAddress) &&
			  (((cmdbuf_obj->cmdbuf_bus_address -
			   base_ddr_addr + cmdbuf_obj->cmdbuf_size) >
			    exe_cmdbuf_busAddress)))
				 && (cmdbuf_obj->cmdbuf_run_done == 0))
				break;
			new_cmdbuf_node = new_cmdbuf_node->next;
		}
	}
	base_cmdbuf_node = new_cmdbuf_node;
	new_cmdbuf_node = new_cmdbuf_node->previous;

	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
	if (dev->sw_cmdbuf_rdy_num != 0) {
		enc_vcmd_reset_current_asic(dev);
		enc_vcmd_start(dev, base_cmdbuf_node); /*restart new command*/
	}
	spin_unlock_irqrestore(dev->spinlock, flags);
	if (*cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	return 0;
}

static int enc_vcmd_irq_cmderr_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	size_t base_ddr_addr = 0;
	size_t exe_cmdbuf_busAddress;
	bi_list_node *base_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();
	base_ddr_addr = enc_data->enc_base_ddr_addr;

	/* command error,don't need to reset */
	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	if (dev->hw_version_id > HW_ID_1_0_C) {
		new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
		if (new_cmdbuf_node == NULL) {
			gb_printf(KERN_ERR, "%s: error, new_cmdbuf_node == NULL\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	} else {
		exe_cmdbuf_busAddress =
			VCMDGetAddrRegisterValue((const void *)dev->hwregs,
				dev->reg_mirror, HWIF_VCMD_EXECUTING_CMD_ADDR);
		/* find the cmderror cmdbuf */
		while (1) {
			if (new_cmdbuf_node == NULL) {
				spin_unlock_irqrestore(dev->spinlock, flags);
				return -1;
			}
			cmdbuf_obj =
			 (struct cmdbuf_obj *)new_cmdbuf_node->data;
			if ((((cmdbuf_obj->cmdbuf_bus_address -
			 base_ddr_addr) <= exe_cmdbuf_busAddress)
			  && (((cmdbuf_obj->cmdbuf_bus_address -
			   base_ddr_addr + cmdbuf_obj->cmdbuf_size)
			    > exe_cmdbuf_busAddress)))
				 && (cmdbuf_obj->cmdbuf_run_done == 0))
				break;
			new_cmdbuf_node = new_cmdbuf_node->next;
		}
	}
	base_cmdbuf_node = new_cmdbuf_node;
	/* this cmdbuf and cmdbufs prior to itself, run_done = 1 */
	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	new_cmdbuf_node = base_cmdbuf_node;
	if (new_cmdbuf_node != NULL) {
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_CMDERR;
	}
	base_cmdbuf_node = base_cmdbuf_node->next;
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
	if (dev->sw_cmdbuf_rdy_num != 0)
		enc_vcmd_start(dev, base_cmdbuf_node); /*restart new command*/
	spin_unlock_irqrestore(dev->spinlock, flags);
	if (*cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	return 0;
}

static int enc_vcmd_irq_endcmd_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	bi_list_node *base_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();

	/* end command interrupt */
	new_cmdbuf_node = dev->list_manager.head;
	dev->working_state = WORKING_STATE_IDLE;
	if (dev->hw_version_id > HW_ID_1_0_C) {
		new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
		if (new_cmdbuf_node == NULL) {
			gb_printf(KERN_ERR, "%s: error, new_cmdbuf_node == NULL\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	} else {
		/* find the end cmdbuf */
		while (1) {
			if (new_cmdbuf_node == NULL) {
				spin_unlock_irqrestore(dev->spinlock, flags);
				return -1;
			}
			cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
			if ((cmdbuf_obj->has_end_cmdbuf == 1)
				&& (cmdbuf_obj->cmdbuf_run_done == 0))
				break;
			new_cmdbuf_node = new_cmdbuf_node->next;
		}
	}
	base_cmdbuf_node = new_cmdbuf_node;
	/* this cmdbuf and cmdbufs prior to itself, run_done = 1 */
	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	base_cmdbuf_node = base_cmdbuf_node->next;
	enc_vcmd_delink_cmdbuf(dev, base_cmdbuf_node);
	enc_vcmd_link_cmdbuf(dev, base_cmdbuf_node);
	if (dev->sw_cmdbuf_rdy_num != 0)
		enc_vcmd_start(dev, base_cmdbuf_node); /*restart new command*/

	spin_unlock_irqrestore(dev->spinlock, flags);
	if (*cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	return 0;
}

static int enc_vcmd_irq_intcmd_handle(struct enc_hantrovcmd_dev *dev,
	u32 cmdbuf_id, unsigned long flags, u32 *cmdbuf_processed_num)
{
	bi_list_node *new_cmdbuf_node = NULL;
	u32 tmp_num = *cmdbuf_processed_num;
	struct cmdbuf_obj *cmdbuf_obj = NULL;
	struct ENC_PRIVATE_DATA *enc_data = NULL;

	enc_data = get_enc_private_handle();

	if (dev->hw_version_id <= HW_ID_1_0_C) {
		if (cmdbuf_id >= TOTAL_DISCRETE_CMDBUF_NUM) {
			gb_printf(KERN_ERR, "%s: error, cmdbuf_id greater than ceiling\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return -1;
		}
	}
	new_cmdbuf_node = enc_data->enc_global_cmdbuf_node[cmdbuf_id];
	if (new_cmdbuf_node == NULL) {
		gb_printf(KERN_ERR, "%s: error cmdbuf_id !!\n", __func__);
		spin_unlock_irqrestore(dev->spinlock, flags);
		return -1;
	}
	// interrupt cmdbuf and cmdbufs prior to itself, run_done = 1
	while (1) {
		if (new_cmdbuf_node == NULL)
			break;
		cmdbuf_obj = (struct cmdbuf_obj *)new_cmdbuf_node->data;
		if ((cmdbuf_obj->cmdbuf_run_done == 0)) {
			cmdbuf_obj->cmdbuf_run_done = 1;
			cmdbuf_obj->executing_status = CMDBUF_EXE_STATUS_OK;
			tmp_num++;
			*cmdbuf_processed_num = tmp_num;
		} else
			break;
		new_cmdbuf_node = new_cmdbuf_node->previous;
	}
	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18))
irqreturn_t enc_hantrovcmd_isr(int irq, void *dev_id,
	struct pt_regs *regs)
#else
irqreturn_t enc_hantrovcmd_isr(int irq, void *dev_id)
#endif
{
	unsigned int handled = 0;
	struct enc_hantrovcmd_dev *dev = (struct enc_hantrovcmd_dev *)dev_id;
	u32 irq_status = 0;
	int ret_val;
	unsigned long flags;
	u32 cmdbuf_processed_num = 0;
	u32 cmdbuf_id = 0;

/*
 * If core is not reserved by any user,
 * but irq is received, just clean it
 */
	spin_lock_irqsave(dev->spinlock, flags);
	if (dev->list_manager.head == NULL) {
		gb_printf(KERN_INFO, "%s: received IRQ but core has nothing to do.\n",
		 __func__);
		irq_status = vcmd_read_reg((const void *)dev->hwregs,
			VCMD_REGISTER_INT_STATUS_OFFSET);
		vcmd_write_reg((const void *)dev->hwregs,
			VCMD_REGISTER_INT_STATUS_OFFSET, irq_status);
		spin_unlock_irqrestore(dev->spinlock, flags);
		return IRQ_HANDLED;
	}

	//gb_printf(KERN_INFO, "%s: received IRQ to do something!!\n", __func__);
	irq_status = vcmd_read_reg((const void *)dev->hwregs,
		VCMD_REGISTER_INT_STATUS_OFFSET);
	if (!irq_status) {
		spin_unlock_irqrestore(dev->spinlock, flags);
		return IRQ_HANDLED;
	}

	//gb_printf(KERN_INFO, "%s: dev->core_id : %d, irq_status:%x\n",
	//	__func__, dev->core_id, irq_status);
	vcmd_write_reg((const void *)dev->hwregs,
		VCMD_REGISTER_INT_STATUS_OFFSET, irq_status);
	dev->reg_mirror[VCMD_REGISTER_INT_STATUS_OFFSET / 4] = irq_status;
	if ((dev->hw_version_id > HW_ID_1_0_C) && (irq_status & 0x3f)) {
		/* if error,read from register directly. */
		cmdbuf_id = vcmd_get_register_value((const void *)dev->hwregs,
			dev->reg_mirror, HWIF_VCMD_CMDBUF_EXECUTING_ID);
		if (cmdbuf_id >= TOTAL_DISCRETE_CMDBUF_NUM) {
			gb_printf(KERN_ERR, "%s: cmdbuf_id great than ceiling error\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return IRQ_HANDLED;
		}
	} else if (dev->hw_version_id > HW_ID_1_0_C) {
		/* read cmdbuf id from ddr */
		cmdbuf_id = *(dev->vcmd_reg_mem_virtual_address +
			EXECUTING_CMDBUF_ID_ADDR);
		//gb_printf(KERN_INFO, "%s: cmdbuf_id %d from virtual!!\n",
		// __func__, cmdbuf_id);
		if (cmdbuf_id >= TOTAL_DISCRETE_CMDBUF_NUM) {
			gb_printf(KERN_ERR, "%s: error cmdbuf_id greater than ceiling\n",
			 __func__);
			spin_unlock_irqrestore(dev->spinlock, flags);
			return IRQ_HANDLED;
		}
	}

	if (vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_RESET)) {
		enc_vcmd_irq_reset_handle(dev);
		handled++;
		spin_unlock_irqrestore(dev->spinlock, flags);
		return IRQ_HANDLED;
	}
	if (vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_ABORT)) {
		ret_val = enc_vcmd_irq_abort_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
		return IRQ_HANDLED;
	}
	if (vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_BUSERR)) {
		ret_val = enc_vcmd_irq_buserr_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
		return IRQ_HANDLED;
	}
	if (vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_TIMEOUT)) {
		ret_val = enc_vcmd_irq_timeout_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
		return IRQ_HANDLED;
	}
	if (vcmd_get_register_mirror_value(dev->reg_mirror,
		HWIF_VCMD_IRQ_CMDERR)) {
		ret_val = enc_vcmd_irq_cmderr_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
		return IRQ_HANDLED;
	}
	if (vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_ENDCMD)) {
		ret_val = enc_vcmd_irq_endcmd_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
		return IRQ_HANDLED;
	}
	if (dev->hw_version_id <= HW_ID_1_0_C)
		cmdbuf_id = vcmd_get_register_mirror_value(dev->reg_mirror,
			HWIF_VCMD_IRQ_INTCMD);
	if (cmdbuf_id) {
		ret_val = enc_vcmd_irq_intcmd_handle(dev, cmdbuf_id,
			flags, &cmdbuf_processed_num);
		if (ret_val != 0)
			return IRQ_HANDLED;
		handled++;
	}

	spin_unlock_irqrestore(dev->spinlock, flags);
	if (cmdbuf_processed_num)
		wake_up_interruptible_all(dev->wait_queue);
	if (!handled)
		gb_printf(KERN_INFO, "%s: IRQ received, but not hantro's!\n", __func__);
	return IRQ_HANDLED;
}


